package AFS::Instant::CellDB;

use 5.006;
use Carp qw(cluck);
use strict;
use warnings;
use Carp;
use Fb::Cache;
use AFS::Instant::Toolz qw(getCellName);
use Fb::Out qw(out $VERBOSE);
use Fb::Data qw(parse_storagesize parse_timeinterval $MAXINT32 addToSet removeFromSet setEquality);
use Fb::IO qw(iaexec);
use Fb::Network qw(getFirstIP);

use AFS::Instant::aCall;

use Time::Local;
use POSIX ":sys_wait_h";
use Data::Dumper;
use DBI;
use XML::LibXML;
use XML::LibXML::XPathContext;
use Digest;

require Exporter;
our @ISA=qw(Exporter);
our $VERSION='3.0';
our $xmlnsci='http://instantafs.cbs.mpg.de/xmlns/cellinfo/1.0';

# This is the time in seconds after which an information is considered unreliable.
# Since data partitions are actively being disabled by the address update processes
# on missing connection, this value can be very high (5 hours).
our $ONLINETIMEOUT=5*60*60;

=head1 NAME

AFS::Instant::CellDB - SQL-Database-accelerated *read* access to AFS-information

=head1 SYNOPSIS

 use AFS::Instant::CellDB;
  
 my $celldb=AFS::Instant::CellDB->new();

 # This has to be done for every process (i.e. after fork()ing
 $celldb->ensureDB() || die "Unable to connect to database";

 my $servers=$celldb->serverInfo('uuidlist');
 my $volref=$celldb->getVolume('root.cell');
 my $rootvolumes=$celldb->listVolumes(pattern => '^root\..*');
 my $isonline=$celldb->instanceInfo('accessible',$volref->{'rw'});
  
 my $serveruuid=$celldb->serverInfo('address2uuid','myserver.my.domain');
 my $partitions=$celldb->getPartition($serveruuid,'a');
 my @partitions=@{$celldb->getPartitions(serveruuid => $serveruuid,partition => 'a',format => 'arrayref')};

 # Location (=Partition) group handling
 my $mypartitiongroup=$celldb->getLocationGroup('raidpartitions');
 my $allpartitions=$celldb->expandLocationPattern('*/*',format => 'hashtree');

 # Volume filtering
 my $filtertree=$celldb->parseVolumeFilter("root.* rocount>1");
 my $volumeshash=$celldb->filterVolumes($filtertree);

 # Volume groups (as defined in /a/a/etc/cell.xml)
 my $volumegroups=$celldb->listVolumeGroups();
 my $volumeshash2=$celldb->getVolumeGroup('rootvolumes');
 my $volref2=$celldb->getVolume('root.cell');
 my $membershipshash=$celldb->getVolumeMemberships($volref2);
  
=head1 REQUIRES

XML::LibXML, XML::LibXML::XPathContext, AFS::Instant::Toolz, DBI
DBD::Mysql, Time::Local;

=head1 METHODS

=head2 Constructor

=over 4

=item $celldb=new(%options)

Creates a celldb-object. Several options may be given:

=over 4

=item B<cache>

Enables using caching for volume information fetched from DB
Cache validity time is given in seconds. Default: B<0> (don't cache).

=item B<cellname>

Name of the cell to connect to. Defaults to AFS::Instant::Toolz's default
cell which is usually the local workstation's cell but might be
changed by command line parameters or environment variables.

=item B<timeout>

Time (in seconds) after which an information in the DB is considered outdated.
The only objects carrying timestamps in the DB are partition records which are
always joined with physical instances. This way, reliability of an instance's
data cab be tested like this:

 if ( ( time() - $instance->{'infotime'} ) < $self->{onlinetimeout} ) { ... }

This value overrides the default timeout which is read from I<$ONLINETIMEOUT>.

=back

=item B<localdb>

Defines a filename of a SQLite database to be used as DB.

=item B<localdbschema>

Usually the schema at the default location is being used to populate the
database before using it.  This parameter can used to override it.

=item B<cellinfofile>

Read cellinfo from the given file.

=item B<dnslookups>

If servernames are looked up in DNS and request to DB actually done by IP.

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self={
		cellname => undef,
		db => undef,
		dbro => undef,
		dbpid => undef,
		localdb => undef,
		localdbschema => undef,
		serverinfo => {'infotime' => 0},
		volumecache => undef,
		use_cache => 0,
		cellinfo => {},
		spof => {},
		userpref => {},
		dnslookups => 1,
		onlinetimeout => $ONLINETIMEOUT,
		
	};
	my $me=bless($self,$package);
	if ( $options{'cache'} ) {
		$self->{use_cache}=1;
		if ( $self->{use_cache} ) {
			$self->{volumecache}=Fb::Cache->new({default_expires_in => $options{cache}});
		}
	}
	if ( defined($options{'dnslookups'}) ) {
		$self->{dnslookups}=$options{'dnslookups'};
	}
	if ( defined($ENV{'INSTANTAFS_LOCALDB'}) ) {
		$self->{localdb}=$ENV{'INSTANTAFS_LOCALDB'};
	}
	if ( defined($options{'localdb'}) ) {
		$self->{localdb}=$options{'localdb'};
	}
	if ( defined($options{'localdbschema'}) ) {
		$self->{localdbschema}=$options{'localdbschema'};
	}
	if ( defined($options{'timeout'}) ) {
		$self->{onlinetimeout}=$options{'timeout'};
	}
	if ( defined($options{'cellinfofile'}) ) {
		$self->{cellinfofile}=$options{'cellinfofile'};
	}
	if ( defined($options{'cellname'}) ) {
		$self->{cellname}=$options{'cellname'};
	} else {
		$self->{cellname}=getCellName() || return undef;
	}
	$self->ensureDB() || return undef;
	return $me;
}

=head2 Object Methods

=over 4

=item $celldb->ensureCellinfo();

Reads cell.xml from the local cell's /admin/etc directory.

=cut

sub ensureCellinfo {
	my $self=shift;
	my $cellxml;
	if ( $self->{cellinfofile} ) {
		$cellxml=$self->{cellinfofile};
	} else {
		my $dot='';
		if ( $ENV{'DOT'} ) {
			$dot='.';
		}
		$cellxml="/afs/" . $self->{cellname} . "/admin/${dot}etc/cell.xml";
	}
	my @stat=stat($cellxml);
	if ( defined($self->{cellinfo}->{'mtime'}) ) {
		if ( $self->{cellinfo}->{'mtime'} == $stat[9] ) {
			return 1;
		}
		out('d',"Cellinfo XML file was changed, reading again");
	} else {
		out('d',"Reading cellinfo XML file '$cellxml'");
	}
	if ( ! -r $cellxml ) {
		out('e',"Cellinfo XML file '$cellxml' not found or is not readable");
		return 0;
	}
	my $parser=XML::LibXML->new();
	$parser->load_ext_dtd(0);
	$parser->validation(0);
	out('d',"Reading cellinfo '$cellxml'");
	my $doc=$parser->parse_file($cellxml);
	my $root=$doc->getDocumentElement;
	my $xc=XML::LibXML::XPathContext->new($doc);
	$xc->registerNs('ci',$xmlnsci);
	$self->{cellinfo}={
		'root' => $root,
		'xc' => $xc,
		'doc' => $doc,
		'mtime' => $stat[9],
	};
	return 1;
}


=item $celldb->ensureDB()

Connects to the current cell's SQL server which is used to speed up things. Servername B<instantafsdb> is mandatory as SQL server name or alias. Database name is always B<instantafs>.

Access credentials to the Database read from:

=over 4

=item 1. B<~/.instantafs/>I<cellname>B<.db>

=item 2. B</afs/>I<cellname>B</admin/etc/secret/instantafs/instantafs.db>

=item 3. B</etc/instantafs/>I<cellname>B<.db>

=item 4. B</etc/instantafs/instantafs.db>

=back

First readable file access wins. The file is expected to contain two lines:

=over 4

=item Name of the database user

=item Password of the database user

=back

If no file could be accessed, connection will be established as the B<instantafsro> user with an
empty password. The database should grant this user SELECT permissions only.   

=cut

sub ensureDB {
   my $self=shift;
   if ( defined($self->{'db'}) ) {
   	if ( $self->{dbpid} == $$ ) {
	      return 1;
	   }
   }
	if ( $self->{'localdb'} ) {
		# Switch to local file based DB. This is for functional tests
		my $createschema=0;
		my @schemata=();
		my $filename="$self->{'localdb'}" . $self->{cellname} . ".sqlite";
		if ( ! -e $filename ) {
			out('d',"Creating new sqlite DB '$filename'");
			$createschema=1;
			my $schemafile="/usr/share/libafs-instant-perl/instantafs.sql";
			if ( $self->{localdbschema} ) {
				$schemafile=$self->{localdbschema};
			}
			my $fh;
			if ( ! open($fh,"< $schemafile") ) {
				out('e',"Unable to read DB schema from '$schemafile'");
				return 0;
			}
			my $schema='';
			while ( my $l=<$fh> ) {
				if ( $l =~ /^\s*(SET|UNIQUE|KEY|ENGINE|COMMENT|--|\/\*)/ ) {
					# ignore
				} elsif ( $l =~ /\s*;/ ) {
					push(@schemata,$schema);
					$schema='';
				} else {
					$schema .= $l;
				}
			}
			print $schema;
			close($fh);
		} else {
			out('d',"Accessing existing sqlite DB '$filename'");
		}
		my $dbh=DBI->connect("dbi:SQLite:dbname=$filename");
		if ( $createschema ) {
			for my $schema (@schemata) {
				my $query=$dbh->prepare($schema);
				if ( ! $query ) {
					out('e',"Unable to apply instantafs DB schema to local DB file");
					unlink($filename);
					return 0;				
				}
				if ( ! $query->execute() ) {
					out('e',"Unable to apply instantafs DB schema to local DB file");
					unlink($filename);
					return 0;
				}
			}
		}
		$dbh->do("PRAGMA synchronous = OFF");
		$dbh->do("PRAGMA cache_size = 10000");
		$self->{db}=$dbh;
		$self->{dbpid}=$$;
		return 1;
	}
	my $fh;my $login='instantafsro';my $passwd='';
	
	my @files=();
	if ( defined($ENV{'HOME'}) ) {
		push(@files,"$ENV{'HOME'}/.instantafs/$self->{cellname}.db");
	}
	push(@files,
   	"/afs/$self->{cellname}/admin/etc/secret/instantafs/instantafs.db",
   	"/etc/instantafs/$self->{cellname}.db",
   	'/etc/instantafs/instantafs.db'
   );
   for my $file (@files) {
	   if ( open($fh,"< $file") ) {
	   	out('dd',"Reading DB-credentials from '$file'");
			$login=<$fh>;chomp($login);  
			$passwd=<$fh>;chomp($passwd);
			close($fh);
			last;
		}
	}
	my $dbh;
	if ( $login eq 'instantafsro' ) {
		$self->{dbro}=1;
	} elsif ( $login eq 'instantafs' ) {
		$self->{dbro}=0;
	} else {
		$self->{dbro}=undef;
	}
	my $servername="instantafsdb";
	$dbh=DBI->connect("DBI:mysql:instantafs:$servername",$login,$passwd);
	if ( ! $dbh ) {
		out('e',"Unable to connect to Instantafs cache database");
		return 0;
	}
	out('d',"Opened connection to instantafs database");
	$dbh->{mysql_auto_reconnect}=1;
	$self->{db}=$dbh;
	$self->{dbpid}=$$;
	return 1;
}

sub ensureDBw {
	my $self=shift;
	return $self->ensureDB(@_);
}

=item $query=$celldb->sqlquery($sqlquery,$errormessage)

Returns a query object to the given I<$sqlquery>. On failure the
I<$errormessage> is displayed and B<undef> is returned.

=cut

sub sqlquery {
	my ($self,$sqlquery,$errormessage)=@_;
	my $query=$self->{db}->prepare($sqlquery);
	if ( ! $query ) {
		out('e',$errormessage);
		out('d',"Invalid query was: $sqlquery");
		return undef;
	}
	if ( ! $query->execute() ) {
		out('e',$errormessage);
		if ( $VERBOSE > 3 ) {
			out('d',"Query was: $sqlquery");
		}
		return undef;
	} else {
		if ( $VERBOSE > 3 ) {
			out('d',"Query was: $sqlquery");
		}
		return $query;
	}
}

=item $query=$celldb->sqlqueryw($sqlquery,$errormessage)

Returns a query object to the given I<$sqlquery>. On failure the
I<$errormessage> is displayed and B<undef> is returned.  This method must be
used for write accesses to DB.

=cut

sub sqlqueryw {
	my ($self,$sqlquery,$errormessage)=@_;
	my $query=$self->{db}->prepare($sqlquery);
	if ( ! $query ) {
		out('e',$errormessage);
		out('d',"Invalid query was: $sqlquery");
		return undef;
	}
	my $result=$query->execute();
	if ( ! $result ) {
		out('e',$errormessage);
		out('d',"Query was: $sqlquery");
		return undef;
	} else {
		if ( $VERBOSE > 2 ) {
			out('d',"Query was: $sqlquery");
		}
		return $query;
	}
}


sub sqlquote {
	my ($self,$string)=@_;
	return $self->{db}->quote($string);
}

=back

=head2 Gathering Information

=over 4

=item $celldb->serverInfo($function,$value)

serverInfo is a function family for getting information about AFS-fileservers. Caching is being used to minimize
SELECT-queries to the SQL DB. Available functions are:

=over 4

=item B<uuid2name>(I<$value>)

Resolves given UUID I<$value> to the primary name of the associated AFS-fileserver.

=item B<uuid2ioaddresses>(I<$value>)

Returns all accessible interfaces of the AFS-fileserver identified by its UUID in I<$value>.

=item B<uuid2ioaddress>(I<$value>)

Returns an arbitrary accessible interface of the AFS-fileserver identified by its UUID in I<$value>
If no interface is accessible, B<undef> is returned.

=item B<online>(I<$value>)

Tests, if the server given by its UUID is currently considered online.

=item B<address2uuid>(I<$value>)

Tries to resolve the given IP address or name I<$value> into a server UUID. B<undef> is returned, if
the address is unknown. If a name is given, it is tried to be resolved to an IP before DB-lookup.

=item B<uuidonline>(I<$value>)

Returns, if the given AFS-Fileserver identified by its UUID I<$value> has at least one accessible
interface. B<undef> is return, if the UUID is unknown.

=item B<addressonline>(I<$value>)

Returns, if the given address is currently accessible. Returns B<undef>, if the address is not associated
to an actual fileserver of the cell. If I<$value> is not an IP, the address will be resolved before.

=item \@list=B<uuidlist>()

Returns an array reference to the list of UUIDs of the cell's servers.

=back

=cut

sub serverInfo {
	my ($self,$type,$value)=@_;
	$self->ensureDB() || return undef;
	my $time=time();
	if ( ( $time - $self->{serverinfo}->{'infotime'} ) > 10*60 ) {
		# 
		my $query=$self->sqlquery("SELECT * FROM addresses","Unable to get address resolve list from server");
		if ( ! $query ) {
			return undef;
		}
		my %si=('infotime' => $time);
		$self->{serverinfo}=\%si;
		while ( my $row=$query->fetchrow_hashref() ) {
			if ( $row->{'viewname'} ) {
				$si{'byuuid'}->{$row->{'uuid'}}->{'viewname'}=$row->{'address'};
			} else {
				push(@{$si{'byuuid'}->{$row->{'uuid'}}->{'addresses'}},$row);
				if ( $row->{'io'} ) {
					push(@{$si{'byuuid'}->{$row->{'uuid'}}->{'io'}},$row);
					$si{'byuuid'}->{$row->{'uuid'}}->{'online'}=1;
				} elsif ( ! defined($si{'byuuid'}->{$row->{'uuid'}}->{'online'}) ) {
					$si{'byuuid'}->{$row->{'uuid'}}->{'online'}=0;
				}
				$si{'byaddr'}->{$row->{'address'}}=$row;
			}
		}
	}
	my $si=$self->{serverinfo};
	if ( $type eq 'uuid2name' ) {
		return $si->{'byuuid'}->{$value}->{'viewname'};
	} elsif ( $type eq 'uuid2ioaddresses' ) {
		return [sort map { $_->{'address'} } @{$si->{'byuuid'}->{$value}->{'io'}}];
	} elsif ( $type eq 'uuid2addresses' ) {
		return [sort map { $_->{'address'} } @{$si->{'byuuid'}->{$value}->{'addresses'}}];
	} elsif ( $type eq 'uuid2ioaddress' ) {
		my @addr=map { $_->{'address'} } @{$si->{'byuuid'}->{$value}->{'io'}};
		if ( @addr ) {
			return $addr[0];
		} else {
			return undef;
		}
	} elsif ( $type eq 'online' ) {
		my @addr=map { $_->{'address'} } @{$si->{'byuuid'}->{$value}->{'io'}};
		if ( @addr ) {
			return 1;
		} else {
			return 0;
		}
	} elsif ( $type eq 'address2uuid' ) {
		my $ip=$value;
		if ( $value !~ /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ ) {
			$ip=getFirstIP($value);
			if ( ! defined($ip) ) {
				return undef;
			}
		}
		return $si->{'byaddr'}->{$ip}->{'uuid'};
	} elsif ( $type eq 'uuidonline' ) {
		return $si->{'byuuid'}->{$value}->{'online'};
	} elsif ( $type eq 'addressonline' ) {
		my $ip=$value;
		if ( $value !~ /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ ) {
			$ip=getFirstIP($value);
			if ( ! defined($ip) ) {
				return undef;
			}
		}
		return $si->{'byaddr'}->{$ip}->{'io'};
	} elsif ( $type eq 'uuidlist' ) {
		return [sort keys %{$si->{'byuuid'}}];
	} elsif ( $type eq 'dump' ) {
		return $si;
	} else {
		out('e',"serverInfo(): Invalid request '$type'");
		return undef;
	}
}

=item \%volumes=$celldb->listVolumes(%options)

Returns a list of volumes of the cell. Constraints given via I<%options> may be applied:

=over 4

=item B<pattern> = I<PCRE-Pattern>

A PCRE-expression to filter the volume list. PCRE-Patterns are more powerful but slower

=item B<prefix> = I<prefix>

A prefix string, returned volume names must satisfy.

=item B<reference>

Set the data source of the volume. Possible Values are:

=over 4

=item B<vldb> Volume list of the VLDB
=item B<servers> List of volume names of volume instances on fileservers
=item B<onlineservers> List only volumes names, whose RWs are on servers that are online (not implemented, yet)
=item B<all> List all volume from VLDB and from servers

=back

=back

B<undef> is returned, in case of any errors.

=cut

sub listVolumes {
	my ($self,%options)=@_;
	my $reference='all';
	if ( $options{'reference'} ) {
		$reference=$options{'reference'};
	}
	
	# Take prefix option into account
	my $prefixconstraint='';my $prefixconstraintadd='';
	if ( defined($options{prefix}) ) {
		$prefixconstraint=" WHERE volumename LIKE " . $self->{db}->quote($options{prefix} . '%');
		$prefixconstraintadd=" AND volumename LIKE " . $self->{db}->quote($options{prefix} .'%');
	} elsif ( defined($options{pattern}) ) {
		# This mode is not testable with SQLITE
		$prefixconstraint=" WHERE volumename RLIKE " . $self->{db}->quote($options{pattern});
		$prefixconstraintadd=" AND volumename RLIKE " . $self->{db}->quote($options{pattern});
	}
	my $querycmd;
	if ( $reference eq 'vldb' ) {
		$querycmd="SELECT volumename FROM volumes$prefixconstraint";
	} elsif ( $reference eq 'servers' ) {
		$querycmd="SELECT DISTINCT volumename FROM instances$prefixconstraint";
	} elsif ( $reference eq 'onlineservers' ) {
		$querycmd="SELECT DISTINCT volumename FROM instances WHERE instancetype = 0$prefixconstraintadd AND serveruuid IN (SELECT uuid FROM addresses WHERE io = 1)";
	} elsif ( $reference eq 'all' ) {
		$querycmd="SELECT volumename FROM volumes$prefixconstraint UNION SELECT DISTINCT volumename FROM instances$prefixconstraint";
	} else {
		out('e',"listVolumes(): invalid reference ID '$reference'");
		return undef;
	}
	my $query=$self->sqlquery($querycmd,"Unable to query for volumes");
	if ( ! $query ) {
		out('e',"Unable to get list of volumes of the cell");
		return undef;
	}
	my %volumes=();
	while ( my @row=$query->fetchrow_array() ) {
		$volumes{$row[0]}=1;
	}
	return \%volumes;
}

=item $result=$celldb->isServerUuid($name)

Returns true, if the given name I<$name> looks like an AFS Fileserver UUID

=cut

sub isServerUuid {
	my ($self,$name)=@_;
	if ( $name =~ /^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{2}-[a-f0-9]{2}-[a-f0-9]{12}$/ ) {
		return 1;
	} else {
		return 0;
	}
}

=item $result=$celldb->expandLocationPattern($pattern,%options)

Returns locations ( {'serveruuid' => $server,'partition' => $partitionname} ) that are
matching the given location pattern I<$pattern>. This includes
   * someserver/* (any partition on given server)
   * */a all partitions labeled 'a' on all servers.
   * location group names as defined in cell.xml.

I<$result> is B<undef> on any error and an array reference to the location hashes. Certain
I<%options> may be given:

=over 4

=item B<format>

Default format is B<arrayref>. B<hashtree> may be chosen to get a hash tree like this:

 $result={'server1' => {'a' => 1, 'b' => 1},'server2' => {'c' => 1}}
 
=item B<hashtree>

By default, the hashtree contains just the locations of the given pattern. Using this option, a former tree
may be given, locations returned by the expression should be merged into.

The tree is then returned as I<$result>

=item B<activeonly>

Only return locations that are currently online.

=back

=cut

sub expandLocationPattern {
	my ($self,$pattern,%options)=@_;
	my $format='hashtree';
	if ( $options{'format'} ) {
		$format=$options{'format'};
	}
	if ( ! $options{'resolvedepth'} ) {
		$options{'resolvedepth'}=1;
	}
	if ( $options{'resolvedepth'} > 5 ) {
		out('e',"expandLocationPattern(): Resolve depth exceeded.");
		return undef;
	}
	my $activeonly=0;
	if ( $options{'activeonly'} ) {
		$activeonly=1;
	}
	$self->ensureDB() || return undef;
	if ( $pattern =~ /,/ ) {
		my @patterns=split /,/,$pattern;
		my $locations;
		if ( $options{'hashtree'} ) {
			$locations=$options{'hashtree'};
		}
		for my $p (@patterns) {
			my $result=$self->expandLocationPattern($p,
				format => 'hashtree',
				hashtree => $locations,
				activeonly => $activeonly,
				resolvedepth => $options{'resolvedepth'}
			);
			if ( ! defined($result) ) {
				return undef;
			}
		}
		if ( $format eq 'arrayref' ) {
			my @locations;
			for my $uuid (sort keys %{$locations}) {
				for my $partition (sort keys %{$locations ->{$uuid}}) {
					push(@locations,$locations->{$uuid}->{$partition});
				}
			}
			return \@locations;
		} else {
			return $locations;
		}
	}
	my $query;
	if ( $pattern =~ /^([^\/]+)\/(.*)$/ ) {
		my ($server,$partition)=($1,$2);
		my $cond='';
		my $byaddr=0;
		if ( $server ne '*' ) {
			if ( $self->isServerUuid($server) ) {
				$cond="AND partitions.serveruuid=" . $self->{db}->quote($server);
			} else {
				$byaddr=1;
				my $serverip;
				if ( $self->{dnslookups} ) {
					$serverip=getFirstIP($server);
				} else {
					$serverip=$server;
				}
				if ( ! defined($serverip) ) {
					out('e',"expandLocationPattern(): Unable to resolve IP of '$server'");
					return undef;
				}
				$cond="AND addresses.address=" . $self->{db}->quote($serverip);
			}
		}
		if ( $partition ne '*' ) {
			$cond .= " AND partitions.`partition`=" . $self->{db}->quote($partition);
		}
		my @fields=map { 'partitions.' . $_ } ('partsize','partfree','quotasum','quotasumrw','sizesum','sizesumrw','filessum','filessumrw','partinstances','partinstancesrw','partvolumes','ponline','scanlength','maxmtimerw','infotime','serveruuid','partition');
		if ( $byaddr ) {
			my $querystring=
				"SELECT " . join(', ',(@fields,'a2.address')) .
				" FROM addresses" .
				" INNER JOIN partitions" .
				" ON addresses.uuid = partitions.serveruuid" .
				" LEFT JOIN addresses AS a2" .
				" ON addresses.uuid = a2.uuid" .
				" WHERE a2.viewname = 1 $cond";
			$query=$self->sqlquery($querystring,"Unable to query database for locations") || return undef;
		} else {
			my $querystring=
				"SELECT " . join(', ',(@fields,'addresses.address')) .
				" FROM partitions" .
				" INNER JOIN addresses" .
				" ON addresses.uuid = partitions.serveruuid" .
				" WHERE addresses.viewname = 1 $cond";
			$query=$self->sqlquery($querystring,"Unable to query database for locations") || return undef;
		}
		my $locations;
		if ( $options{'hashtree'} ) {
			$locations=$options{'hashtree'};
		}
		my $dbresults=0;
		while ( my $row=$query->fetchrow_hashref() ) {
			$dbresults++;
			my $add=1;
			if ( $activeonly ) {
				if ( ! $self->partitionInfo('online',$row) ) {
					$add=0;
				}
			}
			if ( ! defined($row->{'partition'}) ) {
				$add=0;
			}
			if ( $add ) {
				$locations->{$row->{'serveruuid'}}->{$row->{'partition'}}=$row;
			}
		}
		if ( ! $dbresults ) {
			out('d',"expandLocationPattern(): No results for pattern '$pattern'");
		}
		if ( $format eq 'hashtree' ) {
			return $locations;
		} elsif ( $format eq 'arrayref' ) {
			my @locations;
			for my $uuid (sort keys %{$locations}) {
				for my $partition (sort keys %{$locations->{$uuid}}) {
					push(@locations,$locations->{$uuid}->{$partition});
				}
			}
			return \@locations
		} else {
			out('e',"expandLocationPattern(): Invalid format '$format'");
			return undef;
		}
	} else {
		# Test, if it's a locationgroup.
		return $self->getLocationGroup($pattern,%options);
	}
	return undef;
}

=item $result=getLocationGroup($name,%options)

Tries to resolve a location group, returning the locations matched.

I<$result> is B<undef> on any error and an array reference to the location hashes.
I<$options{'format'}> may be set to B<'hashtree'> to return a tree like this:

 $result={'server1' => {'a' => 1, 'b' => 1},'server2' => {'c' => 1}}

Otherwise $result is an array reference to hashes containing the B<serveruuid> and
the B<partition> of one location each.

Possible I<%options> are:

=over 4

=item B<resolvedepth>

The current depth of location group resolving. Since groups/expressions can be nested, this
field was implement as a security measure to prevent "infinite" resolving.

=item B<activeonly>

Return online partitions only.

=back

=cut

sub getLocationGroup {
	my ($self,$name,%options)=@_;
	if ( ! $self->ensureCellinfo() ) {
		return undef;
	}
	out('d',"Looking for locationgroup '$name'");
	my $activeonly=0;
	if ( $options{'activeonly'} ) {
		$activeonly=1;
	}
	if ( ! $options{'resolvedepth'} ) {
		$options{'resolvedepth'}=1;
	}
	my $lg_=$self->{cellinfo}->{'xc'}->find("//ci:locationgroup[\@name='$name']");
	if ( ! $lg_ ) {
		# This is not a location group name
		return undef;
	}
	my $locations={};
	if ( $options{'hashtree'} ) {
		$locations=$options{'hashtree'};
	}
	$lg_=$lg_->[0];
	for my $location (@{$lg_->getElementsByTagNameNS($xmlnsci,'location')}) {
		my $host=$location->getAttribute('server');
		my $partition=$location->getAttribute('partition');
		out('d',"Resolving location '$host/$partition' from locationgroup '$name'");
		my $lresult=$self->expandLocationPattern(
			"$host/$partition",
			'activeonly' => $activeonly,
			'format' => 'hashtree',
			'hashtree' => $locations,
			'resolvedepth' => $options{'resolvedepth'} + 1
		);
		if ( ! defined($lresult) ) {
			out('v',"getLocationGroup(): Pattern '$host/$partition' in locationgroup '$name' didn't return any results");
		}
	}
	my $format='hashtree';
	if ( $options{'format'} ) {
		$format=$options{'format'};
	}
	if ( $format eq 'hashtree' ) {
		return $locations;
	} elsif ( $format eq 'arrayref' ) {
		my @locations=();
		for my $uuid (sort keys %{$locations}) {
			for my $partition (sort keys %{$locations->{$uuid}}) {
				push(@locations,$locations->{$uuid}->{$partition});
			}
		}
		return \@locations
	} else {
		out('e',"getLocationGroup(): Invalid format '$format'");
		return undef;
	}
}

##### To be removed in favor of AFS::Instant::Toolz::checkXmlAuthorisation()

sub checkXmlAuthorisation {
	my ($self,$k5user,$rule)=@_;
	my $k4user=$k5user;
	$k4user =~ s/\//\./;
	for my $principal ( @{$rule->getElementsByTagName('principal')} ) {
		if ( $principal->hasAttribute('name') ) {
			my @names=split /\s+/,$principal->getAttribute('name');
			for my $name (@names) {
				if ( $name eq $k5user ) {
					return 1;
				}
			}
		}
		if ( $principal->hasAttribute('pattern') ) {
			my $pattern=$principal->getAttribute('pattern');
			if ( $k5user =~ /^$pattern$/ ) {
				return 1;
			}
		}
	}
	my $users_groups;
	for my $group ( @{$rule->getElementsByTagName('group')} ) {
		if ( $group->hasAttribute('name') ) {
			if ( ! defined($users_groups) ) {
				$users_groups=$self->userInfo('memberships+',$k4user);
			}
			my @names=split /\s+/,$group->getAttribute('name');
			for my $name (@names) {
				if ( $users_groups->{$name} ) {
					return 1;
				}
			}
		}
		if ( $group->hasAttribute('pattern') ) {
			out('i',"Group membership patterns not supported, yet");
		}
	}
	return 0;
}


=item \%location=$celldb->selectLocation($lgroup,\@rules,%options);

Selects a location from a location group or pattern according to a given set
of rules.  The first rule that can be satisfied wins.

I<%location> is the selected afs-partition-record or I<\%location> ==
B<undef> if none was selectable.

I<@rules>=is an array hashref containg one rule per hashref.
Each hash contains at least a 'type' and a 'value' key/value pair.
Possible types are:

=over 4

=item B<mostrw>

Select the partition containing most of the rw instances matched by the PCRE
in the value field.

=item B<mostro>

Select the partition containing most of the ro instances matched by the PCRE
in the value field.

=item B<leastused>

Select the partition with the most free space.

=back

No matching rule will cause a random location to be selected.

Possible I<%options> are:

=over 4

=item B<activeonly>

Only select locations that are currently online.

=back

=cut

sub hashMax {
	my ($hashref,$filter)=@_;
	my %filter=();
	for my $f (@{$filter}) {
		$filter{$f}=1;
	}
	my $max=0;my $maxkey;
	for my $key (keys %{$hashref}) {
		if ( $hashref->{$key} > $max ) {
			if ( $filter{$key} ) {
				$max=$hashref->{$key};
				$maxkey=$key;
			}
		}
	}
	return $maxkey;
}

sub partitionWinner {
	my ($self,$partitiontree,$min)=@_;
	my $first=1;
	my $currentwinner;my $currentvalue;
	for my $serveruuid (keys %{$partitiontree}) {
		for my $partition (keys %{$partitiontree->{$serveruuid}}) {
			my $value=$partitiontree->{$serveruuid}->{$partition};
			if ( $first ) {
				$currentwinner={'serveruuid' => $serveruuid,'partition' => $partition};
				$currentvalue=$value;
				$first=0;
			} else {
				if ( $min ) {
					if ( $value < $currentvalue ) {
						$currentwinner={'serveruuid' => $serveruuid,'partition' => $partition};
						$currentvalue=$value;
					}
				} else {
					if ( $value > $currentvalue ) {
						$currentwinner={'serveruuid' => $serveruuid,'partition' => $partition};
						$currentvalue=$value;
					}
				}
			}
		}
	}
	return $currentwinner;
}

sub selectLocation {
	my ($self,$pattern,$rules,%options)=@_;
	my $activeonly=0;
	if ( $options{'activeonly'} ) {
		$activeonly=1;
	}
	my $locations=$self->expandLocationPattern($pattern,'activeonly' => $activeonly,'format' => 'arrayref');
	if ( ! defined($locations) ) {
		return undef;
	}
	if ( ! @{$locations} ) {
		if ( ! $options{test} ) {
			out('w',"selectLocation(): No locations.");
		}
		return undef;
	}
	if ( scalar(@{$rules}) == 0 ) {
		out('dd',"selectLocation(): No rules, return first location.");
		return $locations->[0];
	}
	out('d',"selectLocation(): Processing rules: " . join(',',map { $_->{'type'} } @{$rules}));
	for my $rule (@{$rules}) {
		my $type=$rule->{'type'};
		if ( $type eq 'mostrw' ) {
			my %lgroup=();
			for my $location (@{$locations}) {
				my %options=('serveruuid' => $location->{'serveruuid'},'partition' => $location->{'partition'},'instancetype' => 'rw');
				if ( $rule->{'value'} eq 'pattern' ) {
					$options{'pattern'}=$rule->{'pattern'};
				}
				my $volumes=$self->listVolumes(%options);
				$lgroup{$location->{'serveruuid'}}->{$location->{'partition'}}=scalar(@{$volumes});
			}
			my $winner=$self->partitionWinner(\%lgroup,0);
			if ( defined($winner) ) {
				return $winner;
			}
		} elsif ( $type eq 'mostro' ) {
			my %lgroup=();
			for my $location (@{$locations}) {
				my %options=('serveruuid' => $location->{'serveruuid'},'partition' => $location->{'partition'},'instancetype' => 'readonly');
				if ( $rule->{'value'} eq 'pattern' ) {
					$options{'pattern'}=$rule->{'pattern'};
				}
				my $volumes=$self->listVolumes(%options);
				$lgroup{$location->{'serveruuid'}}->{$location->{'partition'}}=scalar(@{$volumes});
			}
			my $winner=$self->partitionWinner(\%lgroup,0);
			if ( defined($winner) ) {
				return $winner;
			}
		} elsif ( $type eq 'leastused' ) {
			my %lgroup=();
			for my $location (@{$locations}) {
				my $partrec=$self->getPartition($location->{'serveruuid'},$location->{'partition'});
				if ( $partrec ) {
					$lgroup{$location->{'serveruuid'}}->{$location->{'partition'}}=$partrec->{'partfree'};
				}
			}
			my $winner=$self->partitionWinner(\%lgroup,0);
			if ( defined($winner) ) {
				return $winner;
			}
		} elsif ( $type eq 'any' ) {
			return $locations->[0];
		}
	}
	return undef;
}

=item $result=$celldb->ensureSPOFData()

Makes sure, information about single Points of failure is read from cell.xml
to memory and available to calcInstanceRedundancy() .

I<$result> == 1, only if SPOF information are available now.

=cut

sub ensureSPOFData {
	my $self=shift;
	$self->ensureCellinfo() || return 0;
	if ( defined($self->{spof}->{infotime}) ) {
		if ( $self->{spof}->{infotime} >= $self->{cellinfo}->{mtime} ) {
			return 1;
		}
	}
	my $spofgroups_=$self->{cellinfo}->{xc}->find("//ci:spofgroup");
	my %groups=();
	for my $spofgroup_ (@{$spofgroups_}) {
		my @spoflocations=();
		for my $spoflocation_ (@{$self->{cellinfo}->{xc}->find("./ci:spoflocation",$spofgroup_)}) {
			my $tree={};
			for my $s (split /,/,$spoflocation_->getAttribute('locationgroup')) {
				$self->expandLocationPattern($s,'format' => 'hashtree','hashtree' => $tree);
			}
			push(@spoflocations,$tree);
		}
		$groups{$spofgroup_->getAttribute('name')}=\@spoflocations;
	}
	$self->{spof}={
		'infotime' => time(),
		'groups' => \%groups,
	};
	return 1;
}

=item $redundancy=$celldb->calcInstanceRedundancy($volumereference,%options)

Returns the number of independent RO instances of a given volume. "independent"
is determined by taking Single-Point-of-Failure-Information into account.

The result is *not* the redundancy but redundancy + 1.

Possible I<%options> are:

=over 4

=item B<includerw>

Usually, only RO instances are counted. This option set to B<1> will take RW
instances into account, too.

=item B<includerostate2>

This options defines, if RO-instances in RO-state 2 are taken into account, too.
See APPENDIX L</"RO States"> for information about the RO state 2.

=item B<spoffilter>

Optional hashref. Keys contain names of SPOF groups that should be included
in the calculation.  Due to the author's lack of algorithmic skills, only a
single key is allowed, yet.

=item B<includenovldb>

Take RO copies into account that are unknown to VLDB but exist physically.

=back

All instances outside of the celldb's SPOF group are always considered independent.

=cut

sub calcInstanceRedundancy {
	my ($self,$volref,%options)=@_;
	$self->ensureSPOFData() || return undef;
	my $includerw=0;
	if ( defined($options{'includerw'}) ) {
		$includerw=1;
	}
	my $includero2=0;
	if ( defined($options{'includerostate2'}) ) {
		$includero2=$options{'includerostate2'};
	}
	my $includenovldb=0;
	if ( defined($options{'includenovldb'}) ) {
		$includenovldb=$options{'includenovldb'};
	}

	
	my @spofgroups=keys %{$self->{spof}->{groups}};
	if ( defined($options{'spoffilter'}) ) {
		@spofgroups=keys %{$options{'spoffilter'}};
		if ( scalar(@spofgroups) > 1 ) {
			out('e',"calcInstanceRedundancy(): Only a single SPOF-Group can be given via 'spoffilter'");
			return undef;
		}
	} else {
		if ( @spofgroups > 1 ) {
			out('e',"calcInstanceRedundancy(): Only a single SPOF-Group is allowed at this time. Please adjust cell.xml");
			return undef;
		}
	}
	# Get the locations of the given volume
	my @vollocations=();
	my %ctree=();
	for my $ro (@{$volref->{'ro'}}) {
		my $inc=0;
		if ( defined($ro->{'rostate'}) ) {
			if ( $includero2 or ( $ro->{'rostate'} < 2 ) ) {
				$inc=1;
			}
		} elsif ( $includenovldb ) {
			$inc=1;
		}
		if ( $inc ) {
			$ctree{$ro->{'serveruuid'}}->{$ro->{'partition'}}=1;
			push(@vollocations,$ro);
		}
	}
	my $rw=$volref->{'rw'};
	if ( defined($rw) and $includerw ) {
		# AFS doesn't allow multiple ROs on the same server or even partition
		# we need to check the collision-tree only for RW-instances
		if ( ! $ctree{$rw->{'serveruuid'}}->{$rw->{'partition'}} ) {
			$ctree{$rw->{'serveruuid'}}->{$rw->{'partition'}}=1;
			push(@vollocations,$volref->{'rw'});
		}
	}

	if ( @spofgroups == 0 ) {
		return scalar(@vollocations);
	}
	my $group=$spofgroups[0];
	out('d',"SPOF group '$group'");
	my $spoflocations=$self->{spof}->{groups}->{$group};
	my %spoflocations_used=();
	my $outside=0;
	for my $location (@vollocations) {
		my $inside=0;
		for(my $n=0;$n<scalar(@{$spoflocations});$n++) {
			my $tree=$spoflocations->[$n];
			if ( $tree->{$location->{'serveruuid'}}->{$location->{'partition'}} ) {
				$spoflocations_used{$n}=1;
				$inside=1;
			}
		}
		if ( ! $inside ) {
			$outside++;
		}
	}
	my $inside=scalar(keys %spoflocations_used);
	out('d',"SPOF test of Volume '$volref->{'name'}': inside=$inside, outside=$outside");
	return ($inside+$outside);
}



=item \@instances=$celldb->getInstances(%options)

Returns all volume instances of the cell. Constraints may be (and usually
are) given via the I<%options> hash:

=over 4

=item B<serveruuid>

Only return instances residing on this server

=item B<partition>

Only return instances residing on a partition of the given name.

=item B<volume>

Only return instances of the given volume

=item B<instancetype>

Only return instances of the given type. See APPENDIX "Instance Types" for valid IDs.

=item B<pattern>

Only return instances of volumes matching the given pattern (slower than the volume-constraint).

=item B<reference>

##### Remove that.

=back

=cut

sub getInstances {
	my ($self,%options)=@_;
	my $cond='';my $cond2='';
	my $reference='servers';
	if ( $options{'reference'} ) {
		$reference=$options{'reference'};
	}
	if ( $options{'serveruuid'} or $options{'partition'} or $options{'instancetype'} or $options{'volume'} ) {
		my @cond=();
		if ( $options{'serveruuid'} ) {
			push(@cond,"serveruuid=" . $self->{db}->quote($options{'serveruuid'}));
		}
		if ( $options{'partition'} ) {
			push(@cond,"`partition`=" . $self->{db}->quote($options{'partition'}));
		}
		if ( $options{'volume'} ) {
			push(@cond,"volumename=" . $self->{db}->quote($options{'volume'}));
		}
		if ( $options{'instancetype'} ) {
			my $instancetype=$self->instanceTypeToId($options{'instancetype'});
			if ( ! defined($instancetype) ) {
				out('e',"getInstances(): Invalid instancetype '$options{'instancetype'}'");
				return undef;
			}
			push(@cond,"instancetype=$instancetype");
		}
		$cond="WHERE " . join(' AND ',map { "i.$_" } @cond);
		$cond2=join(' AND',map { "v.$_"} @cond);
		if ( $reference ne 'all' ) {
			$cond2 = "WHERE $cond2";
		}
		
	}
	my $query;
	if ( $reference eq 'vldb' ) {
		$query=$self->sqlquery(
			"SELECT v.volumename, v.instancetype, v.serveruuid, v.partition, v.rostate, i.size, i.quota, i.files, i.ctime, i.mtime, i.atime, i.id, i.online, i.needssalvage, i.isbusy, partitions.*" .
			" FROM vldbinstances v LEFT JOIN instances i ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" LEFT JOIN partitions ON v.serveruuid = partitions.serveruuid AND v.partition = partitions.`partition`" .
			" $cond2",
			"Unable to get volume instances (ref=vldb)");

	} elsif ( $reference eq 'servers' ) {
		$query=$self->sqlquery(
			"SELECT i.*, v.rostate, partitions.*" .
			" FROM instances i LEFT JOIN vldbinstances v ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" LEFT JOIN partitions ON i.serveruuid = partitions.serveruuid AND i.partition = partitions.`partition`" .
			" $cond",
			"Unable to get volume instances (ref=servers)");
	} elsif ( $reference eq 'all' ) {
		$query=$self->sqlquery(
			# Get all the instances plus their VLDB volume data
			"SELECT i.volumename, i.instancetype, i.serveruuid, i.partition, v.rostate, i.size, i.quota, i.files, i.ctime, i.mtime, i.atime, i.id, i.online, i.needssalvage, i.isbusy, p.partsize, p.partfree, p.infotime, p.quotasum, p.ponline, p.quotasumrw, p.sizesumrw" .
			" FROM instances i" .
			" LEFT JOIN vldbinstances v ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" INNER JOIN partitions p ON i.serveruuid = p.serveruuid AND i.partition = p.partition" .
			" $cond" .
			" UNION ALL " .
			# Make fake instances from VLDB data when there's no real volume instances
			"SELECT v.volumename, v.instancetype, v.serveruuid, v.partition, v.rostate, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, p.partsize, p.partfree, p.infotime, p.quotasum, p.ponline, p.quotasumrw, p.sizesumrw" .
			" FROM vldbinstances v" . 
			" LEFT JOIN instances i ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" LEFT JOIN partitions p ON v.serveruuid = p.serveruuid AND v.partition = p.partition" .
			" WHERE i.volumename IS NULL AND $cond2",
			"Unable to get volume instances (ref=all)");
	} else {
		out('e',"getInstances(): Invalid reference type '$options{'reference'}'");
		return undef;
	}
	if ( ! $query ) {
		return undef;
	}
	my @instances=();
	if ( $options{'pattern'} ) {
		my $p=$options{'pattern'};
		while ( my $row=$query->fetchrow_hashref() ) {
			if ( $row->{'volumename'} =~ /^$p$/ ) {
				push(@instances,$row);
			}
		}
	} else {
		while ( my $row=$query->fetchrow_hashref() ) {
			push(@instances,$row);
		}
	}
	return \@instances;
}

=item \%instance=$celldb->getInstance($instancename,%options)

Returns the instance hash of the volume instance specified by I<$instancename>. This might just
be the volumename or i.e. volume.readonly or volume.backup.

The full I<$instancename> syntax is I<volumename>.I<instancetype>@I<servername>/I<partition>.

which primary makes sense for RO-instances. I<servername> and I<partition> might be replaced
by the wildcard B<*>.

On error or the volume not existing, B<undef> is returned instead of a hashref.

=cut

sub getInstance {
	my ($self,$instancespec,%options)=@_;
	if ( $instancespec !~ /^([^\/]+?)(?:\.(readonly|backup))?(?:\@(\S+?)(?:\/(\S+))?)?$/ ) {
		out('e',"Invalid volume instance identifier '$instancespec'");
		out('i',"Syntax is [volume]{.[instancetype]}{\@[server]{/[partition]}}");
		return undef;
	}
	my $volume=$1;my $instancetype='rw';
	if ( defined($2) ) {
		$instancetype=$2;
	}
	my $server=$3;my $partition=$4;
	my %opt;
	if ( defined($server) && $server ne '*' ) {
		my $uuid;
		if ( $self->isServerUuid($server) ) {
			$uuid=$server;
		} else {
			$uuid=$self->serverInfo('address2uuid',$server);
			if ( ! defined($uuid) ) {
				out('e',"Unable to extract serveraddress from '$server'");
				return undef;
			}
		}
		if ( ! defined($uuid) ) {
			return undef;
		}
		$opt{serveruuid}=$uuid;
	}
	if ( defined($partition) && $partition ne '*' ) {
		$opt{partition}=$partition;
	}
	my $instances=$self->getInstances('volume' => $volume,'instancetype' => $instancetype,%options,%opt);
	if ( ! defined($instances) ) {
		return undef;
	}
	if ( @{$instances} == 0 ) {
		# no matching instance not found
		return undef;
	} elsif ( @{$instances} == 1 ) {
		return $instances->[0];
	} elsif ( $instancetype eq 'readonly' ) {
		out('e',"getInstance(): Volume '$volume' has multiple RO copies and location wasn't specific enought to extract exactly one");
		return undef;
	} else {
		out('e',"getInstance(): Multiple non-RO-instances found for volume '$volume', strange...");
		return undef;
	}
}

=item \@instances=$celldb->iterateInstances($volumeref,%options)

Returns all instances included in the volume hashref tree I<$volumeref> as a
array reference. No additional checks are done - the I<$volumeref> has to be
fetched by $celldb->getVolume() before. Sort order is always

=over 4

=item 1. RW

=item 2. RO

=item 3. Backup

=back

Possible I<%options> are:

=over 4

=item B<types>

Hashref of the instance types, that should be returned. Types are defined as keys
of the hash. Refer to the "Instance Types" APPENDIX for information about possible types.

=item B<vldbonly>

If set, only volume instances Backed in VLDB will be returned. This doesn't effect backup-instances - they will
always be returned.

=item B<realonly>

If set, only volume instances actually existing on fileservers will be returned.

=back

=cut

sub iterateInstances {
	my ($self,$volref,%options)=@_;
	my %instances=('rw' => 1, 'readonly' => 1, 'backup' => 1);
	if ( $options{'types'} ) {
		if ( ref($options{'types'}) eq 'ARRAY' ) {
			%instances=();
			for my $i (@{$options{'types'}}) {
				$instances{$i}=1;
			}
		} else {
			%instances=%{$options{'types'}};
		}
	}
	my %inst=();
	for my $i (keys %instances) {
		$inst{$self->instanceTypeToId($i)}=1;
	}
	my $vldbonly=0;
	if ( $options{vldbonly} ) {
		$vldbonly=1;
	}
	my $realonly=0;
	if ( $options{realonly} ) {
		$realonly=1;
	}
	my @result=();
	if ( $volref->{'rw'} and $inst{0} ) {
		if ( ! $vldbonly || defined($volref->{'rw'}->{'rostate'}) ) {
			push(@result,$volref->{'rw'});
		}
	}
	if ( $inst{1} ) {
		for my $ro (@{$volref->{'ro'}}) {
			if ( ! $vldbonly || defined($ro->{'rostate'}) ) {
				if ( ! $realonly || ! defined($ro->{rostate}) || $ro->{rostate} != 2 ) {
					push(@result,$ro);
				}
			}
		}
	}
	if ( $volref->{'backup'} and $inst{2} ) {
		push(@result,$volref->{'backup'});
	}
	return \@result;
}

=item \%volumeref=$celldb->getVolume($volumename,%options)

Returns a hashtree-reference to information about all a certain volume plus all instances
of it. Included information are described in APPENDIX "Volume Reference". Possible
I<%options> are:

=over 4

=item B<reference> = I<$reference>

I<$reference> has one of three values:

=over 4

=item B<vldb> Return only volumes that are registered in VLDB.

=item B<servers> Return only volume with actual physical instances on servers.

=item B<all> (default) Return all volumes - those found in VLDB and those only virtually existing via unregistered instances on fileservers.

=back

=item B<test>

If set to B<1>, a nonexisting volume is not considered an error condition.

=back

On error or invalid options, B<undef> is returned instead.

=cut

sub getVolume {
	my ($self,$volume,%options)=@_;
	$self->ensureDB() || return undef;
	my $time=time();
	if ( $self->{use_cache} ) {
		my $entry=$self->{volumecache}->get($volume);
		if ( defined($entry) ) {
			return $entry;
		}
	}
	my $volumeq=$self->{db}->quote($volume);
	my $reference='all';
	if ( defined($options{'reference'}) ) {
		$reference=$options{'reference'};
	}
	my $query;
	if ( $reference eq 'vldb' ) {
		$query=$self->sqlquery(
			"SELECT v.volumename, v.instancetype, v.serveruuid, v.partition, v.rostate, i.size, i.quota, i.files, i.ctime, i.mtime, i.atime, i.id, i.online, i.needssalvage, i.isbusy, partitions.*" .
			" FROM vldbinstances v LEFT JOIN instances i ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" LEFT JOIN partitions ON v.serveruuid = partitions.serveruuid AND v.partition = partitions.`partition`" .
			" WHERE v.volumename=$volumeq",
			"Unable to get volume instances (ref=vldb)");

	} elsif ( $reference eq 'servers' ) {
		$query=$self->sqlquery(
			"SELECT i.*, v.rostate, partitions.*" .
			" FROM instances i LEFT JOIN vldbinstances v ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" LEFT JOIN partitions ON i.serveruuid = partitions.serveruuid AND i.partition = partitions.`partition`" .
			" WHERE i.volumename=$volumeq",
			"Unable to get volume instances (ref=servers)");
	} elsif ( $reference eq 'all' ) {
		$query=$self->sqlquery(
			"SELECT i.volumename, i.instancetype, i.serveruuid, i.partition, v.rostate, i.size, i.quota, i.files, i.ctime, i.mtime, i.atime, i.id, i.online, i.needssalvage, i.isbusy, p.partsize, p.partfree, p.infotime, p.quotasum, p.ponline, p.quotasumrw, p.sizesumrw" .
			" FROM instances i" .
			" LEFT JOIN vldbinstances v ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" INNER JOIN partitions p ON i.serveruuid = p.serveruuid AND i.partition = p.partition" .
			" WHERE i.volumename=$volumeq" .
			" UNION ALL " .
			"SELECT v.volumename, v.instancetype, v.serveruuid as serveruuid, v.partition AS part, v.rostate, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, p.partsize, p.partfree, p.infotime, p.quotasum, p.ponline, p.quotasumrw, p.sizesumrw" .
			" FROM vldbinstances v" . 
			" LEFT JOIN instances i ON i.volumename = v.volumename AND i.instancetype = v.instancetype AND i.serveruuid = v.serveruuid" .
			" LEFT JOIN partitions p ON v.serveruuid = p.serveruuid AND v.partition = p.partition" .
			" WHERE v.volumename=$volumeq AND i.volumename IS NULL",
			"Unable to get volume instances (ref=all)");
	} else {
		out('e',"getVolume(): Invalid reference type '$reference'");
		return undef;
	}
	
	if ( ! $query ) {
		return 0;
	}
	my %v=('ro' => []);
	my $instances=0;
	while ( my $r=$query->fetchrow_hashref() ) {
		$r->{'cellname'}=$self->{cellname};
		if ( $r->{'instancetype'} == 0 ) {
			$v{'rw'}=$r;
			$v{'id'}=$r->{id};
		} elsif ( $r->{'instancetype'} == 1 ) {
			push(@{$v{'ro'}},$r);
		} elsif ( $r->{'instancetype'} == 2 ) {
			$v{'backup'}=$r;
		}
		$instances++;
	}
	if ( ! $instances ) {
		if ( ! $options{test} ) {
			out('d',"No instances of volume '$volume' in this cell");
		}
		if ( $self->{use_cache} ) {
			$self->{volumecache}->remove($volume);
		}
		return undef;
	}
	$query=$self->{db}->prepare("SELECT * FROM volumes WHERE volumename=$volumeq");
	if ( ! $query->execute() ) {
		return 0;
	}
	$v{'reference'}=$reference;
	my $r=$query->fetchrow_hashref();
	$v{cellname}=$self->{cellname};
	if ( defined($r) ) {
		$v{id}=$r->{rwid};
		$v{name}=$r->{volumename};
		for my $field ('roid','backupid','locked','synctry_ro','synctry_backup') {
			$v{$field}=$r->{$field};
		}
	} else {
		$v{name}="($volume)";
	}
	if ( $self->{use_cache} ) {
		$self->{volumecache}->set($volume,\%v);
	}
	return \%v;
}


=item \%partitionref=$celldb->getPartition($serveruuid,$partition)
 
Returns a hash reference to information about a partition specified by its
server's UUID I<$serveruuid> and the partition name I<$partition>.

On error, B<undef> is returned instead.

=cut

sub getPartition {
	my ($self,$serveruuid,$partition)=@_;
	my $serveruuidq=$self->{db}->quote($serveruuid);
	my $partitionq=$self->{db}->quote($partition);
	my $query=$self->sqlquery("SELECT * FROM partitions WHERE serveruuid=$serveruuidq AND `partition`=$partitionq","Unable to get partitionrec of '$serveruuid'/'$partition'");
	if ( ! $query ) {
		return undef;
	}
	my $row=$query->fetchrow_hashref();
	return $row;
}


=item $data=$celldb->getPartitions(%options);

Returns an array reference to all data partitions of the cell. Constraints
and options may be given via the I<%options> hash:

=over 4

=item B<serveruuid>

Only return partitions of this AFS-Fileserver

=item B<partition>

Only return partitions of this name

=item B<online>=I<boolean>

Only return partitions in the given online-state.

=item B<format>

Speficy format, the partitions are returned in. Default is B<arrayref>. The alternative
format B<hashtree> looks like this:

 $data={
   '00834802-5971-1b19-99-6f-0101007faa77' => {
      'a' => { ... }
      'b' => { ... }
   },
   '00809cce-b019-1dcb-b2-98-0101007faa77' => {
      'x' => { ... }
   }
 }

On error, undef is returned instead.

=back

=cut

sub getPartitions {
	my ($self,%options)=@_;
	my $cond='';
	my @cond=();
	if ( $options{'serveruuid'} ) {
		push(@cond,"serveruuid=" . $self->{db}->quote($options{'serveruuid'}));
	}
	if ( $options{'partition'} ) {
		push(@cond,"partitions=" . $self->{db}->quote($options{'partition'}));
	}
	if ( defined($options{'online'}) ) {
		my $online=int($options{'online'});
		push(@cond,"ponline=$online");
	}
	if ( @cond ) {
		$cond=" WHERE " . join(' AND ',@cond);
	}
	my $query=$self->{db}->prepare("SELECT * FROM partitions $cond");
	if ( ! $query->execute() ) {
		return undef;
	}
	my @results=();
	while ( my $row=$query->fetchrow_hashref() ) {
		push(@results,$row);
	}
	my $format='arrayref';
	if ( defined($options{'format'}) ) {
		$format=$options{'format'};
	}
	if ( $format eq 'arrayref' ) {
		return \@results;
	} elsif ( $format eq 'hashtree' ) {
		my %result;
		for my $r (@results) {
			$result{$r->{'serveruuid'}}->{$r->{'partition'}}=$r;
		}
		return \%result,
	} else {
		out('e',"getPartitions(): Invalid format '$format'");
		return undef;
	}
}

=item $data=$cellb->cellInfo($function)

Return statistical data about the cell. Available functions are:

=over 4

=item B<free>

Free space (in Bytes) on data partitions of the cell.

=item B<size>

Size of all data partitions of the cell (in Bytes).

=item B<rwsize>

Size of RW-Data of the cell (-> net data without redundancy or backups)

=item B<rwfiles>

Number of Inodes in RW volumes of the cell.

=back

=cut

sub cellInfo {
	my ($self,$function)=@_;
	my $query;
	if ( $function eq 'free' ) {
		$query=$self->{db}->prepare("SELECT SUM(partfree) FROM partitions");
		if ( ! $query->execute() ) {
			out('e',"Unable to read free space of the cell");
			return undef;
		}
		my $row=$query->fetchrow_array();
		return $row;
	} elsif ( $function eq 'size' ) {
		$query=$self->{db}->prepare("SELECT SUM(partsize) FROM partitions");
		if ( ! $query->execute() ) {
			out('e',"Unable to read storage information of the cell");
			return undef;
		}
		my $row=$query->fetchrow_array();
		return $row;
	} elsif ( $function eq 'rwsize' ) {
		$query=$self->{db}->prepare("SELECT SUM(size) FROM instances WHERE instancetype=0");
		if ( ! $query->execute() ) {
			out('e',"Unable to read storage information of the cell");
			return undef;
		}
		my $row=$query->fetchrow_array();
		return $row;
	} elsif ( $function eq 'rwfiles' ) {
		$query=$self->{db}->prepare("SELECT SUM(files) FROM instances WHERE instancetype=0");
		if ( ! $query->execute() ) {
			out('e',"Unable to read storage information of the cell");
			return undef;
		}
		my $row=$query->fetchrow_array();
		return $row;
	}
}

=item $result=$celldb->partitionInfo($function,\%partitionref)

Returns Information about a given partition. A I<%partitionref> is a hashref
containing at least two fields: B<serveruuid> and B<partition>.  Available
functions are:

=over 4

=item B<online>

Returns a boolean value, stating if the partition is currently accessible.

=item B<hr>

Returns a human readable form of the partition information ( "$servername/$partition" ).

=back

=cut


sub partitionInfo {
	my ($self,$function,$instance_or_location)=@_;
	if ( $function eq 'online' ) {
		my $serveruuidq=$self->{db}->quote($instance_or_location->{'serveruuid'});
		my $partitionq=$self->{db}->quote($instance_or_location->{'partition'});
		my $query=$self->{db}->prepare("SELECT infotime, ponline FROM partitions WHERE serveruuid=$serveruuidq AND `partition`=$partitionq");
		if ( ! $query->execute() ) {
			return undef;
		}
		my $row=$query->fetchrow_hashref();
		if ( defined($row) ) {
			if ( ! $row->{'ponline'} ) {
				return 0;
			}
			if ( ( time() - $row->{'infotime'} ) < $self->{onlinetimeout} ) {
				return 1;
			} else {
				return 0;
			}
		} else {
			out('d',"Partition $serveruuidq/$partitionq wasn't found in DB");
			return 0;
		}
	} elsif ( $function eq 'hr' ) {
		my $server=$self->serverInfo('uuid2name',$instance_or_location->{'serveruuid'});
		if ( defined($server) ) {
			return "$server/" . $instance_or_location->{'partition'};
		}
	} elsif ( $function eq 'hr/short' ) {
		my $server=$self->serverInfo('uuid2name',$instance_or_location->{'serveruuid'});
		if ( defined($server) ) {
			return substr($server,0,3) . "/" . $instance_or_location->{'partition'};
		}
	} else {
		out('e',"partitionInfo(): invalid function '$function'");
		return undef;
	}
}

=item $data=$celldb->instanceInfo($function,$instancereference,%options)

Returns information about an instance given via its I<$instancereference>. Available
functions are:

=over 4

=item B<accessible>, B<online> (deprecated)

Returns, if the instance is currently accessible via Fileserver-Access. This include checking:

=over 4

=item ... if the instance physically exists,
=item ... if there's a VLDB record of this instance. In case of a backup instance, this is not checked.
=item ... if the information about the host partition is not older than a a certain amount of time. The critical limit can be defined while new()ing the CellDB object.
=item ... if the volume is un-busy
=item ... if the volume is online

=back

=item B<storageleft>

Returns the amount of data that can be added in the instance. This is the
minimum of these two values:

=over 4

=item quota left on the instance
=item free space on the partition, the instance is located on.

=back

=item B<errors>

Returns a hashref of errors detected on the given volume instances. Possible errors are:

=over 4

=item B<offline>

The given instances is marked offline on the fileserver. This is not about
accessibility in general. A fileserver that just went offline will not
trigger this error for its hosted instances.

=item B<needssalvage>

The given instances needs salvaging.

=item B<partexist>

The partition, the volume instance refers to, doesn't exist. This is the
case, if a partition was removed after the instance record was create in
VLDB.

=back

=item B<exists>

Returns true, if the instance physically exists on the fileserver (not just in VLDB).

=item B<vldbexists>

Returns true, if the instance has a VLDB record. This test will always return true
in case of a backup instance.

=item B<hrname>

Returns unique name of the given instance.

=back

On error or an invalid method, B<undef> is returned.

=cut

sub instanceInfo {
	my ($self,$function,$instance,%options)=@_;
	if ( ( $function eq 'online' ) or ( $function eq 'accessible' ) ) {
		if ( $instance->{'instancetype'} != 2 ) {
			if ( ! defined($instance->{'rostate'}) ) {
				# Doesn't exist in VLDB ergo no accessible
				return 0;
			}
			if ( $instance->{'instancetype'} == 1 ) {
				if ( $instance->{'rostate'} == 2 ) {
					# Unreleased RO copies are not considered online
					return 0;
				}
			}
		}
		if ( ! defined($instance->{'infotime'}) ) {
			# No information about this partition available
			return 0;
		}
		if ( ! $instance->{'ponline'} ) {
			return 0;
		}
		if ( ( time() - $instance->{'infotime'} ) >= $self->{onlinetimeout} ) {
			# Information about this partition is too old
			return 0;
		}
		
		# Busy, to-be-salvaged or offline
		if ( $instance->{'isbusy'} or $instance->{'needssalvage'} or ! $instance->{'online'} ) {
			return 0;
		}
		return 1;
	} elsif ( $function eq 'errors' ) {
		my %result=();
		if ( defined($instance->{'online'}) && ! $instance->{'online'} ) {
			$result{'offline'}=1;
		}
		if ( $instance->{'needssalvage'} ) {
			$result{'needssalvage'}=1;
		}
		if ( ! defined($instance->{'infotime'}) ) {
			$result{'partexist'}=1;
		}
		return \%result;
	} elsif ( $function eq 'storageleft' ) {
		if ( ! defined($instance->{'quota'}) or ! defined($instance->{'size'}) ) {
			return undef;
		}
		my $quotaleft=$instance->{'quota'}-$instance->{'size'};
		if ( $quotaleft < 0 ) {
			$quotaleft=0;
		}
		if ( $quotaleft > $instance->{'partfree'} ) {
			return $instance->{'partfree'};
		} else {
			return $quotaleft;
		}
	} elsif ( $function eq 'vldbexists' ) {
		if ( $instance->{'instancetype'} == 2 ) {
			# Backup instances always exist in VLDB since we have no better definition
			# using vos
			return 1;
		}
		if ( defined($instance->{'rostate'}) ) {
			return 1;
		}
		return 0;
	} elsif ( $function eq 'exists' ) {
		if ( defined($instance->{'online'}) ) {
			return 1;
		}
		return 0;
	} elsif ( $function eq 'hrname' ) {
		my $name=$instance->{volumename};
		if ( $instance->{instancetype} == 1 ) {
			$name .= ".readonly";
		} elsif ( $instance->{instancetype} == 2 ) {
			$name .= ".backup";
		}
		$name .= "\@" . $self->serverInfo('uuid2name',$instance->{serveruuid}) . "/" . $instance->{partition};
		return $name;
	} else {
		out('e',"instanceInfo(): Invalid function '$function'");
		return undef;
	}
}

=item \($volumename,$instance)=$celldb->resolveVolumeId($id)

Tries to resolve a volume/instance-I<$id> to a I<$volumename> and an numerical
I<$instancetype>.

Returns both information or B<undef> in case of failure to resolve the
I<$id>.

=cut

sub resolveVolumeId {
	my ($self,$id)=@_;
	my $query=$self->sqlquery("SELECT volumename, rwid, roid, backupid FROM volumes WHERE rwid=$id OR roid=$id OR backupid=$id");
	if ( ! $query ) {
		# Error asking DB for ID
		return undef;
	}
	my $data=$query->fetchrow_hashref();
	if ( ! defined($data) ) {
		# ID not found
		return undef;
	}
	my $instancetype;
	if ( $data->{'rwid'} == $id ) {
		$instancetype=0;
	} elsif ( $data->{'roid'} == $id ) {
		$instancetype=1;
	} elsif ( $data->{'backupid'} == $id ) {
		$instancetype=2;
	}
	return [$data->{'volumename'},$instancetype];
}

=item $data=$celldb->volumeInfo($function,$volumeref_or_name,%options)

Returns information about the volume given via either a volume reference
or its name. Available Functions are:

=over 4

=item B<exists>

Returns, if the volume exists. This only makes sense, if a volume name is
given via I<$volumeref_or_name> . On inability to determine the existence,
B<undef> is returned.

=item B<rouptodate>

Returns B<1>, if *all* RO instances of the given volume are currently representing
an up-to-date snapshot of the given volume's RW-instance.

For volumes without RO instances, B<undef> is returned unless the option
B<nozero> is given in which case this value is returned.

To only test for sync necessity of VLDB registered instances, set I<$options{vldbonly}>
to something true.

=item B<hasoldro>

Returns true, if there are any RO instances of this volumes marked as either
"Old Release" or "Not Released".

=item B<backupuptodate>

Returns, if the Backup instance of the given volume is currently representing
an up-to-date snapshot of the given volume's RW-instance.

For volumes without Backup instances, undef is returned unless the option
B<nozero> is given in which case this value is returned.

=item B<roage>

Returns the most recent timestamp at which all the RO instances were in sync with the
RW. It takes sync-tries into account but not VLDB-only (non-released) ROs. Specific
I<%options> are:

=over 4

=item B<createsync>

Treat instances at RW-instance creating time as in-sync.

=back

=item B<backupage>

Returns the most recent timestamp at which the backup instance was in sync with the RW.
It takes sync-tries into account.


=item B<releasable>

Tests the "releasability" of the given volume. These checks are performed:
   * are all RO-sites online (including the ones with currently non-existing RO-clones on)?
   * does an RW exist?
   * is the RW-site online?
   * Test, if there's no RO on a different partition on the RW's server. This would cause
     an "invalid cross device link error".

This function returns > B<1> if all tests are passed. It doesn't check, if a
release is actually necessary.

=item B<backupable>

Returns B<1>, if the volume can be backup synced (=if the RW exists and is online)

=item B<hasrwro>

Returns B<1>, if there's an RO-copy on the RW's location (a cheap RO) - B<0> otherwise.
The B<vldb>-option set to B<1> makes this subfunction only return success, if the RO
is registered in VLDB.

=item B<hasrw>

Returns B<1>, if the volume has an RW instance. The B<vldb>-option set to
B<1> makes this subfunction only return success, if the RW is registered in
VLDB.

=item B<hasbackup>

Return B<1>, if the volume has a backup instance.

=item B<rwro>

Returns the RWRO (Cheap RO instance) of a volume as an instanceref. B<undef> is returned
in case there isn't any. The B<vldb>-option set to B<1> makes this subfunction only return
an RWRO, if it's registered in VLDB.

=item B<haslostinstances>

Returns true, if there are any non-VLDB-registered instances on the cell's fileservers.

=item B<partdiff>

Returns true, if there's a RO instance on the RW instance's server but on a different partition.

=item B<guesssyncsize>

Guesses the amount of data that has to be copied for a RO-release of this
volume. This can be used for an ETA guess or release processes.

=back

Possible I<%options> are:

=over 4

=item B<vldb>

Used by some subfunctions to restrict their results to volume instances that are
registered in VLDB.

=item B<onzero>

Defines the value returned in case no backup/ro instance exists and the B<backupuptodate>
or B<rouptodate> subfunction is called.

=back

=cut

sub volumeInfo {
	my ($self,$function,$volref,%options)=@_;

	# It's ok to just give the volume name...
	if ( ref($volref) eq '' ) {
		if ( $function eq 'exists' ) {
			my $volumeq=$self->sqlquote($volref);
			my $query=$self->sqlquery("SELECT volumename FROM volumes WHERE volumename=$volumeq");
			if ( ! $query ) {
				out('e',"Error testing for existense of volume '$volref'");
				return undef;
			}
			my $row=$query->fetchrow_arrayref();
			if ( $row ) {
				return 1;
			} else {
				return 0;
			}
		}
		my $volref_=$self->getVolume($volref);
		if ( ! defined($volref) ) {
			out('e',"Unable to read information about volume '$volref'");
			return undef;
		}
		$volref=$volref_;
	}
	if ( $function eq 'locked' ) {
		return $volref->{locked};
	} elsif ( $function eq 'rouptodate' ) {
		# This will test, if all the RO copies were synchronised after the last
		# write operation to the RW. Unreleased ROs are considered non-synced
		if ( ! @{$volref->{'ro'}} ) {
			# This volume doesn't have an RO
			if ( defined($options{'onzero'}) ) {
				return $options{'onzero'};
			}
			return 0;
		}
		if ( ! $volref->{'rw'}->{'mtime'} ) {
			# No RW-instance available. By definition this is *not* uptodate since there's no way to check it
			return 0;
		}
		for my $ro (@{$volref->{'ro'}}) {
			# VLDB-release state has to be "New release".
			if ( defined($ro->{rostate}) && $ro->{rostate} >= 1 ) {
				return 0;
			}
			if ( ! $options{vldbonly} || defined($ro->{rostate}) ) {
				# This will test orphan-instances (without VLDB-record), too
				if ( ! defined($ro->{'ctime'}) ) {
					# RO only exists in VLDB
					return 0;
				}
				if ( $ro->{'ctime'} < $volref->{'rw'}->{'mtime'} ) {
					return 0;
				}
			}
		}
		return 1;
	} elsif ( $function eq 'hasoldro' ) {
		for my $ro (@{$volref->{'ro'}}) {
			if ( defined($ro->{rostate}) && $ro->{rostate} >= 1 ) {
				return 1;
			}
		}
		return 0;
	} elsif ( $function eq 'haslostinstances' ) {
		for my $instance (@{$volref->{ro}}) {
			if ( ! defined($instance->{rostate}) ) {
				return 1;
			}
		}
		return 0;
	} elsif ( $function eq 'errors' ) {
		my %errors;
		my %roservers;
		my $rwserver;
		my $rwpartition;
		if ( $volref->{rw} ) {
			$rwserver=$volref->{rw}->{serveruuid};
			$rwpartition=$volref->{rw}->{partition};
		} else {
			$errors{'norw'}=1;
		}
		for my $ro (@{$volref->{'ro'}}) {
			if ( $roservers{$ro->{serveruuid}} ) {
				$errors{'multiple_ro'}=1;
			}
			$roservers{$ro->{serveruuid}}=1;
			if ( defined($rwserver) && $rwserver eq $ro->{serveruuid} && $rwpartition ne $ro->{partition} ) {
				$errors{'rw_crossdevice'}=1;
			}
		}
		return \%errors;
	} elsif ( $function eq 'partdiff' ) {
		if ( ! $volref->{rw} ) {
			return 0;
		}
		for my $ro (@{$volref->{ro}}) {
			if ( $ro->{serveruuid} eq $volref->{rw}->{serveruuid} ) {
				if ( $ro->{partition} ne $volref->{rw}->{partition} ) {
					return 1;
				}
			}
		}
		return 0;
	} elsif ( $function eq 'backupuptodate' ) {
		if ( ! $volref->{'backup'} ) {
			if ( defined($options{'onzero'}) ) {
				return $options{'onzero'};
			}
			return 0;
		}
		if ( ! $volref->{'rw'} ) {
			# No RW-instance available. By definition this is *not* uptodate since there's no way to check it
			return 0;
		}
		if ( $volref->{'backup'}->{'ctime'} < $volref->{'rw'}->{'mtime'} ) {
			return 0;
		}
		return 1;
	} elsif ( $function eq 'roage' ) {
		my $minage;
		if ( $self->volumeInfo('rouptodate',$volref) ) {
			return time();
		}
		for my $ro (@{$volref->{'ro'}}) {
			if ( defined($ro->{'rostate'}) ) {
				if ( $ro->{'rostate'} < 2 ) {
					if ( defined($minage) ) {
						if ( defined($ro->{'ctime'}) ) {
							if ( $ro->{'ctime'} < $minage ) {
								$minage=$ro->{'ctime'};
							}
						}
					} else {
						$minage=$ro->{'ctime'};
					}
				}
			}
		}
		if ( ! defined($minage) ) {
			if ( $options{'createsync'} and $volref->{'rw'}->{'ctime'} ) {
				$minage=$volref->{'rw'}->{'ctime'};
			} else {
				return undef;
			}
		}
		if ( $volref->{'synctry_ro'} > $minage ) {
			return $volref->{'synctry_ro'};
		}
		return $minage;
	} elsif ( $function eq 'backupage' ) {
		if ( ! $volref->{'backup'} ) {
			return undef;
		}
		if ( ! $volref->{'rw'} ) {
			return undef;
		}
		if ( $volref->{'backup'}->{'ctime'} > $volref->{'rw'}->{'mtime'} ) {
			return time();
		}
		if ( $volref->{'synctry_backup'} > $volref->{'backup'}->{'ctime'} ) {
			return $volref->{'synctry_backup'};
		} else {
			return $volref->{'backup'}->{'ctime'};
		}
	} elsif ( $function eq 'releasable' ) {
		my $foundro=0;
		if ( ! $volref->{'rw'} ) {
			# No RW exists
			return 0;
		}
		$self->instanceInfo('online',$volref->{'rw'}) || return 0;
		for my $ro (@{$volref->{'ro'}}) {
			if ( defined($ro->{rostate}) ) {
				$foundro=1;
				$self->partitionInfo('online',$ro) || return 0;
			}
		}
		return $foundro;
	} elsif ( $function eq 'backupable' ) {
		return  $self->instanceInfo('online',$volref->{'rw'});
	} elsif ( $function eq 'hasrwro' ) {
		my $ro=$self->volumeInfo('rwro',$volref,%options);
		return defined($ro);
	} elsif ( $function eq 'hasrw' ) {
		if ( defined($volref->{'rw'}->{'online'}) ) {
			if ( $options{'vldb'} ) {
				if ( ! defined($volref->{'rw'}->{'rostate'}) ) {
					return 0;
				}
			}
			return 1;
		} else {
			return 0;
		}
	} elsif ( $function eq 'rwro' ) {
		if ( ! defined($volref->{'rw'}->{'online'}) ) {
			# No RW means there's no RWRO-instance
			return undef;
		}
		for my $ro (@{$volref->{'ro'}}) {
			if ( ( $ro->{'serveruuid'} eq $volref->{'rw'}->{'serveruuid'} ) and ( $ro->{'partition'} eq $volref->{'rw'}->{'partition'} ) ) {
				if ( $options{'vldb'} ) {
					if ( defined($ro->{'rostate'}) ) {
						return $ro;
					} else {
						return undef;
					}
				} else {
					return $ro;
				}
			}
		}
		return undef;
	} elsif ( $function eq 'hasbackup' ) {
		if ( ! defined($volref->{'backup'}->{'online'}) ) {
			return 1;
		} else {
			return 0;
		}
	} elsif ( $function eq 'guesssyncsize' ) {
		my $rw=$volref->{rw};
		if ( ! defined($rw->{size}) ) {
			return undef;
		}
		my $remoteupdate=0;my $remotefull=0;my $rosize=0;
		my $guess=0;
		for my $ro (@{$volref->{ro}}) {
			if ( $ro->{serveruuid} ne $rw->{serveruuid} || $ro->{partition} ne $rw->{partition} ) {
				if ( defined($ro->{rostate}) && $ro->{rostate} == 0 ) {
					$remoteupdate++;
					if ( $rw->{size} > $ro->{size} ) {
						$guess += ($rw->{size}-$ro->{size});
					}
				} else {
					$guess += $rw->{size};
				}
			} else {
				$guess += 1;
			}
		}
		return $guess;
	} else {
		out('e',"volumeInfo(): Invalid function '$function'");
		return undef;
	}
}

=item $data=$celldb->volumeStat($function,\@volumenames,%options)

Returns statistical information about the given list of volumes I<@volumes>. These functions
are available:

=over 4

=item B<rwsize>

Returns an array ref containing
   * size of all rw instances
   * minimum size
   * maximum size

=item I<instancetype>I<timetype>

Returns min and maximum time values. For valid instancetypes see APPENDIX "Instance Types". Valid I<timetype>s are:
   * B<mtime> time of last change to data the instance
   * B<atime> time of last access to this instance
   * B<ctime> time, the instance was create (or updated in case of RO- or Backup-instances).

=back

=cut

sub volumeStat {
	my ($self,$function,$volumes,%options)=@_;
	if ( $function eq 'rwsize' ) {
		if ( @{$volumes} == 0 ) {
			return [0,0,0];
		}
		my $query=$self->{db}->prepare("SELECT sum(size), min(size), max(size) FROM `instances` WHERE instancetype=0 AND volumename IN (" . 
			join(',',map { $self->{db}->quote($_) } @{$volumes}) . ")");
		if ( ! $query->execute() ) {
			out('e',"Unable to determine rw size of given volumegroup via DB");
			return undef;
		}
		my @sizes=$query->fetchrow_array();
		return \@sizes;
	} elsif ( $function =~ /^(\S+)(mtime|ctime|atime)$/ ) {
		my ($voltype,$timetype)=($1,$2);
		my $typeid=$self->instanceTypeToId($voltype);
		out('d',"volumeStat(): doing statistics on volume of type '$typeid'");
		my $query=$self->{db}->prepare("SELECT min($timetype), max($timetype) FROM `instances` WHERE instancetype=$typeid AND volumename IN (" . 
			join(',',map { $self->{db}->quote($_) } @{$volumes}) . ")");
		if ( ! $query->execute() ) {
			out('e',"Unable to determine times of given volumegroup via DB");
			return undef;
		}
		my @sizes=$query->fetchrow_array();
		return \@sizes;
	} else {
		out('e',"volumeStat(): Invalid function '$function'");
		return undef;
	}
}

=item \%list=$celldb->listPtdb(%options)

Returns a hashref containing all PTDB objects as keys. Possible I<%options> are:

=over 4

=item B<type>

Defines an PTDB object type to limit the list scope. Possible Values are:

=over 4

=item B<users> Only list users

=item B<groups> Only list PTDB groups

=item B<all> List everything (default)

=back

=back


=cut

sub listPtdb {
	my ($self,%options)=@_;
	my $constraint='';
	if ( defined($options{'type'}) ) {
		my $type=$options{'type'};
		if ( $type eq 'users' ) {
			$constraint=' WHERE afsid > 0';
		} elsif ( $type eq 'groups' ) {
			$constraint=' WHERE afsid < 0';
		} elsif ( $type eq 'all' ) {
			#
		} else {
			out('e',"listPtdb(): Invalid type given: '$type'");
			return undef;
		}
	}
	my %result=();
	my $query=$self->sqlquery("SELECT name FROM afsids$constraint");
	while ( my @row=$query->fetchrow_array() ) {
		$result{$row[0]}=1;
	}
	return \%result;
}

=item $data=$celldb->ptdbInfo($function,$object,%options)

Returns information about the given PTDB object I<$object> (user or group). Supported functions are:

=over 4

=item B<id2name>

Returns the name belonging to the I<$object> which was given as AFSID.

=item B<name2id>

Returns the AFSID belonging to the I<$object> which was given as Name.

=item B<hashid2name>

Returns a hash based on I<%{$object}> with every key (which is assumed to be
a AFSID) replaced by the associated PTDB object name. Accepted I<%options> are:

=over 4

=item  B<include> (default: B<all>)

Defines, thich kind of objects are to be included in the hash. Possible values are:

=over 4

=item B<all>

Include users and groups.

=item B<groups>

Include groups only.

=item B<users>

Include users only.

=back

=back

=item B<ismemberof>

Returns B<1> if the given user or group is member of the given group. The group to test membership against has to be provided
as an I<%option> with the key B<group>. Example:

 $celldb->ptdbInfo('ismemberof','someuser','group' => 'somegroup');

Only direct membership counts.

=item B<ismember+of>

Returns B<1> if the given user or group is member of the given group or any nested group. The group to test membership against has to be provided
as an I<%option> with the key B<group>. Example:

 $celldb->ptdbInfo('ismember+of','someuser','group' => 'somegroup');

=item B<list>

Returns hashref with all AFSIDs of the cell's PTDB.

=back

=cut

sub ptdbInfo {
	my ($self,$function,$object,%options)=@_;
	my $query;
	if ( $function eq 'id2name' ) {
		$query=$self->sqlquery("SELECT name FROM afsids WHERE afsid=$object");
		if ( ! $query ) {
			out('e',"Unable to resolve object $object into a name");
			return undef;
		}
		my $row=$query->fetchrow_arrayref();
		if ( defined($row) ) {
			return $row->[0];
		}
		return undef;
	} elsif ( $function eq 'exists' ) {
		my $nameq=$self->sqlquote($object);
		$query=$self->sqlquery("SELECT afsid FROM afsids WHERE name=$nameq");
		if ( ! $query ) {
			out('e',"Unable to resolve object $object into an AFSID");
			return undef;
		}
		my $row=$query->fetchrow_arrayref();
		if ( defined($row) ) {
			return 1;
		}
		return 0;
	} elsif ( $function eq 'name2id' ) {
		my $nameq=$self->{db}->quote($object);
		$query=$self->sqlquery("SELECT afsid FROM afsids WHERE name=$nameq");
		if ( ! $query ) {
			out('e',"Unable to resolve object $object into an AFSID");
			return undef;
		}
		my $row=$query->fetchrow_arrayref();
		if ( defined($row) ) {
			return $row->[0];
		}
		return undef;
	} elsif ( $function eq 'hashid2name' ) {
		if ( defined($object) ) {
			my %results=();
			my @ids;
			if ( ! defined($options{'include'}) ) {
				@ids=keys %{$object};
			} elsif ( $options{'include'} eq 'all' ) {
				@ids=keys %{$object};
			} elsif ( $options{'include'} eq 'groups' ) {
				@ids=grep { $_ < 0 } keys %{$object};
			} elsif ( $options{'include'} eq 'users' ) {
				@ids=grep { $_ > 0 } keys %{$object};
			} else {
				out('e',"ptdbInfo(): Invalid value for 'include'-option");
				return undef;
			}
			for my $id (@ids) {
				my $name=$self->ptdbInfo('id2name',$id);
				$results{$name}=$object->{$id};
			}
			return \%results;
		} else {
			return undef;
		}
	} elsif ( $function eq 'ismemberof' ) {
		my $objectid=$self->ptdbInfo('name2id',$object);
		if ( ! $objectid ) {
			out('d',"PTDB object '$object' doesn't exist");
			return undef;
		}
		my $groupid=$self->groupInfo('name2id',$options{'group'});
		if ( ! $groupid ) {
			out('d',"PTDB group '$options{'group'}' doesn't exist");
			return undef;
		}
		$query=$self->sqlquery("SELECT afsgroup FROM memberships WHERE member=$objectid AND afsgroup=$groupid");
		if ( ! $query ) {
			return undef;
		}
		my $row=$query->fetchrow_arrayref();
		return defined($row);
	} elsif ( $function eq 'ismember+of' ) {
		my $objectid=$self->ptdbInfo('name2id',$object);
		my $groupid=$self->groupInfo('name2id',$options{'group'});
		my $members=$self->groupInfo('idmembers+',$groupid);
		if ( $members->{$objectid} ) {
			return 1;
		} else {
			return 0;
		}
	} elsif ( $function eq 'list' ) {
		$query=$self->{db}->prepare("SELECT name FROM afsids");
		if ( ! $query->execute() ) {
			out('e',"Unable to get list of AFS IDs from DB");
			return undef;
		}
		my %result;
		while ( my $row=$query->fetchrow_arrayref() ) {
			$result{$row->[0]}=1;
		}
		return \%result;
	} else {
		out('e',"ptdbInfo(): Invalid function '$function'");
		return undef;
	}
}

=item $data=$celldb->groupInfo($function,$group,%options)

groupInfo is a function family for getting information about PTDB groups I<$group>. Available functions are:

=over 4

=item B<name2id>

Convert given group name to its ID. B<undef> is returned in case of a group
not existing or the given name being assigned to an AFS user.

=item B<exists>

Returns true, if the given group exists, false if it doesn't and B<undef> if the assumption
of I<$group> being a group is wrong.

=item B<id2name>

Convert given group id to its name. B<undef> is returned  if the given id belongs to a user or the id doesn't exist.

=item B<members>

Returns hashref of names of users *and* groups being members of the given group.

=item B<members+>

Returns hashref of names of users being members of the given group. This includes users being members of nested
groups up to a level of 3. Values of the hashref are arrayrefs containing the ids of the (possibly nested) groups
being responsible fo the user being included.

=item B<idmembers>

Returns hashref of ids of users *and* groups being members of the given group.

=item B<idmembers+>

Returns hashref of ids of users being members of the given group. This includes users being members of nested
groups up to a level of 3. Values of the hashref are arrayrefs containing the id of the (possibly nested) groups
being responsible fo the user being included. I<$group> has to be given as ID.

=item B<memberships>

Returns a hashref of names of groups, this group is direct member of.

=item B<memberships+>

Returns the hashref of names of groups, this group is member of. This includes groups, it's member of by nested
grouping.

=item B<idmeberships>

Returns a hashref of ids of groups, this group is direct member of. I<$group> has to be given as ids.

=item B<idmemberships+>

Returns a hashref of ids of groups, this group is member of.This includes groups, its member of by nested
grouping. I<$group> has to be given as id. This function will *not* test, if the given group is actually
a group or even if it exists.

=item B<list>

Returns list of all group names as hashref.

=back

=cut

sub groupInfo {
	my ($self,$function,$group,%options)=@_;
	my $query;
	if ( $function eq 'name2id' ) {
		my $id=$self->ptdbInfo('name2id',$group);
		if ( ! defined($id) ) {
			return undef;
		}
		if ( $id > 0 ) {
			out('d',"groupInfo(): AFSIDs > 0 are not group ids");
			return undef;
		}
		return $id;
	} elsif ( $function eq 'exists' ) {
		my $id=$self->ptdbInfo('name2id',$group);
		if ( ! defined($id) ) {
			return 0;
		}
		if ( $id > 0 ) {
			return undef;
		}
		return 1;
	} elsif ( $function eq 'id2name' ) {
		if ( $group > 0 ) {
			out('d',"groupInfo(): AFSIDs > 0 are not group ids");
			return undef;
		}
		return $self->ptdbInfo('id2name',$group);
	} elsif ( $function eq 'members' ) {
		my $id=$self->groupInfo('name2id',$group);
		if ( ! defined($id) ) {
			return undef;
		}
		$query=$self->sqlquery("SELECT member FROM memberships WHERE afsgroup=$id");
		my %members=();
		if ( ! $query ) {
			out('e',"Unable to read members of group '$group'");
			return undef;
		}
		while ( my $row=$query->fetchrow_arrayref() ) {
			$members{$row->[0]}=1;
		}
		return $self->ptdbInfo('hashid2name',\%members,%options);
	} elsif ( $function eq 'members+' ) {
		my $uid=$self->groupInfo('name2id',$group);
		if ( ! defined($uid) ) {
			return undef;
		}
		return $self->ptdbInfo('hashid2name',$self->groupInfo('idmembers+',$uid,%options));
	} elsif ( $function eq 'idmembers' ) {
		my $id=$self->groupInfo('name2id',$group);
		if ( ! defined($id) ) {
			return undef;
		}
		my $groupnameq=$self->{db}->quote($group);
		$query=$self->sqlquery("SELECT member FROM memberships WHERE afsid=$id");
		my %members=();
		if ( ! $query ) {
			out('e',"Unable to read members of group '$group'");
			return undef;
		}
		while ( my $row=$query->fetchrow_arrayref() ) {
			$members{$row->[0]}=1;
		}
		return \%members;
	} elsif ( $function eq 'memberships' ) {
		my $uid=$self->groupInfo('name2id',$group);
		if ( ! defined($uid) ) {
			return undef;
		}
		return $self->ptdbInfo('hashid2name',$self->groupInfo('idmemberships',$uid,%options));
	} elsif ( $function eq 'idmemberships' ) {
		my $query=$self->{db}->prepare("SELECT afsgroup FROM memberships WHERE member=$group");
		if ( ! $query->execute() ) {
			out('e',"Unable to read membership information for '$group' from DB");
			return undef;
		}
		my %groups;
		while ( my $row=$query->fetchrow_arrayref() ) {
			$groups{$row->[0]}=[];
		}
		return \%groups;
	} elsif ( $function eq 'memberships+' ) {
		my $uid=$self->groupInfo('name2id',$group);
		if ( ! defined($uid) ) {
			return undef;
		}
		return $self->ptdbInfo('hashid2name',$self->groupInfo('idmemberships+',$uid,%options));
	} elsif ( $function eq 'idmemberships+' ) {
		if ( ! $options{'resolvedepth'} ) {
			$options{'resolvedepth'}=1;
		}
		if ( $options{'resolvedepth'} > 5 ) {
			out('e',"Group nesting too deep");
			return undef;
		}
		my $query=$self->sqlquery("SELECT afsgroup FROM memberships WHERE member=$group");
		if ( ! $query ) {
			out('e',"Unable to read membership information for '$group' from DB");
			return undef;
		}
		my %groups;
		while ( my $row=$query->fetchrow_arrayref() ) {
			$groups{$row->[0]}=[];
		}
		for my $group (keys %groups) {
			my $lresult=$self->groupInfo('idmemberships+',$group,%options);
			if ( ! defined($lresult) ) {
				return undef;
			}
			# %groups=(%groups,%{$lresult});
			for my $item (keys %{$lresult}) {
				if ( ! exists $groups{$item} ) {
					$groups{$item}=[$group];
				}
			}
		}
		return \%groups;
	} elsif ( $function eq 'idmembers+' ) {
		if ( ! $options{'resolvedepth'} ) {
			$options{'resolvedepth'}=1;
		}
		if ( $options{'resolvedepth'} > 5 ) {
			out('e',"Group nesting too deep");
		}
		my $groupq=$self->{db}->quote($group);
		my $query=$self->sqlquery("SELECT member FROM memberships WHERE afsgroup=$groupq");
		if ( ! $query ) {
			out('e',"Unable to read members for group '$group' from DB");
			return undef;
		}
		my %groups;
		while ( my $row=$query->fetchrow_arrayref() ) {
			$groups{$row->[0]}=[];
		}
		for my $group (keys %groups) {
			my $lresult=$self->groupInfo('idmembers+',$group,%options);
			if ( ! defined($lresult) ) {
				return undef;
			}
			for my $item (keys %{$lresult}) {
				if ( ! exists $groups{$item} ) {
					$groups{$item}=[$group];
				}
			}
		}
		return \%groups;
	} elsif ( $function eq 'list' ) {
		$query=$self->sqlquery("SELECT name,afsid FROM afsids WHERE afsid<0");
		if ( ! $query ) {
			out('e',"Unable to get list of AFS groups from DB");
			return undef;
		}
		my %result;
		while ( my $row=$query->fetchrow_arrayref() ) {
			$result{$row->[0]}=$row->[1];
		}
		return \%result;
	} else {
		out('e',"groupInfo(): Invalid function '$function'");
		return undef;
	}
}


=item $data=$celldb->userInfo($function,$user,%options)

userInfo is a function family for getting information about the PTDB user I<$user>. Available functions are:

=over 4

=item B<name2id>

Convert given AFS user name to its ID. B<undef> is returned in case of the user not existing or the given name being assigned to an AFS group.

=item B<exists>

Returns true, if the given user exists, false if it doesn't and B<undef> if the assumption
of I<$user> being a user is wrong.

=item B<id2name>

Convert given user ID to its name. B<undef> is returned  if the given ID belongs to a group or the id doesn't exist.

=item B<memberships>

Returns a hashref of names of groups, this user is *direct* member of.

=item B<memberships+>

Returns the hashref of names of groups, the user is member of. This includes groups, it's member of by nested
grouping.

=item B<idmemberships>

Returns a hashref of ids of groups, the user is direct member of. I<$user> has to be given as ids.

=item B<idmemberships+>

Returns a hashref of ids of groups, this user is member of. This includes groups, its member of by nested
grouping. I<$user> has to be given as id. This function will *not* test, if the given user is actually
a user or even if it exists.

=item B<ismemberof>

Returns B<1> if the given user is member of the given group. Group has to be given via the B<group> option
like this:

 $member=$celldb->userInfo('ismemberof','someuser','group' => 'somegroup')

=item B<list>

Returns list of all user names as hashref.

=item B<ischangeable>

Returns, if the given I<$user> is mentioned in the cell wide blacklist. I<$user> is interpreted as
Kerberos5-User string here.

=back

=cut

sub userInfo {
	my ($self,$function,$user,%options)=@_;
	my $query;
	if ( $function eq 'name2id' ) {
		my $id=$self->ptdbInfo('name2id',$user);
		if ( ! defined($id) ) {
			return undef;
		}
		if ( $id < 0 ) {
			out('d',"userInfo(): AFSIDs < 0 are not user ids");
			return undef;
		}
		return $id;
	} elsif ( $function eq 'exists' ) {
		my $id=$self->ptdbInfo('name2id',$user);
		if ( ! defined($id) ) {
			return 0;
		}
		if ( $id < 0 ) {
			return undef;
		}
		return 1;
	} elsif ( $function eq 'id2name' ) {
		if ( $user < 0 ) {
			out('d',"userInfo(): AFSIDs < 0 are not user ids");
			return undef;
		}
		return $self->ptdbInfo('id2name',$user);	
	} elsif ( $function eq 'memberships' ) {
		my $uid=$self->userInfo('name2id',$user);
		if ( ! defined($uid) ) {
			return undef;
		}
		return $self->ptdbInfo('hashid2name',$self->userInfo('idmemberships',$uid,%options));
	} elsif ( $function eq 'idmemberships' ) {
		return $self->groupInfo('idmemberships',$user,%options);
	} elsif ( $function eq 'memberships+' ) {
		my $uid=$self->userInfo('name2id',$user);
		if ( ! defined($uid) ) {
			return undef;
		}
		return $self->ptdbInfo('hashid2name',$self->userInfo('idmemberships+',$uid,%options));
	} elsif ( $function eq 'idmemberships+' ) {
		return $self->groupInfo('idmemberships+',$user,%options);
	} elsif ( $function eq 'ismemberof' ) {
		my $objectid=$self->userInfo('name2id',$user);
		if ( ! defined($objectid) ) {
			return undef;
		}
		my $groupid=$self->groupInfo('name2id',$options{'group'});
		if ( ! defined($groupid) ) {
			return undef;
		}
		$query=$self->sqlquery("SELECT afsgroup FROM memberships WHERE member=$objectid AND afsgroup=$groupid");
		if ( ! $query ) {
			return undef;
		}
		my $row=$query->fetchrow_arrayref();
		return defined($row);
	} elsif ( $function eq 'list' ) {
		$query=$self->sqlquery("SELECT name,afsid FROM afsids WHERE afsid>=0");
		if ( ! $query ) {
			out('e',"Unable to get list of PTDB users from DB");
			return undef;
		}
		my %result;
		while ( my $row=$query->fetchrow_arrayref() ) {
			$result{$row->[0]}=$row->[1];
		}
		return \%result;
	} elsif ( $function eq 'ischangeable' ) {
		$self->ensureCellinfo() || return undef;
		my $utrs_=$self->{cellinfo}->{xc}->find("//ci:untouchable");
		for my $utr_ (@{$utrs_}) {
			my $pattern=$utr_->getAttribute('pattern');
			if ( defined($pattern) ) {
				if ( $user =~ /$pattern/ ) {
					return 0;
				}
			}
			my $group=$utr_->getAttribute('group');
			if ( defined($group) ) {
				if ( $self->userInfo('ismemberof',k4name($user),group => $group) ) {
					return 0;
				}
			}
		}
		return 1;
	} else {
		out('e',"userInfo(): Invalid function '$function'");
		return undef;
	}
}

=item $id=$celldb->instanceTypeToId($type)

Converts Instance type names to instance type IDs. See APPENDIX "Instance Types" for more information.

=cut

sub instanceTypeToId {
	my ($self,$type)=@_;
	if ( ! defined($type) ) {
		return undef;
	}
	my %type2id=(
		'rw' => 0,
		'readonly' => 1,
		'ro' => 1,
		'backup' => 2,
		0 => 0,
		1 => 1,
		2 => 2,
	);
	return $type2id{$type}
}

=item $type=$celldb->instanceIdToType($id)

Converts an instance ID to the associated Type name. See APPENDIX "Instance Types" for more information.

=cut

sub instanceIdToType {
	my ($self,$id)=@_;
	if ( ! defined($id) ) {
		return undef;
	}
	my %id2type=(
		0 => 'rw',
		1 => 'readonly',
		2 => 'backup',
		'rw' => 'rw',
		'readonly' => 'readonly',
		'ro' => 'readonly',
		'backup' => 'backup',
	);
	return $id2type{$id};
}





=item $filternode=$celldb->parseVolumeFilter($expressions,%options)

Parses the given filter constraints, builds a filter tree to efficiently
test volumes for constraint matches.  This method is rather stupid and will
only create a two-branch-basenode (id 0, "AND") which branches further into
all pattern matches OR-tested tested on one side and all other constraints
AND-tested on the other side.  Sorry, I'm not that good in writing parsers
:-( .  Someone els might create a filter-tree with more complex behavior
(nested OR and AND statements, ...) - the matchVolumesComplete() function
will be able to work with it. I<$expressions> may be given in one of
three different forms:

=over 4

=item I<single string>

In this case the expression contains a list of \s+ -separated constraints.

=item I<arrayref>

contains the list of constraints pre-split.

=item I<hashref>

contains the list of constraints pre-split as keys of an hashref.

=back

Possible I<%options> are:

=over 4

=item B<comments>

Contains a hashref assigning comment strings to constraint strings.

=back

=cut

sub normalize_op {
	my ($op,$arz)=@_;
	my $neg=0;
	if ( $arz ) {
		$neg=1;
	}
	if ( $op eq '!=' ) {
		return (5,$neg);
	} elsif ( $op eq '<=' ) {
		return (-2,$neg);
	} elsif ( $op eq '>=' ) {
		return (2,$neg);
	} elsif ( $op eq '=' ) {
		return (0,$neg);
	} elsif ( $op eq '>' ) {
		return (1,$neg);
	} elsif ( $op eq '<' ) {
		return (-1,$neg);
	} else {
		out('e',"Invalid operator '$op'");
		return undef;
	}
}

sub parseVolumeFilter {
	our ($self,$expression,%options)=@_;
	my @expressions;
	if ( ref($expression) eq 'ARRAY' ) {
		@expressions=@{$expression};
	} elsif ( ref($expression) eq 'HASH' ) {
		# Sorting is not really necessary but reduces code paths in case of debugging
		@expressions=sort keys %{$expression};
	} else {
		@expressions=split /\s+/,$expression;
	}

	sub getInstanceTypes {
		my ($opt,$default)=@_;
		my %types;
		if ( defined($opt->{'itype'}) ) {
			for my $t (split /,/,$opt->{itype}) {
				if ( $t eq 'all' || $t eq 'any' ) {
					%types=(0 => 1,1 => 1,2 => 1);
				} else {
					my $id=$self->instanceTypeToId($t);
					if ( ! defined($id) ) {
						out('e',"Invalid instance type '$t' given.");
						return ();
					}
					$types{$id}=1;
				}
			}
		} else {
			%types=%{$default};
		}
		return \%types;
	}
	sub getMatch {
		my ($opt,$default)=@_;
		my $match;
		if ( $opt->{match} ) {
			if ( $opt->{match} eq 'all' ) {
				$match='all';
			} elsif ( $opt->{match} eq 'any' ) {
				$match='any';
			} else {
				out('e',"Invalid match mode '$opt->{match}' given in constraint");
				return undef;
			}
		} else {
			$match=$default;
		}
		return $match;
	}
	sub getVldb {
		my ($opt,$default)=@_;
		my $vldbmode;
		if ( $opt->{vmode} ) {
			if ( $opt->{vmode} eq '1' ||  $opt->{vmode} eq 'vldb' ) {
				$vldbmode=1;
			} elsif ( $opt->{vmode} eq '0' || $opt->{vmode} eq 'servers' ) {
				$vldbmode=0;
			} else {
				out('e',"Invalid vmode '$opt->{vmode}' given in location constraint");
				return undef;
			}
		} else {
			$vldbmode=$default;
		}
		return $vldbmode;
	}
	sub getOnzero {
		my ($opt,$default)=@_;
		my $onzero;
		if ( defined($opt->{onzero}) ) {
			if ( $opt->{onzero} ) {
				$onzero=1;
			} else {
				$onzero=0;
			}
		} else {
			$onzero=$default;
		}
		return $onzero;
	}
	sub getValue {
		my ($value)=@_;
		if ( $value eq 'all' ) {
			$value=-1;
		} elsif ( $value =~ /^[0-9]$/ ) {
			$value=int($value);
		} else {
			out('e',"Invalid numeric value '$value' given in constraint");
			return undef;
		}
		return $value;
	}
	my %volpatterns=('type' => 2,'sub' => [],'neg' => 0);
	my %constraints=('type' => 3,'sub' => [],'neg' => 0);
	my %filter=('type' => 3,'sub' => [],'neg' => 0);
	for my $e (@expressions) {
		my %c=('expression' => $e);
		if ( $options{'comments'}->{$e} ) {
			$c{'comment'}=$options{'comments'}->{$e};
		}
		if ( $e =~ /^(\!)?(.*?)(=|<=|>=|\!=|<|>)(.*)$/ ) {
			my ($arz,$field,$op_,$value)=($1,$2,$3,$4);
			my ($op,$neg)=normalize_op($op_,$arz);
			my %opt=();
			if ( $field ne 'pcre' && $field ne 'pcrename' && $value =~ /^([^:]+):(.*)$/ ) {
				$value=$1;
				for my $item (split /:/,$2) {
					if ( $item =~ /^(.*?)=(.*)$/ ) {
						$opt{$1}=$2;
					}
				}
			} else {
				$opt{$value}=1;
			}
			my $match;
			if ( $field eq 'pcre' || $field eq 'pcrename' ) {
				if ( $options{'nonames'} ) {
					out('e',"No volumes name selection allowed in this filter");
					return undef;
				}
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowed in pcrename constraint");
					return undef;
				}
				$match = $value;
			} elsif ( $field eq 'globname' ) {
				if ( $options{'nonames'} ) {
					out('e',"No volumes name selection allowed in this filter");
					return undef;
				}
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowd in globber expression");
					return undef;
				}
				$match=$value;
				$match =~ s/\.(?:backup|readonly)$//;
				$match =~ s/\./\\./g;
				$match =~ s/\?/\./g;
				$match =~ s/\*/\.\*/g;
			}
			if ( $match ) {
				# ignore everything except 'globname' and 'pcrename'
				push(@{$volpatterns{'sub'}},{'type' => 10,'value' => qr/^$match$/,'neg' => $neg,%c});
			} elsif ( $field eq 'location' || $field eq 'l' ) {
				my $vldb=0;
				if ( defined($1) ne '' ) {
					$vldb=1;
				}
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowed in location match");
					return undef;
				}
				my $types=getInstanceTypes(\%opt,{0 => 1,1 => 1,2 => 1}) || return undef;
				my $match=getMatch(\%opt,'any');
				my $vldbmode=getVldb(\%opt,1);
				my $onzero=getOnzero(\%opt,0);
				if ( ! defined($types) || ! defined($onzero) or ! defined($match) or ! defined($vldbmode) ) { return undef; }
				my $locations={};
				for my $l (split /,/,$value) {
					my $locationset=$self->expandLocationPattern($l,'format' => 'hashtree','hashtree' => $locations);
					if ( ! defined($locationset) ) {
						out('e',"Unable to interpret location '$value', invalid filter");
						return undef;
					}
				}
				push(@{$constraints{'sub'}},{'type' => 11,'instancetypes' => $types,'neg' => $neg,'value' => $locations,'vldb' => $vldbmode,match => $match,%c});
			} elsif ( $field eq 'count' || $field eq 'c' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1,1 => 1,2 => 1});
				my $vldbmode=getVldb(\%opt,0);
				if ( ! defined($types) or ! defined($vldbmode) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 12,'instancetypes' => $types,'op' => $op,'neg' => $neg,'value' => $value,'vldb' => $vldbmode,%c});
			} elsif ( $field eq 'flags' || $field eq 'f' ) {
				my @flags=split /,/,$value;
				my %flags;
				for my $flag (@flags) {
					if ( $flag =~ /^(\!?)(locked|releasable|rouptodate|backupuptodate|hasrwro|hasoldro|haslostinstances)$/ ) {
						my $v=0;
						if ( $1 eq '!' ) {
							$v=1;
						}
						$flags{$2}=$v;
					} else {
						out('e',"Invalid flag '$flag'");
						return undef;
					}
				}
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowed in flags-constraint");
					return undef;
				}
				if ( ! $value ) {
					$neg=! $neg;
				}
				push(@{$constraints{'sub'}},{'type' => 13,'neg' => $neg,value => \%flags,%c});
			} elsif ( $field eq 'errors' || $field eq 'e' ) {
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowed in ierror-constraint");
					return undef;
				}
				my $types=getInstanceTypes(\%opt,{0 => 1}) || return undef;
				my $match=getMatch(\%opt,'any');
				if ( ! defined($types) || ! defined($match) ) { return undef; }
				my %possibleerrors=('offline' => 1,'needssalvage' => 1,'partexist' => 1,'norw' => 1,'multiple_ro' => 1,'rw_crossdevice' => 1);
				my %errors;
				for my $e (split /,/,$value) {
					if ( $e eq 'all' ) {
						%errors=%possibleerrors;
					} elsif ( $possibleerrors{$e} ) {
						$errors{$e}=1;
					} else {
						out('e',"Invalid error '$e' given");
						return undef;
					}
				}
				push(@{$constraints{'sub'}},{'type' => 14,'instancetypes' => $types,'value' => \%errors,'match' => $match,'neg' => $neg,%c});
			} elsif ( $field eq 'online' || $field eq 'o' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1});
				my $vldbmode=getVldb(\%opt,0);
				my $value_=getValue($value);
				if ( ! defined($types) || ! defined($vldbmode) or ! defined($value_) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 15,'instancetypes' => $types,'op' => $op,'neg' => $neg,'value' => $value_,'vldb' => $vldbmode,%c});
			} elsif ( $field eq 'partonline' || $field eq 'po' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1});
				my $vldbmode=getVldb(\%opt,0);
				$value=getValue($value);
				if ( ! defined($types) or ! defined($vldbmode) or ! defined($value) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 16,'instancetypes' => $types,'op' => $op,'neg' => $neg,'value' => $value,'vldb' => $vldbmode,%c});
			} elsif ( $field eq 'redundancy' || $field eq 'red' || $field eq 'r' ) {
				# This options hash will be forwarded verbatimly to calcVolumeRedundancy()
				my %opt_=('includerostate2' => 0,'includerw' => 1,'includenovldb' => 1);
				if ( defined($opt{'spof'}) ) {
					for my $spof (split /,/,$opt{'spof'}) {
						$opt_{'spoffilter'}={$spof => 1};
					}
				}
				my $vldbmode=getVldb(\%opt,1);
				if ( ! defined($vldbmode) ) {
					return undef;
				}
				if ( ! $vldbmode ) {
					$opt_{'includenovldb'}=0;
				}
				if ( $opt{'ro'} ) {
					$opt_{'includerw'}=0;
				}
				if ( $opt{'unreleased'} ) {
					$opt_{'includerostate2'}=1;
				}
				
				push(@{$constraints{'sub'}},{'type' => 17,'op' => $op,'neg' => $neg,'value' => $value,'options' => \%opt_,%c});
			} elsif ( $field eq 'files' ) {
				my $value_=parse_storagesize($value);
				my $types=getInstanceTypes(\%opt,{0 => 1});
				my $match=getMatch(\%opt,'all');
				my $vldbmode=getVldb(\%opt,0);
				my $onzero=getOnzero(\%opt,0);
				if ( ! defined($types) || ! defined($onzero) or ! defined($match) or ! defined($vldbmode) or ! defined($value_) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 19,'instancetypes' => $types,'match' => $match,'op' => $op,'neg' => $neg,'value' => $value_,'vldb' => $vldbmode,%c});
			} elsif ( $field eq 'quota' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1});
				my $match=getMatch(\%opt,'any');
				my $onzero=getOnzero(\%opt,0);
				my $percent=0;
				my $value_;
				if ( $value =~ /^([0-9]+)%/ ) {
					$value_=$1;
					$percent=1;
				} else {
					$value_=parse_storagesize($value) || return undef;
				}
				if ( ! defined($types) || ! defined($onzero) or ! defined($match) or ! defined($value_) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 20,'instancetypes' => $types,'match' => $match,'op' => $op,'neg' => $neg,'value' => $value_,'relative' => $percent,%c});
			} elsif ( $field eq 'size' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1});
				my $match=getMatch(\%opt,'any');
				my $onzero=getOnzero(\%opt,0);
				my $percent=0;
				my $value_;
				if ( $value =~ /^([0-9]+)%/ ) {
					$value_=int($1);
					$percent=1;
				} else {
					$value_=parse_storagesize($value) || return undef;
				}
				if ( ! defined($types) || ! defined($onzero) or ! defined($match) or ! defined($value) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 21,'instancetypes' => $types,'match' => $match,'op' => $op,'neg' => $neg,'value' => $value_,'match' => $match,'onzero' => $onzero,'relative' => $percent,%c});
			} elsif ( $field eq 'free' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1});
				my $match=getMatch(\%opt,'any');
				my $onzero=getOnzero(\%opt,0);
				my $value_=parse_storagesize($value);
				if ( ! defined($types) || ! defined($onzero) or ! defined($match) or ! defined($value_) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 22,'op' => $op,'neg' => $neg,'value' => $value_,'match' => $match,'onzero' => $onzero,%c});
			} elsif ( $field =~ /^(?:(rw|backup|ro|readonly))?([mca])(time|diff)$/ ) {
				my $loctype=$1;
				if ( ! defined($loctype) ) {
					$loctype='rw';
				}
				my $tstype=$2;
				my $diff=0;
				if ( $3 eq 'diff' ) {
					$diff=1;
					my $int=parse_timeinterval($value);
					if ( ! defined($int) ) {
						out('e',"Wrong timeinterval expression '$value'");
						return undef;
					}
					$value=time() - $int;
					$op = - $op;
				} elsif ( $value =~ /^([0-9]{1,2})\.([0-9]{1,2})\.([0-9]{4})(?:\/([0-9]{1,2}):([0-9]{1,2})(?::([0-9]{1,2}))?)?(u?)$/ ) {
					my ($day,$month,$year,$hour,$minute,$second,$utc)=($1,$2,$3,$4,$5,$6,$7);
					if ( ! defined($second) ) {
						$second=0;
						if ( ! defined($hour) ) {
							$hour=0;
							$minute=0;
						}
					}
					if ( $utc ) {
						$value=timegm($second,$minute,$hour,$day,$month-1,$year-1900);
					} else {
						$value=timelocal($second,$minute,$hour,$day,$month-1,$year-1900);
					}
				} else {
					out('e',"Invalid date format '$value'. Use 'DD.MM.YYYY/hh:mm:ss' !");
					return undef;
				}
				my %types=();
				if ( $loctype eq 'rw' ) {
					$types{0}=1;
				}
				if ( $loctype =~ /^(ro|readonly)$/ ) {
					$types{1}=1;
				}
				if ( $loctype eq 'backup' ) {
					$types{2}=1;
				}
				push(@{$constraints{'sub'}},{
					'type' => 23 ,'op' => $op,'neg' => $neg,'value' => $value,
					'instancetypes' => \%types,'field' => "${tstype}time",%c
				});
			} elsif ( $field eq 'sync' || $field eq 's' ) {
				my $types=getInstanceTypes(\%opt,{1 => 1});
				my $value_=getValue($value);
				my $onunknown=0;
				if ( defined($opt{onunknown}) ) {
					$onunknown=int($opt{onunknown});
				}
				if ( ! defined($types) || ! defined($onunknown) or ! defined($value_) ) { return undef; }
				my $unreleasedro=0;
				if ( $opt{onunreleased} ) {
					$unreleasedro=1;
				}
				delete $types->{0};
				my $grace;
				if ( $opt{'creategrace'} ) {
					$grace=parse_timeinterval($opt{'creategrace'});
					if ( ! defined($grace) ) {
						out('e',"Unable to parse creategrace interval '$opt{creategrace}'");
						return undef;
					}
				}
				push(@{$constraints{'sub'}},{'type' => 24,'op' => $op,'neg' => $neg,'instancetypes' => $types, onunknown => $onunknown,'neg' => $neg,'value' => $value_,creategrace => $grace,onunreleased => $unreleasedro,%c});
			} elsif ( $field =~ /^(backup|ro|readonly)age$/ ) {
				my $type=$1;
				if ( $type eq 'readonly' ) {
					$type='ro';
				}
				my %opt_=();
				if ( $opt{'createsync'} ) {
					$opt_{'createsync'}=1;
				}
				my $int=parse_timeinterval($value);
				if ( ! defined($int) ) {
					out('e',"Wrong timeinterval expression '$value'");
					return undef;
				}
				$value=$int;
				push(@{$constraints{'sub'}},{'type' => 25,'op' => $op,'neg' => $neg,'value' => $value,'typefunction' => "${type}age",%c,'options' => \%opt_});
			} elsif ( $field =~ /^(rw|ro|readonly|backup)?part(?:ition)s??free$/ ) {
				my $loctype=$1;
				if ( ! defined($loctype) ) {
					$loctype=''
				}
				my %types=();
				if ( $loctype =~ /^(rw|)$/ ) {
					$types{0}=1;
				}
				if ( $loctype =~ /^(ro|readonly|)$/ ) {
					$types{1}=1;
				}
				if ( $loctype =~ /^(backup|)$/ ) {
					$types{2}=1;
				}
				my $value_=parse_storagesize($value);
				if ( ! defined($value_) ) {
					return undef;
				}
				push(@{$constraints{'sub'}},{'type' => 26,'instancetypes' => \%types,'op' => $op,'neg' => $neg,'value' => $value_,%c});
			} elsif ( $field eq 'locationcount' || $field eq 'lc' ) {
				my $types=getInstanceTypes(\%opt,{0 => 1,1 => 1});
				my $vldbmode=getVldb(\%opt,0);
				my $value_=getValue($value);
				push(@{$constraints{'sub'}},{'type' => 27,'op' => $op,'neg' => $neg,'instancetypes' => $types,vldb => $vldbmode,'value' => $value_,%c});
			} elsif ( $field eq 'set' ) {
				if ( $value !~ /^([0-9]+)\/([0-9]+)$/ ) {
					out('e',"Invalid constraint format. Use set=A/B with A>0 and B>0 and B>=A .");
					return undef;
				}
				my ($a,$b)=($1,$2);
				if ( $a > $b ) {
					out('e',"Invalid constraint value. set=A/B: A must be less or equal B");
					return undef;
				}
				if ( $a == 0 || $b == 0 ) {
					out('e',"Invalid constraint value. set=A/B: neither A nor B must be zero");
					return undef;
				}
				my $salt=undef;
				if ( defined($opt{salt}) ) {
					if ( $opt{salt} =~ /^[a-z0-9A-Z]{1,16}$/ ) {
						$salt=$opt{salt};
					} else {
						out('e',"set-Salt has to be alphanumeric, less than 16 chars long");
						return undef;
					}
				}
				push(@{$constraints{'sub'}},{'type' => 28,'neg' => $neg,value => [$a-1,$b],salt => $salt});
			} else {
				out('e',"Invalid constraint type '$field'");
				return undef;
			}
		} else {
			if ( $options{'nonames'} ) {
				out('e',"No volumes name selection allowed in this filter");
				return undef;
			}
			$e =~ /^(\!?)(.*)/;
			my $neg=($1 ne '');
			my $match=$2;
			
			$match =~ s/\./\\./g;
			$match =~ s/\?/\./g;
			$match =~ s/\*/\.\*/g;
			push(@{$volpatterns{'sub'}},{'type' => 10,'value' => qr/^$match$/,'neg' => $neg,%c});
		}
	}
	if ( ! @{$volpatterns{'sub'}} ) {
		push(@{$volpatterns{'sub'}},{'type' => 1,'neg' => 0});
	}
	push(@{$filter{'sub'}},\%volpatterns);
	push(@{$filter{'sub'}},\%constraints);
	return \%filter;
}

=item $result=$celldb->matchVolume($filternode,$volumereference,\%options)

Matches a volume (represented by a complete I<$volumereference> as provided
by getVolume() ) against a filternode I<$filternode> as created by parseVolumeFilter().

I<$result> == 1, only if the volume matches the filternode (incl. all
possibly existing subnodes).

This method depends on reading volume/instance information from the database. The actual
testing is done locally. Someone smart might be able to write a replacement method
for filterVolumesComplete() which is able to offload all the testing to the SQL-server.
Currently the biggest problem is the missing PCRE matching support in Mysql.

Possible I<%options> are:

=over 4

=item B<comments>

An arrayref that will be filled with comments associated to matching nodes. The existence
of this field will disabled OR- and AND-shortcuts. Depending on the B<commentfalse> flag
the behaviour changes.

=item B<commentfalse>

Makes matchVolume() add comments on failed matches - not on successful ones.

=back

=cut

sub cond_check {
	my ($value,$reference,$op,$neg)=@_;
	if ( $op == 0 ) {
		return ($value == $reference) ^ $neg;
	} elsif ( $op == -1 ) {
		return ($value < $reference) ^ $neg;
	} elsif ( $op == 1 ) {
		return ($value > $reference) ^ $neg;
	} elsif ( $op == 5 ) {
		return ($value != $reference) ^ $neg;
	} elsif ( $op == -2 ) {
		return ($value <= $reference) ^ $neg;
	} elsif ( $op == 2 ) {
		return ( $value >= $reference) ^ $neg;
	}
}

sub matchVolumeAndComment {
	my ($self,$node,$volref,$options)=@_;
	my $result=$self->matchVolume($node,$volref,$options);
	my $onfalse=defined($options->{'commentfalse'});
	if ( $onfalse != $result ) {
		if ( $node->{'comment'} ) {
			$options->{'comments'}->{$node->{'expression'}}=$node->{'comment'};
		}
	}
	return $result;
}

sub matchVolume {
	my ($self,$node,$volref,$options)=@_;
	my $false=$node->{'neg'};
	my $true=! $false;
	my $type=$node->{'type'};
	if ( $type == 1 ) {
		return $true;
	} elsif ( $type == 2 ) {
		# OR
		my $true_=0;
		for my $n_ (@{$node->{'sub'}}) {
			if ( defined($options->{'comments'}) ) {
				if ( $self->matchVolumeAndComment($n_,$volref,$options) ) {
					$true_=1;
				}
			} else {
				if ( $self->matchVolume($n_,$volref,$options) ) {
					return $true;
				}
			}
		}
		if ( $true_ ) {
			return $true;
		}
		return $false;
	} elsif ( $type == 3 ) {
		# AND
		my $false_=0;
		for my $n_ (@{$node->{'sub'}}) {
			if ( defined($options->{'comments'}) ) {
				if ( ! $self->matchVolumeAndComment($n_,$volref,$options) ) {
					$false_=1;
				}
			} else {
				if ( ! $self->matchVolume($n_,$volref,$options) ) {
					return $false;
				}
			}
		}
		if ( $false_ ) {
			return $false;
		}
		return $true;
	} elsif ( $type == 9 ) {
		if ( $node->{'value'}->{$volref->{'name'}} ) {
			return $true;
		} else {
			return $false;
		}
	} elsif ( $type == 10 ) {
		return ( $volref->{'name'} =~ $node->{'value'} ) ^ $false;
	} elsif ( $type == 11 ) {
		my $icount=0;
		if ( $node->{match} eq 'all' ) {
		 	for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'},'realonly' => ! $node->{'vldb'})}) {
		 		$icount++;
	 			my $result=$node->{'value'}->{$instance->{'serveruuid'}}->{$instance->{'partition'}};
	 			if ( ! $result ) {
	 				return $false;
	 			}
	 		}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
	 		return $true;
	 	} else {
		 	for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'},'realonly' => ! $node->{'vldb'})}) {
		 		$icount++;
	 			my $result=$node->{'value'}->{$instance->{'serveruuid'}}->{$instance->{'partition'}};
	 			if ( $result ) {
	 				return $true;
	 			}
	 		}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
	 		return $false
	 	}
	} elsif ( $type == 12 ) {
		my $instances=$self->iterateInstances($volref,'types' => $node->{'instancetypes'},vldbonly => $node->{vldb});
		return cond_check(scalar(@{$instances}),$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 13 ) {
		my $result=0;
		for my $flag (keys %{$node->{value}}) {
			# The flag test returning negative is considered the same as undefined
			# Possible cases: Test for locked volume without VLDB entry but with fileserver instances
			my $vi=$self->volumeInfo($flag,$volref) || 0;
			my $lresult=$vi ^ $node->{value}->{$flag};
			$result |= $lresult;
		}
		return $result ^ $false;
	} elsif ( $type == 14 ) {
		my $verrors=$self->volumeInfo('errors',$volref);
		my %errors=%{$verrors};
		for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'})}) {
			my $ierrors=$self->instanceInfo('errors',$instance);
			%errors=(%errors,%{$ierrors});
		}
		if ( $node->{'match'} eq 'any' ) {
			for my $error (keys %errors) {
				if ( $node->{'value'}->{$error} ) {
					return $true;
				}
			}
			return $false;
		} else {
			for my $error (keys %{$node->{value}}) {
				if ( ! $errors{$error} ) {
					return $false;
				}
			}
			return $true;
		}
	} elsif ( $type == 15 ) {
		my $count=0;my $onlinecount=0;
		for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
			$count++;
			if ( $self->instanceInfo('accessible',$instance) ) {
				$onlinecount++;
			}
		}
		my $ref=$count;
		if ( $node->{'value'} != -1 ) {
			$ref=$node->{'value'};
		}
		return cond_check($onlinecount,$ref,$node->{'op'},$node->{'neg'});
	} elsif ( $type == 16 ) {
		my $count=0;my $onlinecount=0;
		for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
			$count++;
			if ( $self->partitionInfo('online',$instance) ) {
				$onlinecount++;
			}
		}
		my $ref=$count;
		if ( $node->{'value'} != -1 ) {
			$ref=$node->{'value'};
		}
		return cond_check($onlinecount,$ref,$node->{'op'},$node->{'neg'});
	} elsif ( $type == 17 ) {
		my $redundancy=$self->calcInstanceRedundancy($volref,%{$node->{'options'}},'includerw' => $node->{'rw'},'spoffilter' => $node->{'spoffilter'})-1;
		return cond_check($redundancy,$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 19 ) {
		# number of files
		my $icount=0;
		if ( $node->{'match'} eq 'any' ) {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				if ( defined($instance->{files}) ) {
					$icount++;
					if ( cond_check($instance->{'files'},$node->{'value'},$node->{'op'},0) ) {
						return $true;
					}
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $false;
		} else {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				if ( defined($instance->{files}) ) {
					$icount++;
					if ( ! cond_check($instance->{'files'},$node->{'value'},$node->{'op'},0) ) {
						return $false;
					}
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $true;
		}
	} elsif ( $type == 20 ) {
		# quota
		my $icount=0;
		if ( $node->{'match'} eq 'any' ) {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				if ( $node->{'relative'} ) {
					my $value=int($instance->{'size'}*$node->{'value'}/100);
					if ( cond_check($instance->{'quota'},$value,$node->{'op'},$node->{'neg'}) ) {
						return $true;
					}
				} else {
					if ( cond_check($instance->{'quota'},$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
						return $true;
					}
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $false;
		} else {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				if ( $node->{'relative'} ) {
					my $value=int($instance->{'size'}*$node->{'value'}/100);
					if ( ! cond_check($instance->{'quota'},$value,$node->{'op'},$node->{'neg'}) ) {
						return $false;
					}
				} else {
					if ( ! cond_check($instance->{'quota'},$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
						return $false;
					}
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $true;
		}
	} elsif ( $type == 21 ) {
		# size
		my $icount=0;
		if ( $node->{'match'} eq 'any' ) {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				if ( $node->{'relative'} ) {
					my $value=int($instance->{'quota'}*$node->{'value'}/100);
					if ( cond_check($instance->{'size'},$value,$node->{'op'},$node->{'neg'}) ) {
						return $true;
					}
				} else {
					if ( cond_check($instance->{'size'},$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
						return $true;
					}
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $false;
		} else {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				if ( $node->{'relative'} ) {
					my $value=int($instance->{'quota'}*$node->{'value'}/100);
					if ( ! cond_check($instance->{'size'},$value,$node->{'op'},$node->{'neg'}) ) {
						return $false;
					}
				} else {
					if ( ! cond_check($instance->{'size'},$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
						return $false;
					}
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $true;
		}
	} elsif ( $type == 22 ) {
		# storage left (quota free,partfree)
		my $icount=0;
		if ( $node->{'match'} eq 'any' ) {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				my $free=$self->instanceInfo('storageleft',$instance);
				if ( ! defined($free) ) {
					next;
				}
				if ( cond_check($free,$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
					return $true;
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $false;
		} else {
			for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'},'vldbonly' => $node->{'vldb'})} ) {
				my $free=$self->instanceInfo('storageleft',$instance);
				if ( ! defined($free) ) {
					next;
				}
				if ( ! cond_check($free,$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
					return $false;
				}
			}
	 		if ( $icount == 0 ) {
	 			if ( $node->{onzero} ) {
	 				return $true
	 			} else {
	 				return $false;
	 			}
	 		}
			return $true;
		}
	} elsif ( $type == 23 ) {
		for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'})}) {
			if ( cond_check($instance->{$node->{'field'}},$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
				return $true;
			}
		}
		return $false;
	} elsif ( $type == 24 ) {
		my $count=0;my $refcount=0;
		my $rw=$volref->{rw};
		if ( ! defined($rw->{mtime}) ) {
			if ( $node->{onunknown} ) {
				return $true;
			} else {
				return $false;
			}
		}
		for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'})}) {
			if ( ! defined($instance->{rostate}) ) {
				# Ignore VLDB-unknown ROs
				next;
			}
			$refcount++;
			if ( ! $instance->{rostate} > 1 ) {
				if ( ! $node->{onunreleased} ) {
					# Count unreleased as not in sync
					next;
				} else {
					$count++;
					next;
				}
			}
			if ( ! defined($instance->{ctime}) ) {
				if ( $node->{onunknown} ) {
					$count++;
				}
				next;
			}
			if ( $instance->{ctime} >= $rw->{mtime} ) {
				$count++;
			}
		}
		my $ref=$refcount;
		if ( $node->{'value'} != -1 ) {
			$ref=$node->{'value'};
		}
		my $result=cond_check($count,$ref,$node->{'op'},0);
		if ( ! $result ) {
			if ( defined($node->{creategrace}) ) {
				if ( defined($rw->{ctime}) ) {
					if ( ( time() - $rw->{ctime} ) < $node->{creategrace} ) {
						return $true;
					}
				}
			}
			return $false;
		} else {
			return $true;
		}
	} elsif ( $type == 25 ) {
		my $syncortryts=$self->volumeInfo($node->{'typefunction'},$volref,%{$node->{'options'}});
		if ( ! defined($syncortryts) ) {
			return $false;
		}
		my $age=time() - $syncortryts; 
		return cond_check($age,$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 26 ) {
		for my $instance (@{$self->iterateInstances($volref,'types' => $node->{'instancetypes'})}) {
			if ( ! cond_check($instance->{'partfree'},$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
				return $false;
			}
		}
		return $true;
	} elsif ( $type == 27 ) {
		my %locations=();
		for my $instance (@{$self->iterateInstances($volref)}) {
			if ( ! $node->{vldb} || defined($instance->{rostate}) ) {
				$locations{$instance->{'serveruuid'} . $instance->{'partition'}}=1;
			}
		}
		if ( cond_check(scalar(keys %locations),$node->{'value'},$node->{'op'},$node->{'neg'}) ) {
			return $true;
		} else {
			return $false;
		}
	} elsif ( $type == 28 ) {
		my $digest=Digest->new("MD5");
		if ( defined($node->{salt}) ) {
			$digest->add($node->{salt});
		}
		$digest->add($volref->{name});
		my $hash32=unpack("L",$digest->digest());
		if ( ( $hash32 % $node->{'value'}->[1] ) == $node->{'value'}->[0] ) {
			return $true;
		} else {
			return $false;
		}
	} else {
		out('e',"Invalid operation code '$type'");
		return 0;
	}
	out('e',"No result returnd by condition (id='$type')");
	return undef;
}

=item \%volumegroups=$celldb->listVolumeGroups()

Returns a hash containing all volume groups as keys.

=cut

sub listVolumeGroups {
	my ($self)=@_;
	$self->ensureCellinfo() || return undef;
	my $groups_=$self->{cellinfo}->{xc}->find("//ci:volgroup");
	my %groups=();
	for my $group_ (@{$groups_}) {
		$groups{$group_->getAttribute('id')}=1;
	}
	return \%groups;
}


=item \%volumes=$celldb->getVolumeGroup($volumegroupid)

Expands the name of a volume group I<$volumegroupid> as defined in cell.xml to a hashref
with keys being volume names and values being volume reference hashes.

On error or unknown I<$volumegroupid>, B<undef> is returned instead.

=cut


sub getVolumeGroup {
	my ($self,$id)=@_;
	$self->ensureCellinfo() || return undef;
	my $group_=$self->{cellinfo}->{xc}->find("//ci:volgroup[\@id='$id']/*");
	if ( ! $group_ ) {
		out('d',"getVolumeGroup(): Unknown group '$id'");
		return undef;
	}
	my %volumes=();
	for my $tag_ (@{$group_}) {
		if ( $tag_->nodeName eq 'include' ) {
			if ( $tag_->hasAttribute('conditions') ) {
				my $filter=$self->parseVolumeFilter($tag_->getAttribute('conditions')) || return undef;
				my $volumes=$self->filterVolumes($filter);
				addToSet(\%volumes,$volumes);
			} else {
				out('e',"Error in cell.xml, missing conditions in volgroup '$id' 's include rule");
				return undef;
			}
		} elsif ( $tag_->nodeName eq 'exclude' ) {
			if ( $tag_->hasAttribute('conditions') ) {
				my $filter=$self->parseVolumeFilter($tag_->getAttribute('conditions')) || return undef;
				my $volumes=$self->filterVolumes($filter);
				removeFromSet(\%volumes,$volumes);
			} else {
				out('e',"Error in cell.xml, missing conditions in volgroup '$id' 's exclude rule");
			}
		} else {
			out('d',"getVolumeGroup(): Unknown subtag '" . $tag_->nodeName . "' below volgroup '$id'");
			return undef;
		}
	}
	return \%volumes;
}

=item \%groups=$celldb->getVolumeMemberships(\@volrefs,%options);

Determines the volume groups, the volumes given via their I<@volrefs> are
member of. By default, the returned hash contains the groups, any
volume in I<@volrefs> is member of. Possible I<%options> are:

=over 4

=item B<mode> (default: B<any>)

Defines the matching mode. If B<all> is given, only the groups are return that
all given volumes are member of.

=back

=cut

sub getVolumeMemberships {
	my ($self,$volrefs,%options)=@_;
	$self->ensureCellinfo() || return undef;
	my $groups=$self->listVolumeGroups();
	my %memberships_any=();my %memberships_all=();
	for my $group (keys %{$groups}) {
		my $group_=$self->{cellinfo}->{xc}->find("//ci:volgroup[\@id='$group']/*");
		my $included_any=0;my $included_all=1;
		for my $volref (@{$volrefs}) {
			my $vincluded=0;
			for my $tag_ (@{$group_}) {
				if ( $tag_->nodeName eq 'include' ) {
					if ( $tag_->hasAttribute('conditions') ) {
						my $filter=$self->parseVolumeFilter($tag_->getAttribute('conditions'));
						$vincluded |= $self->matchVolume($filter,$volref);
					}
				} elsif ( $tag_->nodeName eq 'exclude' ) {
					if ( $tag_->hasAttribute('conditions') ) {
						my $filter=$self->parseVolumeFilter($tag_->getAttribute('conditions'));
						$vincluded &= ! $self->matchVolume($filter,$volref);
					}
				}
			}
			$included_any |= $vincluded;
			$included_all &= $vincluded;
		}
		if ( $included_any ) {
			$memberships_any{$group}=1;
		}
		if ( $included_all ) {
			$memberships_all{$group}=1;
		}
	}
	if ( defined($options{mode}) && $options{mode} eq 'all' ) {
		return \%memberships_all;
	} else {
		return \%memberships_any;
	}
}

=item \%volumes=$celldb->filterVolumesComplete($filterbasenode)

Returns a hashref (keys being volume names, values being volume reference hashes)
of volumes matching the given I<$filterbasenode> as provided by parseVolumeFilter().

=cut

sub filterVolumesComplete {
	my ($self,$filterbasenode,%options)=@_;
	out('d',"Using correct volume filter engine");
	my %volumes=();
	for my $volume (keys %{$self->listVolumes()}) {
		my $volref=$self->getVolume($volume);
		my $options;
		if ( $options{'comments'} ) {
			$options={'comments' => {},'commentfalse' => $options{'commentfalse'}}
		}
		if ( $self->matchVolume($filterbasenode,$volref,$options) ) {
			$volumes{$volume}=$volref;
		}
		if ( $options{'comments'} ) {
			$options{'comments'}->{$volume}=$options->{'comments'};
		}
	}
	return \%volumes
}

=item \%volumes=$celldb->filterVolumes($filterbasenode,%options)

Returns a hashref (keys being volume names, values being volume reference hashes)
of volumes matching the given I<$filterbasenode> as provided by parseVolumeFilter().

This function makes the assumption, that the I<$filterbasenode> was provided
by parseVolumeFilter() which always sorts the name pattern matches in a unique
way.  This way, it's possible to pre-filter the volume names, before all
information have to be read about them.  This usually speeds up the matching
process tremendously.

It's always safe to use this optimized function for nodes created by parseVolumeFilter().

These I<%options> may be given:

=over 4

=item B<nodata>

Only return volume data (as returned by getVolume()) if it's already available. Otherwise,
just use B<1> as hash value of volumes.

=item B<comments>

...

=back

=cut

sub filterVolumes {
	my ($self,$filterbasenode,%options)=@_;
	out('d',"Using slightly optimised volume filter engine");
	if ( $filterbasenode->{'sub'}->[0]->{'type'} == 2 ) {
		# Seems to be a pattern matching tree. We try to
		# speed things up by only applying conditions to matched volumes
		my %volumes=();
		my $volumes=$self->listVolumes();
		for my $volume (keys %{$volumes}) {
			my $match=0;
			my $options;
			if ( $options{'comments'} ) {
				$options={'comments' => {},'commentfalse' => $options{'commentfalse'}}
			}
			if ( $self->matchVolume($filterbasenode->{'sub'}->[0],{'name' => $volume},$options) ) {
				if ( $options{'nodata'} ) {
					$match=1;
				} else {
					$match=$self->getVolume($volume);
				}
			}
			if ( $match ) {
				if ( @{$filterbasenode->{'sub'}->[1]->{'sub'}} > 0 ) {
					my $volref=$self->getVolume($volume);
					if ( $self->matchVolume($filterbasenode->{'sub'}->[1],$volref,$options) ) {
						$match=$volref;
					} else {
						$match=0;
					}
				}
			}
			if ( $options{'comments'} ) {
				$options{'comments'}->{$volume}=$options->{'comments'};
			}
			if ( $match ) {
				$volumes{$volume}=$match;
			}
		}
		return \%volumes;
	} else {
		return $self->filterVolumesComplete($filterbasenode,%options);
	}
}

=item $result=$celldb->checkVolumes($testid,\%groupcache,%options)

Runs test specified by the ID I<$testid>. I<%groupcache> is a hash
containing groups (key=groupname,value=getVolumeGroup(groupname)).  This
cache will be modified by checkVolumes() if necessary and used if possible. 
Possible I<%options> are:

=over 4

=item B<testnumber>

Optional number of the test node. This will be used in message being output.

=item B<singlevolume>

A single volume may be given as value of this key to just run the test against
this specific one.

=item B<result>

An optional hashref that will be used to store results of the test. These data will
be written:

=over 4

=item B<volumecount>

Number of volumes being tested

=item B<successcount>

Number of volumes passing the test

=item B<failedcount>

Number of volumes failing the test

=item B<failedvolumes>

Hashref containing the failed volumes as keys.

=back

=back

=cut

sub checkVolumes {
	my ($self,$testnode_,$groupcache,%options)=@_;
	$self->ensureCellinfo() || return undef;
	my $testid;
	if ( $testnode_->hasAttribute('id') ) {
		$testid="'" . $testnode_->getAttribute('id');
		if ( defined($options{'testnumber'}) ) {
			$testid .= "' (#$options{'testnumber'})";
		}
	} else {
		if ( defined($options{'testnumber'}) ) {
			$testid="(#$options{'testnumber'})";
		}
	}
	out('v',"Running volume test $testid");

	my $thisgroup={};
	my $groupcount;
	if ( $options{'singlevolume'} ) {
		out('v',"Testing single volume '$options{'singlevolume'}' against this rule");
		$thisgroup={$options{'singlevolume'} => 1};
		$groupcount=1;
	} else {
		for my $group (split /[,\s]+/,$testnode_->getAttribute('group')) {
			if ( ! $groupcache->{$group} ) {
				my $members=$self->getVolumeGroup($group);
				if ( ! $members ) {
					out('e',"Unable to get volume group '$group'");
					return -1;
				}
				$groupcache->{$group}=$members;
				out('d',"Including group '$group'");
				addToSet($thisgroup,$members);
			}
		}
		if ( $testnode_->hasAttribute('excludegroup') ) {
			for my $group (split /[,\s]+/,$testnode_->getAttribute('excludegroup')) {
				if ( ! $groupcache->{$group} ) {
					my $members=$self->getVolumeGroup($group);
					if ( ! $members ) {
						out('e',"Unable to get volume group '$group'");
						return -1;
					}
					$groupcache->{$group}=$members;
					out('d',"Excluding group '$group'");
					removeFromSet($thisgroup,$members);
				}
			}
		}
		if ( $testnode_->hasAttribute('excludevolumes') ) {
			for my $volume (split /[,\s]+/,$testnode_->getAttribute('excludevolumes')) {
				if ( $thisgroup->{$volume} ) {
					delete $thisgroup->{$volume};
				}
			}
		}
		$groupcount=scalar(keys %{$thisgroup});
		out('v',"Test group contains '$groupcount' volumes");
	}
	
	my $conditions;
	my %comments;
	my $conditions_=$self->{cellinfo}->{xc}->find("./ci:condition",$testnode_);
	my $docomments=0;
	if ( $conditions_ ) {
		$conditions=[];
		$docomments=1;
		for my $condition_ (@{$conditions_}) {
			$comments{$condition_->getAttribute('condition')}=$condition_->getAttribute('comment');
			push(@{$conditions},$condition_->getAttribute('condition'));
		}
	} else {
		$conditions=$testnode_->getAttribute('conditions');
	}
	my $filtertree=$self->parseVolumeFilter($conditions,'nonames' => 1,'comments' => \%comments);
	if ( ! $filtertree ) {
		out('e',"Unable to parse filter conditions of test $testid");
		return -1;
	}
	
	# This is used to set the severity of an error (i.e. for nagios)
	my $errorlevel=1;
	if ( $testnode_->hasAttribute('level') ) {
		$errorlevel=$testnode_->getAttribute('level');
	}
	
	# The filter tree will be modified to take the test's group as sole location source
	# Filter rule 10 is reserved for this kind of location set injection
	$filtertree->{'sub'}->[0]->{'sub'}=[{
		'type' => 9,
		'value' => $thisgroup,
		'neg' => 0,
	}];
	
	# Constraints will be negated to sieve out all locations not satisfying the test's constraints
	$filtertree->{'sub'}->[1]->{'neg'}=1;
	
	my $failedvolumes;
	my %commentresults=();
	if ( $docomments ) {
		$failedvolumes=$self->filterVolumes($filtertree,'comments' => \%commentresults,'commentfalse' => 1);
	} else {
		$failedvolumes=$self->filterVolumes($filtertree);
	}
	my $failedcount=scalar(keys %{$failedvolumes});
	my $successcount=$groupcount-$failedcount;
	if ( $options{'result'} ) {
		$options{'result'}->{'objectcount'}=$groupcount;
		$options{'result'}->{'successcount'}=$successcount;
		$options{'result'}->{'failedcount'}=$failedcount;
		$options{'result'}->{'failedobjects'}=$failedvolumes;
		$options{'result'}->{'testid'}=$testid;
		$options{'result'}->{'comments'}={};
		for my $volume (keys %commentresults) {
			if ( %{$commentresults{$volume}} ) {
				my $volumecomments=join(", ",values %{$commentresults{$volume}});
				$options{'result'}->{'comments'}->{$volume}=$volumecomments;
			}
		}
	}
	if ( $failedcount ) {
		out('v',"Test $testid: '$failedcount' of '$groupcount' volumes failed");
		return $errorlevel;
	} else {
		out('v',"Test $testid successfull");
	}
	return 0;
}

sub checkLocations {
	my ($self,$testnode_,%options)=@_;
	$self->ensureCellinfo() || return undef;
	my $testid;
	if ( $testnode_->hasAttribute('id') ) {
		$testid="'" . $testnode_->getAttribute('id');
		if ( defined($options{'testnumber'}) ) {
			$testid .= "' (#$options{'testnumber'})";
		}
	} else {
		if ( defined($options{'testnumber'}) ) {
			$testid="(#$options{'testnumber'})";
		}
	}
	out('v',"Running volume test $testid");

	my $thisgroup={};
	my $groupcount=0;
	if ( $options{'singlelocation'} ) {
		my $locations=$self->expandLocationPattern($options{singlelocation},format => 'arrayref') || return -1;
		if ( @{$locations} != 1 ) {
			out('e',"Location expression '$options{'singlelocation'}' didn't return exactly one location.");
			return -1;
		}
		out('v',"Testing single location '$options{'singlelocation'}' against this rule");
		$thisgroup={$locations->[0]->{serveruuid} => {$locations->[0]->{partition} => $locations->[0]}};
		$groupcount=1;
	} else {
		for my $group (split /[,\s]+/,$testnode_->getAttribute('group')) {
			my $locations=$self->expandLocationPattern($group,format => 'arrayref') || return -1;
			for my $location (@{$locations}) {
				$thisgroup->{$location->{serveruuid}}->{$location->{partition}}=$location;
				$groupcount++;
			}
			out('d',"Including location expression '$group'");
		}
		if ( $testnode_->hasAttribute('excludegroup') ) {
			for my $group (split /[,\s]+/,$testnode_->getAttribute('excludegroup')) {
				my $locations=$self->expandLocationPattern($group,format => 'arrayref') || return -1;
				for my $location (@{$locations}) {
					delete $thisgroup->{$location->{serveruuid}}->{$location->{partition}};
				}
				out('d',"Excluding location expression '$group'");
			}
		}
		my $groupcount=0;my $servers=0;
		for my $server (keys %{$thisgroup}) {
			$groupcount += scalar(keys %{$thisgroup->{$server}});
		}
		out('v',"Test group contains '$groupcount' locations on '$servers' servers.");
	}
	my $collatedlocations;
	if ( $testnode_->hasAttribute('collate') ) {
		if ( $testnode_->getAttribute('collate') eq '1' ) {
		
			my $stat=$self->collatePartitionStatistics($thisgroup);
			$stat->{name}='*';
			$collatedlocations->{'x'}->{'x'}=$stat;
			$thisgroup={'x' => { 'x' => $stat }};
		}
	}
	
	my $conditions;
	my %comments;
	my $conditions_=$self->{cellinfo}->{xc}->find("./ci:condition",$testnode_);
	my $docomments=0;
	if ( $conditions_ ) {
		$conditions=[];
		$docomments=1;
		for my $condition_ (@{$conditions_}) {
			$comments{$condition_->getAttribute('condition')}=$condition_->getAttribute('comment');
			push(@{$conditions},$condition_->getAttribute('condition'));
		}
	} else {
		$conditions=[$testnode_->getAttribute('conditions')];
	}
	my $filtertree=$self->parsePartitionFilter($conditions,'comments' => \%comments);
	if ( ! $filtertree ) {
		out('e',"Unable to parse filter conditions of test $testid");
		return -1;
	}
	
	my $errorlevel=1;
	if ( $testnode_->hasAttribute('level') ) {
		$errorlevel=$testnode_->getAttribute('level');
	}
	
	# The filter tree will be modified to take the test's group as sole location source
	if ( ! $collatedlocations ) {
		$filtertree->{'sub'}->[0]->{'sub'}=[{
			'type' => 10,
			'value' => $thisgroup,
			'neg' => 0,
		}];
	}
	
	# Constraints will be negated to sieve out all volumes not satisfying the test's constraints
	$filtertree->{'sub'}->[1]->{'neg'}=1;
	
	my $failedlocations;
	my %commentresults=();
	if ( $docomments ) {
		$failedlocations=$self->filterPartitions($filtertree,'comments' => \%commentresults,'commentfalse' => 1,format => 'arrayref',partitionstats => $collatedlocations);
	} else {
		$failedlocations=$self->filterPartitions($filtertree,format => 'arrayref',partitionstats => $collatedlocations);
	}
	if ( ! defined($failedlocations) ) {
		return -1;
	}
	my $failedcount=scalar(@{$failedlocations});
	my $successcount=$groupcount-$failedcount;
	if ( $options{'result'} ) {
		$options{'result'}->{'objectcount'}=$groupcount;
		$options{'result'}->{'successcount'}=$successcount;
		$options{'result'}->{'failedcount'}=$failedcount;
		$options{'result'}->{'failedobjects'}=$failedlocations;
		$options{'result'}->{'testid'}=$testid;
		$options{'result'}->{'comments'}={};
		for my $volume (keys %commentresults) {
			if ( %{$commentresults{$volume}} ) {
				my $locationcomments=join(", ",values %{$commentresults{$volume}});
				$options{'result'}->{'comments'}->{$volume}=$locationcomments;
			}
		}
	}
	if ( $failedcount ) {
		if ( $collatedlocations ) {
			out('v',"Test $testid failed");
		} else {
			out('v',"Test $testid: '$failedcount' of '$groupcount' locations failed");
		}
		return $errorlevel;
	} else {
		out('v',"Test $testid successfull");
	}
	return 0;
}



sub parsePartitionFilter {
	our ($self,$expression,%options)=@_;
	my @expressions;
	if ( ref($expression) eq 'ARRAY' ) {
		@expressions=@{$expression};
	} elsif ( ref($expression) eq 'HASH' ) {
		# Sorting is not really necessary but reduces code paths in case of debugging
		@expressions=sort keys %{$expression};
	} else {
		@expressions=split /\s+/,$expression;
	}
	sub getInstanceTypesKey {
		my ($opt,$default,%options)=@_;
		my %types;
		if ( defined($opt->{'itype'}) ) {
			for my $t (split /,/,$opt->{itype}) {
				if ( $t eq 'all' ) {
					%types=(0 => 1,1 => 1,2 => 1);
				} else {
					my $id=$self->instanceTypeToId($t);
					if ( ! defined($id) ) {
						out('e',"Invalid instance type '$t' given.");
						return ();;
					}
					$types{$id}=1;
				}
			}
		} else {
			%types=%{$default};
		}
		my $key;
		if ( setEquality(\%types,{0 => 1}) ) {
			$key='rw';
		} elsif ( setEquality(\%types,{0 => 1,1 => 1, 2 => 2}) ) {
			$key='';
		} else {
			out('e',"Invalid itype specification");
			return 0;
		}
		if ( $options{prefix} ) {
			$key=$options{prefix} . $key;
		}
		return $key;
	}
	my %patterns=('type' => 2,'sub' => [],'neg' => 0);
	my %constraints=('type' => 3,'sub' => [],'neg' => 0);
	my %filter=('type' => 3,'sub' => [],'neg' => 0);
	for my $e (@expressions) {
		my %c=('expression' => $e);
		if ( $options{'comments'}->{$e} ) {
			$c{'comment'}=$options{'comments'}->{$e};
		}
		if ( $e =~ /^(\!)?(.*?)(=|<=|>=|\!=|<|>)(.*)$/ ) {
			my ($arz,$field,$op_,$value)=($1,$2,$3,$4);
			my ($op,$neg)=normalize_op($op_,$arz);
			my %opt=();
			if ( $value =~ /^([^:]+):(.*)$/ ) {
				$value=$1;
				for my $item (split /:/,$2) {
					if ( $item =~ /^(.*?)=(.*)$/ ) {
						$opt{$1}=$2;
					}
				}
			} else {
				$opt{$value}=1;
			}
			my $match;
			if ( $field eq 'location' ) {
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowed in globber expression");
					return undef;
				}
				my $locations=$self->expandLocationPattern($value,format => 'hashtree');
				push(@{$patterns{'sub'}},{type => 10,value => $locations,
					'neg' => $neg,%c
				});
			} elsif ( $field eq 'count' ) {
				my $itype=getInstanceTypesKey(\%opt,{0 => 1,1 => 1,2 =>1},prefix => 'partinstances');
				if ( ! defined($itype) ) { return undef; }
				push(@{$constraints{'sub'}},{
					type => 12,field => $itype,
					op => $op,neg => $neg,
					value => $value,%c
				});
			} elsif ( $field eq 'online' ) {
				if ( $value !~ /^[01]$/ ) {
					out('e',"Invalid value for 'online' comparision");
					return undef;
				}
				if ( $op % 5 ) {
					out('e',"Operator '$op_' not allowed in 'online' constraint");
					return undef;
				}
				push(@{$constraints{'sub'}},{
					type => 16, value => ($value ^ int($op == 5)),neg => $neg,%c
				});
			} elsif ( $field eq 'files' ) {
				my $value_=parse_storagesize($value);
				my $itype=getInstanceTypesKey(\%opt,{0 => 1,1 => 1,2 =>1},prefix => 'filessum');
				if ( ! defined($itype) or ! defined($value_) ) { return undef; }
				push(@{$constraints{'sub'}},{
					type => 19,field => $itype,
					op => $op,neg => $neg,
					value => $value_,%c
				});
			} elsif ( $field eq 'free' ) {
				my $value_=parse_storagesize($value);
				if ( ! defined($value_) ) { return undef; }
				push(@{$constraints{'sub'}},{'type' => 26,op => $op,neg => $neg,value => $value_,%c});
			} elsif ( $field eq 'size' ) {
				my $value_=parse_storagesize($value);
				if ( ! defined($value_) ) { return undef;	}
				push(@{$constraints{'sub'}},{'type' => 28,op => $op,neg => $neg,value => $value_,%c});
			} elsif ( $field eq 'used' ) {
				my $value_=parse_storagesize($value);
				if ( ! defined($value_) ) {	return undef; }
				push(@{$constraints{'sub'}},{'type' => 29,op => $op,neg => $neg,value => $value_,%c});
			} else {
				out('e',"Invalid constraint type '$field'");
				return undef;
			}
		} else {
			my $locations=$self->expandLocationPattern($e,format => 'hashtree');
			push(@{$patterns{'sub'}},{type => 10,value => $locations,'neg' => 0,%c});
		}
	}
	if ( ! @{$patterns{'sub'}} ) {
		push(@{$patterns{'sub'}},{'type' => 1,'neg' => 0});
	}
	push(@{$filter{'sub'}},\%patterns);
	push(@{$filter{'sub'}},\%constraints);
	return \%filter;
}

sub matchPartitionAndComment {
	my ($self,$node,$volref,$options)=@_;
	my $result=$self->matchPartition($node,$volref,$options);
	my $onfalse=defined($options->{'commentfalse'});
	if ( $onfalse != $result ) {
		if ( $node->{'comment'} ) {
			$options->{'comments'}->{$node->{'expression'}}=$node->{'comment'};
		}
	}
	return $result;
}


sub matchPartition {
	my ($self,$node,$partref,$options)=@_;
	my $false=$node->{'neg'};
	my $true=! $false;
	my $type=$node->{'type'};
	if ( $type == 1 ) {
		return $true;
	} elsif ( $type == 2 ) {
		# OR
		my $true_=0;
		for my $n_ (@{$node->{'sub'}}) {
			if ( defined($options->{'comments'}) ) {
				if ( $self->matchPartitionAndComment($n_,$partref,$options) ) {
					$true_=1;
				}
			} else {
				if ( $self->matchPartition($n_,$partref,$options) ) {
					return $true;
				}
			}
		}
		if ( $true_ ) {
			return $true;
		}
		return $false;
	} elsif ( $type == 3 ) {
		# AND
		my $false_=0;
		for my $n_ (@{$node->{'sub'}}) {
			if ( defined($options->{'comments'}) ) {
				if ( ! $self->matchPartitionAndComment($n_,$partref,$options) ) {
					$false_=1;
				}
			} else {
				if ( ! $self->matchPartition($n_,$partref,$options) ) {
					return $false;
				}
			}
		}
		if ( $false_ ) {
			return $false;
		}
		return $true;
	} elsif ( $type == 10 ) {
		if ( $node->{'value'}->{$partref->{serveruuid}}->{$partref->{partition}} ) {
			return $true;
		} else {
			return $false;
		}
	} elsif ( $type == 12 ) {
		return cond_check($partref->{$node->{field}},$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 16 ) {
		if ( $partref->{ponline} == $node->{value} ) {
			return $true;
		} else {
			return $false;
		}
	} elsif ( $type == 19 ) {
		return cond_check($partref->{$node->{field}},$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 26 ) {
		return cond_check($partref->{partfree},$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 28 ) {
		return cond_check($partref->{partsize},$node->{'value'},$node->{'op'},$node->{'neg'});
	} elsif ( $type == 29 ) {
		return cond_check($partref->{partsize}-$partref->{partfree},$node->{'value'},$node->{'op'},$node->{'neg'});
	} else {
		out('e',"Invalid operation code '$type'");
		return undef;
	}
}


sub collatePartitionStatistics {
	my ($self,$partitions)=@_;
	my %r=(
		partsize => 0,
		partfree => 0,
		sizesum => 0,
		sizesumrw => 0,
		filessum => 0,
		filessumrw => 0,
		partinstances => 0,
		partinstancesrw => 0,
		quotasum => 0,
		quotasumrw => 0,
		partvolumes => 0,
		partname => 'collated',
		ponline => 1,
		maxmtimerw => 0,
	);
	my @partitions;
	if ( ref($partitions) eq 'ARRAY' ) {
		@partitions=@{$partitions};
	} else {
		for my $s (keys %{$partitions}) {
			for my $p (keys %{$partitions->{$s}}) {
				push(@partitions,$partitions->{$s}->{$p});
			}
		}
	}
	for my $p_ (@partitions) {
		for my $f ('partsize','partfree','quotasum','quotasumrw','sizesum','sizesumrw','filessum','filessumrw','partinstances','partinstancesrw','partvolumes') {
			$r{$f} += $p_->{$f};
		}
		
		for my $f ('maxmtimerw') {
			if ( $r{$f} < $p_->{$f} ) {
				$r{$f}=$p_->{$f};
			}
		}
		$r{ponline} &= $p_->{ponline};
	}
	return \%r;
}


=item $result=$celldb->filterPartitions($querytree,%options);

This method will calculate some statistics about all partitions
and matches them against a I<$querytree> as created by L<parsePartitionFilter>.
Possible I<%options> are:

=over 4

=item B<partitionstats>

Gives a partition statistics tree as returned by L<getPartitionStatistics()>
to save time building it.

=back

=cut

sub filterPartitions {
	my ($self,$filterbasenode,%options)=@_;
	my $format='hashtree';
	if ( $options{format} ) {
		$format=$options{format};
	}
	my $stats;
	if ( $options{partitionstats} ) {
		$stats=$options{partitionstats};
	} else {
		$stats=$self->getPartitions(format => 'hashtree') || exit 2;
	}
	my %partitions;my @partitions;
	for my $serveruuid (keys %{$stats}) {
		for my $partition (keys %{$stats->{$serveruuid}}) {
			my $partref=$stats->{$serveruuid}->{$partition};
			my $options;
			if ( $options{'comments'} ) {
				$options={'comments' => {},'commentfalse' => $options{'commentfalse'}};
			}
			if ( $self->matchPartition($filterbasenode,$partref,$options) ) {
				$partitions{$serveruuid}->{$partition}=$partref;
				push(@partitions,$partref);
			}
			if ( $options{'comments'} ) {
				$options{'comments'}->{$partref}=$options->{'comments'};
			}
		}
	}
	if ( $format eq 'hashtree' ) {
		return \%partitions;
	} elsif ( $format eq 'arrayref' ) {
		return \@partitions;
	} else {
		out('e',"Invalid format '$format'");
		return undef;
	}
}


sub isExtendedPartitionFilter {
	my ($self,$filternode)=@_;
	my %extended=(19 => 1);
	my %normal=(10 => 1,12 => 1,16 => 1,26 => 1,28 => 1,29 => 1);
	my $extended=0;
	if ( $filternode->{type} == 2 || $filternode->{type} == 3 ) {
		for my $node (@{$filternode->{'sub'}}) {
			$extended |= $self->isExtendedPartitionFilter($node);
		}
	}
	return $extended;
}

=item $result=$celldb->runVolumeHooks($volume,$action)

Runs hooks associated to action I<$action> for volume I<$volume> according to
cell.xml.

I<$result> == 1, only if none of the hooks failed (which is true, if there aren't any).

=cut

sub runVolumeHooks {
	my ($self,$volume,$action,%options)=@_;
	if ( ! $self->ensureCellinfo() ) {
		return undef;
	}
	my $hooks_=$self->{cellinfo}->{'xc'}->find("//ci:hook[\@action='$action']");
	my $result=1;
	if ( ! $hooks_ ) {
		out('d',"No '$action'-hooks for volume '$volume'");
		return 1;
	}
	for my $hook_ (@{$hooks_}) {
		my $volinclude=$hook_->getAttribute('volinclude');
		if ( $volume =~ /^$volinclude$/ ) {
			my $test=$hook_->getAttribute('test');
			$test =~ s/\{volume\}/$volume/g;
			my $lresult = iaexec($test,'test');
			my $label=$hook_->getAttribute('label');
			if ( ! $lresult ) {
				out('w',"'$action'-hook ('$label') for volume '$volume' failed.");
				$result=0;
			} else {
				out('v',"'$action'-hook ('$label') for volume '$volume' succeeded.");
			}
		}
	}
	return $result;
}

=back

=head1 APPENDIX

=head2 Instance Types

Volume Instance types can be given to AFS::Instant::CellDB-Methods either
numerically or by a symbolic type id. Numerical Types are:

=over 4

=item B<0> RW-Instance

=item B<1> RO-Instance

=item B<2> Backup-Instance

=back

Type-ID-conversion is done via the methods instanceTypeToId() and instanceId2Type() .

=head2 Volume Reference

A volume reference is a hash(ref) containg these fields:

 B<locked>          True if the volume is locked
 B<id>              ID of the volume and the RW instance
 B<roid>            The ID of this volumes RO-instances
 B<backupid>        The ID of this volumes Backup-instance
 B<name>            The volume's name
 B<synctry_backup>  The last time (epoch), this volume's backup instance would've been 
                 synced with the RW but wasn't because it was already up-to-date
                 This value is useful to test, if an automated backup-job is working
                 correctly.
 B<synctry_ro>      The last time (epoch), this volume's RO instances would've been
                 synced with the RW but were not because they were already up-to-date
                 This value is useful to test, if an automated release-job is working
                 correctly.
 B<rw>              A hash reference to the RW instance data of this volume
                 (See APPENDIX "Instance Reference" for details)
 B<ro>              An array reference to RO-instance references of this volume
                 (See APPENDIX "Instance Reference" for details)
 B<backup>          A hash reference to the backup instance data of this volume (if any)
                 (See APPENDIX "Instance Reference" for details)

=head2 Instance Reference

An instance hash(ref) contains these fields:

 B<serveruuid>      The UUID of the AFS-Fileserver, this instance resides on
 B<partition>       The partition, this instance resides on
 B<instancetype>    Numericall type ID of the instance (see APPENDIX "Instance Types")
 B<id>              ID of the volume instance
 B<online>          Boolean which represents the "Online"-flag of the instance.
                 Do not use this value to determine, if the instance is
                 accessible.
                 To check, if the instance is actually accessible to AFS-Clients,
                 use I<$celldb>->instanceInfo('B<accessible>',I<$instanceref>)
                 which takes the infotime value into account and tests for
                 VLDB-Coverage.
 B<needssalvage>    If the instance is in need of salvaging
 B<files>           Number of Inodes contained in this instance (Files+Folders)
 B<size>            Contained user data in this volume instances (This doesn't
                 necessary equal the space needed on its partition). The value
                 is in Bytes.
 B<quota>           The maximum size (volume quota) of the instance in Bytes.
 B<rostate>         The RO state of an RO instance (see APPENDIX L</"RO States">)
 B<ctime>           Create time of the instance. RO- and Backup- Instances's
                 Synchronisation will always cause this value to be set
                 to now ( time() )
 B<mtime>           Time of last change to the instance
 B<atime>           Time of the last access to this instance.
 B<infotime>        Time of last update of the volume instance's partition's
                 information in SQL DB.
 B<partfree>        Free space on the partition, the instance resides on
 B<partsize>        Size of the partition, the instance resides on
 B<quotasum>        Sum of quota assigned to all volume instances on the
                 instances partition
 B<quotasumrw>        Sum of all RW instances quota on the partition
 B<sizesumrw>        Sum of all RW instances sizes
 B<rostate>         Despite its name, this is relevant for RW instances as well.
                 Possible states are:
                 B<undef> the instances exists only physical but not in VLDB
                 B<0> for ROs: the instance is marked a "new release".
                      for RWs: the instance exists physically
                 B<1> the instance is marked an "old release"
                 B<2> the instance hasn't been released yes
                      (doesn't exist physically but in VLDB)
 
All storage values are given in Bytes. They are always multiples of 4k since
AFS uses 4k Blocks as management units.  All time values are given in epoch.

=head2 RO States

An instance's RO state is a numerical value which primarily makes sense for RO instances.
For all other volume instance types, this value is alway B<0>
Possible values are:

=over 4

=item B<0>

The instance was successfully synced with the RW instance during the last "vos release"

=item B<1>

The instance is an old release which usually means, there are more recent
RO instances of the associated volume available.

=item B<2>

The instance hasn't been synced, yet. This is normal directly after the
RO was added using "vos addsite"

=back

This value does not make any statement about the date in the RO being
up-to-date (compared to the RW instance's content). You should use this method for that:

 $celldb->volumeInfo('rouptodate',$volumereference)

Since this value is read from VLDB, it being B<undef> states that there's no VLDB-record
about a given instance.

=head2 Filter IDs

This is a reference about filter node IDs used by parseVolumeFilter() and filterVolumes().
Every node may contain an explicit negation ( {'neg' => 1} ). If a node contains subnodes,
they are included as an ordered list in {'sub' => [...]}.

=over 4

=item B<1>

Constant true.

=item B<2>

This node matches, if all the subnodes match (AND). Short-circuit evaluation is being used.

=item B<3>

This node matches, if at least one subnode matches (OR). Short-circuit evaluation is being used.

=item B<9> "volume list"

Matches any volume that is included in the hash %{$ref->{'value'}}. This rule is used to for volume
checks when a preselected list of volumes is to be tested against further constraints.

=item B<10>

Test for a given volume pattern. Volume names will be tested against the given PCRE.

=item B<11> (location)

Test, if the instance type defined by {'type'} is located in a location represented in
the location hash tree in {'value'}. If {'vldb'} is B<1>, the constraint is tested against
VLDB data.

=item B<12> (rw_count, ro_count, backup_count, count) 

Tests the number of existing instances. Non-released instances are counted.

=item B<13> volumeInfo-test

Call a $celldb->volumeInfo() subfunction and use the result as test.

=item B<14> (rw_errors, ro_errors, backup_errors, errors)

Tests, if *any* of the given errors (given has keys in the hashref in
{'value'}) can be found on the specified volume instances.  If the filter
should look for any known error, {'any'} has to be B<1>.

=item B<15> (rwonline, roonline, backuponline, online, vldbrwonline, vldbroonline, vldbonline)

Tests the number of volume instances online (in the given category or all volume instances).
The B<vldb> prefix limits testing to volume instances known to VLDB.

=item B<16>

Tests, if the given number of hosts partitions of instances of the given volume are online. 

=item B<17> (redundancy)

Tests the number of independent instances.

=item B<19> (files)

Tests the number of files in the RW instance.

=item B<20> (quota)

Tests the quota of the RW instance. The flag {'relative' => 1} is used to do
a test against a percentage of the rw instance's size.

=item B<21> (size)

Tests the current size of the RW instance. The flag {'relative' => 1} is
used to do a test against a percentage of the rw instance's quota.

=item B<22> (rw_free, free)

Tests the usable space of the RW instance. This includes testing for free quota and
free space on the partition.

=item B<23> (mtime, ctime, atime, mdiff, cdiff, adiff, ro_... backup_...)

Tests the given timestamp. Testing an RO timestamp will use an implicit ALL-quantifier.

=item B<24> (rosync, backupsync)

Tests, if the Backup- or RO-Instances are in sync. Testing applied to RO-instances
implies an ALL-quantifier.

=item B<25> (backupage, roage)

Tests for the time difference to the last backup/ro sync or sync try. This is actually a test
of the sync-process of the cell.

=item B<26> (partitionfree, rwpartitionfree, ropartitionsfree, backuppartitionfree)

Tests for free space on instance's partitions.

=item B<27> (locationcount)

Tests for the number if locations, instances of this volume reside on.

=back

=cut

1;
