package AFS::Instant::Cred;

use 5.006;
use strict;
use warnings;
use Carp;
use Data::Dumper;

my $libauthenkrb5=0;
eval "use Authen::Krb5;";
if ( $@ eq '' ) {
	$libauthenkrb5=1;
}

use Fb::Out;
use Fb::Data qw(bash_quote max Dump);
use Fb::IO qw(iaexec);

use AFS::Instant::Toolz qw(getCellName getRealmName k4name);
use AFS::Instant::Distri;

use Expect;
use POSIX ":sys_wait_h";
use Time::Local;
use File::Temp qw(tempfile);
use filetest 'access';

require Exporter;
our @ISA=qw(Exporter);


=head1 NAME

AFS::Instant::Cred - Gathers information about the current user's credentials

=head1 SYNOPSIS

use AFS::Instant::Cred;

my $credentials=AFS::Instant::Cred->new();

$credentials->fetchKerberosTgt('admin','mycell.com','password' => 'secret_password')

if ( $credentials->getTokenTime('mycell.com') > 0 ) {
   print ($credentials->getTokenTime('mycell.com')) . " seconds left.\n";
} else {
   print STDERR "No (valid) ticket for 'mycell.com'.\n";
}

=head1 REQUIRES

AFS::Instant::Toolz, AFS::KTC_PRINCIPAL, AFS::KTC_TOKEN

=head1 DESCRIPTION

This class represents credentials of the current user which covers Kerberos
tickets and AFS-tokens.

=head1 METHODS

=head2 Constructor

=over 4

=item $credentials=new(%options)

Creates a new object. Some options may be given:

=over 4

=item B<maxinfoage>

Override default maximum tolerable age of cell data (in seconds).
Default is B<2> seconds.

=item B<realmname>

Overrides the default realmname as provided by Toolz.lm

=item B<cellname>

Overrides the default cellname as provided by Toolz.pm.

=item B<ccfile> = I<filename>

Override default credential cache file (the one given in $KRB5CCNAME)

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self_={
		maxinfoage => 2,
		ticket_cache => {},
		token_cache => {},
		token_infotime => 0,
		ticket_infotime => 0,
		tgtservice => undef,
		tgtprincipal => undef,
		tgtrealm => undef,
		cellname => undef,
		realmname => undef,
		defaultlifetime => 26*60*60,
		kinit => '/usr/bin/kinit',
		klist => '/usr/bin/klist',
		kdestroy => '/usr/bin/kdestroy',
		tokens => '/usr/bin/tokens',
		aklog => "/usr/bin/aklog",
		ktutil => '/usr/bin/ktutil',
		unlog => '/usr/bin/unlog',
		uselib => $libauthenkrb5,
		ftest => 0,
		# Possible modes:
		#  'untouched' : CC was externally initialized, time bias is 60
		#  'renewed' : TGT was renewed, Tokens were re-fetched, time bias is 120
		#  'destroyed' : TGT and tokens were destroy, time bias is 180
		#  'new' : TGT was re-initialized via virtual kinit, tokens were re-fetched, time bias is 240
		#          One additional token was fetched as well for cell "more.$defaultcell"
		ftest_mode => undef,
	};
	
	my $self=bless($self_,$package);
	if ( $options{ftest} ) {
		out('i',"Cred.pm: Switching to ftesting");
		$self->{ftest}=1;
		if ( defined($options{ftest_basetime}) ) {
			$self->{ftest_basetime}=$options{ftest_basetime};
		} else {
			# We use the current time as default. Since we fake time(),
			# the time being left on tickets/tokens is constant anyway
			$self->{ftest_basetime}=time();
		}
		if ( defined($options{ftest_principal}) ) {
			$self->{ftest_principal}=$options{ftest_principal};
		} else {
			$self->{ftest_principal}="exampleuser";
		}
		$self->{ftest_time}=$self->{ftest_basetime} + 60; # The current virtual time in the ftest
		$self->{ftest_mode}='untouched';
		$self->{ccfile}="/tmp/krb5cc_ftestfake";
	} elsif ( $options{ccfile} ) {
		$self->{ccfile}=$options{ccfile};
	} else {
		if ( ! defined($ENV{'KRB5CCNAME'}) ) {
			$self->{ccfile}=undef;
		} elsif ( $ENV{'KRB5CCNAME'} =~ /^FILE:(.*)$/ ) {
			$self->{ccfile}=$1;
		} else {
			out('w',"Unusual content of \$KRB5CCNAME '$ENV{'KRB5CCNAME'}', falling back to 'donttouch'");
			$self->{ccfile}=undef;
		}
	}
	if ( defined($options{maxinfoage}) ) {
		$self->{maxinfoage}=$options{maxinfoage};
	}
	if ( defined($options{'cellname'}) ) {
		$self->{cellname}=$options{'cellname'};
	} else {
		$self->{cellname}=getCellName();
	}
	if ( defined($options{'realmname'}) ) {
		$self->{realmname}=$options{'realmname'};
	} else {
		$self->{realmname}=getRealmName() || return undef;
	}
	return $self;
}


=item $filename=$cred->getCCFile()

Returns the CC file name associated to this I<$cred>. This is
the KRB5CCNAME by default but can be overriden on object
instantiation.

=cut

sub getCCFile {
	my $self=$_[0];
	if ( defined($self->{ccfile}) ) {
		return $self->{ccfile};
	} elsif ( defined($ENV{'KRB5CCNAME'}) ) {
		if ( $ENV{'KRB5CCNAME'} =~ /^FILE:(.*)$/ ) {
			return $1;
		} else {
			out('w',"Unusual content of \$KRB5CCNAME.");
			out('d'," Content is '$ENV{KRB5CCNAME}'");
			return undef;
		}
	}
	return "/tmp/krb5cc_" . $>;
}

=item $result=$cred->krbWrap(\&call);

Call I<&call()> within I<$cred>'s kerberos context.

I<$result> is the result of the call.

=cut

sub krbWrap {
	my ($self,$call)=@_;
	my $switch=1;
	out('dd',"Kerberos wrapper called. File is: '" . $self->getCCFile() . "'");
	if ( ! defined($self->{ccfile}) ) {
		$switch=0;
	}
	my $ccbackup;
	if ( $switch ) {
		$ccbackup=$ENV{'KRB5CCNAME'};
		if ( defined($self->{ccfile}) ) {
			$ENV{'KRB5CCNAME'}="FILE:" . $self->{ccfile};
		} else {
			delete $ENV{'KRB5CCNAME'};
		}
	}
	my $result=&{$call}();

	if ( $switch ) {
		if ( ! defined($ccbackup) ) {
			delete $ENV{'KRB5CCNAME'};
		} else {
			$ENV{'KRB5CCNAME'}=$ccbackup;
		}
	}
	return $result;
}


=item $result=$cred->krbExec($cmdline,$mode,%options)

This method is a wrapper for I<Fb::IO::iaexec()> which will set the kerberos environment
according to the I<$cred> object.

It work like this:

I<$result>=iaexec($cmdline,$mode,%options);

=cut

sub krbExec {
	my ($self,$cmdline,$mode,%options)=@_;
	return $self->krbWrap(sub { iaexec($cmdline,$mode,%options) });
}

=item $result=$cred->aklog()

Calls aklog wrapped in I<$cred>'s Kerberos context.

I<$result> == B<1>, only i aklog was called successfully.

=cut

sub aklog {
	my $self=$_[0];
	my @cmd=($self->{aklog},"-c",$self->{cellname},"-k",$self->{realmname});
	if ( $self->{ftest} ) {
		out('i',"Faking aklog call");
		out('i',"Call is: " . join(' ',@cmd));
		$self->{tokens_infotime}=0;
		return 1;
	}
	return $self->krbExec(
		\@cmd,
		'exec',
		description => "Getting AFS tokens",
	);
}


=head2 Gathering information

=over 4

=item \%ticketinfo=$cred->getTicket($service);

Returns information about a given ticket. I<$service> must be a full
service name or a service name without the realm part in which case
the default realm is being used.

I<\%ticketinfo> is a hashref containing information about the ticket.

=cut

sub getTicket {
	my ($self,$service)=@_;
	$self->readTicketCache() || return undef;
	if ( $service !~ /\@/ ) {
		$service="$service\@$self->{realmname}";
	}
	return $self->{ticket_cache}->{$service};
}

=item \%ticketinfo=$cred->getTgt()

Returns the I<%ticketinfo>-hash(ref) associated with the TGT
of I<$cred>'s kerberos context.

=cut

sub getTgt {
	my $self=shift;
	$self->readTicketCache() || return undef;
	if ( ! defined($self->{tgtservice}) ) {
		return undef;
	}
	return $self->{ticket_cache}->{$self->{tgtservice}};
}

=item \@servicelist=$credentials->listTickets();

Returns names of service tickets contained in current CC.

=cut

sub listTickets {
	my $self=shift;
	$self->readTicketCache() || return undef;
	my @names=sort keys %{$self->{ticket_cache}};
	return \@names;
}

=item [$principal,$realm]=$cred->getOwner()

Returns the I<$principal> name and the I<$realm> for whom the
TGT was issued.

=cut

sub getOwner {
	my $self=shift;
	$self->readTicketCache() || return undef;
	return [$self->{tgtprincipal},$self->{tgtrealm}];
}

=item $result=$cred->readTicketCache()

Reads the CC of the I<$cred>'s context. This method is an abstraction
which will

=over 4

=item Read the ticket cache via L<Authent::Krb5>

=item Read the ticket cache via calling B<klist>

=item Read fake data for functional testing.

=back

=cut

sub readTicketCache {
	my $self=$_[0];
	if ( ( $self->_virtualTime() - $self->{ticket_infotime} ) <= $self->{maxinfoage} ) {
		return 1;
	}
	if ( $self->{ftest} ) {
		return $self->_readTicketCacheFake()
	} elsif ( $self->{uselib} ) {
		return $self->krbWrap(sub { $self->readTicketCacheViaLib() });
	} else {
		return $self->krbWrap(sub { $self->readTicketCacheViaCmd() });
	}
}

=item $result=$cred->_readTicketCacheFake()

Puts fake ticket data into the cache and returns success.
This method is not to be called directly. Use L<readTicketCache()> instead.

=cut


sub _readTicketCacheFake {
	my $self=$_[0];
	$self->{ticket_cache}={};
	my $user=$self->{ftest_principal};
	my $tgtservice="krbtgt/$self->{realmname}\@$self->{realmname}";
	my $afsservice="afs/$self->{cellname}\@$self->{realmname}";
	my $someservice="host/somehost.somedomain\@$self->{realmname}";
	if ( $self->{ftest_mode} eq 'untouched' ) {
		$self->{tgtprincipal}=$user;
		$self->{tgtrealm}=$self->{realmname};
		$self->{tgtservice}=$tgtservice;
		$self->{ticket_cache}={
			$tgtservice => {
				realm => $self->{realmname},
				service => "krbtgt/$self->{realmname}",
				start => $self->{ftest_basetime},
				end => $self->{ftest_basetime} + 24*3600,
				renew => $self->{ftest_basetime} + 3*24*3600,
			},
			$afsservice => {
				realm => $self->{realmname},
				service => "afs/$self->{cellname}",
				start => $self->{ftest_basetime},
				end => $self->{ftest_basetime} + 24*3600,
				renew => $self->{ftest_basetime} + 3*24*3600, # doesn't matter
			},
			$someservice => {
				realm => $self->{realmname},
				service => "host/somehost.somedomain",
				start => $self->{ftest_basetime},
				end => $self->{ftest_basetime} + 3*3600,
				renew => $self->{ftest_basetime} + 6*3600, # doesn't matter
			},
		};
		$self->{ticket_infotime}=$self->{ftest_basetime}+60;
	} elsif ( $self->{ftest_mode} eq 'renewed' ) {
		$self->{tgtprincipal}=$user;
		$self->{tgtrealm}=$self->{realmname};
		$self->{tgtservice}=$tgtservice;
		$self->{ticket_cache}={
			$tgtservice => {
				realm => $self->{realmname},
				service => "krbtgt/$self->{realmname}",
				start => $self->{ftest_basetime} + 120,
				end => $self->{ftest_basetime} + 24*3600 + 120,
				renew => $self->{ftest_basetime} + 3*24*3600,
			},
			$afsservice => {
				realm => $self->{realmname},
				service => "afs/$self->{cellname}",
				start => $self->{ftest_basetime} + 120,
				end => $self->{ftest_basetime} + 24*3600 + 120,
				renew => $self->{ftest_basetime} + 3*24*3600, # doesn't matter
			},
		};
		$self->{ticket_infotime}=$self->{ftest_basetime}+120;
	} elsif ( $self->{ftest_mode} eq 'destroyed' ) {
		$self->{tgtprincipal}=undef;
		$self->{tgtrealm}=undef;
		$self->{tgtservice}=undef;
		$self->{ticket_infotime}=$self->{ftest_basetime}+180;
	} elsif ( $self->{ftest_mode} eq 'new' ) {
		$self->{tgtprincipal}=$user;
		$self->{tgtrealm}=$self->{realmname};
		$self->{tgtservice}=$tgtservice;
		$self->{ticket_cache}={
			$tgtservice => {
				realm => $self->{realmname},
				service => "krbtgt/$self->{realmname}",
				start => $self->{ftest_basetime} + 240,
				end => $self->{ftest_basetime} + 24*3600 + 240,
				renew => $self->{ftest_basetime} + 3*24*3600 + 240,
			},
			$afsservice => {
				realm => $self->{realmname},
				service => "afs/$self->{cellname}",
				start => $self->{ftest_basetime} + 240,
				end => $self->{ftest_basetime} + 24*3600 + 240,
				renew => $self->{ftest_basetime} + 3*24*3600 + 240, # doesn't matter
			},
		};
		$self->{ticket_infotime}=$self->{ftest_basetime}+240;
	}	
	return 1;
}

sub readTicketCacheViaLib {
	my $self=shift;
	my $a=Authen::Krb5::init_context() || return 0;
	my $cc=Authen::Krb5::cc_default();
	my $c=$cc->start_seq_get();
	$self->{tgtprincipal}=undef;
	$self->{tgtrealm}=undef;
	$self->{ticket_infotime}=time(); # That's OK since it's never called when ftest is in use
	$self->{ticket_cache}={};
	my $first=1;
	while ( my $t=$cc->next_cred($c) ) {
		my $service=$t->server();
		my ($principal,$realm)=split /\@/,$service;
		my %ticket=(
			start => $t->starttime(),
			end => $t->endtime(),
			renew => $t->renew_till(),
			service => $principal,
			realm => $realm,
		);
		$self->{ticket_cache}->{$service}=\%ticket;
		if ( $first ) {
			my ($principal,$realm_)=split /\@/,$t->client();
			$self->{tgtprincipal}=$principal;
			$self->{tgtrealm}=$realm_;
			$self->{tgtservice}=$service;
			$first=0;
		}
	}
	$cc->end_seq_get($c);
	Authen::Krb5::free_context();
	return 1;
}

=item $result=$credentials->readTicketCacheViaCmd()

Reads tickets from current CC via klist

=cut

sub readTicketCacheViaCmd {
	my $self=shift;
	my %flags=(
		'F' => 'forwardable',
		'f' => 'forwarded',
		'P' => 'proxiable',
		'I' => 'initial',
		'R' => 'renewable'
	);
	$self->{tgtprincipal}=undef;
	$self->{tgtrealm}=undef;
	$self->{ticket_infotime}=time(); # That's OK since it's never called when ftest is in use
	$self->{ticket_cache}={};
	my $old_lc=$ENV{LC_ALL};
	$ENV{LC_ALL}='C';
	# A call to this method is already wrapped - krbExec isn't necessary
	my $output=iaexec(
		[$self->{'klist'},'-fe'],
		'readpipe',
		description => "Reading local kerberos ticket cache",
		format => 'arrayref',
		stderr => 'hide',
	);
	if ( defined($old_lc) ) {
		$ENV{LC_ALL}=$old_lc;
	}
	if ( ! $output ) {
		return 0;
	}
	push(@{$output},'');
	my $first=1;
		my $ticket={};	
		for my $l (@{$output}) {
			my $found=0;
			if ( ( $l eq '' ) or exists($ticket->{keyinfo}) ) {
				if ( defined($ticket->{'service'}) ) {
					$self->{ticket_cache}->{"$ticket->{'service'}\@$ticket->{'realm'}"}=$ticket;
					if ( $first ) {
						$self->{tgtservice}="$ticket->{'service'}\@$ticket->{'realm'}";
						$first=0;
					}
					$ticket={};
				}
			}
			if ( ( $l =~ /^(Ticket|Valid)/ ) or ( $l eq '' ) ) {
				# ignored
				$found=1;
			}
			if ( $l =~ /^Default principal: ([^\@]+)\@(\S+)$/ ) {
				$found=1;
				$self->{tgtprincipal}=$1;
				$self->{tgtrealm}=$2;
			}
			if ( $l =~ /([0-9\/]+ [0-9:]+)\s+([0-9\/]+ [0-9:]+)\s+(\S+)\@(\S+)/ ) {
				$found=1;
				$ticket={};
				$ticket->{start}=$self->klist_str2time($1);
				$ticket->{end}=$self->klist_str2time($2);
				$ticket->{service}=$3;
				$ticket->{realm}=$4;
			}
			if ( $l =~ /renew until ([0-9\/]+ [0-9:]+)/ ) {
				$found=1;
				$ticket->{'renew'}=$self->klist_str2time($1);
			}
			if ( $l =~ /Flags: (\S+)/ ) {
				$found=1;
				my $flags=$1;
				$ticket->{flags}={};
				for(my $n=0;$n < length($flags);$n++) {
					my $flag=substr($flags,$n,1);
					if ( defined($flags{$flag}) ) {
						$ticket->{'flags'}->{$flags{$flag}}=1;
					}
				}
			}
			if ( $l =~ /Etype \(skey, tkt\): ([^ ]+), ([^ ]+)/ ) {
				my ($keymode,$ticketmode)=($1,$2,$3);
				$ticket->{'keyinfo'}=$keymode;
				$ticket->{'ticketinfo'}=$ticketmode;
			}
		}
	return 1;	
}

=item $time=$credentials->$klist_str2time($string);

Converts a string as supplied by B<klist> to an epoch time value.

=cut

sub klist_str2time {
	my ($self,$string)=@_;
	my $result=-1;
	if ( $string =~ /^([0-9]+)\/([0-9]+)\/([0-9]+) ([0-9]+):([0-9]+):([0-9]+)$/ ) {
		$result=timelocal($6,$5,$4,$2,$1-1,$3);
	}
	return $result;
}


=item \%tokeninfo=$credentials->getToken($cellname);

Returns information about a token, accessible to the current process. I<$cellname>
defines, which AFS-cell's token to look for.

Returns a hashref I<\%tokeninto> containing these fields as key/value pairs:

=over 4

=item B<end>

An epoch value of the last second of thi token's validity time

=item B<afsid>

The numerical AFSID assigned to this token

=item B<cellname>

The name of the cell, this token is assigned to.

=back

A value of B<undef> is returned if no token exists for the given cell.

=cut

sub getToken {
	my ($self,%options)=@_;
	my $cellname=$self->{cellname};
	if ( $options{$cellname} ) {
		$cellname=$options{$cellname};
	}
	$self->readTokens() || return undef;
	if ( defined($self->{token_cache}->{$cellname}) ) {
		return $self->{token_cache}->{$cellname};
	} else {
		return undef;
	}
}

sub listTokens {
	my $self=$_[0];
	$self->readTokens() || return undef;
	return [sort keys %{$self->{token_cache}}];
}

=item $seconds=$cred->getTokenTimeLeft(%options);

Returns the time in I<$seconds> left for the token of
the default cell. If there's no token for the cell, B<0> is
returned. On error B<undef> is returned. Possible I<%options>
are:

=over 4

=item B<cellname> => I<$cellname>

Check for tokens of cell I<$cellname> instead of the default cell.

=back

=cut

sub getTokenTimeLeft {
	my ($self,%options)=@_;
	my $cellname=$self->{cellname};
	if ( $options{cellname} ) {
		$cellname=$options{cellname};
	}
	$self->readTokens() || return undef;
	if ( defined($self->{token_cache}->{$cellname}) ) {
		
		return $self->_timeLeft($self->{token_cache}->{$cellname});
	}
	return 0;
}


=item $seconds=$cred->getTgtTimeLeft(%options)

Returns the number of seconds left on the TGT. Possible I<%options>
are:

=over 4

=item B<includerenew> => B<1>

Returns the "possible time" left (which means, incl. max renew
time).

=back

B<0> is returned, if there's no TGT.

=cut

sub getTgtTimeLeft {
	my ($self,%options)=@_;
	$self->readTicketCache() || return undef;
	if ( ! $self->{tgtservice} ) {
		return 0;
	}
	my $ticket=$self->{ticket_cache}->{$self->{tgtservice}} || return 0;
	return $self->_timeLeft($ticket,%options);
}



=item $seconds=$cred->getTicketTimeLeft(%options)

Returns the time left on the TGT. Possible I<$options> are:

=over 4

=item B<service> => I<$servicename>

The name of a service to be checked instead of the TGT. I<$servicename>
must either be the full principal name or or the name with
the realm part (in which case the default realm is assumed).

=item B<includerenew> => B<1>

Returns the "possible time" left (which means, incl. max renew
time).

=back

=cut

sub getTicketTimeLeft {
	my ($self,%options)=@_;
	$self->readTicketCache() || return undef;
	my $service=$options{service};
	if ( ! defined($service) ) {
		return $self->getTgtTimeLeft(includerenew => $options{includerenew});
	}
	if ( $service !~ /\@/ ) {
		$service="$service\@$self->{realmname}";
	}
	my $ticket=$self->{ticket_cache};
	if ( ! $ticket ) {
		# Ticket doesn't exist (yet)
		return 0;
	}
	return $self->_timeLeft($ticket,includerenew => $options{includerenew});
}


=item $seconds=$cred->getAfsTicketTimeLeft(%options)

Returns the number of I<$seconds> of validity left on the
the token of the default AFS cell. Possible I<$options> are:

=over 4

=item B<cellname> => I<$cellname>

Request time for tokens of the cell I<$cellname> instead of the
default cell.

=back

=cut

sub getAfsTicketTimeLeft {
	my ($self,%options)=@_;
	my $cellname=$self->{cellname};
	if ( $options{cellname} ) {
		$cellname=$options{cellname};
	}
	$self->readTicketCache() || return undef;
	my $realmname=$self->{realmname};
	for my $service (sort keys %{$self->{ticket_cache}}) {
		if ( $service eq "afs\@$realmname" || $service eq "afs/$cellname\@$realmname" ) {
			my $ticket=$self->{ticket_cache}->{$service};
			return $self->_timeLeft($ticket);
		}
	}
	return 0;
}

=item $seconds=$cred->_timeLeft(\%ticket,%options)

Abstraction for retrieving a zero-cut number of
seconds left on either the a I<%ticket> (which might be a token hashref) 's
validity or on the ticket's max renew time
(if I<$options{includerenew}> == B<1>) .

=cut

sub _timeLeft {
	my ($self,$ticket,%options)=@_;
	if ( ! defined($ticket) ) {
		return 0;
	}
	my $end=$ticket->{end};
	if ( $options{includerenew} ) {
		$end=max($end,$ticket->{renew});
	}
	my $timeleft=$end-$self->_virtualTime();
	if ( $timeleft < 0 ) {
		return 0;
	} else {
		return $timeleft;
	}
}

=item $epoch=$cred->_virtualTime();

Returns the current epoch or the one being faked by
ftesting. No L<time()> should be done in Cred.pm
outside of this method.

=cut

sub _virtualTime {
	my $self=$_[0];
	if ( $self->{ftest} ) {
		return $self->{ftest_time};
	} else {
		return time();
	}
}

=back

=head2 Reading information from kernel/disk

=over 4

=item $result=readTokens()

Reads information about tokens accessible to the current process from kernel.

I<$result> == B<1>, only if all token information were read successfully.

=cut

sub readTokens {
	my $self=shift;
	if ( ( $self->_virtualTime() - $self->{token_infotime} ) <= $self->{maxinfoage} ) {
		return 1;
	}
	if ( $self->{ftest} ) {
		if ( $self->{ftest_mode} eq 'untouched' ) {
			$self->{token_cache}->{$self->{cellname}}={
				end => $self->{ftest_basetime} + 24*3600,
				afsid => 1000,
				afs_principal => "afs",
				cellname => $self->{cellname},
			};
			$self->{token_infotime}=$self->{ftest_basetime} + 60;
		} elsif ( $self->{ftest_mode} eq 'renewed' ) {
			$self->{token_cache}->{$self->{cellname}}={
				end => $self->{ftest_basetime} + 24*3600 + 120,
				afsid => 1000,
				afs_principal => "afs",
				cellname => $self->{cellname},
			};
			$self->{token_infotime}=$self->{ftest_basetime} + 120;
		} elsif ( $self->{ftest_mode} eq 'destroyed' ) {
			$self->{token_cache}={};
			$self->{token_infotime}=$self->{ftest_basetime} + 180;
		} elsif ( $self->{ftest_mode} eq 'new' ) {
			$self->{token_cache}={
				$self->{cellname} => {
					end => $self->{ftest_basetime} + 24*3600 + 240,
					afsid => 1000,
					afs_principal => "afs",
					cellname => $self->{cellname},
				},
				"mode.$self->{cellname}" => {
					end => $self->{ftest_basetime} + 24*3600 + 240,
					afsid => 2000,
					afs_principal => "afs",
					cellname => "more.$self->{cellname}",
				},
			};
			$self->{token_infotime}=$self->{ftest_basetime} + 240;
		} else {
			out('e',"readTokens(): Invalid ftest mode '$self->{ftest_mode}'");
			$self->{token_infotime}=0;
			return 0;
		}
		return 1;
	}
	my @months=('Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec');
	my %months=();
	for(my $n=0;$n<=$#months;$n++) {
		$months{$months[$n]}=$n;
	}
	$self->{token_cache}={};
	my $tokenoutput=iaexec(
		[$self->{tokens}],
		'readpipe',
		description => "Reading tokens from kernel",
		format => 'arrayref',
	);
	if ( $tokenoutput ) {
		for my $l (@{$tokenoutput}) {
			if ( $l =~ /^(?:User's \(AFS ID ([0-9]+)\)(?: rxkad)? )?[tT]okens for (?:([^\@]+)\@)?(\S+) \[Expires ([a-zA-Z]+) +([0-9]+) ([0-9]+):([0-9]+)\]$/ ) {
				my $dummy;
				my ($afsid,$afs_principal,$cellname)=($1,$2,$3);
				if ( ! defined($afsid) ) {
					out('w',"get_tokens(): AFSID of current token is not existing. You should created it.");
				}
				if ( ! defined($afs_principal) ) {
					$afs_principal='(unknown)';
				}
				my ($min2,$hour2,$day2,$month2)=($7,$6,$5,$4);
				($dummy,my $min1,my $hour1,my $day1,my $month1,my $year1,$dummy,$dummy,$dummy) = localtime(time);
				$year1 += 1900; $month1 += 1;
				if ( defined($months{$month2}) ) {
					$month2=$months{$month2} + 1;
				} else {
					out('e',"get_tokens_commands(): Unable to parse 'tokens'-output's month-value '$month2'");
					return 0;
				}
				my $year2;
				if ( $month2 < $month1 ) {
					$year2=$year1+1;
				} else {
					$year2=$year1;
				}
				my $time=timelocal(0,$min2,$hour2,$day2,$month2-1,$year2);
				my $tokeninfo={
					end => $time,
					afsid => $afsid,
					afs_principal => $afs_principal,
					cellname => $cellname
				};
				$self->{token_cache}->{$cellname}=$tokeninfo;
			} elsif ( $l =~ /^RxkadTokens/ ) {
				# Tokens from another cell that we can't interpret
			} elsif ( $l !~ /^(|   --End of list--|Tokens held by the Cache Manager:)$/ ) {
				out('w',"get_tokens_commands(): Unknown output from 'tokens', ignored.");
				out('i',"Output was: '$l'");
			}
		}
	} else {
		return 0;
	}
	$self->{token_infotime}=time(); # That's OK since it's never reached when ftest is being used
	return 1;
}

=back

=head2 Fetching tickets/tokens

=over 4

=item $result=$credentials->fetchKerberosTgt($principal,%options)

Fetches a Kerberos5-TGT for the given I<$principal> of the
default realm. These I<%options> can be given:

=over 4

=item B<password>

Password to be used when fetching the initial TGT

=item B<keytab>

Keytab file to be used for fetching the TGT. Either principal (exclusive)or
password have to be given.

=item B<lifetime>

Requested ticket lifetime in seconds.

=back

=cut

sub fetchKerberosTgt {
	my ($self,$principal,%options)=@_;
	if ( ( defined($options{'password'}) + defined($options{'keytab'}) ) != 1 ) {
		out('e',"fetch_tgt(): Either password *x*or keytab must be given. Internal error.");
		return 0;
	}
	if ( $principal =~ /\@/ ) {
		out('e',"fetchKerberosTgt(): Principal '$principal' contains a '\@' sign.");
		return 0;
	}
	if ( $self->{uselib} && ! defined($options{lifetime}) ) {
		return $self->krbWrap(sub { $self->fetchKerberosTgtViaLib($principal,%options) });
	} else {
		return $self->fetchKerberosTgtViaShellcommand($principal,%options);
	}
}
	

sub fetchKerberosTgtViaLib {
	my ($self,$principal,%options)=@_;
	out('d',"Getting Kerberos-TGT using libauthen-krb5-perl");
	if ( $self->{ftest} ) {
		out('i',"Faking getting Kerberos-TGT via libauthen-krb5-perl");
		$self->{ticket_infotime}=0;
		$self->{ftest_time}=$self->{ftest_basetime} + 240;
		$self->{ftest_mode}='new';
		$self->{ftest_principal}=$principal;
		return 1;
	}
	my $realm=$self->{realmname};
	$a=Authen::Krb5::init_context() || return 0;
	# Make a principal object from the Principal name
	my $princ=Authen::Krb5::parse_name("$principal\@$realm");

	# Get the CC and clean it up
	my $cc=Authen::Krb5::cc_default();
	$cc->initialize($princ);

	my $result;
	my $cred;
	if ( defined($options{'keytab'}) ) {
		my $keytab=Authen::Krb5::kt_resolve($options{'keytab'});
		if ( ! defined($keytab) ) {
			out('w','Keytab $keytab not found or not accessible');
			Authen::Krb5::free_context();
			return 0;
		}
		$cred=Authen::Krb5::get_init_creds_keytab($princ,$keytab);
	} else {
		$cred=Authen::Krb5::get_init_creds_password($princ,$options{'password'});
	}
	if ( $cred ) {
		$result=$cc->store_cred($cred);
	} else {
		$result=0;
	}
	Authen::Krb5::free_context();
	if ( ! $result ) {
		my $err=Authen::Krb5::error();
		out('e',"Krb5-Error: $err");
		if ( $err eq 'Decrypt integrity check failed' ) {
			if ( defined($options{'keytab'}) ) {
				out('i',"Your keytab most probably contains a wrong, outdated or wrong versioned key");
			} else {
				out('i',"This most likely means 'Wrong password' or 'Typo in username'");
			}
		} elsif ( $err eq 'Bad format in credentials cache' ) {
			out('i',"This most likely means 'Your root partition is full'.");
		}
		return 0;
	}
	return 1;
}

sub fetchKerberosTgtViaShellcommand {
	my ($self,$principal,%options)=@_;
	my $realm=$self->{realmname};
	my $lifetime=$self->{defaultlifetime};
	if ( defined($options{'lifetime'}) ) {
		$lifetime=$options{'lifetime'};
	}

	out('d',"Getting Kerberos-TGT using kinit");
	$self->{ticket_infotime}=0;
	if ( defined($options{'keytab'}) ) {
		my @cmd=($self->{'kinit'},'-f','-P',"-l",$lifetime . "s",'-k','-t',$options{'keytab'},"$principal\@$realm");
		if ( $self->{ftest} ) {
			$self->{ftest_mode}='new';
			$self->{ftest_principal}=$principal;
		$self->{ftest_time}=$self->{ftest_basetime} + 240;
			out('i',"Faking kinit call: " . join(' ',@cmd));
			return 1;
		}
		if ( ! $self->krbExec(\@cmd,'exec',description => "Getting Kerberos-TGT using a keytab") ) {
			return 0;
		}
	} elsif ( defined($options{password}) ) {
		my @cmd=($self->{'kinit'},'-l',$lifetime . "s",'-p',"$principal\@$realm");
		if ( $self->{test} ) {
			$self->{ftest_mode}='new';
			$self->{ftest_principal}=$principal;
			$self->{ftest_time}=$self->{ftest_basetime} + 240;
			out('i',"Faking kinit call, piping password: " . join(' ',@cmd));
			return 1;
		}
		my $fh=$self->krbExec(
			\@cmd,
			'writepipe',
			description => "Getting Kerberos-TGT using a password",
			stderr => 'hide',
			stdout => 'hide',
		);
		if ( ! $fh ) {
			return 0;
		}
		print $fh "$options{password}\n";
		close($fh);
		if ( ( $? >> 8 ) != 0 ) {
			return 0;
		}
	} else {
		out('e',"fetchKerberosTgtViaShellcommand(): Neither password nor keytab option given");
		return 0;
	}
	return 1;
}

=item $result=$cred->renewTgt(%options)

Tries to renew the TGT. Possible I<%options> are:

=over 4

=item B<lifetime>

Ticket lifetime in seconds to request from the KDC.

=back

=cut

sub renewTgt {
	my ($self,%options)=@_;
	my $lifetime=$self->{defaultlifetime};
	if ( defined($options{'lifetime'}) ) {
		$lifetime=$options{'lifetime'};
	}
	$self->{ticket_infotime}=0;
	if ( $self->{ftest} ) {
		out('i',"Faking renewing TGT.");
		$self->{ftest_mode}='renewed';
		$self->{ftest_time}=$self->{ftest_basetime}+120;
		return 1;
	}
	my $result=$self->krbExec(
		[$self->{'kinit'},'-R','-l${lifetime}s'],
		'exec',
		description => "Renewing TGT via kinit -R."
	);
	out('d',"Renewing Kerberos ticket for '" . human_readable($lifetime,'timeinterval') . "'");
	return $result;
}

=item $result=$cred->switchCC(%options)

Creates a new Credentials cache file and sets the environment variable
accordingly.

Warning: This method affects the whole process, not just the I<$self>-object.

Possible I<%options> are:

=over 4

=item B<cctemplate>

defines, which template to use for the CC-file. Possible values are:

=over 4

=item B<default> (default)

Use B<krb5_cc_>I<$uid>_I<...> which is default for Kerberos 5.

=item B<ia>

Use instantafs specific template which is explicitely non-default.
This is will make gssd skip these CCs which enables NFS with GSSAPI
while multiple AFS identities are being used.

=back

=back

=cut

sub switchCC {
	my ($self,%options)=@_;
	if ( $self->{ftest} ) {
		out('i',"Faking creating new CC for not operating on the original one.");
		$self->{ccfile}="/tmp/krb5cc_1000_fake_switchCC";
		return 1;
	}
	my $newcc;my $fh;
	my $euid=$>;
	my $template;
	my $dir='/tmp';
	if ( ! defined($options{cctemplate}) || $options{cctemplate} eq 'default' ) {
		$template="krb5cc_${euid}_XXXX";
	} elsif ( $options{cctemplate} eq 'ia' ) {
		$template="ia-krb5cc_${euid}_XXXX";
		my $test="/run/user/$<";
		if ( -w $test ) {
			$dir=$test;
		}
	} else {
		out('e',"switchCC(): Invalid 'cctemplate' option given: '$options{template}'");
		return 0;
	}
	($fh,$newcc)=tempfile($template,DIR => $dir);
	if ( ! $newcc ) {
		out('e',"Unable to create temporary cc-file.");
		return 0;
	}
	close($fh);
	out('d',"Switching Kerberos CC to '$newcc'");
	$self->{ccfile}=$newcc;
	return 1;
}

=back

=head2 Removing credentials

=over 4

=item $result=$credentials->removeTickets()

Removes the current kerberos CC.

=cut

sub removeTickets {
	my ($self,%options)=@_;
	my $cc=$self->getCCFile() || return 0;
	$self->{ticket_infotime}=0;
	if ( $options{dryrun} ) {
		out('i',"Would've destroyed Kerberos CC in '$cc' now");
		return 1;
	}
	if ( $self->{ftest} ) {
		out('i',"Faking destroying CC '$cc'");
		$self->{ftest_mode}='destroyed';
		$self->{ftest_time}=$self->{ftest_basetime} + 180;
		return 1;
	}
	out('d',"Destroying Kerberos CC '$cc'");
	my $result=$self->krbExec([$self->{'kdestroy'}],'exec',description => "Cleaning up Kerberos5-CC");
	return $result;
}



=item $result=$cred->writeKeytab($user,$password,$filename,%options)

Creates a keytab I<$filename> and puts keys for I<$user> in it
(which are derived from I<$password>). Possible I<%options> are:

=over 4

=item B<chown>

If defined: change owner of the keytab file to the given I<$user>.

=back

I<$result> == B<1>, only if no error occured.

=cut

sub writeKeytab {
	my ($self,$filename,$passwd,$kvno,$user,%options)=@_;
	if ( ! defined($options{encryption}) ) {
		out('e',"writeKeytab(): no encryption option given.");
		return 0;
	}
	my $filenameq=bash_quote($filename);
	my $userq=quotemeta($user);
	# Just in case the keytab existed before. Wa can't rely on
	# ktutil overwriting it - ktutil will always append new keys
	unlink($filename);

	my $success=0;
	out('v',$self->{ktutil},"Generating a user keytab using expect");
	my $ktutil=Expect->spawn($self->{ktutil});
	$Expect::Log_Stdout=0;
	if ( $ktutil->expect(1,"ktutil:") ) {
		print $ktutil "add_entry -password -p $user -k $kvno -e $options{encryption}\n";
		if ( $ktutil->expect(1,"Password for ") ) {
			print $ktutil "$passwd\n";
			if ( $ktutil->expect(1,"ktutil:") ) {
				print $ktutil "write_kt $filename\n";
				if ( $ktutil->expect(2,"ktutil:") ) {
					print $ktutil "exit\n";
					if ( defined($options{'chown'}) ) {
						fexec("chown $userq $filenameq","Changing owner of new keytab file");
					}
					$success=1;
				} else {
					out('e',"ktutil-session: no prompt after write_kt.");
				}
			} else {
				out('e',"ktutil-session: no prompt after password was entered.");
			}
		} else {
			out('e',"ktutil-session: no password-prompt.");
		}
	} else {
		out('e',"ktutil-session: no prompt after program start.");
	}
	$ktutil->soft_close();
	return $success;
}

=item \%principalrec=$cred->parseK5Principal($principal)

Parses the given I<$principal> string and returns the given
I<%principalrec> hashref which contains these fields as key/value pairs:

=over 4

=item B<principal>

The principal name.

=item B<instance>

The principal's instance. If there's no instance, the value is B<undef>.

=item B<realm>

The realm. If no realm is given in I<$principal>, uc($self->{cellname}) is returned.

=back

=cut

sub parseK5Principal {
	my ($self,$text)=@_;
	if ( $text !~ /^([^\/\@]+)(?:\/([^\@]+))?(?:\@(\S+))?$/ ) {
		out('e',"Invalid K5 principal syntax '$text'");
		return undef;
	}
	my $principal=$1;
	my $instance=$2;
	my $realm;
	if ( defined($3) ) {
		$realm=uc($3);
	} else {
		$realm=uc($self->{cellname});
	}
	my %rec=(
		'principal' => $principal,
		'instance' => $instance,
		'realm' => $realm,
	);
	return \%rec;
}

=item \%principalrec=$cred->normalizeK5Principall($principal)

Returns a normalized K5-principal expression of I<$principal> in I<$nprinc>
(always containing a REALM part).

=cut

sub normalizeK5Principal {
	my ($self,$principal)=@_;
	my $rec=$self->parseK5Principal($principal);
	if ( ! defined($rec) ) {
		return undef;
	}
	if ( ! defined($rec->{'instance'}) ) {
		return "$rec->{'principal'}\@$rec->{'realm'}";
	} else {
		return "$rec->{'principal'}/$rec->{'instance'}\@$rec->{'realm'}";
	}
}



=item $filename=$cred->getKeytabFilename($principal)

Determines, which keytab file to use. $principal is a K5-principal
which might contain a REALM part.

I<$filename> the first keytab file which was found and readable.

Warning: There's no check on the contained key itself.

See B<Finding Keytabs> in the APPENDIX section for more details.

=cut

sub getKeytabFilename {
	my ($self,$principal)=@_;
	my $defaultrealm=uc($self->{cellname});
	my $princrec=$self->parseK5Principal($principal);
	my $nprinc4=k4name($self->normalizeK5Principal($principal));
	if ( $self->{ftest} ) {
		return "/etc/instantafs/ftest/$nprinc4.keytab";
	}
	my $pi4=$princrec->{'principal'};
	my $pi5=$princrec->{'principal'};
	if ( defined($princrec->{'instance'}) ) {
		$pi4=$princrec->{'principal'} . "." . $princrec->{'instance'};
		$pi5=$princrec->{'principal'} . "/" . $princrec->{'instance'};
	}
	$pi5 .= '@' . $princrec->{'realm'};
	out('d',"Looking for keytab for '$pi5' (K4='$nprinc4',defaultrealm='$defaultrealm')");
	if ( defined($ENV{'HOME'}) && -r "$ENV{'HOME'}/.instantafs/$nprinc4.keytab" ) {
		out('d',"Found keytab in current user's homedir");
		return "$ENV{'HOME'}/.instantafs/$nprinc4.keytab";
	} elsif ( defined($ENV{'HOME'}) && ( $princrec->{'realm'} eq $defaultrealm ) and ( -r "$ENV{'HOME'}/.instantafs/$pi4.keytab" ) ) {
		out('d',"Found keytab in current user's homedir (norealm)");
		return "$ENV{'HOME'}/.instantafs/$pi4.keytab";
	} elsif ( -r "/etc/instantafs/$nprinc4.keytab" ) {
		out('d',"Found keytab in /etc/instantafs");
		return "/etc/instantafs/$nprinc4.keytab";
	} elsif ( ( $princrec->{'realm'} eq $defaultrealm ) and ( -e "/etc/instantafs/$pi4.keytab" ) ) {
		return "/etc/instantafs/$pi4.keytab";
	}
	out('d',"Looked for '$principal' keytab in '$ENV{'HOME'}/.instantafs' and '/etc/instantafs', no success.");
	return undef;
}


=head1 SEE ALSO

B<instantafs>

=head1 AUTHOR

Frank Burkhardt <burk@cbs.mpg.de>

=cut

1;
__END__
