package AFS::Instant::Distri;

use 5.006;
use strict;
use warnings;
use Carp;

use Fb::Out qw(out);
use Fb::Data qw(bash_quote);

require Exporter;

use vars qw(@ISA $VERSION);

our @ISA=qw(Exporter);
our $VERSION='3.0';

our @EXPORT_OK=qw(
	get_distri
	get_distri_name
	get_distri_version
	get_kernel
	packages_installed
	packages_installed_version
	is_pag
);	

=head1 NAME

AFS::Instant::Distri - Functions related to special Linux distributions or Operating Systems

=head1 SYNOPSIS

 my $distri=get_distri_name();
 my $version=get_distri_version();
 my $distrifull=get_distri();
 my $ok=packages_installed(0,'samba');
 my $ok=packages_installed_version(1,'samba' => '>= 3');

=head1 DESCRIPTION

This module contains functions for detecting the local Linux distribution
and installed packages.

=head1 FUNCTIONS

=head2 Distribution detection

=over 4

=cut

my $local_distribution;
my $local_distribution_version;
my $local_kernel;

=item $distri=get_distri_name()

Returns the name of the local Computer's Linux distribution.

I<$distri> eq 'B<UNKNOWN>' whenever the distribution can not be
detected.

=cut



sub get_distri_name {
	if ( defined($local_distribution) ) {
		return $local_distribution;
	}
	if ( -e '/etc/debian_version' ) {
		$local_distribution='debian';
		if ( open(I,'< /etc/debian_version') ) {
			$local_distribution_version=<I>;
			chomp($local_distribution_version);
			close(I);
			if ( $local_distribution_version eq 'testing/unstable' ) {
				# unstable before etch being stable
				$local_distribution_version=4.0;
			} elsif ( $local_distribution_version =~ /^\S+\/sid$/ ) {
				$local_distribution_version=999;
			} elsif ( $local_distribution_version =~ /^([0-9]+\.[0-9]+)/ ) {
				$local_distribution_version=$1 + 0;
			} else {
				out('e',"Detected distribution 'debian', but version doesn't seem to be valid.");
				$local_distribution_version=0;
			}
		}
	} elsif ( -d '/Library/StartupItems' ) {
	  $local_distribution='macosx';
	  $local_distribution_version='10.x';
	} else {
		$local_distribution='UNKNOWN';
	}
	return $local_distribution;
}

=item $version=get_distri_version()

Returns the version number of the local Linux distribution.

I<$version> eq 'B<0>' whenever the distribution can not be detected.

=cut

sub get_distri_version {
	if ( ! defined($local_distribution) ) {
		get_distri();
	}
	return $local_distribution_version;
}

=item $fulldistri=get_distri()

Returns the name and version number of the local Linux distribution.

I<$fulldistri> eq 'UNKNOWN-0' whenever the distribution can not be detected.

=cut

sub get_distri {
	if ( ! defined($local_distribution) ) {
		get_distri_name();
	}
	return "${local_distribution}-$local_distribution_version";
}

=item $kernel=get_kernel()

Returns the current kernel type.

=cut

sub get_kernel {
	if (  defined($local_kernel) ) {
		return $local_kernel;
	}

	my $uname=`uname`;
	chomp($uname);
	if ( $uname eq 'Darwin' ) {
		$local_kernel='darwin';
	} elsif ( $uname eq 'Linux' ) {
		$local_kernel='linux';
	} else {
		$local_kernel='UNKNOWN';
	}
	return $local_kernel;
}

	
=back

=head2 Package functions

=over 4

=item $result=package_installed($showerrors,@packagelist)

Tests, if all (possibly 'virtual') packages in I<@packagelist> are installed
on the local system.

I<$showerrors> : Show missing packages

I<$result> == B<1>, only if *all* packages are installed.

=cut

sub packages_installed {
	my $distri=get_distri();
	if ( $distri eq 'debian' ) {
		package_installed_debian(@_);
	} else {
		return 0;
	}
}

my $debian_packages_read=0;
my %debian_packages=();
my %debian_packages_virtual=();

sub read_packages_debian {
	%debian_packages=();
	%debian_packages_virtual=();
	if ( open(I,"< /var/lib/dpkg/status") ) {
		my $name;
		my $status_ok=0;
		my $version;
		my $provided;
		while (my $l=<I>) {
			chomp($l);
			if ( $l =~ /^Package: (\S+)$/ ) {
				$name=$1;
			} elsif ( $l =~ /^Status: (.*)$/ ) {
				my $status=$1;
				if ( $status =~ /^(install|hold) ok installed$/ ) {
					$status_ok=1;
				}
			} elsif ( $l =~ /^Version: (\S+)$/ ) {
				$version=$1;
			} elsif ( $l =~ /^Provides: (.*)$/ ) {
				$provided=$1;
			} elsif ( $l eq '' ) {
				if ( defined($name) and $status_ok and defined($version) ) {
					$debian_packages{$name}=$version;
					if ( defined($provided) ) {
						my @provided=split /[\s,]+/,$provided;
						foreach my $provided (@provided) {
							$debian_packages_virtual{$provided}=1;
						}
					}
				}
				$name=undef;
				$status_ok=0;
				$version=undef;
				$provided=undef;
			}
		}
		close(I);
	} else {
		out('e',"Unable to read from '/var/lib/dpkg/status': $!");
		return 0;
	}
}

sub packages_installed_debian {
	my ($showerrors,@packages)=@_;
	if ( ! $debian_packages_read ) {
		if ( ! read_packages_debian() ) {
			return 0;
		}
	}
	my $result=1;
	foreach my $package (@packages) {
		if (
		 ! defined($debian_packages{$package}) and
		 ! defined($debian_packages_virtual{$package})
		) {
			if ( $showerrors ) {
				out('e',"Package '$package' is not installed.");
			}
			$result=0;
		}
	}
	return $result;
}


sub packages_installed_version {
	my $distri=get_distri();
	if ( $distri eq 'debian' ) {
		packages_installed_version_debian(@_);
	} else {
		return 0;
	}
}

sub packages_installed_version_debian {
	my ($showerrors,%packages)=@_;
	if ( ! $debian_packages_read ) {
		if ( ! read_packages_debian() ) {
			return 0;
		}
	}
	my $result=0;
	foreach my $package (keys %packages) {
		my ($operator,$version)=split /\s+/,$packages{$package};
		# Virtual packages will always cause errors here.
		if ( defined($debian_packages{$package}) ) {
			if ( ! fexec("dpkg --compare-versions " .
			 bash_quote($debian_packages{$package}) . " " .
			 bash_quote($operator) . " " .
			 bash_quote($version))
			) {
				if ( $showerrors ) {
					out('e',"Package ${package}'s version doesn't match (needed $packages{$package}, got $debian_packages{$package})");
				}
				$result=0;
			}
		}
	}
	return $result;
}

=back

=head2 PAG functions

=over 4

=item $result=is_pag($default)

Checks, if the current process is inside a PAG. The value I<$default> is returned,
if the PAG membership is not determinable on the local kernel.

I<$result> == B<1>, only if the current process is inside a pag.

=cut

sub is_pag {
	my $default=shift;
	if ( get_kernel() =~ /linux|darwin$/ ) {
		return is_pag_linux()
	} else {
		out('e',"Determining PAG membership is not supported for kernel '" . get_kernel() . "'");
		return $default;
	}
}


=item $result=is_pag_linux()

See is_pag().

This function is stolen from unpagsh.c, but probably it's part of the
openafs-source, thank god: C-Source is comparable to perl's :-)

=cut

sub is_pag_linux {
	my @groups=split(' ',$( );
	
	# Check for 32bit-PAGs
	for my $group (@groups) {
		if ( ( $group >> 24 ) == 0x41 ) {
			return 1;
		}
	}
   
   # Check for old 16bit-PAGs
   my $g0=0;my $g1=0;
   my $h=0;my $l=0;my $ret=0;

   if ( @groups < 2 ) { return 0 };

   $g0 = $groups[$#groups] & 0xffff;
   $g1 = $groups[$#groups-1] & 0xffff;
   $g0 -= 0x3f00;
   $g1 -= 0x3f00;

   if ( ( $g0 < 0xc000 ) && ( $g1 < 0xc000 ) ) {
		$l = (($g0 & 0x3fff) << 14) | ($g1 & 0x3fff);
		$h = ($g0 >> 14);
		$h = ($g1 >> 14) + 3*$h;
		$ret = (($h << 28) | $l);

      if ((($ret >> 24) & 0xff) == 65) {
			return 1;
      }
   }
   return 0;
}

=back

=head1 AUTHOR

Frank Burkhardt <burk@cbs.mpg.de>

=cut


1;
__END__
