package AFS::Instant::Manage;

use 5.006;
use Carp qw(cluck);
use strict;
use warnings;
use Carp;
use Fb::Data qw(hash_dir setEquality human_readable);
use Fb::Out qw(out $VERBOSE);
use Fb::IO qw(hash_mkdir iaexec);
use AFS::Instant::Toolz qw(make_mount remove_mount set_acl);
use AFS::Instant::aCall;
use AFS::Instant::CellDB;
use Time::Local;
use POSIX ":sys_wait_h";
use Data::Dumper;


require Exporter;
our @ISA=qw(Exporter);
our $VERSION='1.0';
our $xmlnsci='http://instantafs.cbs.mpg.de/xmlns/cellinfo/1.0';
our $ONLINETIMEOUT=10*60;

=head1 NAME

AFS::Instant::Manage - Management of volumes, volume instances, users and groups. Requires special privileges.

=head1 SYNOPSIS

use AFS::Instant::CellDB;

use AFS::Instant::Manage;

my $celldb=AFS::Instant::CellDB->new();

my $manage=AFS::Instant::Manage->new(celldb => $celldb,auth => 'local');


# Volume lifecycle

$manage->createVolume('newvolume',$partitionrec);

$manage->createInstanceRo('newvolume',$partitionrec);

$manage->renameVolume('newvolume','someothervolume');

$manage->setVolumeQuota('someothervolume',10*1024*1024); # set quota to 10MiB

my $volref=$celldb->getVolume('someothervolume');

$manage->removeInstance($volref->{'ro'}->[0]);

$manage->removeInstance($volref->{'rw'});


# Group/User-Management

$manage->createGroup('newgroup',owner => 'admin',id => -1000221);

$manage->createUser('newuser');

$manage->setPtdbMembership('newuser','newgroup',1);

$manage->setGroupMembers(

	'newgroup',

	{'someuser' => 1,'someotheruser' => 1}

);

$manage->removeGroup('newgroup');

$manage->removeUser('newuser');

=head1 REQUIRES

AFS::Instant::Toolz, AFS::Instant::CellDB

=head1 GENERIC OPTIONS

All methods of this class take these I<%options>:

=over 4

=item B<dryrun>

This options causes all write operations to be simmulated. Information is
shown about what would have been done. This is for debugging of the calling
application or to do "what-if"s.

=item B<auth>

##### ...

=back

=head1 METHODS

=head2 Constructor

=over 4

=item $manage=new(%options)

Creates a AFS-Cell-Management-object. Several options may be given:

=over 4

=item B<celldb>

A celldb object to be used

=item B<auth>

Defines, how to authenticate against AFS-Servers. Possible values are

=over 4

=item B<token>

Use the token in the AFS-Cachemanager to authenticate. Membership in
the B<system:administrators> group is mandatory in this case.

=item B<local>

Use B<-localauth> authentication. This needs access to the Cell-Masterkey
stored in B</etc/openafs/server> in B<KeyFile>, B<rxkad.keytab> or B<KeyFileEx>.

=back

=back

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self={
		celldb => undef,
		vos => '/usr/bin/vos',
		pts => '/usr/bin/pts',
		auth => 'token',
		error => 0,
	};
	if ( defined($options{'celldb'}) ) {
		$self->{celldb}=$options{'celldb'};
	} else {
		$self->{celldb}=AFS::Instant::CellDB->new();
	}
	if ( defined($options{'auth'}) ) {
		$self->{auth}=$options{'auth'};
	}
	$self->{celldb}->ensureDB() || return undef;
	my $me=bless($self,$package);
	return $me;
}

=head2 Object Methods

=over 4

=item $result=$manage->moveInstance($instanceref,$targetlocation,%options);

Moves a volume instance given by its I<$instanceref> to a given I<$targetlocation> (either a partition ref
or another instanceref which contains the same information). Possible I<%options> are:

=over 4

=item B<skipbackup>

In case of a backup instance, no error is returned but the move process is
silently successfully skipped.

=item B<roreference> => I<$sourceinstance>

In case of an RO instance, I<$sourceinstance> is used to create the target RO from. Otherwise data
of the instance referred to by I<$instanceref> is being used.

=item B<progress>

If true, a simple progress bar will be shown.

=item B<fast>

No clone is created before moving. This will render the moved instance
ununsable for AFS-clients during the move.

=item B<vldbonly>

Just create the new RO-instance as VLDB entry on the I<$targetlocation>.

=back

=cut

sub moveInstance {
	my ($self,$instanceref,$targetlocation,%options)=@_;
	out('w',"moveInstance(): hinting not implemented, yet.");
	my @iae;
	$self->calcAuth(\%options,iae => \@iae) || return 0;
	if ( $instanceref->{instancetype} == 2 ) {
		out('e',"There's no way to move a backup instance around, sorry.");
		return 0;
	}
	my $sourceaddr=$self->{celldb}->serverInfo('uuid2name',$instanceref->{'serveruuid'});
	my $targetaddr=$self->{celldb}->serverInfo('uuid2name',$targetlocation->{'serveruuid'});
	
	if ( $instanceref->{instancetype} == 0 ) {
		my @liveadd;
		if ( $options{'fast'} ) {
			@liveadd=('-live');
		}
		if ( $options{dryrun} ) {
			out('i',"Would've moved instance '" . $self->{celldb}->instanceInfo('hrname',$instanceref) . "'");
			return 1;
		}
		out('v',"Moving RW-instance of '$instanceref->{volumename}'");
		if ( iaexec(
			[$self->{vos},'move',$instanceref->{'id'},$sourceaddr,$instanceref->{'partition'},$targetaddr,
				$targetlocation->{'partition'},@liveadd,@iae],
			'exec',
		) ) {
			return 1;
		} else {
			out('e',"Error moving RW-instance of '$instanceref->{volumename}'");
			return 0;
		}
	} elsif ( $instanceref->{instancetype} == 1 ) {
		my $volref=$self->{celldb}->getVolume($instanceref->{volumename});
		my $targetcontainsro=0;
		for my $ro (@{$volref->{ro}}) {
			if ( $targetlocation->{serveruuid} eq $ro->{serveruuid} &&
				$targetlocation->{partition} eq $ro->{partition}
			) {
				$targetcontainsro=1;	
			}	
		}
		
		if ( ! $targetcontainsro ) {
			# Add RO instance VLDB entry for target location (if non exists, yet)
			if ( $options{dryrun} ) {
				out('i',"Would've added new RO on '$targetaddr/$targetlocation->{partition}'");
			} else {
				if ( ! iaexec(
					[$self->{vos},'addsite','-server',$targetaddr,'-partition',$targetlocation->{'partition'},'-id',
						$instanceref->{'volumename'},@iae],
					'exec',
					stdout => 'try',
					stderr => 'include',
					description => "Adding new RO-instance's location to VLDB"
				) ) {
					return 0;
				}
			}
			if ( ! $options{vldbonly} ) {
				if ( $options{dryrun} ) {
					out('i',"Would've released RO instances of '$volref->{name}'");
				} else {
					# Sync all RO instances (since 'vos clone'/'vos shadow' won't provide better alternatives)
					if ( ! iaexec(
						[$self->{vos},'release','-id',$volref->{id},@iae],
						'exec',
						description => "Releasing RO-instances"
					) ) {
						return 0;
					}
				}
			}
		}
		
		# Remove source location's RO instance
		if ( $options{dryrun} ) {
			out('i',"Would've removed source instance '" . $self->{celldb}->instanceInfo('hrname',$instanceref) . "'");
		} else {
			if ( ! iaexec(
				[$self->{vos},'remove',$sourceaddr,$instanceref->{'partition'},$instanceref->{'id'}],
				'exec',
				description => "Removing old RO-instance from VLDB and source server",
			) ) {
				return 0;
			}
		}
		return 1;
	}
	return undef;
}






=item $authtype=$manage->calcAuth(\%optionshash,%options);

Extracts the default authentication method id (key: B<auth>) from a standard
I<%optionshash>. Possible I<%options> are:

=over 4

=item B<localauth>

A scalar ref that will be set to '-localauth' for easy vos/pts command execution
if the auth method is B<local>.

=item B<iae>

A arrayref containing data that is to be appended to the commandline array
sent to AFS::Instant::Toolz::iaexec().

=item B<acallok>

If true, the auth method B<acall> is considered OK. In this case the calling function
will try to accomplish its task by using an acall RPC.

=back

As fallback, the default authentication method I<$manage->{auth}> is being used.

=cut

sub calcAuth {
	my ($self,$optionshash,%options)=@_;
	my $auth=$self->{auth};
	my %ok=('token' => 1,'local' => 1);
	if ( $options{'acallok'} ) {
		$ok{'acall'}=1;
	}
	if ( $optionshash->{'auth'} ) {
		if ( ! $ok{$optionshash->{'auth'}} ) {
			out('e',"Invalid auth parameter '$optionshash->{'auth'}'");
			return undef;
		}
		$auth=$optionshash->{'auth'};
	}
	if ( defined($options{'iae'}) ) {
		@{$options{'iae'}}=('-cell',$self->{celldb}->{cellname});
	}
	if ( $auth eq 'local' ) {
		if ( ! -r "/etc/openafs/server/KeyFile" && ! -r "/etc/openafs/server/rxkad.keytab" && ! -r "/etc/openafs/server/KeyFileEx" ) {
			out('w',"calcAuth(): No read access to a AFS server keyfile or file doesn't exist");
			out('w',"calcAuth(): Since 'local' auth was chosen, this won't end well");
		}
		if ( defined($options{'localauth'}) ) {
			$options{'localauth'}=' -localauth ';
		}
		if ( defined($options{'iae'}) ) {
			push(@{$options{'iae'}},'-localauth');
		}
	} else {
		if ( ( $auth eq 'acall' ) and ! $options{'acallok'} ) {
			out('e',"object-wide default auth 'acall' is not suitable for this operation'");
			return undef;
		}
	}
	return $auth;
}

=item $type=$manage->removeInstance($instanceref,%options)

Removes the volume instance given by its I<$instanceref>. These options are accepted:

=over 4

=item B<auth>

Do authentication to the fileserver - either via (system:administratos-)
B<token> or B<local> (~ localauth option on vos commands, needs root access
to an AFS server).

=item B<dryrun>

Don't actually remove the volume, just simmulate it.

=item B<terrors>

Tolerated errors. Possible values are:

=over 4

=item B<hints>

Don't return error when DB hinting is not working correctly.

=back

=back

After successful removal, a hint is placed in the SQL DB.

=cut

sub removeInstance {
	my ($self,$instance,%options)=@_;
	if ( ! defined($instance) ) {
		out('e',"removeInstance(): Undefined instance");
		return 0;
	}
	my $volume=$instance->{'volumename'};
	my $serveruuid=$instance->{'serveruuid'};
	my $partition=$instance->{'partition'};
	my $instancetype=$self->{celldb}->instanceIdToType($instance->{'instancetype'});
	
	my $server=$self->{celldb}->serverInfo('uuid2ioaddress',$serveruuid);
	if ( ! defined($server) ) {
		out('e',"Unable to determine io address for removing volume instance of '$volume'");
		return 0;
	}
	my @iae;
	$self->calcAuth(\%options,iae => \@iae) || return 0;
	my $instancesuffix='';
	if ( $instancetype ne 'rw' ) {
		$instancesuffix="." . $instancetype;
	}
	
	my $success=1;
	my $servername=$self->{celldb}->serverInfo('uuid2name',$serveruuid);
	if ( ! $options{'dryrun'} ) {
		if ( defined($instance->{rostate}) ) {
			if ( ! iaexec(
				[$self->{vos},'remove','-server',$server,'-partition',$partition,
					'-id',"$volume$instancesuffix",@iae],
				'exec',
				description => "Removing volume instance '$volume$instancesuffix' from '$servername/$partition'"
			) ) {
				return 0;
			}
		} else {
			if ( ! iaexec(
				[$self->{vos},'zap','-server',$server,'-partition',$partition,
					'-id',$instance->{id},@iae],
				'exec',
				description => "Removing (zapping) VLDB-unregistered volume instance '$volume$instancesuffix' from '$servername/$partition'"
			) ) {
				return 0;
			}
		}
	} else {
		out('i',"Would've removed instance '$instancetype' of volume '$volume' from '$servername/$partition' now");
		return 1;
	}
	if ( $success ) {
		out('d',"Placing remove-hint for instance '$volume'/$instancetype in DB");
		my $query;
		my $instancetypeid=$self->{celldb}->instanceTypeToId($instancetype);
		my $volumeq2=$self->{celldb}->sqlquote($volume);
		my $serverq2=$self->{celldb}->sqlquote($serveruuid);
		$query=$self->{celldb}->sqlqueryw(
			"DELETE FROM instances WHERE volumename=$volumeq2 AND instancetype=$instancetypeid AND serveruuid=$serverq2",
			"Unable to place remove-hint for $volume/$instancetype in DB"
		);
		if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
			return 0;
		}
		$query=$self->{celldb}->sqlqueryw(
			"DELETE FROM vldbinstances WHERE volumename=$volumeq2 AND instancetype=$instancetypeid AND serveruuid=$serverq2",
			"Unable to place remove-hint for VLDB entry of $volume/$instancetype in DB"
		);
		if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
			return 0;
		}
		if ( $instancetypeid == 0 ) {
			# Removing an RW-instance always wipes a potentially existing backup Instance, too
			$query=$self->{celldb}->sqlqueryw(
				"DELETE FROM instances WHERE volumename=$volumeq2 AND instancetype=2 AND serveruuid=$serverq2",
				"Unable to place remove-hint for $volume/$instancetype in DB"
			);
			if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
				return 0;
			}
		}
		
		# test, if there are instances Left for this volume
		$query=$self->{celldb}->sqlquery(
			"SELECT count(volumename) FROM instances WHERE volumename=$volumeq2",
			"Unable to test for additional volume instances"
		);
		if ( ! $query ) {
			return 0;
		} else {
			my $row=$query->fetchrow_arrayref();
			if ( $row->[0] == 0 ) {
				out('d',"Placing remove-hint for volume '$instance->{'volumename'}' in DB");
				$query=$self->{celldb}->sqlqueryw(
					"DELETE FROM volumes WHERE volumename=$volumeq2",
					"Unable to place remove hint for volume '$volume' in DB"
				);
				if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
					return 0;
				}
			}
		}
	}
	return 1;
}

=item $result=$manage->removeVolume($volref,%options)

Removes the volume given by its I<$volref>. Possible I<%options> are the same as
the ones of the removeInstance() function since this is basically a wrapper
for it.

=cut

sub removeVolume {
	my ($self,$volref,%options)=@_;
	out('d',"Attempt to remove volume '$volref->{name}'");
	my $success=0;
	for my $iref (@{$self->{celldb}->iterateInstances($volref,types => {0 => 1,1 => 1},realonly => 1)}) {
		my $isuccess=$self->removeInstance($iref,%options);
		$success &= $isuccess;
	}
	return $success;
}


=item $result=$manage->createVolume($volumename,$partitionref,%options)

Creates a volume with the name I<$volumename> on the partition specified by I<$partitionref> (hashref containing at least the
keys B<serveruuid> and B<partition>). These I<%options> are accepted:

=over 4

=item B<terrors>

A hashref containing "Tolerated Errors" which will *not* cause createVolume() to return an error. Possible terrors are:

=over 4

=item B<volhash>

The volhash-directory is used to i.e. set ACLs and to chown. 

=item B<backup>

if the backup instance cannot be created

=item B<ro>

if the given RO instances could not be created.

=item B<acls>

if the given ACLs could not be applied.

=item B<chown>

if the volume owner could not be set.

=item B<mountpoint>

if the target mountpoint could not be created

=item B<hints>

if the DB hints could not be written.

=back

=item B<acls>

A hashref containg users/groups as keys and a permission string (i.e. B<rlwidka>) as values.

=item B<nacls>

Negative ACLs - similar to B<acls>.

=item B<owner>

Volume owner to be set (a PTDB name).

=item B<dryrun>

Set to a true value, this option will cause the whole creation process to be simmulated only.

=item B<auth>

Use this method to authenticate to the fileserver. Possible values are B<token> to use a system:administrators-token
(which is the only really usefull choice) and B<local> to use -localauth on a AFS-Server as root.

=item B<backup>

If set, a backup instance will be created.

=item B<rolocations>

A hashtree of RO locations. This tree should look like this:

 $options{'rolocations'}={
   '002c89d6-9978-1e3f-a4-1c-113ca8c0aa77' => { 'a' => 1 },
 }

Alternatively, the RO locations can be given as an array reference of locations hashrefs.

=item B<wait>

Wait for the given amount of seconds after RW instance was create before starting operations in AFS filespace.

=item B<quota>

The amount of quota to assign to the RW instance (in Bytes).

=item B<volhash> (default: B<1>)

Defines, if createVolume() should try to create generic volume mountpoint to the new RW instance in
B</afs/>I<cellname>B</admin/volumes>. This is necessary for acls to be set and chmod to be done.

=item B<mountpoint>

A path that should be a mountpoint to an instance or instances of the newly create volume. The
mountpoint will point to the RW-instance by default.

=item B<mountpointtype> (default: B<rw>)

Defines the type of mountpoint to be created. Possible values are: B<rw> and B<magic>.

=back

=cut

# terrors: volhash, backup, chown,acls,mountpoint,ro
sub createVolume {
	my ($self,$volume,$rwlocation,%options)=@_;
	my $serveruuid=$rwlocation->{'serveruuid'};
	my $partition=$rwlocation->{'partition'};
	my $auth=$self->calcAuth(\%options) || return 0;
	my $do_volhash=1;
	if ( defined($options{'volhash'}) ) {
		$do_volhash=$options{'volhash'};
	}
	if ( ! $do_volhash and ($options{'acls'} or $options{'nacls'} or $options{'owner'}) ) {
		out('e',"createVolume(): Unable to set acls or ownership without volhash");
		return 0;
	}
	if ( ( $auth eq 'local' ) and ($options{'acls'} or $options{'nacls'} or $options{'owner'}) ) {
		out('e',"Unable to set acls or ownership with localauth authentication");
		return 0;
	}
	if ( ! $self->createVolumeSimple($volume,$serveruuid,$partition,%options) ) {
		return 0;
	}

	my $volhash_success=0;
	my $voltemp_hash_=hash_dir($volume,2) || return 0;
	my $voltemp_base="/afs/" . $self->{celldb}->{cellname} . "/admin/volumes";
	my $voltemp_hash=join('/',@{$voltemp_hash_});
	my $path="$voltemp_base/rw/$voltemp_hash/$volume";

	my $volref=$self->{celldb}->getVolume($volume);
	if ( ! $volref ) {
		out('e',"Unable to get volume ref after creation of '$volume'");
		return 0;
	}

	if ( $do_volhash ) {
		if ( hash_mkdir($volume,"$voltemp_base/rw",2,'dryrun' => $options{'dryrun'}) ) {
			if ( make_mount($path,$volume,'rw' => 1,'idempotent' => 1,'dryrun' => $options{'dryrun'},'checkvolumes' => 3) ) {
				$volhash_success=1;
			} else {
				out('w',"Unable to create temporary mount point for volume '$volume' ('$path').");
				if ( ! $options{'terrors'}->{'volhash'} ) {
					return 0;
				}
			}
		} else {
			if ( ! $options{'terrors'}->{'volhash'} ) {
				return 0;
			}
			out('w',"Unable to create directories leading to temporary mountpoint ('$voltemp_base/rw')");
		}
	}
	
	if ( $options{'acls'} or $options{'nacls'} or $options{'owner'} ) {
		if ( ! $volhash_success ) {
			out('w',"Unable to continue without temporary mountpoint");
			$self->removeInstance($volref->{rw},%options);
			return 0;
		}
		if ( ! $options{'dryrun'} ) {
			if ( ! chmod(0755,$path) ) {
				out('w',"Changing posix permission to '0755' failed on volume '$volume'");
				out('d',"Path was '$path'");
				if ( ! $options{'terrors'}->{'chmod'} ) {
					remove_mount($path);
					$self->removeInstance($volref->{rw},%options);
					return 0;
				}
			}
		} else {
			out('i',"Would've set unix-permission 0755 now");
		}
		if ( $options{'acls'} ) {
			if ( ! set_acl($options{'acls'},$path,'dryrun' => $options{'dryrun'}) ) {
				out('w',"Unable to set initial ACL for volume '$volume'.");
				if ( ! $options{'terrors'}->{'acls'} ) {
					remove_mount($path,'dryrun' => $options{'dryrun'});
					$self->removeInstance($volref->{rw},%options);
					return 0;
				}
			}
		}
		if ( $options{'owner'} ) {
			if ( ! $options{'dryrun'} ) {
				if ( ! iaexec(
					['/bin/chown',$options{'owner'},$path],
					'exec',
					description => "Setting owner of volume '$volume'",
				) ) {
					if ( ! $options{'terrors'}->{'owner'} ) {
						remove_mount($path,'dryrun' => $options{'dryrun'});
						$self->removeInstance($volref->{rw},%options);
						return 0;
					}
				}
			} else {
				out('i',"Would've chown()ed the new volume now");
			}
		}
	}
	my $volumeq=quotemeta($volume);
	if ( $options{'backup'} ) {
		if ( ! $self->syncVolume($volume,'backup','dryrun' => $options{'dryrun'}) ) {
			out('w',"Unable to create backup instance for volume '$volume'");
			if ( ! $options{'terrors'}->{'backup'} ) {
				$self->removeInstance($volref->{rw},%options);
				if ( $volhash_success ) {
					remove_mount($path,'dryrun' => $options{'dryrun'});
				}
				return 0;
			}
		}
	}

	if ( $options{'mountpoint'} ) {
		my $rw=1;
		my $type='rw';
		if ( defined($options{mountpointtype}) ) {
			$type=$options{mountpointtype};
		}
		if ( $type ne 'rw' ) {
			$rw=0;
		}
		my $vsuffix='';
		if ( $type eq 'backup' ) {
			$vsuffix=".backup";
		} elsif ( $type eq 'readonly' ) {
			$vsuffix=".readonly";
		}
		if ( ! make_mount($options{'mountpoint'},$volume,'rw' => $rw,'idempotent' => 1,'dryrun' => $options{'dryrun'},'checkvolumes' => 3) ) {
			out('w',"Unable to create mountpoint '$options{'mountpoint'}' to volume '$volume'(rw)");
			if ( ! $options{'terrors'}->{'mountpoint'} ) {
				$self->removeInstance($volref->{rw},%options);
				if ( $volhash_success ) {
					remove_mount($path,'dryrun' => $options{'dryrun'});
				}
				return 0;
			}
		} else {
			out('v',"Successfully created mountpoint '$options{'mountpoint'}' to volume '$volume'");
		}
	}
	if ( $options{'rolocations'} ) {
		my $rolocations={};
		if ( ref($options{'rolocations'}) eq 'ARRAY' ) {
			for my $rol (@{$options{'rolocations'}}) {
				$rolocations->{$rol->{'serveruuid'}}->{$rol->{'partition'}}=1;
			}
		} else {
			$rolocations=$options{'rolocations'};
		}
		my $ro_success=1;
		my $ro_success_one=0;
		for my $serveruuid_ro (keys %{$rolocations}) {
			# Iterating over serveruuids of the partition hashtree...
			my @partitions=sort keys %{$rolocations->{$serveruuid_ro}};
			if ( @partitions > 1 ) {
				out('w',"Attempting to create more than one RO on '$serveruuid_ro' for volume '$volume', ignoring this server");
			} elsif ( @partitions == 1 ) {
				my $lsuccess=$self->createInstanceRo($volume,{'serveruuid' => $serveruuid_ro,'partition' => $partitions[0]},%options);
				if ( ! $lsuccess ) {
					if ( ! $options{'terrors'}->{'volhash'} ) {
						$self->removeInstance($volref->{rw},%options);
						if ( $volhash_success ) {
							remove_mount($path,'dryrun' => $options{'dryrun'});
						}
						return 0;
					}
				}
				$ro_success &= $lsuccess;
				$ro_success_one=1;
			}
		}
		if ( $ro_success_one ) {
			if ( $do_volhash ) {
				if ( ! hash_mkdir($volume,"$voltemp_base/ro",2) ) {
					out('w',"Unable to create volhash mountpoint for $volume.readonly");
					if ( ! $options{'terrors'}->{'volhash'} ) {
						$self->removeInstance($volref->{rw},%options);
						if ( $volhash_success ) {
							remove_mount($path,'dryrun' => $options{'dryrun'});
						}
						return 0;
					}
				}
			}
		}
		return 1;
	}
	return 1;
}

=item $result=$manage->createInstanceRo($volumename,$location,%options);

Creates an RO instance of the volume I<$volumename> at the given I<$location> (which is a hashref
as returned by AFS::Instant::CellDB::getPartition() or an instance hashref). Possible I<%options> are:

=over 4

=item B<fillup>

This option set to B<1> will cause the new RO instance to be populated with data instead of just
being created as an additional volume instance entry in VLDB. In this case, the RO will be online
after createInstanceRo() is finished.

=item B<terrors>

A hashref containing the list of "tolerated errors" as keys. Possible "terrors" are:

=over 4

=item B<hints>

Don't return failure on failed Database hinting updates.

=item B<freespace>

Don't return failure if target partition's free space is less than the volume's RW's current
space allocation.

=item B<fillup>

Don't return failure if the *physical* volume instance on the target partition could not be created
(-> if the release afterwards failed)



=back

=back

=cut

sub createInstanceRo {
	my ($self,$volume,$location,%options)=@_;
	my $serveruuid=$location->{'serveruuid'};
	my $partition=$location->{'partition'};
	my $serveraddress=$self->{celldb}->serverInfo('uuid2ioaddress',$serveruuid);
	if ( ! defined($serveraddress) ) {
		out('e',"Unable to find a working address of server '$serveruuid'");
		return 0;
	}
	my $servername=$self->{celldb}->serverInfo('uuid2name',$serveruuid);
	if ( ! defined($servername) ) {
		$servername=$serveruuid;
	}

	my @iae;
	$self->calcAuth(\%options,iae => \@iae) || return 0;

	my $partitionref=$self->{celldb}->getPartition($serveruuid,$partition);
	if ( ! $partitionref ) {
		out('e',"createInstanceRo(): Target partition '$servername/$partition' is unknown to DB");
		return 0;
	}
	my $volref=$self->{celldb}->getVolume($volume);
	if ( $volref->{'rw'} ) {
		if ( $partitionref->{'partfree'} < $volref->{'rw'}->{'size'} ) {
			if ( ! $options{'terrors'}->{'freespace'} ) {
				out('e',"createInstanceRo(): Target partition '$servername/$partition' 's free space is less than '$volume' 's RW instance's current size");
				return 0;
			}
		}
	}

	if ( $options{'dryrun'} ) {
		out('i',"Would've created RO instance of volume '$volume' on '$servername/$partition'");
		return 1;
	}

	my $rwservername=$self->{celldb}->serverInfo('uuid2name',$volref->{'rw'}->{serveruuid});
	my @filledupadd;my $filleduprostate=2;

	# For DB-Hints...
	my $volumeq=$self->{celldb}->sqlquote($volume);
	my $serveruuidq=$self->{celldb}->sqlquote($serveruuid);
	my $partitionq=$self->{celldb}->sqlquote($partition);


	if ( $options{'fillup'} ) {
		out('e',"This option is not working, yet");
		return 0;
		if ( ! $volref->{'rw'} ) {
			out('e',"createInstanceRo(): Unable to populate '$volume' 's new RO instance with data from non-existing RW instance");
			return 0;
		}
		my $rwserveraddress=$self->{celldb}->serverInfo('uuid2ioaddress',$volref->{'rw'}->{serveruuid});
		if ( ! defined($rwserveraddress) ) {
			out('e',"Unable to find a working address of server '$serveruuid' which hosts the RW instance");
			return 0;
		}
		my $result=iaexec(
			[$self->{vos},'copy','-id',$volref->{'rw'}->{'id'},'-fromserver',$rwserveraddress,
				'-frompartition',$volref->{'rw'}->{'partition'},'-toname',$volume,'-toserver',$serveraddress,
				'-topartition',$partition,'-readonly',@iae],
			'exec',
			description => "Copying content of RW-instance of '$volume' to '$servername/$partition'",
		);
		if ( ! $result ) {
			if ( ! $options{'terrors'}->{'fillup'} ) {
				return 0;
			}
		} else {
			@filledupadd=('-valid');
			$filleduprostate=0;
			my $time=time();
			my $query=$self->{celldb}->sqlqueryw(
				"INSERT INTO instances (volumename,instancetype,serveruuid,`partition`,size,quota,files,ctime,mtime,atime,id,online,needssalvage,isbusy) VALUES " .
					"($volumeq,1,$serveruuidq,$partitionq,$volref->{'rw'}->{'size'},$volref->{'rw'}->{'quota'},$time,$volref->{'rw'}->{'mtime'},0,0,1,0,0)",
				"Unable to put hint for new RO instance of volume '$volume' 's into DB"
			);
			if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
				return 0;
			}
		}
	}
	if ( ! iaexec(
		[$self->{vos},'addsite','-server',$serveraddress,'-partition',$partition,'-id',$volume,@filledupadd,@iae],
		'exec',
		description => "Adding RO-site to VLDB"
	) ) {
		return 0;
	}

	if ( ! $options{nohint} ) {
		# DB-Hinting...
		my $query=$self->{celldb}->sqlqueryw(
			"INSERT INTO vldbinstances (volumename,instancetype,serveruuid,`partition`,rostate) VALUES ($volumeq,1,$serveruuidq,$partitionq,$filleduprostate)",
			"Unable to place hint for new RO instance'VLDB entry of '$volume' in DB"
		);
		if ( ! $query ) {
			if ( ! $options{'terrors'}->{'hints'} ) {
				return 0;
			}
		}
	}
	return 1;
}

=item $result=$manage->createVolumeSimple($volumename,$serveruuid,$partition,%options)

Creates a new volume plus a RW instance named I<$volumename> on the specified server
I<$serveruuid> at the given I<$partition>. Possible I<%options> are:

=over 4

=item B<quota>

The quota of the new volume in Bytes. Use B<0> for no quota.

=item B<terrors>

A hashref containing the "tolerated errors" that won't trigger an unsuccessful return of
this method. Possible "terrors" are:

=over 4

=item B<hints>

Failed Database hinting updates for newly created stuff won't trigger an error.

=back

=back

=cut


sub createVolumeSimple {
	my ($self,$volume,$serveruuid,$partition,%options)=@_;
	my $quota=5000*1024;
	if ( $options{'quota'} ) {
		$quota=$options{'quota'};
	}
	my $query;
	my $vref=$self->{celldb}->getVolume($volume);
	if ( $vref ) {
		out('w',"Volume '$volume' already exists");
		return 0;
	}
	my $partref=$self->{celldb}->getPartition($serveruuid,$partition);
	if ( ! defined($partref) ) {
		out('e',"Target partition is unknown to DB");
		return 0;
	}
	if ( ! $self->{celldb}->partitionInfo('online',$partref) ) {
		out('e',"Partition of the new volume is currently offline");
		return 0;
	}
	if ( $options{'dryrun'} ) {
		out('i',"Would've created volume '$volume' now");
		return 1;
	}
	my $fh;
	my $quotakb=int($quota/1024);

	my @iaesuffix;
	$self->calcAuth(\%options,iae => \@iaesuffix) || return 0;
	my $serveraddress=$self->{celldb}->serverInfo('uuid2ioaddress',$serveruuid);
	my $servername=$self->{celldb}->serverInfo('uuid2name',$serveruuid);
	if ( ! defined($servername) ) {
		$servername='';
	} else {
		$servername="($servername)";
	}
	if ( ! defined($serveraddress) ) {
		out('e',"Unable to find a working address of server '$serveruuid'$servername");
		return 0;
	}
	my $partitionq=quotemeta($partition);
	my @output;
	if ( ! iaexec(
		[$self->{vos},'create','-server',$serveraddress,'-partition',$partition,'-name',$volume,'-maxquota',$quotakb,@iaesuffix],
		'readpipe',
		stderr => 'include',
		stdout => 'try',
		format => \@output,
		description => "Creating volume '$volume'",
	) ) {
		return 0;
	}
	out('v',"Created volume '$volume'");
	my $rwid=0;
	for my $l (@output) {
		if ( $l =~ /^Volume testtest already exists/ ) {
			# CellDB didn't catch this one, it already exists
			return 0;
		} elsif ( $l =~ /^Volume ([0-9]+) created on partition/ ) {
			$rwid=$1;
		}
	}
	if ( ! $rwid ) {
		out('e',"Error creating RW instance of volume '$volume'");
		return 0;
	}
	my $volumeq2=$self->{celldb}->sqlquote($volume);
	my $partitionq2=$self->{celldb}->sqlquote($partition);
	my $serveruuidq2=$self->{celldb}->sqlquote($serveruuid);
	my $time=time();
	$query=$self->{celldb}->sqlqueryw(
		"INSERT INTO volumes (volumename,rwid,roid,backupid,rostate,locked,synctry_ro,synctry_backup) VALUES ($volumeq2,$rwid,0,0,0,0,0,0)",
		"Unable to put hint for new volume '$volume' into DB"
	);
	if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
		return 0;
	}
	$query=$self->{celldb}->sqlqueryw(
		"INSERT INTO instances (volumename,instancetype,serveruuid,`partition`,size,quota,files,ctime,mtime,atime,id,online,needssalvage,isbusy) VALUES " .
			"($volumeq2,0,$serveruuidq2,$partitionq2,2048,$quota,0,$time,$time,$time,$rwid,1,0,0)",
		"Unable to put hint for new volume '$volume' 's RW instance into DB"
	);
	if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
		return 0;
	}
	
	$query=$self->{celldb}->sqlqueryw(
		"INSERT INTO vldbinstances (volumename,instancetype,serveruuid,`partition`,rostate) VALUES " .
			"($volumeq2,0,$serveruuidq2,$partitionq2,0)",
		"Unable to put hint for new volume '$volume' 's VLDB-RW-Record into DB"
	);
	if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
		return 0;
	}
	return 1;
}

=item $result=$manage->syncVolume($volume,$instancetype,%options)

Update either the RO- or backup instances (depending on I<$instancetype>) of
the given I<$volume>. I<$volume> is either a volumename or a hasref as returned by
getVolume(). If I<$volume> is a hashref, syncVolume() will rely on that data completely.
These I<%options> are interpreted:

=over 4

=item B<force>

Usually syncVolume() tries to be smart about which volume is actually synced. This options
overrides this behaviour. Possible values are:

=over 4

=item >= B<1>

Try to sync the given I<$instancetype>, even if they seem to be up-to-date. This might be necessary
if a change was done recently and the RW's information haven't been updated, yet. It's usually a good idea
to use this option in a script directly after a change.

=item >= B<2>

This option is only effective for RO-copies. It enables the B<-force>
commandline option in the B<vos release> command.

=back

=item B<checkhooks>

Run release-hooks on I<$volume>.

=item B<forcehooksoverride>

Ignore (negative) result of release hooks.

=item B<auth>

Authentication method to be used. Possible values are:

=over 4

=item B<token>

Use system:administrators token.

=item B<local>

Use localauth on an AFS-Server as Root.

=item B<acall>

Use acall (the B<instantafs.acall-setquota> function on the B<acall-db> -server).

=back

=item B<terrors>

A hashref containing keys defining, which errors will be tolerated:

=over 4

=item B<ropartoffline>

A partition, a (potential or real) RO copy of the volume resides on seems to beoffline.

=item B<noro>

Missing VLDB-RO-records are not considered an error. Success is returned in this case.

=item B<hints>

Don't return error if hinting fails.

=item 

=back

=item B<try>

Consider this function call a try. Everything works the same but no error message
is shown. This is being used by syncVolumes() to determine a working auth method.

=back

=cut

sub syncVolume {
	my ($self,$volume,$instancetype_,%options)=@_;
	my $volref;
	if ( ref($volume) eq '' ) {
		$volref=$self->{celldb}->getVolume($volume);
	} else {
		$volref=$volume;
		$volume=$volref->{'name'};
	}
	my $force=0;
	if ( $options{'force'} ) {
		$force=$options{'force'};
	}
	my $instancetype=$self->{celldb}->instanceIdToType($instancetype_);
	if ( ! defined($instancetype) ) {
		out('e',"syncVolume('${volume}'): Invalid Instancetype '$instancetype_'");
		return 0;
	}
	if ( $instancetype eq 'rw' ) {
		out('e',"syncVolume('${volume}'): Unable to sync rw instance - it's the actual sync source");
		return 0;
	}
	my @iae;
	my $method=$self->calcAuth(\%options,iae => \@iae,acallok => 1) || return 0;

	my $success=1;

	out('v',"syncVolume('${volume}'): instance_type=$instancetype, auth_method=$method");
	if ( $volume =~ /^-/ ) {
		out('w',"syncVolume(): Unable to sync volume '$volume' which starts with '-'");
		return 0;
	}
	if ( ! $options{force} && $options{checkhooks} ) {
		if ( ! $self->{celldb}->runVolumeHooks($volume,"release/" . $instancetype) ) {
			if ( ! $options{forcehooksoverride} ) {
				return 0;
			}
		}
	}
	if ( $instancetype eq 'readonly' ) {
		my $foundro=0;
		my $exit=0;
		if ( ! $self->{celldb}->volumeInfo('hasrw',$volref) ) {
			out('e',"syncVolume('${volume}'): no RW instance found");
			$exit=1;
		}
		if ( ! $self->{celldb}->instanceInfo('accessible',$volref->{'rw'}) ) {
			if ( ! $options{'terrors'}->{'partitionoffline'} ) {
				out('e',"syncVolume('${volume}'): RW partition is offline");
				$exit=1;
			}
		}
		my $offlinecount=0;
		for my $ro (@{$volref->{'ro'}}) {
			if ( ! defined($ro->{rostate}) ) {
				next;
			}
			$foundro=1;
			if ( ! $self->{celldb}->partitionInfo('online',$ro) ) {
				$offlinecount++;
			}
		}
		if ( $offlinecount ) {
			if ( ! $options{'terrors'}->{'partitionoffline'} ) {
				out('e',"syncVolume('${volume}'): '$offlinecount' RO-partition(s) is/are offline");
				$exit=1;
			}
		}
		if ( ! $foundro ) {
			if ( ! $options{'terrors'}->{'noro'} ) {
				out('e',"syncVolume('${volume}'): no RO instance found, not syncing.");
				$exit=1;
			} else {
				return 1;
			}
		}
		if ( $exit ) {
			return 0;
		}
	} elsif ( $instancetype eq 'backup' ) {
		if ( ! $self->{celldb}->volumeInfo('backupable',$volref) ) {
			if ( ! $options{'terrors'}->{'partitionoffline'} ) {
				out('w',"syncVolume('${volume}'): Unable to backup-Sync volume, RW partition is offline");
				return 0;
			}
		}
	}
	if ( $method eq 'acall' ) {
		if ( ! $options{'dryrun'} ) {
			my $ac=AFS::Instant::aCall->new('role' => 'db') || return 0;
			my $function;
			my $updatetest;
			if ( $instancetype eq 'readonly' ) {
				$function='instantafs.acall-release';
				$updatetest='rouptodate';
			} elsif ( $instancetype eq 'backup' ) {
				$function='instantafs.acall-vosbackup';
				$updatetest='backupuptodate';
			} else {
				out('e',"syncVolume('${volume}'): Invalid sync type '$instancetype', internal error'");
				return 0;
			}
			if ( ! $force && $self->{celldb}->volumeInfo($updatetest,$volref,vldbonly => 1) ) {
				return 1;
			}
			my ($acresult,$result,$text)=$ac->call($function,[$volume]);
			if ( $acresult ) {
				if ( $result ) {
					# acall function resulted in an error
					if ( $VERBOSE > 0 ) {
						print STDERR $text;
					}
					$success=0;
				}
			} else {
				out('w',"Error talking to acall-db server");
				# error talking to acall server
				$success=0;
			}
		} else {
			out('i',"Would've synchronized volume '$volume' 's $instancetype instance(s) via acall now");
		}
	} else {
		if ( $instancetype eq 'readonly' ) {
			my $doit=$force;
			if ( defined($volref) ) {
				my $uptodate=$self->{celldb}->volumeInfo('rouptodate',$volref,vldbonly => 1);
				$doit |= ! $uptodate;
				if ( $uptodate and ! $options{'dryrun'} and ! $force ) {
					# Instance class is uptodate, marking this fact in DB
					$self->updateSyncTry($volume,1);
				}
			} elsif ( $options{'dryrun'} ) {
				out('i',"syncVolume(): no information about volume '$volume' during dryrun, fake syncing anyway");
				$doit=1;
			}
			if ( $doit ) {
				if ( ! $options{'dryrun'} ) {
					my $volumeq=quotemeta($volume);
					my @forceadd;
					if ( $force > 1 ) {
						@forceadd=('-force');
					}
					if ( iaexec(
						[$self->{vos},'release','-id',$volume,@forceadd,@iae],
						($options{'try'} ? 'test' : 'exec'),
						description => "Releasing volume '$volume'"
					) ) {
						if ( ! $self->doSyncHinting($volref,1) ) {
							if ( ! $options{'terrors'}->{'hints'} ) {
								$success=0;
							}
						}
					} else {
						$success=0;
					}
				} else {
					out('i',"Would've released volume '$volume' s RO copies now");
				}
			}
		} elsif ( $instancetype eq 'backup' ) {
			my $doit=$force;
			if ( defined($volref) ) {
				my $uptodate=$self->{celldb}->volumeInfo('backupuptodate',$volref);
				$doit |= ! $uptodate;
				if ( $uptodate and ! $options{'dryrun'} and ! $force ) {
					$self->updateSyncTry($volume,2);
				}
			} elsif ( $options{'dryrun'} ) {
				out('i',"syncVolume(): no information about volume '$volume' during dryrun, fake syncing anyway");
				$doit=1;
			}
			if ( $doit ) {
				if ( ! $options{'dryrun'} ) {
					my $volumeq=quotemeta($volume);
					my $time=time();
					if ( iaexec(
						[$self->{vos},'backup','-id',$volume,@iae],
						($options{'try'} ? 'test' : 'exec'),
						description => "Synchronising backup-instance of volume '$volume'"
					) ) {
						if ( ! $self->doSyncHinting($volref,2) ) {
							if ( ! $options{'terrors'}->{'hints'} ) {
								$success=0;
							}
						}
					} else {
						$success=0;
					}
				} else {
					out('i',"Would've synchronized backup instance of volume '$volume' now");
				}
			}
		}
	}
	return $success;
}

=item $result=$manage->doSyncHinting($volref,$synctype,$timestamp)

Updates DB to reflect the RO- or Backup-sync which was started at I<$timestamp> (epoch value).

=cut

sub doSyncHinting {
	my ($self,$volref,$synctype,$timestamp)=@_;
	out('d',"doSyncHinting(): not implemented, yet");
	return 1;
	if ( $synctype == 2 ) {
		if ( $self->{celldb}->volumeInfo('hasbackup',$volref) ) {

		}
	}
}

=item $result=$manage->syncVolumes(\@volumes,$instancetype,%options)

Update either the RO- or backup instances (depending on I<$instancetype>) of
the given I<@volumes>.  These I<%options> are interpreted:

=over 4

=item B<auth>

Authentication methods to be used to be used. This is an arrayref containing
an ordered list.  Methods are tried, the first one working will be used for
remaining volumes.  Possible methods are.

=over 4

=item B<token>

Use system:administrators token.

=item B<local>

Use localauth on an AFS-Server as Root.

=item B<acall>

Use acall (the B<instantafs.acall-setquota> function on the B<acall-db> -server).

=back

By default, only the default auth method of the I<$manage>-object is used.

=item B<terrors>

A hashref containing keys defining, which errors will be tolerated:

=over 4

=item B<ropartoffline>

A partition, a (potential or real) RO copy of the volume resides on seems to beoffline.

=item B<noro>

It's not considered an error if there's no RO-instance in VLDB and RO-sync is requested.

=back

=item B<trylimit> (default: B<4>)

Maximum number of volumes to try before considering all methods failed. Defaults to B<4>.

=item B<results>

A hashref which will be filled with data about the process. These fields are important:

=over 4

=item B<errors>

A hashref that will be filled with names of volumes (as keys) which are not synced successfully by any method.

=back

=item B<parallel>

*Reserved*

=item B<preload> (default: B<0>)

Loads all volume-records before attempting to sync the first one. This is sometimes helpful, if the load on
fileservers it too high during sync. In this case, the instantafs.update-process might be unable to
update the DB fast enough.

=item B<force>

Do sync even if the volume is in synced state already.

=item B<forcehooksoverride>

Ignore (negative) result of release hook tests.

=back

=cut

sub syncVolumes {
	my ($self,$volumes,$instancetype_,%options)=@_;
	my @methods=($self->{'auth'});
	if ( $options{'auth'} ) {
		if ( ref($options{'auth'}) eq '' ) {
			# This should not happen but it might
			@methods=($options{'auth'});
		} elsif ( ref($options{'auth'}) eq 'ARRAY' ) {
			@methods=@{$options{'auth'}};
		} else {
			out('e',"syncVolumes(): Invalid reference found in 'auth'-options");
			return 0;
		}
	}
	if ( ! @methods ) {
		out('e',"syncVolumes(): No methods given, aborting");
		return 0;
	}
	my $errorlimit=4;
	if ( defined($options{'trylimit'}) ) {
		$errorlimit=$options{'trylimit'};
	}
	my $preload=1;
	if ( defined($options{'preload'}) ) {
		$preload=$options{'preload'};
	}
		
	my @errors=();
	my $chosen_method;
	my %suboptions=();
	for my $option ('force','terrors','dryrun') {
		if ( $options{$option} ) {
			$suboptions{$option}=$options{$option};
		}
	}
	if ( $options{'results'} ) {
		if ( ! $options{'results'}->{'errors'} ) {
			$options{'results'}->{'errors'}={};
		}
	}
	my $errors=0;
	
	# We preload all the volume data. A sync-storm might cloud partition information later
	my %volrefs=();
	my $filteredvolumes=[];
	if ( $preload ) {
		my $instancetype=$self->{celldb}->instanceTypeToId($instancetype_);
		out('v',"syncVolumes(): Preloading " . scalar(@{$volumes}) . " volume record(s)");
		for my $volume (@{$volumes}) {
			$volrefs{$volume}=$self->{celldb}->getVolume($volume);
			if ( $volrefs{$volume} ) {
				if ( $options{checkhooks} ) {
					if ( ! $self->{celldb}->runVolumeHooks($volume,"release/" . $self->{celldb}->instanceIdToType($instancetype)) ) {
						if ( ! $options{forcehooksoverride} ) {
							$errors++;
							next;
						}
					}
				}
				if ( $options{'force'} ) {
					push(@{$filteredvolumes},$volume);
				} else {
					my $updatetest='rouptodate';
					if ( $instancetype == 2 ) {
						$updatetest='backupuptodate';
					}
					my $uptodate=$self->{celldb}->volumeInfo($updatetest,$volrefs{$volume},vldbonly => 1);
					if ( defined($uptodate) ) {
						if ( $uptodate ) {
							if ( ! $options{'dryrun'} ) {
								$self->updateSyncTry($volume,$instancetype);
							}
						} else {
							push(@{$filteredvolumes},$volume);
						}
					} else {
						if ( ! $options{'terrors'}->{'noro'} ) {
							out('e',"syncVolumes(): Unable to sync volume '$volume', unknown sync state");
							$errors++;
						}
					}
				}
			}
		}
		out('v',"Post-preload-filter: '" . scalar(@{$filteredvolumes}) . "' out of '" . scalar(@{$volumes}) . "' will be sync'ed");
	} else {
		$filteredvolumes=$volumes;
	}
	foreach my $volume (@{$filteredvolumes}) {
		my $volref;
		if ( $preload ) {
			$volref=$volrefs{$volume};
		} else {
			$volref=$self->{celldb}->getVolume($volume);
		}
		if ( ! defined($volref) ) {
			out('e',"syncVolumes(): No data available about volume '$volume', skipping this one");
			# This counts as an error but not one relevant for method selection ($errors++)
			push(@{$options{'results'}->{'errors'}},$volume);
			next;
		}
		if ( ! defined($chosen_method) ) {
			foreach my $method (@methods) {
				out('d',"Trying method '$method'");
				if ( $self->syncVolume($volref,$instancetype_,'auth' => $method,'try' => 1,%suboptions) ) {
					if ( scalar(@{$volumes}) > 1 ) {
						out('v',"Authentication-method '$method' succeeded on volume '$volume'. Will use it for all volumes...");
					} else {
						out('v',"Authentication method '$method' succeeded");
					}
					$chosen_method=$method;
					last;
				} else {
					out('d',"No success with authentication method '$method' for volume '$volume'");
				}
			}
			if ( ! defined($chosen_method) ) {
				if ( $options{'results'} ) {
					push(@{$options{'results'}->{'errors'}},$volume);
				}
				$errors++;
				push(@errors,$volume);
				if ( $errors > $errorlimit ) {
					out('e',"No authentication-method worked on " . scalar(@errors) ." different volumes ('" . join("','",@errors). "'), giving up.");
					return 0;
				}
			}
		} else {
			out('d',"Using formerly succesfully used method '$chosen_method'");
			if ( ! $self->syncVolume($volref,$instancetype_,'auth' => $chosen_method,%suboptions) ) {
				if ( $options{'results'} ) {
					push(@{$options{'results'}->{'errors'}},$volume);
					$errors++;
				}
			}
		}
	}
	
	return ( $errors == 0 );
}


=item $result=$celldb->updateSyncTry($volumename,$instancetypeid)

Updates the synctry-field in the SQL DB. This shows that a given volume's RO
or Backup instance "would have been updated" now but the RW instance hadn't
been changed.  This way, it's easy to filter volumes, no syncronisation
process took care of.

I<$instancetypeid> has to be numerical (1 for "vos release", 2 for "vos backup").

I<$result> == 1, only if the synctry-field was updated successfully in DB.

=cut

sub updateSyncTry {
	my ($self,$volume,$instancetypeid)=@_;
	out('d',"Logging sync try of type '$instancetypeid' for volume '$volume'");
	my $time=time();
	my $fieldname;
	if ( $instancetypeid == 1 ) {
		$fieldname='synctry_ro';
	} elsif ( $instancetypeid == 2 ) {
		$fieldname='synctry_backup';
	} else {
		out('e',"updateSyncTry(): invalid instance type id '$instancetypeid'");
		return 0;
	}
	my $volumeq=$self->{celldb}->sqlquote($volume);
	my $query=$self->{celldb}->sqlqueryw(
		"UPDATE volumes SET ${fieldname}=$time WHERE volumename=$volumeq",
		"Unable to document synchronisation try (type=$instancetypeid) on volume '$volume'"
	);
	if ( $query ) {
		return 0;
	}
	out('d',"Successfully documented synchronisation try (type=$instancetypeid) on volume '$volume'");
	return 1;
}


=item $result=$manage->setVolumeQuota($volumename,$newquota,%options)

Sets the volume I<$volumename> 's quota to I<$newquota> (which is given in Bytes). Possible options are:

=over 4

=item B<auth>

Authentication method to be used for setting the quota. Possible values are:

=over 4

=item B<token>

Use system:administrators token.

=item B<local>

Use localauth on an AFS-Server as Root.

=item B<acall>

Use acall (the B<instantafs.acall-setquota> function on the B<acall-db> -server).

=back

=item B<dryrun>

Just simmmulate setting the quota.

=item B<terrors>

Defines a set of tolerated errors by ID.

=over 4

=item B<offline>

Try to set quota even if the RW instance seems to be currently offline.

=item B<hints>

Don't return failure if the hinting update of the DB didn't work.

=back

=item B<force>

Try to set quota even if the same value is already applied to the volume's RW.

=item B<try>

Consider this function call a try. Everything works the same but no error message
is shown.

=back

=cut

sub setVolumeQuota {
	my ($self,$volume,$newquota,%options)=@_;
	$self->{celldb}->ensureDB() || return 0;
	my @iae;
	my $method=$self->calcAuth(\%options,iae => \@iae,acallok => 1);
	if ( ! $method ) {
		return 0;
	}
	if ( $method eq 'acall' ) {
		if ( ! $options{'dryrun'} ) {
			my $ac=AFS::Instant::aCall->new('role' => 'db');
			my ($acresult,$result,$text)=$ac->call('instantafs.acall-setquota',[$volume,$newquota]);
			if ( $acresult ) {
				if ( $result ) {
					print STDERR $text;
					return 0;
				} else {
					return 1;
				}
			} else {
				return 0;
			}
		} else {
			out('i',"Would've set quota of volume '$volume' to '$newquota' via acall. Assuming success.");
			return 1;
		}
	}
	my $volref=$self->{celldb}->getVolume($volume);

	if ( $newquota < 1024 ) {
		# Never set quota to zero, if it's not actually zero - it would
		# turn off quotacheck.
		$newquota=1024;
	}
	
	if ( ! defined($volref) ) {
		if ( ! $options{'terrors'}->{'novolume'} ) {
			if ( ! $options{'try'} ) {
				out('e',"setQuota(): Unknown volume '$volume'");
			}
			return 0;
		} else {
			out('v',"Setting quota of volume '$volume' 's RW instance, although the volume doesn't seem to exist");
		}
	} else {
		if ( ! $volref->{'rw'} ) {
			if ( ! $options{'terrors'}->{'norw'} ) {
				if ( ! $options{'try'} ) {
					out('e',"setQuota(): Volume '$volume' doesn't seem to have an RW instance");
				}
				return 0;
			} else {
				out('v',"Setting quota of volume '$volume' 's RW instance, although no RW seem to exist");
			}
		} else {
			if ( $volref->{'rw'}->{'quota'} == $newquota ) {
				if ( ! $options{'force'} ) {
					out('v',"Skipping setting quota to the same value");
					return 1;
				} else {
					out('v',"Setting quota to already set value.");
				}
			}
		}
	}
	my $quotainkb=int($newquota/1024);
	if ( $options{'dryrun'} ) {
		out('i',"Would've set quota of volume '$volume' to '$newquota' using method '$method'");
		return 1;
	}
	if ( ! $self->{celldb}->instanceInfo('accessible',$volref->{'rw'}) ) {
		if ( ! $options{'terrors'}->{'offline'} ) {
			if ( ! $options{'try'} ) {
				out('e',"setVolumeQuota(): Volume '$volume' 's RW instance is currently not online");
			}
			return 0;
		} else {
			out('v',"Setting quota of volume '$volume' 's RW instance, although it seems to be offline");
		}
	}
	if ( ! iaexec(
		[$self->{vos},'setfields','-id',$volume,'-maxquota',$quotainkb,@iae],
		($options{'try'} ? 'test' : 'exec'),
		description => "Setting quota of volume $volume to " . human_readable($newquota,'memory') . "B"
	) ) {
		return 0;
	}
	my $volumeq2=$self->{celldb}->sqlquote($volume);
	my $query=$self->{celldb}->sqlqueryw(
		"UPDATE instances SET quota=$newquota WHERE volumename=$volumeq2 AND instancetype=0",
		"Unable to place hint about new quota of volume '$volume' in DB"
	);
	if ( ! $query ) {
		if ( ! $options{'terrors'}->{'hints'} ) {
			return 0;
		}
	}
	return 1;
}

=item $result=$manage->renameVolume($oldname,$newname,%options)

Renames the given volume I<$oldname> to I<$newname>. These I<%options> may be given:

=over 4

=item B<dryrun>

Only simmulate renaming the volume.

=back

=cut

sub renameVolume {
	my ($self,$oldname,$newname,%options)=@_;
	my $oldref=$self->{celldb}->getVolume($oldname);
	my $newref=$self->{celldb}->getVolume($newname);
	if ( ! $oldref ) {
		out('e',"Volume '$oldname' doesn't exist");
		return 0;
	}
	if ( $newref ) {
		out('e',"Volume '$newname' already exists, unable to rename '$oldname' to that");
		return 0;
	}

	my @iae;
	$self->calcAuth(\%options,iae => \@iae) || return 0;
	if ( $options{'dryrun'} ) {
		out('i',"Would've renamed volume '$oldname' to '$newname' now");
		return 1;
	}
	
	if ( ! iaexec(
		[$self->{vos},'rename','-oldname',$oldname,'-newname',$newname,@iae],
		'exec',
		description => "Renaming volume '$oldname' to '$newname'"
	) ) {
		return 0;
	}
	
	# To Push the new name into local caches...
	$oldref->{'name'}=$newname;
	my $instances=$self->{celldb}->iterateInstances($oldref);
	if ( defined($instances) ) {
		for my $iref (@{$instances}) {
			$iref->{'volumename'}=$newname;
		}
	} else {
		out('w',"renameVolume(): error doing iterateInstances() for cache-hinting");
	}
	my $oldnameq2=$self->{celldb}->sqlquote($oldname);
	my $newnameq2=$self->{celldb}->sqlquote($newname);
	my $query=$self->{celldb}->sqlqueryw(
		"UPDATE volumes SET volumename=$newnameq2 WHERE volumename=$oldnameq2",
		"Unable to place hint about volume rename in DB"
	);
	if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
		return 0;
	}
	$query=$self->{celldb}->sqlqueryw(
		"UPDATE instances SET volumename=$newnameq2 WHERE volumename=$oldnameq2",
		"Unable to place hint about volume instance rename in DB"
	);
	if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
		return 0;
	}
	$query=$self->{celldb}->sqlqueryw(
		"UPDATE vldbinstances SET volumename=$newnameq2 WHERE volumename=$oldnameq2",
		"Unable to place hint about VLDB volume instance record rename in DB"
	);
	if ( ! $query and ! $options{'terrors'}->{'hints'} ) {
		return 0;
	}
	return 1;
}

=item $result=$manage->createGroup($groupname,%options)

Create a PTDB group I<$groupname>. These options are possible:

=over 4

=item B<members>

A array reference to names of AFS users or groupss to be included as members in the new group.

=item B<owner>

The name of the owner (user or PTDB group) of the new group.

=item B<id>

The id of the new group

=item B<dryrun>

Just simulate creating the group.

=item B<terrors>

Defines hashref of error-ids that are to be tolerated (I<$result> == 1 even if the occur):

=over 4

=item B<hints>

Don't fail if DB-hinting fails

=item B<owner>

Don't fail if owner cannot be set

=back

=back

=cut

sub createGroup {
	my ($self,$name,%options)=@_;
	my %members=();
	if ( $options{'members'} ) {
		if ( ref($options{'members'}) eq 'ARRAY' ) {
			for my $member (@{$options{'members'}}) {
				$members{$member}=1;
			}
		} else {
			%members=%{$options{'members'}};
		}
		my $error=0;
		for my $member (keys %members) {
			if ( ! $self->{celldb}->ptdbInfo('name2id',$member) ) {
				out('e',"createGroup(): unable to add new member '$member', doesn't exist");
				$error=1;
			}
		}
		if ( $error ) {
			return 0;
		}
	}
	
	# This makes the group create operation truely idempotent
	if ( defined($self->{celldb}->groupInfo('name2id',$name)) ) {
		my $oldmembers=$self->{celldb}->groupInfo('members',$name);
		if ( ! setEquality($oldmembers,\%members) ) {
			out('e',"Group '$name' exists but desired memberships of the group and current ones don't match");
			return 0;
		}
		if ( $options{'id'} ) {
			my $oldid=$self->{celldb}->groupInfo('name2id',$name);
			if ( $options{'id'} != $oldid ) {
				out('e',"Group '$name' exists but old AFSID doesn't match requested new ID");
				return 0;
			}
		}
		##### Ownership test to be implemented. Need some database-table-extension for that
		out('v',"createGroup(): Group '$name' already exists but since the existing and the requested groups are identical, I'll return success");
		return 1;
	}

	my @iae;
	my $method=$self->calcAuth(\%options,iae => \@iae,acallok => 0);

	my $query;
	my $newid;
	if ( ! $options{'dryrun'} ) {
		my $nameq=quotemeta($name);
		my $fh;
		my @idadd=();
		if ( $options{'id'} ) {
			@idadd=('-id',$options{'id'});
		}
		my $output;
		if ( ! iaexec(
			[$self->{pts},'creategroup',$name,@idadd,@iae],
			'readpipe',
			format => \$output,
			description => "Creating PTDB group '$name'",
		) ) {
			return 0;
		}
		
		if ( $output =~ /^group \S+ has id (-[0-9]+)/ ) {
			$newid=$1;
		}
		$nameq=$self->{celldb}->sqlquote($name);
		$query=$self->{celldb}->sqlqueryw(
			"INSERT INTO afsids (afsid,name) VALUES ($newid,$nameq)",
			"Unable to place group create hint in DB"
		);
		if ( ! $query ) {
			if ( ! $options{'terrors'}->{'hints'} ) {
				return 0;
			}
		}
	} else {
		out('i',"Would've created group '$name' now");
	}
	
	my $addsuccess=1;
	if ( %members ) {
		my $nameq=quotemeta($name);
		for my $member (sort keys %members) {
			if ( ! $options{'dryrun'} ) {
				my $memberq=quotemeta($member);
				if ( ! iaexec(
					[$self->{pts},'add',$member,$name,@iae],
					'exec',
					description => "Joining PTDB object '$member' into group '$name'",
				) ) {
					$addsuccess=0;
				} else {
					my $memberid=$self->{celldb}->ptdbInfo('name2id',$member);
					if ( defined($memberid) ) {
						$query=$self->{celldb}->sqlqueryw(
							"INSERT INTO memberships (afsgroup,member) VALUES ($newid,$memberid)",
							"Unable to place hint about new group membership in DB"
						);
					} else {
						out('w',"Unable to get AFSID of user '$member' after successfull group membership update - strange...");
					}
				}
			} else {
				out('i',"Would've added user '$member' to group '$name' now");
			}
		}
	}
	if ( $options{'owner'} ) {
		if ( ! $options{'dryrun'} ) {
			my $owner=$options{'owner'};
			if ( ! iaexec(
				[$self->{pts},'chown',$name,$owner,@iae],
				'exec',
				description => "Changing ownership of PTDB group '$name' to '$owner'",
			) ) {
				out('e',"Unable to set owner of '$owner'");
				if ( ! $options{'terrors'}->{'owner'} ) {
					return 0;
				}
			}
		} else {
			out('i',"Would've set owner of group '$name' to '$options{'owner'}'");
		}
	}
	return $addsuccess;
}

=item $result=$manage->removeGroup($groupname,%options)

Removes the given PTDB group I<$groupname>. Possible options are:

=over 4

=item B<dryrun>

Simmulate removing the group.

=item B<terrors>

Contains a hashref, defining which actions may fail without removeGroup() failing.

=over 4

=item B<hints>

=back

=back

=cut

sub removeGroup {
	my ($self,$name,%options)=@_;
	my $id=$self->{celldb}->ptdbInfo('name2id',$name);
	if ( ! defined($id) ) {
		out('d',"Group '$name' doesn't exist. removeGroup() will idempotently return success");
		return 1;
	} elsif ( $id > 0 ) {
		out('d',"User name '$name' exists");
		return 0;
	}
	if ( $options{'dryrun'} ) {
		out('i',"Would've remove group '$name' now");
		return 1;
	}
	my @iae;
	my $method=$self->calcAuth(\%options,iae => \@iae,acallok => 0);
	if ( ! iaexec(
		[$self->{pts},'delete',$name,@iae],
		'exec',
		description => "Removing PTDB group '$name'",
	) ) {
		out('e',"Unable remove group '$name'");
		return 0;
	}
	my $query=$self->{celldb}->sqlqueryw(
		"DELETE FROM afsids where afsid=$id",
		"Unable to place hint about group removal in DB"
	);
	if ( ! $query ) {
		if ( ! $options{'terrors'}->{'hints'} ) {
			return 0;
		}
	}
	return 1;
}

=item $result=$manage->setGroupMembers($groupname,\%memberships,%options);

Sets the whole I<%membership> set of a given I<$group> at once. I<%memberships> is a hash.
Each future member (either a group or a user) is represented by a key with a true value.
Possible I<%options> are:

=over 4

=item B<dryrun>

Only simmulate setting the memberships.

=back

=cut

sub setGroupMembers {
	my ($self,$group,$members,%options)=@_;
	my $exists=$self->{celldb}->groupInfo('exists',$group);
	if ( ! defined($exists) ) {
		out('e',"'$group' is not a group or error determining existence of '$group'");
		return 0;
	}
	if ( ! $exists ) {
		out('e',"Group '$group' doesn't exist");
		return 0;
	}
	my $result=1;
	my $oldmembers=$self->{celldb}->groupInfo('members',$group);
	for my $member (sort keys %{$members}) {
		if ( ! $oldmembers->{$member} ) {
			$result &= $self->setPtdbMembership($member,$group,1,%options);
		}
	}
	for my $member (sort keys %{$oldmembers}) {
		if ( ! $members->{$member} ) {
			$result &= $self->setPtdbMembership($member,$group,0,%options);
		}
	}
	return $result
}

=item $result=$manage->setPtdbMembership($member,$group,$include,%options)

Makes sure, the given ptdb object I<$member> is a member of the given I<$group>
( I<$include> == B<1> ) or not ( I<$include> == B<0> ). Possible I<%options>
are:

=over 4

=item B<dryrun>

Just simmulate everything.

=item B<terrors>

A hashref containing IDs of actions that may fail without setPtdbMenbership() failing.

=over 4

=item B<hints>

Do not return error if hinting to DB fails.

=back

=item B<nohints>

Do not try to do any hinting.

=back

=cut

sub setPtdbMembership {
	my ($self,$member,$group,$include,%options)=@_;
	my $membership=$self->{celldb}->ptdbInfo('ismemberof',$member,'group' => $group);
	if ( ! defined($membership) ) {
		if ( ! $self->{celldb}->ptdbInfo('exists',$member) ) {
			out('e',"setPtdbMembership(): PTDB object '$member' doesn't exist");
		}
		if ( ! $self->{celldb}->groupInfo('exists',$group) ) {
			out('e',"setPtdbMembership(): Group '$group' doesn't exist or is not a group");
		}
		return 0;
	}
	my $group_=$self->{celldb}->groupInfo('name2id',$group);
	my $member_=$self->{celldb}->ptdbInfo('name2id',$member);
	if ( $membership == $include ) {
		if ( $include ) {
			out('d',"Not making '$member' a member of group '$group', is already a member");
		} else {
			out('d',"Not revoking '$member' 's membership in group group '$group', is not a member");
		}
		return 1;
	}
	if ( $options{'dryrun'} ) {
		if ( $include ) {
			out('i',"Would've added ptdb object '$member' to group '$group'");
		} else {
			out('i',"Would've removed ptdb object '$member' from group '$group'");
		}
		return 1;
	}
	my @iae;
	$self->calcAuth(\%options,iae => \@iae,acallok => 0) || return 0;
	if ( $include ) {
		if ( ! iaexec(
			[$self->{pts},'adduser',$member,$group,@iae],
			'exec',
			description => "Adding PTDB object '$member' to group '$group'",
		) ) {
			return 0;
		}
		if ( ! $options{nohints} ) {
			my $query=$self->{celldb}->sqlqueryw(
				"INSERT INTO memberships (afsgroup,member) VALUES ($group_,$member_)",
				"Unable to add hint for new membership into DB"
			);
			if ( ! $query ) {
				if ( ! $options{'terrors'}->{'hints'} ) {
					return 0;
				}
			}
		}
		return 1;
	} else {
		if ( ! iaexec(
			[$self->{pts},'removeuser',$member,$group,@iae],
			'exec',
			description => "Removing PTDB object '$member' from group '$group'",
		) ) {
			return 0;
		}
		if ( ! $options{nohints} ) {
			my $query=$self->{celldb}->sqlqueryw(
				"DELETE FROM memberships WHERE afsgroup=$group_ AND member=$member_",
				"Unable to place hint about membership removal into DB"
			);
			if ( ! $query ) {
				if ( ! $options{'terrors'}->{'hints'} ) {
					return 0;
				}
			}
		}
		return 1;
	}
}

sub _switchSTDERR {
	my $self=shift;
	if ( ! defined($self->{stderrbackup}) ) {
		my $fh;
		open($fh,'>&STDERR');
		$self->{stderrbackup}=$fh;
	} else {
		open(STDERR,">&",$self->{stderrbackup});
		delete $self->{stderrbackup};
	}
}

=item $result=$manage->createUser($username,%options)

Creates a new PTDB user record name I<$username>. Possible I<%options> are:

=over 4

=item B<dryrun>

Just simmulate creating the user

=item B<afsid>

Set specific AFS ID.

=item B<idempotent>

Define, if the creation process should be idempotent. Defaults to B<1>.

=back

=cut

sub createUser {
	my ($self,$user,%options)=@_;
	my $formerid=$self->{celldb}->userInfo('name2id',$user);
	if ( $formerid ) {
		if ( exists $options{'idempotent'} and ! $options{'idempotent'} ) {
			out('e',"User already exists");
			return 0;
		}
		if ( defined($options{'afsid'}) ) {
			if ( $options{'afsid'} == $formerid ) {
				out('v',"User already exists");
			} else {
				out('e',"User already exists with different AFSID");
				return 0;
			}
		}
	}
	my @iae;
	$self->calcAuth(\%options,iae => \@iae,acallok => 0) || return 0;

	if ( $options{'dryrun'} ) {
		out('i',"Would've created new AFS user '$user' now");
		return 1;
	}

	my @idadd;
	if ( $options{'afsid'} ) {
		@idadd=('-id',$options{'afsid'});
	}

	my $newid;
	my $output;
	if ( ! iaexec(
		[$self->{pts},'createuser',$user,@idadd,@iae],
		'readpipe',
		format => \$output,
		description => "Creating new AFS user '$user'"
	) ) {
		return 0;
	}
	if ( $output =~ /^User \S+ has id ([0-9]+)/ ) {
		$newid=$1;
	}
	my $userq2=$self->{celldb}->sqlquote($user);
	my $query=$self->{celldb}->sqlqueryw(
		"INSERT INTO afsids (afsid,name) VALUES ($newid,$userq2)",
		"Unable to place user create hint in DB"
	);
	if ( ! $query ) {
		if ( ! $options{'terrors'}->{'hints'} ) {
			return 0;
		}
	}
	return 1;
}

=item $result=$manage->removeUser($username,%options)

Removes the given PTDB user I<$username>. Possible options are:

=over 4

=item B<dryrun>

Simmulate removing the group.

=back

=cut

sub removeUser {
	my ($self,$name,%options)=@_;
	my $id=$self->{celldb}->ptdbInfo('name2id',$name);
	if ( ! defined($id) ) {
		out('d',"User '$name' doesn't exist. removeUser() will idempotently return success");
		return 1;
	} elsif ( $id < 0 ) {
		out('e',"Group name '$name' exists.");
		return 0;
	}
	my @iae;
	$self->calcAuth(\%options,iae => \@iae,acallok => 0) || return 0;
	if ( $options{'dryrun'} ) {
		out('i',"Would've removed user '$name' now");
		return 1;
	}
	if ( ! iaexec(
		[$self->{pts},'delete',$name,@iae],
		'exec',
		description => "Removing user '$name'"
	) ) {
		return 0;
	}
	my $query=$self->{celldb}->sqlqueryw(
		"DELETE FROM afsids where afsid=$id",
		"Unable to place hint about user removal in DB"
	);
	if ( ! $query ) {
		if ( ! $options{'terrors'}->{'hints'} ) {
			return 0;
		}
	}
	return 1;
}



1;
