package AFS::Instant::Output::Partitions;

use 5.006;
use strict;
use warnings;

use Carp qw(cluck);
use Carp;
use Time::Local;
use Data::Dumper;

use Fb::Data qw(human_readable);
use Fb::Out;
use AFS::Instant::Toolz qw(getUserPref);

require Exporter;
our @ISA=qw(Exporter);
our $VERSION='3.0';

=head1 NAME

AFS::Instant::Output - Output functions for volume and instance information

=head1 SYNOPSIS

  use AFS::Instant::Output::Partitions;
  use AFS::Instant::CellDB;
  
  my $celldb=AFS::Instant::CellDB->new();
  my $output=AFS::Instant::Output::Partitions->new(celldb => $celldb,mode => 'bigtable',hr => 1);

  my $partitions=$celldb->expandLocationPattern('*/*',format => 'arrayref');
  print $output->partitions($partitions);
  
=head1 REQUIRES

AFS::Instant::CellDB, AFS::Instant::Toolz;

=head1 METHODS

=head2 Constructor

=over 4

=item $output=new(%options)

Creates a new Output-Object for returning partition information in certain formats.
Possible options are:

=over 4

=item B<celldb>

A AFS::Instant::CellDB object. This is actually mandatory.

=item B<separator>

Defines the separator between partition records. Defaults to B<\n> (newline).

=item B<header>

Defines, if the default header of the given output mode should be returned (B<1>) prior
to the actual data or not (B<0>).

=item B<mode>

Defines the format in which partitions are presented in. Possible formats are:

=over 4

=item B<name>

Only the server + "/" + partitionname.

=item B<bigtable>

A table containing the most vital information about partitions.

=item B<sums>

A table containing statistical information about all partitions. This mode is only
useful when multiple partition are being output since a call to $output->partition()
will return empty.

=item B<table:...>

#####

=back

=item B<namelengthhint>

Defines the maximum length of a partition (or partition group) name. This is used when
a table is displayed and no alignment for names is explicitely given.

=item B<hr> (default: B<1> for mode bigtable)

Defines if the output should be made human readable (value B<1>) or not (value B<0>).
"human readable" means that

=over 4

=item timestamps 

are convertet to local time values and/or timediff-values

=item integers

are shown as kilo/mega/... expressions.

=item memory values

are show as kibi/mibi/gibi/... expressions.

=back

=back

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self={
		celldb => undef,
		separator => "\n",
		header => 0,
		hr => 0,
		mode => 'name',
		sortkey => 'name',
		sortreverse => 0,
	};
	if ( defined($options{'celldb'}) ) {
		$self->{celldb}=$options{'celldb'};
	}
	if ( defined($options{'mode'}) ) {
		$self->{mode}=$options{'mode'};
		if ( $self->{mode} eq 'bigtable' ) {
			$self->{hr}=1;
		}
	}
	if ( defined($options{'separator'}) ) {
		$self->{separator}=$options{'separator'};
	}
	if ( defined($options{'header'}) ) {
		$self->{header}=$options{'header'};
	}
	if ( defined($options{'sortkey'}) ) {
		$self->{sortkey}=$options{'sortkey'};
	}
	if ( defined($options{'sortreverse'}) ) {
		$self->{sortreverse}=$options{'sortreverse'};
	}
	if ( defined($options{'hr'}) ) {
		$self->{hr}=$options{'hr'};
	}
	$self=bless($self,$package);

	if ( ! $options{nopref} ) {
		my $userpref=getUserPref();
		if ( $userpref ) {
			my $formats=$userpref->find("//up:format[\@type='output/partitions'][\@id='$self->{mode}']");
			if ( $formats ) {
				my $format=$formats->[0];
				if ( ! defined($self->{hr}) && $format->hasAttribute('hr') ) {
					$self->{hr}=int($format->hasAttribute('hr'));
				}
				if ( ! defined($self->{header}) && $format->hasAttribute('header') ) {
					$self->{header}=int($format->hasAttribute('hr'));
				}
				out('d',"Using custom format '$self->{mode}'");
				$self->{mode}=$format->getAttribute('format');
			} elsif ( $self->{mode} =~ /^name:/ ) {
				$self->{mode}='name';
			}
		}
	}

	if ( $self->{mode} =~ /^table:(.*)$/ ) {
		my @f=split /\|/,$1;
		my %fielddata=(
			name => [5,'Loc.','Location'],
			size => [-6,'Size','Partition_size'],
			free => [-6,'Free','Free_space'],
			used => [-6,'Used','Space_used'],
			instances => [-5,'#I','Number_of_Instances'],
			instancesrw => [-5,'#I_RW','Number_of_RW-Instances'],
			volumes => [-5,'#V','Number_of_Volumes'],
			filessum => [-5,'#F',"Number_of_Inodes"],
			filessumrw => [-5,'#F_RW',"Number_of_Inodes_in_RW-Instances"],
			quotasum => [-6,"Quota","Sum_of_Quota"],
			quotasumrw => [-6,"Q[RW]","Sum_of_RW-Quota"],
			committedrw => [-6,"C[RW]","Committed_RW_Quota"],
			sizesum => [-6,"Size","Filesize"],
			sizesumrw => [-6,"S[RW]","Filesize_RW"],
			committed => [-6,"Com.","Committed_Quota"],
			maxmtimerw => [-15,"Mtim","Mtime_latest_RW"],
			infotime => [-15,'Info','Partition_Update_Time'],
			comment => [0,'Comment','Comment_if_any'],
		);
		my @tableheader;
		for my $f (@f) {
			my $key=$f;
			my $opt={};
			if ( $f =~ /(\S+?):(.*)$/ ) {
				$f=$1;
				for my $o (split /:/,$2) {
					if ( $o =~ /^(.*?)=(.*)$/ ) {
						$opt->{$1}=$2;
					} else {
						$opt->{$o}=1;
					}
				}
			}
			my $data=$fielddata{$f};
			if ( ! $data ) {
				out('e',"Invalid field specifier '$key'");
				return undef;
			}
			my $align=$data->[0];
			if ( $f =~ /time/ && $opt->{diff} ) {
				$align=-4;
			} elsif ( $f eq 'name' ) {
				if ( ! defined($opt->{align}) && defined($options{namelengthhint}) ) {
					$align=$options{namelengthhint};
				}
			}
			if ( ! defined($opt->{align}) ) {
				$opt->{align}=$align;
			}
			my $rec={%{$opt},name => $f,key => $key};
			push(@{$self->{tablefields}},$rec);
			my $tableheader=$data->[2-int($self->{hr})];
			$self->align(\$tableheader,$opt->{align});
			push(@tableheader,$tableheader);
		}
		$self->{tableheader}=join(' ',@tableheader);
		$self->{mode}='table';
	} elsif ( $self->{mode} eq 'bigtable' ) {
		return new('AFS::Instant::Output::Partitions',%options,mode => 'table:name:status|size|free|used|instances|instancesrw|volumes|filessum|filessumrw|sizesum|sizesumrw|quotasum|quotasumrw|maxmtimerw:diff|infotime:diff|comment');
	} elsif ( $self->{mode} !~ /^(name|dump)$/ ) {
		out('e',"Invalid partition output mode '$self->{mode}'");
		return undef;
	}
	if ( $self->{sortkey} !~ /^(name|name#|quota-usage|partfree|partusage|partsize|none)$/ ) {
		out('e',"Invalid sort key '$self->{sortkey}'");
		return undef;
	}
	return $self;
}

=head2 Object Methods

=over 4

=item $string=$output->HR($value,$hrmode);

Returns the I<$value> in human readable form - if defined via $output->{hr}.
Othervise I<$string> will contain a copy of I<$value>. I<$hrmode> is the mode
given to the AFS::Instant::Toolz::human_readable() method.

=cut

sub HR {
	my ($self,$value,$hrmode,%options)=@_;
	if ( ! defined($value) ) {
		return undef;
	}
	if ( $self->{hr} ) {
		$value=human_readable($value,$hrmode,%options);
		if ( $options{align} ) {
			$self->align(\$value,$options{align});
		}
	}
	return $value;
}

sub align {
	my ($self,$s,$a)=@_;
	if ( ! $self->{hr} ) {
		return;
	}
	$$s=Fb::Data::align($$s,$a);
}


=item $text=$output->partition($partitionrec,%options)

Returns a single partition according to the given parameters of the I<$output> object.

=cut

sub partition {
	my ($self,$partrec,%options)=@_;
	my $commentadd='';
	if ( $options{'comment'} ) {
		if ( ! $self->{'hr'} ) {
			$options{'comment'} =~ s/\s/_/g;
		}
		$commentadd=" " . $options{'comment'}
	}
	my $namefull;my $nameshort;
	my $servername;
	if ( defined($partrec->{serveruuid}) ) {
		$servername=$self->{celldb}->serverInfo('uuid2name',$partrec->{'serveruuid'});
		$namefull="$servername/$partrec->{partition}";
		$nameshort=substr($servername,0,3) . '/' . $partrec->{partition};
	} else {
		$namefull=$partrec->{name};
		$nameshort=substr($partrec->{name},0,3);
	}
	if ( $self->{mode} eq 'name' ) {
		return $namefull . $commentadd;
	} elsif ( $self->{mode} eq 'dump' ) {
		return Dumper($partrec);
	} elsif ( $self->{mode} eq 'table' ) {
		my %fields;my $rwl;my $rols;my @fields;
		for my $field (@{$self->{tablefields}}) {
			my $value=$fields{$field->{key}};
			if ( ! defined($value) ) {
				my $name=$field->{name};
				if ( $name eq 'name' ) {
					if ( $self->{hr} ) {
						if ( abs($field->{align}) > 5 ) {
							$value=$namefull;
						} else {
							$value=$nameshort;
						}
					} else {
						$value=$namefull;
					}
					if ( $field->{status} ) {
						if ( ! $partrec->{ponline} ) {
							$value="!" . $value;
						}
					}
				} elsif ( $name eq 'size' ) {
					$value=$self->HR($partrec->{partsize},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'free' ) {
					$value=$self->HR($partrec->{partfree},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'used' ) {
					$value=$self->HR($partrec->{partsize}-$partrec->{partfree},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'instances' ) {
					$value=$self->HR($partrec->{partinstances},'integer','limit' => 9999,'space' => '');
				} elsif ( $name eq 'instancesrw' ) {
					$value=$self->HR($partrec->{partinstancesrw},'integer','limit' => 9999,'space' => '');
				} elsif ( $name eq 'volumes' ) {
					$value=$self->HR($partrec->{partvolumes},'integer','limit' => 9999,'space' => '');
				} elsif ( $name eq 'filessum' ) {
					$value=$self->HR($partrec->{filessum},'integer','limit' => 9999,'space' => '');
				} elsif ( $name eq 'filessumrw' ) {
					$value=$self->HR($partrec->{filessumrw},'integer','limit' => 9999,'space' => '');
				} elsif ( $name eq 'quotasum' ) {
					$value=$self->HR($partrec->{quotasum},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'quotasumrw' ) {
					$value=$self->HR($partrec->{quotasumrw},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'committedrw' ) {
					$value=int($partrec->{quotasumrw}/$partrec->{partsize}*100) . "%";
				} elsif ( $name eq 'sizesum' ) {
					$value=$self->HR($partrec->{sizesum},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'sizesumrw' ) {
					$value=$self->HR($partrec->{sizesumrw},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'committed' ) {
					$value=int($partrec->{quotasum}/$partrec->{partsize}*100) . "%";
				} elsif ( $name eq 'maxmtimerw' ) {
					$value=$partrec->{maxmtimerw};
					if ( $value ) {
						if ( $field->{diff} ) {
							$value=$self->HR(time()-$value,'timeinterval');
						} else {
							$value=$self->HR($value,'localtime/short');
						}
					} else {
						$value='-';
					}
				} elsif ( $name eq 'online' ) {
					$value=$partrec->{ponline};
				} elsif ( $name eq 'infotime' ) {
					$value=$partrec->{infotime};
					if ( $field->{diff} ) {
						$value=$self->HR(time()-$value,'timeinterval');
					} else {
						$value=$self->HR($value,'localtime/short');
					}
				} elsif ( $name eq 'comment' ) {
					if ( $options{'comment'} ) {
						if ( ! $self->{'hr'} ) {
							$options{'comment'} =~ s/\s/_/g;
						}
						$value=$options{'comment'}
					} elsif ( $self->{hr} ) {
						$value='-';
					}
				}
				if ( ! defined($value) ) {
					$value='?';
				}
				$self->align(\$value,$field->{align});
				$fields{$field->{key}}=$value;
			}
			push(@fields,$value);
		}
		return join(' ',@fields);
	} elsif ( $self->{mode} eq 'sums' ) {
		return '';
	}
}


sub header {
	my ($self)=@_;
	my $mode=$self->{mode};
	if ( $mode eq 'table' ) {
		return $self->{tableheader};
	} elsif ( $mode eq 'name' ) {
		if ( $self->{hr} ) {
			return "Volumename Comments";
		} else {
			return "Volumename Comments_if_any";
		}
	}
	return '';
}

=item (\%sort_hash,$sorttype)=makeSortHash(\%volumes)

Creates a hash with key=I<volumename> and value=I<sortkey>. Additionally, the necessary
comparision operation is returned as I<$sorttype>: B<0> numerical comparision necessary,
B<1> string comparision necessary.

=cut

sub makeSortHash {
	my ($self,$partitions)=@_;
	my %sort_hash=();
	my $sorttype;
	my $sortkey=$self->{sortkey};
	if ( $sortkey eq 'serveruuid' ) {
		$sorttype=1;
	} elsif ( $sortkey eq 'partition' ) {
		$sorttype=1;
	} elsif ( $sortkey eq 'name' ) {
		$sorttype=1;
	} elsif ( $sortkey eq 'name#' ) {
		$sorttype=0;
		$sortkey='name';
	} elsif ( $sortkey eq 'partsize' ) {
		$sorttype=0;
	} elsif ( $sortkey eq 'partfree' ) {
		$sorttype=0;
	} elsif ( $sortkey eq 'partused' ) {
		$sorttype=0;
	} elsif ( $sortkey eq 'quotasum' ) {
		$sorttype=0;
	} elsif ( $sortkey eq 'quotasumrw' ) {
		$sorttype=0;
	}
	for my $partrec (@{$partitions}) {
		my $value;
		if ( $sortkey eq 'name' ) {
			if ( defined($partrec->{serveruuid}) ) {
				$value=$self->{celldb}->serverInfo('uuid2name',$partrec->{'serveruuid'}) . "/" . $partrec->{'partition'};
			} else {
				$value=$partrec->{name};
			}
		} elsif ( $sortkey eq 'serveruuid' ) {
			$value=$partrec->{'serveruuid'}
		} elsif ( $sortkey eq 'partition' ) {
			# Only the partition counts. However, we like to be deterministic and add the uuid
			$value=$partrec->{'partition'} . "/" . $partrec->{'serveruuid'};
		} elsif ( $sortkey eq 'partsize' ) {
			$value=$partrec->{'partsize'};
		} elsif ( $sortkey eq 'partfree' ) {
			$value=$partrec->{'partfree'};
		} elsif ( $sortkey eq 'partused' ) {
			$value=$partrec->{'partsize'}-$partrec->{'partfree'};
		} elsif ( $sortkey eq 'quotasum' ) {
			$value=$partrec->{'quotasum'};
		} elsif ( $sortkey eq 'quotasumrw' ) {
			$value=$partrec->{'quotasumrw'};
		}
		$sort_hash{"$partrec"}=$value;
	}
	return (\%sort_hash,$sorttype);
}

sub partitions {
	my ($self,$partitions,%options)=@_;

	my $text='';
	if ( $self->{header} ) {
		$text=$self->header();
		if ( $text ) {
			$text .= $self->{separator};
		} else {
			$text='';
		}
	}

	my @parts;
	if ( $self->{sortkey} eq 'none' ) {
		@parts=@{$partitions};
	} else {
		# sort_hash assigns creates a hash volume->sortvalue according to given sort mode
		my ($sort_keys,$sort_type)=$self->makeSortHash($partitions);
		if ( $sort_type == 0 ) {
			if ( $self->{sortreverse} ) {
				@parts=sort { $sort_keys->{"$b"} <=> $sort_keys->{"$a"} } @{$partitions};
			} else {
				@parts=sort { $sort_keys->{"$a"} <=> $sort_keys->{"$b"} } @{$partitions};
			}
		} else {
			if ( $self->{sortreverse} ) {
				@parts=sort { $sort_keys->{"$b"} cmp $sort_keys->{"$a"} } @{$partitions};
			} else {
				@parts=sort { $sort_keys->{"$a"} cmp $sort_keys->{"$b"} } @{$partitions};
			}
		}
	}
	
	my $comments={};
	if ( $options{'comments'} ) {
		$comments=$options{'comments'};
	}
	for my $partition (@parts) {
		$text .= $self->partition($partition,'comment' => $comments->{"$partition"}) . $self->{separator};
	}
	return $text;
}


=back

=cut

1;
