package AFS::Instant::Output::Volumes;

use 5.006;
use Carp qw(cluck);
use strict;
use warnings;

use Carp;
use Time::Local;
use Data::Dumper;

use Fb::Data qw($MAXINT32 $MAXINT64 human_readable min max);
use Fb::Out;

use AFS::Instant::Toolz qw(getUserPref);

require Exporter;
our @ISA=qw(Exporter);
our $VERSION='1.0';
our @ROFLAGS=('+','-','_'); # map: instance/rostate->char

=head1 NAME

AFS::Instant::Output - Output functions for volume and instance information

=head1 SYNOPSIS

  use AFS::Instant::Output;
  use AFS::Instant::CellDB;
  
  my $celldb=AFS::Instant::CellDB->new();
  my $output=AFS::Instant::Output::Volumes->new(celldb => $celldb);
  my $rootvol=$celldb->getVolume('root.afs');
  print $output->volume($rootvol,'full','hr' => 1);
  
  my $filter=$celldb->parseVolumeFilter(['root.*']);
  my $rootvolumes=$celldb->filterVolumes($filter);
  print $output->volumes([values %{$rootvolumes}]);
  
=head1 REQUIRES

AFS::Instant::CellDB, AFS::Instant::Toolz;

=head1 METHODS

=head2 Constructor

=over 4

=item $output=new(%options)

Creates a new Volume-Data-Output-Object for returning volume/instance
informationin certain formats. Possible options are:

=over 4

=item B<celldb>

A AFS::Instant::CellDB object. This is actually mandatory.

=item B<separator>

Defines the separator between volume records. Defaults to B<\n> (newline).

=item B<header>

Defines, if the default header of the given output (B<1>) mode should be returned prior
to the actual data or not (B<0>).

=item B<mode>

Defines the format in which volumes are presented in. Possible formats are:

=item B<nopref>

Do not lookup output modes in userpref.xml.

=over 4

=item B<name>

Only the names of the volumes.

=item B<bigtable>

A table containing the most vital information about a volume and it's RW instance.

=item B<stat>

A table containing statistical information about all volumes. This mode is only
useful when multiple volumes are being output since a single call to $output->volume()
will return empty.

=item B<full>

All information about the volume (multiple lines).

=item B<none>

Don't output anything

=back

=item B<hr>

Defines if the output should be made human readable (value B<1>) or not (value B<0>).
"human readable" means that

=over 4

=item timestamps 

are convertet to local time values and/or timediff-values

=item integers

are shown as kilo/mega/... expressions.

=item memory values

are show as kibi/mibi/gibi/... expressions.

=back

=back

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self={
		celldb => undef,
		separator => "\n",
		header => undef,
		hr => 0,
		mode => 'full',
		sortkey => 'name',
		sortreverse => 0,
	};
	if ( defined($options{'celldb'}) ) {
		$self->{celldb}=$options{'celldb'};
	} else {
		out('e',"CellDB-object is missing");
		return undef;
	}
	if ( defined($options{'mode'}) ) {
		$self->{mode}=$options{'mode'};
	}
	if ( defined($options{'separator'}) ) {
		$self->{separator}=$options{'separator'};
	}
	if ( defined($options{'header'}) ) {
		$self->{header}=$options{'header'};
	}
	if ( defined($options{'sortkey'}) ) {
		$self->{sortkey}=$options{'sortkey'};
	}
	if ( defined($options{'sortreverse'}) ) {
		$self->{sortreverse}=$options{'sortreverse'};
	}
	if ( defined($options{'hr'}) ) {
		$self->{hr}=$options{'hr'};
	}
	$self=bless($self,$package);

	if ( ! $options{nopref} ) {
		my $userpref=getUserPref();
		if ( $userpref ) {
			my $formats=$userpref->find("//up:format[\@type='output/volumes'][\@id='$self->{mode}']");
			if ( $formats ) {
				my $format=$formats->[0];
				if ( ! defined($self->{hr}) && $format->hasAttribute('hr') ) {
					$self->{hr}=int($format->hasAttribute('hr'));
				}
				if ( ! defined($self->{header}) && $format->hasAttribute('header') ) {
					$self->{header}=int($format->hasAttribute('hr'));
				}
				out('d',"Using custom format '$self->{mode}'");
				$self->{mode}=$format->getAttribute('format');
			} elsif ( $self->{mode} =~ /^name:/ ) {
				$self->{mode}='name';
			}
		}
	}
	if ( $self->{mode} =~ /^table:(.*)$/ ) {
		my @f=split /\|/,$1;
		my %fielddata=(
			name => [22,'Volumename','Volumename'],
			size => [-6,'Size','Current_Size_RW'],
			quota => [-6,'Quota','Quota_RW'],
			quotafree => [-6,'QFree','Quota_Free'],
			quotausage => [-3,"Q%",'Quota_Percent'],
			rwfree => [-6,'RWFree','RW_Space_usable'],
			instances => [-2,'#I','Number_of_Instances'],
			files => [-4,'File',"Number_of_Inodes_RW"],
			partfree => [-6,'PFree','Partfree_RW'],
			quotasum => [-6,'PQ','Partitionquota'],
			quotasumrw => [-6,'PQ_RW','Partitionquota_RW'],
			free => [-6,'Free','Free_space_for_RW'],
			mtime => [15,'Mtim','Mtime[RW]'],
			location => [7,'RW','RW-Location'],
			rolocations => [23,'ROs','Non-Cheap-RO-Locations'],
			infotime => [15,'Info','Last_Update_RW'],
			comment => [0,'Comment','Comment_if_any'],
			'id/rw' => [0,"ID","Volume_ID"],
			'id/ro' => [0,"ROID","Volume_RO_ID"],
		);
		my @tableheader;
		for my $f (@f) {
			my $key=$f;
			my %options;
			if ( $f =~ /(\S+?):(.*)$/ ) {
				$f=$1;
				for my $o (split /:/,$2) {
					if ( $o =~ /^(.*?)=(.*)$/ ) {
						$options{$1}=$2;
					} else {
						$options{$o}=1;
					}
				}
			}
			my $data=$fielddata{$f};
			if ( ! $data ) {
				out('e',"Invalid field specifier '$key'");
				return undef;
			}
			my $align=$data->[0];
			if ( $f =~ /time/ && $options{diff} ) {
				$align=-4;
			}
			if ( ! defined($options{align}) ) {
				$options{align}=$align;
			}
			my $rec={%options,name => $f,key => $key};
			push(@{$self->{tablefields}},$rec);
			my $tableheader=$data->[2-int($self->{hr})];
			$self->align(\$tableheader,$options{align});
			push(@tableheader,$tableheader);
		}
		$self->{tableheader}=join(' ',@tableheader);
		$self->{mode}='table';
	} elsif ( $self->{mode} eq 'bigtable' ) {
		return new('AFS::Instant::Output::Volumes',%options,mode => 'table:name:showlock=1|size|quota|instances|files|partfree|mtime:diff|location|rolocations|comment',hr => 1,header => 1);
	} elsif ( $self->{mode} eq 'stat' ) {
		if ( ! defined($self->{hr}) ) {
			$self->{hr}=1;
		}
	} elsif ( $self->{mode} !~ /^(name|full|dump|none)$/ ) {
		out('e',"Invalid volume output mode '$self->{mode}'");
		return undef;
	}
	if ( $self->{sortkey} !~ /^(name|quota|size|quotafree|rwfree|mtime|atime|ctime|files|instances|quotausage|quotasum|partfree|partused|partsize|none)$/ ) {
		out('e',"Invalid sort key '$self->{sortkey}'");
		return undef;
	}
	return $self;
}

=head2 Object Methods

=over 4

=item $string=$output->HR($value,$hrmode);

Returns the I<$value> in human readable form - if defined via $output->{hr}.
Othervise I<$string> will contain a copy of I<$value>. I<$hrmode> is the mode
given to the AFS::Instant::Toolz::human_readable() method.

=cut

sub HR {
	my ($self,$value,$hrmode,%options)=@_;
	if ( ! defined($value) ) {
		return undef;
	}
	if ( $self->{hr} ) {
		$value=human_readable($value,$hrmode,%options);
		if ( $options{align} ) {
			$self->align(\$value,$options{align});
		}
	}
	return $value;
}

=item $text=$output->instanceFull($instancerec,$indent,%options)

Returns the given I<$instancerec> as a human readable block. These I<%options> are accepted:

=over 4

=item B<indent>

Place indentation before the field names. Value is the number of tabs.

=item B<format>

Set to B<array> will make showInstance() return all fields as an array containing arrarefs like this:

[I<indentationstring>,I<fieldname>,I<fieldvalue>]

=item B<supress>

A hashref containing elements (as keys) that should not be shown. Possible elements are:

=over 4

=item B<volumename>

The name of the associated volume.

=item B<instancetype>

The type of the volume instance (RW,Backup,Readonly)

=item B<partition>

All data regarding the partition, the volume instance resides on. This includes partition name, server name, partition size, free space on the partition and the partition quota sum.

=item B<id>

The volume instance ID.

=item B<times>

All three timestamps of the volume instance.

=back

=back

=cut

sub instanceFull {
	my ($self,$irec,%options)=@_;
	my $indent=0;
	if ( defined($options{'indent'}) ) {
		$indent=$options{'indent'};
	}
	my @t=();
	my $in="   " x $indent;
	if ( ! $options{'supress'}->{'volumename'} ) {
		push(@t,[$in,"volume",$irec->{'volumename'}]);
	}
	if ( ! $options{'supress'}->{'instancetype'} ) {
		push(@t,[$in,"instancetype",$self->{celldb}->instanceIdToType($irec->{'instancetype'})]);
	}
	if ( ! $options{'supress'}->{'partition'} ) {
		push(@t,[$in,"location",$self->{celldb}->serverInfo('uuid2name',$irec->{'serveruuid'}) . ":/vicep$irec->{'partition'}"]);
		if ( defined($irec->{'partfree'}) ) {
			push(@t,[$in,"partition_free",$self->HR($irec->{'partfree'},'memory') . 'B']);
			push(@t,[$in,"partition_size",$self->HR($irec->{'partsize'},'memory') . 'B']);
			push(@t,[$in,"partition_quotasum",$self->HR($irec->{'quotasum'},'memory') . 'B']);
		}
	}
	if ( ! $options{'supress'}->{'id'} ) {
		if ( defined($irec->{'id'}) ) {
			push(@t,[$in,"instanceid",$irec->{'id'}]);
		}
	}
	if ( defined($irec->{'size'}) ) {
		my $append='';
		if ( $irec->{'quota'} and $self->{'hr'} ) {
			$append=" (" . int($irec->{'size'}/$irec->{'quota'}*1000)/10 . "%)"
		}
		push(@t,[$in,"size",$self->HR($irec->{'size'},'memory') . "B" . $append]);
	}
	if ( defined($irec->{'quota'}) ) {
		if ( $irec->{'quota'} != 0 ) {
			push(@t,[$in,"quota",$self->HR($irec->{'quota'},'memory') . "B"]);
		} else {
			push(@t,[$in,"quota","unlimited"]);
		}
	}
	if ( defined($irec->{'files'}) ) {
		push(@t,[$in,"files",$irec->{'files'}]);
	}
	if ( ! $options{'supress'}->{'times'} ) {
		my $time=time();
		for my $field ('mtime','ctime','atime','infotime') {
			if ( defined($irec->{$field}) ) {
				my $diff='';
				if ( $irec->{$field} < $time ) {
					$diff=" (" . $self->HR($time-$irec->{$field},'timeinterval') . ")"
				}
				push(@t,[$in,$field,$self->HR($irec->{$field},'localtime') . $diff]);
			}
		}
	}

	my @flags=();
	if ( $irec->{'isbusy'} ) {
		push(@flags,'busy');
	}
	if ( ! $self->{celldb}->instanceInfo('accessible',$irec) ) {
		push(@flags,'unaccessible');
	}
	my $errors=$self->{celldb}->instanceInfo('errors',$irec);
	push(@flags,sort keys %{$errors});
	if ( $irec->{'instancetype'} == 1 ) {
		if ( ! defined($irec->{'rostate'}) ) {
			push(@flags,'vldb_unknown');
		} elsif ( $irec->{'rostate'} == 2 ) {
			push(@flags,'unreleased');
		} elsif ( $irec->{'rostate'} == 1 ) {
			push(@flags,'old_copy');
		}
	}
	if ( @flags ) {
		push(@t,[$in,"flags",join(' ',@flags)]);
	}
	if ( $options{'format'} ) {
		if ( $options{'format'} eq 'array' ) {
			return @t;
		}
	}
	my $text='';
	for my $field (@t) {
		$text = $text . $field->[0] . $field->[1] . ": " . $field->[2] . "\n";
	}
	return $text;
}

=item $volumeFull($volref,%options)

Returns the volume given via its volume reference I<$volref> as a human readable full
overview. Some output modes tend to display lots of information about the RW-instances
of the volume.

Possible options are:

=over 4

=item B<indent>

Number of indentation characters to put in front of the output. Defaults to B<0>.

=item B<supress>

A hashref containing elements (as keys) that should not be shown. Possible elements are:

=over 4

=item B<volumename>

The name of the volume

=item B<volumeflags>

Do not show boolean function results of the volume. Those flags are in fact boolean resuls of
various functions working on already shown properties of the volume and its instances.

=back

Additionally, all valid keys of the B<supress> option of the instanceFull() method may be given. They
will be forwarded to calls to instanceFull().

=back

=cut

sub volumeFull {
	my ($self,$volrec,%options)=@_;
	my $indent=0;
	if ( $options{'indent'} ) {
		$indent=$options{'indent'};
	}
	my %supress=();
	if ( $options{'supress'} ) {
		%supress=%{$options{'supress'}};
	}
	my $in="   " x $indent;
	my @t=();
	my $hasrw=0;
	if ( ! $options{'supress'}->{'volumename'} ) {
		push(@t,[$in,"volumename",$volrec->{'name'} . "@" . $volrec->{'cellname'}]);
	}
	push(@t,[$in,"id",$volrec->{'id'}]);
	if ( ! $volrec->{'backup'} and $volrec->{'backupid'} ) {
		push(@t,[$in,"id_backup",$volrec->{'backupid'}]);
	}
	if ( $volrec->{'rw'} ) {
		$hasrw=1;
		push(@t,[$in,"instance","rw"]);
		my %suprw=%supress;
		$suprw{'id'}=1;
		$suprw{'volumename'}=1;
		$suprw{'instancetype'}=1;
		push(@t,$self->instanceFull($volrec->{'rw'},'supress' => \%suprw,'format' => 'array','indent' => $indent+1));
	}
	if ( ! $supress{'times'} ) {
		my $time=time();
		if ( $volrec->{'synctry_ro'} ) {
			my $no=0;
			for my $ro (@{$volrec->{'ro'}}) {
				if ( defined($ro->{'ctime'}) ) {
					if ( $ro->{'ctime'} > $volrec->{'synctry_ro'} ) {
						$no=1;
					}
				}
			}
			if ( ! $no ) {
				my $diff='';
				if ( $volrec->{'synctry_ro'} < $time ) {
					$diff=" (" . $self->HR($time-$volrec->{'synctry_ro'},'timeinterval') . ")"
				}
				push(@t,[$in,"sync_ro_lasttry",$self->HR($volrec->{'synctry_ro'},'localtime') . $diff]);
			}
		}
		if ( $volrec->{'synctry_backup'} ) {
			if ( $volrec->{'backup'} ) {
				if ( $volrec->{'backup'}->{'ctime'} < $volrec->{'synctry_backup'} ) {
					my $diff='';
					if ( $volrec->{'synctry_backup'} < $time ) {
						$diff=" (" . $self->HR($time-$volrec->{'synctry_backup'},'timeinterval') . ")"
					}
					push(@t,[$in,"sync_backup_lasttry",$self->HR($volrec->{'synctry_backup'},'localtime') . $diff]);
				}
			}
		}
	}
	if ( $volrec->{'backup'} ) {
		push(@t,[$in,"instance","backup"]);
		my %supb=%supress;
		$supb{'volumename'}=1;
		$supb{'instancetype'}=1;
		if ( $hasrw ) {
			$supb{'partition'}=1;
		}
		push(@t,$self->instanceFull($volrec->{'backup'},'supress' => \%supb,'format' => 'array','indent' => $indent+1));
	}

	for my $ro (@{$volrec->{'ro'}}) {
		my %supro=%supress;
		$supro{'volumename'}=1;
		$supro{'instancetype'}=1;
		my $add='';
		if ( $hasrw ) {
			if (
				( $ro->{'serveruuid'} eq $volrec->{'rw'}->{'serveruuid'} ) and
				( $ro->{'partition'} eq $volrec->{'rw'}->{'partition'} ) ) {
					$supro{'partition'}=1;	
					$add=' (cheap copy)';
			}
		}
		push(@t,[$in,"instance","readonly$add"]);
		push(@t,$self->instanceFull($ro,'supress' => \%supro,'format' => 'array','indent' => $indent+1));
	}
	my @flags;
	for my $flag ('locked','rouptodate','backupuptodate','releasable','rwro') {
		if ( $self->{celldb}->volumeInfo($flag,$volrec) ) {
			push(@flags,$flag);
		}
	}
	if ( @flags ) {
		push(@t,[$in,"flags",join(' ',@flags)]);
	}
	if ( $options{'comment'} ) {
		push(@t,[$in,"comment",$options{'comment'}]);
	}
	my $text='';
	for my $field (@t) {
		$text = $text . $field->[0] . $field->[1] . ": " . $field->[2] . "\n";
	}
	return $text;
}

=item $flag=roflag($rwinstance,$roinstance)

Returns a flag representing the state of the given I<$roinstance>. 

=cut

sub roflag {
	my ($rw,$ro)=@_;
	if ( ! defined($ro->{'rostate'}) ) {
		# RO is unknown to VLDB
		return '?';
	}
	if ( defined($ro->{'ctime'}) and defined($rw->{'mtime'}) && ( $ro->{rostate} == 0 ) ) {
		if ( $ro->{'ctime'} >= $rw->{'mtime'} ) {
			# RO contains the current state of the RW
			return '*';
		}
	}
	return $ROFLAGS[$ro->{'rostate'}];
}

=item align(\$string,$alignment)

Aligns a I<$string> either to the left (I<$alignment> > 0 ) or to the right
(I<$alignment> < 0) of a space (" ") padded text I<abs($alignment)>
characters in length.

=cut

sub align {
	my ($self,$s,$a)=@_;
	if ( ! $self->{hr} ) {
		return;
	}
	my $l=length($$s);
	if ( $a > 0 ) {
		if ( $l < $a ) {
			$$s = $$s . (" " x ($a-$l));
		} elsif ( $l > $a )  {
			$$s = substr($$s,0,$a);
		}
	} elsif ( $a < 0 ) {
		$a=0-$a;
		if ( $l < $a ) {
			$$s=(" " x ($a-$l)) . $$s;
		}
	}
}

=item $text=$output->volume($volume)

Returns a textified version of a single given volume I<$volume>. I<$volume>
can be either a string containing a volume name or a volume reference as
returned by AFS::Instant::CellDB::getVolume() .

=cut

sub volume {
	my ($self,$volref,%options)=@_;
	my $mode=$self->{mode};
	if ( $mode eq 'none' ) {
		return '';
	}
	if ( $mode eq 'name' ) {
		my $commentadd='';
		if ( $options{'comment'} ) {
			if ( ! $self->{'hr'} ) {
				$options{'comment'} =~ s/\s/_/g;
			}
			$commentadd=" " . $options{'comment'}
		}
		if ( ref($volref) eq '' ) {
			return $volref . $commentadd;
		} else {
			return $volref->{'name'} . $commentadd;
		}
	}
	if ( ref($volref) eq '' ) {
		my $volref_=$self->{celldb}->getVolume($volref);
		if ( ! defined($volref_) ) {
			out('e',"volume(): invalid volume name '$volref'");
			return undef;
		}
		$volref=$volref_;
	}
	if ( $mode eq 'full' ) {
		return $self->volumeFull($volref,%options);
	} elsif ( $mode eq 'dump' ) {
		return Dumper($volref);
	} elsif ( $mode eq 'table' ) {
		my %fields;my $rwl;my $rols;my @fields;
		for my $field (@{$self->{tablefields}}) {
			my $value=$fields{$field->{key}};
			if ( ! defined($value) ) {
				my $name=$field->{name};
				if ( $name eq 'name' ) {
					$value=$volref->{name};
					if ( $fields{$field->{key}}->{options}->{showlock} ) {
						if ( $volref->{locked} ) {
							$value="!$value";
						}
					}
				} elsif ( $name eq 'size' ) {
					$value=$self->HR($volref->{rw}->{size},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'quota' ) {
					$value=$self->HR($volref->{rw}->{quota},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'quotausage' ) {
					$value=int($volref->{rw}->{size}*100/$volref->{rw}->{quota});
				} elsif ( $name eq 'quotafree' ) {
					$value=$volref->{rw}->{quota}-$volref->{rw}->{size};
					if ( $value < 0 ) {
						$value=0;
					}
					$value=$self->HR($value,'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'rwfree' ) {
					$value=$volref->{rw}->{quota}-$volref->{rw}->{size};
					if ( $value < 0 ) {
						$value=0;
					}
					if ( $volref->{rw}->{partfree} < $value ) {
						$value=$volref->{rw}->{partfree};
					}
					$value=$self->HR($value,'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'instances' ) {
					$value=int(exists $volref->{rw}) + int(exists $volref->{backup}) + scalar(@{$volref->{ro}});
				} elsif ( $name eq 'files' ) {
					$value=$self->HR($volref->{rw}->{files},'integer','limit' => 999,'space' => '');
				} elsif ( $name eq 'partfree' ) {
					$value=$self->HR($volref->{rw}->{partfree},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'quotasum' ) {
					$value=$self->HR($volref->{rw}->{quotasum},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'quotasumrw' ) {
					$value=$self->HR($volref->{rw}->{quotasum},'memory','limit' => 9999,'space' => '');
				} elsif ( $name eq 'free' ) {
					if ( $volref->{rw} ) {
						$value=$volref->{rw}->{partfree};
						if ( $volref->{rw}->{quota} ) {
							my $qfree=$volref->{rw}->{quota} - $volref->{rw}->{size};
							if ( $qfree < 0 ) {
								$qfree=0;
							}
							if ( $qfree < $value ) {
								$value=$qfree;
							}
						}
						$value=$self->HR($value,'memory','limit' => 9999,'space' => '');
					}
				} elsif ( $name eq 'mtime' ) {
					$value=$volref->{rw}->{mtime};
					if ( $field->{diff} ) {
						if ( defined($value) ) {
							$value=time()-$value;
						}
						$value=$self->HR($value,'timeinterval');
					} else {
						$value=$self->HR($value,'localtime/short');
					}
				} elsif ( $name eq 'infotime' ) {
					$value=$volref->{rw}->{infotime};
					if ( $field->{diff} ) {
						$value=$self->HR(time()-$value,'timeinterval');
					} else {
						$value=$self->HR($value,'localtime/short');
					}
				} elsif ( $name eq 'location' ) {
					if ( ! defined($rwl) ) {
						($rwl,$rols)=$self->getRWandROs($volref);
					}
					$value=$rwl;
				} elsif ( $name eq 'rolocations' ) {
					if ( ! defined($rols) ) {
						($rwl,$rols)=$self->getRWandROs($volref);
					}
					$value=$rols;
				} elsif ( $name eq 'id/rw' ) {
					$value=$volref->{id};
				} elsif ( $name eq 'id/ro' ) {
					$value=$volref->{roid};
				} elsif ( $name eq 'comment' ) {
					if ( $options{'comment'} ) {
						if ( ! $self->{'hr'} ) {
							$options{'comment'} =~ s/\s/_/g;
						}
						$value=$options{'comment'}
					} elsif ( $self->{hr} ) {
						$value='-';
					}
				}
				if ( ! defined($value) ) {
					$value='?';
				}
				$self->align(\$value,$field->{align});
				$fields{$field->{key}}=$value;
			}
			push(@fields,$value);
		}
		return join(' ',@fields);
	} elsif ( $mode eq 'sums' ) {
		return '';
	}
}

sub getRWandROs {
	my ($self,$volref)=@_;
	my $rwl='?';
	my $rwroadd='';
	if ( $volref->{'rw'}->{'id'} ) {
		my $server=$self->{celldb}->serverInfo('uuid2name',$volref->{'rw'}->{'serveruuid'});
		$server =~ /^([^\.]+)/;
		if ( $self->{'hr'} ) {
			$rwl=substr($1,0,3);
		} else {
			$rwl=$server;
		}
		$rwl=$rwl . '/' . $volref->{'rw'}->{'partition'};
		if ( ! $self->{celldb}->instanceInfo('accessible',$volref->{'rw'}) ) {
			$rwl="!$rwl";
		}
	}
	my @rols=();
	foreach my $ro (@{$volref->{'ro'}}) {
		my $server=$self->{celldb}->serverInfo('uuid2name',$ro->{'serveruuid'});
		my $rwro=0;
		if ( $volref->{'rw'}->{'id'} ) {
			if ( $ro->{'serveruuid'} eq $volref->{'rw'}->{'serveruuid'} ) {
				if ( $ro->{'partition'} eq $volref->{'rw'}->{'partition'} ) {
					my $roflag=roflag($volref->{'rw'},$ro);
					$rwroadd .= $roflag;
					if ( $roflag ne '_' ) {
						# Since '_' (rostate==2) means "there's no physical volume",
						# we don't need another offline-hint
						if ( ! $self->{celldb}->instanceInfo('accessible',$ro) ) {
							$rwroadd="!$rwroadd";
						}
					}
					$rwro=1;
				}
			}
		} else {
			$rwl='-';
		}
		if ( ! $rwro ) {
			my $rol;
			$server =~ /^([^\.]+)/;#
			if ( $self->{'hr'} ) {
				$rol=substr($1,0,3);
			} else {
				$rol=$server;
			}
			$rol = $rol . '/' . $ro->{'partition'} . roflag($volref->{'rw'},$ro);
			if ( ! $self->{celldb}->instanceInfo('accessible',$ro) ) {
				$rol="!$rol";
			}
			push(@rols,$rol);
		}
	}
	if ( $rwl ne '?' ) {
		$rwl .= $rwroadd;
	}
	return ($rwl,join(',',sort @rols));
}

sub header {
	my ($self)=@_;
	my $mode=$self->{mode};
	if ( $mode eq 'table' ) {
		return $self->{tableheader};
	} elsif ( $mode eq 'name' ) {
		return "Volumename";
	}
	return '';
}


=item (\%sort_hash,$sorttype)=makeVolumeSortHash(\%volumes)

Creates a hash with key=I<volumename> and value=I<sortkey>. Additionally, the necessary
comparision operation is returned as I<$sorttype>: B<0> numerical comparision necessary,
B<1> string comparision necessary.

=cut

sub makeSortHash {
	my ($self,$volumes)=@_;
	my %sort_hash=();
	my $sorttype=0;
	# Text based sorting instead of numeric
	if ( $self->{sortkey} eq 'name' ) {
		$sorttype=1;
	}
	for my $volume (@{$volumes}) {
		my $volref=$self->{celldb}->getVolume($volume);
		my $value;
		if ( $self->{sortkey} eq 'name' ) {
			$value=$volume;
		} elsif ( $self->{sortkey} eq 'quota' ) {
			$value=$volref->{'rw'}->{'quota'};
		} elsif ( $self->{sortkey} eq 'quotafree' ) {
			if ( $volref->{'rw'}->{'quota'} > $volref->{'rw'}->{'size'} ) {
				$value=$volref->{'rw'}->{'quota'}-$volref->{'rw'}->{'size'};
			} else {
				$value=0;
			}
		} elsif ( $self->{sortkey} eq 'rwfree' ) {
			if ( $volref->{'rw'}->{'quota'} > $volref->{'rw'}->{'size'} ) {
				$value=$volref->{'rw'}->{'quota'}-$volref->{'rw'}->{'size'};
				if ( $volref->{'rw'}->{'partfree'} < $value ) {
					$value=$volref->{'rw'}->{'partfree'};
				}
			} else {
				$value=0;
			}
		} elsif ( $self->{sortkey} eq 'size' ) {
			$value=$volref->{'rw'}->{'size'};
		} elsif ( $self->{sortkey} eq 'mtime' ) {
			$value=$volref->{'rw'}->{'mtime'};
		} elsif ( $self->{sortkey} eq 'atime' ) {
			$value=$volref->{'rw'}->{'atime'};
		} elsif ( $self->{sortkey} eq 'ctime' ) {
			$value=$volref->{'rw'}->{'ctime'};
		} elsif ( $self->{sortkey} eq 'files' ) {
			$value=$volref->{'rw'}->{'files'};
		} elsif ( $self->{sortkey} eq 'instances' ) {
			$value=0;
			if ( defined( $volref->{'rw'} ) ) {
				$value++;
			}
			if ( defined( $volref->{'backup'} ) ) {
				$value++;
			}
			$value += scalar(@{$volref->{'ro'}});
		} elsif ( $self->{sortkey} eq 'quotausage' ) {
			if ( $volref->{'rw'}->{'quota'} ) {
				$value=$volref->{'rw'}->{'size'}/$volref->{'rw'}->{'quota'};
			} else {
				$value=0;
			}
		} elsif ( $self->{sortkey} eq 'partfree' ) {
			$value=$volref->{rw}->{'partfree'};
		} elsif ( $self->{sortkey} eq 'partused' ) {
			$value=$volref->{rw}->{'partsize'}-$volref->{rw}->{'partfree'};
		} elsif ( $self->{sortkey} eq 'partsize' ) {
			$value=$volref->{rw}->{'partsize'};
		} elsif ( $self->{sortkey} eq 'quotasum' ) {
			$value=$volref->{rw}->{'quotasum'};
		}
		$sort_hash{$volume}=$value;
	}
	return (\%sort_hash,$sorttype);
}

=item $text=$output->volumes(\%volumes,%options)

Returns a I<$text> block containing the output of the given volumes I<%volumes>. All format definitions
were already done when the AFS::Instant::Output object was created.

Possible I<%options> are:

=over 4

=item B<comments>

A hashref containing a volumename to string map. These strings are to be displayed near the volumes
to explain why they are displayed. This is i.e. used by the volume constraint check of
instantafs.check.

=item B<format> (Default: B<string>)

Set returned format. Possible formats are:

=over 4

=item B<arrayref>

=item B<string>

=back

=back

=cut

sub volumes {
	my ($self,$volumes,%options)=@_;
	my $format='string';
	if ( defined($options{format}) ) {
		$format=$options{format};
	}
	if ( $self->{mode} eq 'stat' ) {
		my %idata;
		for my $t (0,1,2) {
			for my $f ('files','quota','size','online','offline') {
				$idata{$t}->{$f}=0;
			}
			$idata{$t}->{servers}={};
			$idata{$t}->{partitions}={};
		}
		my $missing_ros=0;my $missing_rws=0;# Volumes without a certain type of instance in the cell
		my $vldbunknown_ros=0;my $vldbunknown_rws=0; # instances not registered in VLDB
		my $unreleased_ros=0;my $old_ros=0; #####
		my $volumecount=scalar(keys %{$volumes});
		my $rwstorage_min=$MAXINT64;
		my $partfree_min=$MAXINT64;
		my $busy_instances=0;my $salvage_instances=0;
		my $instances=0;my %partitions;my %servers;
		my $roage_min=$MAXINT32;my $roage_max=0;
		my $backupage_min=$MAXINT32;my $backupage_max=0;
		for my $volume (keys %{$volumes}) {
			my $volref;
			if ( ref($volume) eq '' ) {
				$volref=$self->{celldb}->getVolume($volume);
			} else {
				$volref=$volume;
			}
			my $roage=$self->{celldb}->volumeInfo('roage',$volref);
			if ( defined($roage) ) {
				$roage_min=min($roage_min,$roage);
				$roage_max=max($roage_max,$roage);
			}
			my $backupage=$self->{celldb}->volumeInfo('backupage',$volref);
			if ( defined($backupage) ) {
				$backupage_min=min($backupage_min,$backupage);
				$backupage_max=max($backupage_max,$backupage);
			}
			for my $i (@{$self->{celldb}->iterateInstances($volref)}) {
				$partitions{"$i->{serveruuid}/$i->{partition}"}=1;
				$servers{$i->{serveruuid}}=1;
				$idata{$i->{'instancetype'}}->{partitions}->{"$i->{serveruuid}/$i->{partition}"}=1;
				$idata{$i->{'instancetype'}}->{servers}->{$i->{serveruuid}}=1;
				$idata{$i->{'instancetype'}}->{count}++;
				if ( defined($i->{size}) ) {
					if ( ! defined($i->{files}) ) {
						print Dumper($i);
					}
					$idata{$i->{'instancetype'}}->{'files'} += $i->{'files'};
					$idata{$i->{'instancetype'}}->{'quota'} += $i->{'quota'};
					$idata{$i->{'instancetype'}}->{'size'} += $i->{'size'};
					$busy_instances += $i->{'isbusy'};
					$salvage_instances += $i->{'needssalvage'};
				}
				if ( $self->{celldb}->instanceInfo('accessible',$i) ) {
					$idata{$i->{'instancetype'}}->{'online'}++;
				} else {
					$idata{$i->{'instancetype'}}->{'offline'}++
				}
				if ( $i->{instancetype} == 0 ) {
					$rwstorage_min=min($rwstorage_min,$self->{celldb}->instanceInfo('storageleft',$i));
					$partfree_min=min($partfree_min,$i->{partfree});
				}
				if ( ! defined($i->{rostate}) ) {
					if ( $i->{instancetype} == 0 ) {
						$vldbunknown_rws++;
					} elsif ( $i->{instancetype} == 1 ) {
						$vldbunknown_ros++;
					}
				} else {
					if ( $i->{instancetype} == 1 ) {
						if ( $i->{rostate} == 1 ) {
							$old_ros++;
						} elsif ( $i->{rostate} == 2 ) {
							$unreleased_ros++;
						}
					}
				}
				$instances++;
			}
			if ( ! defined($volref->{rw}) ) {
				$missing_rws++;
			}
			if ( ! defined($volref->{ro}) || ! @{$volref->{ro}} ) {
				$missing_ros++;
			}
		}
		my $size=0;my $commited=0;my $accessible_instances=0;my $unaccessible_instances=0;
		my $files=0;
		my $servers=scalar(keys %servers);
		my $partitions=scalar(keys %partitions);
		for my $t (0,1,2) {
			$size += $idata{$t}->{'size'};
			$commited += $idata{$t}->{'quota'},
			$accessible_instances += $idata{$t}->{'online'};
			$unaccessible_instances += $idata{$t}->{'offline'};
			$files += $idata{$t}->{files};
			$idata{$t}->{servers_count}=scalar(keys %{$idata{$t}->{servers}});
			$idata{$t}->{partitions_count}=scalar(keys %{$idata{$t}->{partitions}});
		}
		$size=$self->HR($size,'memory');
		$commited=$self->HR($commited,'memory');
		$accessible_instances=$self->HR($accessible_instances,'integer');
		$unaccessible_instances=$self->HR($unaccessible_instances,'integer');
		$files=$self->HR($files,'integer');
		$missing_rws=$self->HR($missing_rws,'integer');
		$missing_ros=$self->HR($missing_ros,'integer');
		$rwstorage_min=$self->HR($rwstorage_min,'memory');
		$partfree_min=$self->HR($partfree_min,'memory');
		$volumecount=$self->HR($volumecount,'integer');
		for my $t (0,1,2) {
			for my $f ('quota','size') {
				$idata{$t}->{$f}=$self->HR($idata{$t}->{$f},'memory');
			}
			for my $f ('files','online','offline') {
				$idata{$t}->{$f}=$self->HR($idata{$t}->{$f},'integer');
			}
		}
		$backupage_max=time()-$backupage_max;
		$roage_max=time()-$roage_max;
		$backupage_max=$self->HR($backupage_max,'timeinterval');
		$roage_max=$self->HR($roage_max,'timeinterval');
		return "size $size
size_ro $idata{1}->{size}
size_rw $idata{0}->{size}
commited $commited
commited_rw $idata{0}->{quota}
volumes $volumecount
volumes_without_rw $missing_rws
volumes_without_ro $missing_ros
instances $instances
instances_accessible $accessible_instances
instances_accessible_rw $idata{0}->{online}
instances_accessible_ro $idata{1}->{online}
instances_accessible_backup $idata{2}->{online}
instances_unaccessible $unaccessible_instances
instances_unaccessible_rw $idata{0}->{offline}
instances_unaccessible_ro $idata{1}->{offline}
instances_unaccessible_backup $idata{2}->{offline}
instances_no_vldb_rw $vldbunknown_rws
instances_no_vldb_ro $vldbunknown_ros
instances_ro_old $old_ros
instances_ro_unreleased $unreleased_ros
instances_busy $busy_instances
instances_in_need_of_salvage $salvage_instances
files $files
files_rw $idata{0}->{files}
files_ro $idata{1}->{files}
files_backup $idata{2}->{files}
sync_oldest_ro $roage_max
sync_oldest_backup $backupage_max
fileservers $servers
fileservers_rw $idata{0}->{servers_count}
locations $partitions
locations_rw $idata{0}->{partitions_count}
locations_free_rw_min $partfree_min
storage_left_rw_min $rwstorage_min
";
	}

	my @vols;
	if ( $self->{sortkey} eq 'none' ) {
		@vols=keys %{$volumes};
	} elsif ( $self->{sortkey} eq 'name' ) {
		if ( $self->{sortreverse} ) {
			@vols=sort { $b cmp $a } keys %{$volumes};
		} else {
			@vols=sort keys %{$volumes};
		}
	} else {
		# sort_hash assigns creates a hash volume->sortvalue according to given sort mode
		my ($sort_keys,$sort_type)=$self->makeSortHash([keys %{$volumes}]);
		if ( $sort_type == 0 ) {
			if ( $self->{sortreverse} ) {
				@vols=sort { $sort_keys->{$b} <=> $sort_keys->{$a} } keys %{$volumes};
			} else {
				@vols=sort { $sort_keys->{$a} <=> $sort_keys->{$b} } keys %{$volumes};
			}
		} else {
			if ( $self->{sortreverse} ) {
				@vols=sort { $sort_keys->{$b} cmp $sort_keys->{$a} } keys %{$volumes};
			} else {
				@vols=sort { $sort_keys->{$a} cmp $sort_keys->{$b} } keys %{$volumes};
			}
		}
	}
	my @output;
	my $comments={};
	if ( $options{'comments'} ) {
		$comments=$options{'comments'};
	}
	
	foreach my $volume (@vols) {
		my $single=$self->volume($volume,'comment' => $comments->{$volume});
		push(@output,$single);
	}
	if ( @output ) {
		if ( $self->{header} ) {
			my $header=$self->header();
			if ( defined($header) ) {
				unshift(@output,$header);
			}
		}
		if ( $format eq 'string' ) {
			return join($self->{separator},@output) . $self->{separator};
		} elsif ( $format eq 'arrayref' ) {
			return \@output;
		}
	} else {
		if ( $format eq 'string' ) {
			return '';
		} elsif ( $format eq 'arrayref' ) {
			return [];
		}
	}
}

=back

=head1 APPENDIX

=head2 Output Modes

=over 4

=item B<name>

Each record written includes just the volume name

=item B<bigtable>

Each record contains a rich set of information about the volume and its RW instance. These information are included:

=over 4

=item B<volumename>

The name of the volume. A B<!> is appended, whenever the volume is locked in VLDB and HR-mode is effective.

=item B<locked>

This field is only shown in non-HR-mode. It shows B<1> if the volume is locked and B<0> otherwise.

=item B<size>

Current size of the RW instance - in I<X>iB-units if necessary. If no RW instance exists, a "-" is shown. In non-HR-mode the size is given in Bytes.

=item B<quota>

Current quota of the RW instance - in I<X>iB-units if necessary. If no RW instance exists, a "-" is shown. In non-HR-mode the size is given in Bytes.

=item B<instances>

Number of logical instances of the volume - with B<kMGTP> suffix if necessary in HR-mode. This includes all RW-, RO- and Backup instances.

=item B<files>

Number of files in the RW instance - with B<kMGTP> suffix if necessary in HR-mode. More precisely: the number of Inodes. If no RW instance exists, a "-" is shown.

=item B<partfree>

Free space on the RW partition - in I<X>iB-units if necessary. If no RW instance exists, a "-" is shown. In non-HR-mode the size is given in Bytes.

=item B<mdiff>

Time since the last change to the RW instance ( time() - mtime(rw) ). If no RW instance exists, a "-" is shown. In non-HR-mode the value is 

=item B<rwlocation>

Location of the RW instance. The location is shown as the first three
letters of the fileserver's primary name + "/" + the partition name.  If an
RO-instance exists on the RW's partition (a "cheap copy") a flag
representing the RO's state is added.

=over 4

=item "B<?>"

The RO exists physically on the fileserver but VLDB doesn't know about it.
Possible reasons:

=over 4

=item * instance was removed by B<vos remsite> instead of B<vos remove>

=item * Harddisk was physically moved between servers

=item * (Unix-)Mountpoints were exchanged on a server and no B<vos syncvldb> was done

=back

This instance is no accessible by regular AFS-clients.

=item "B<_>"

The RO exists only in VLDB - not on a fileserver's data partition. Possible Reasons:

=over 4

=item * The RO was create via vos addsite and the volume hasn't been released, yet.

=item * The volume instance's data partition is not accessible.

=back

=item "B<->"

The RO is an old copy. RO-Copies are updated one by one when a volume is "released". These
reasons may lead to that flag:

=over 4

=item * The instance on this fileserver hasn't been updated, yet and the release is still in progress (perfectly normal).

=item * The fileserver of this instance was offline during release.

=item * No network connectivity between the RO's and the RW's fileserver.

=item * The release process was canceled. (In this case, the volume is usually still locked)

=back

=item "B<+>"

The RO's state Is perfectly fine. It's doesn't represent the latest state of the RW.

=item "B<*>"

The RO is currently representing the exact state of the RW instance. In
technical terms: the RW's mtime is not after the RO's ctime.

=back

A "B<!>" character before the location string shows that the instance is currently
unavailable. Possible reasons are:

=over 4

=item I<offline> instance flag

The volume instance was taken offline - i.e. by B<vos offline>.

=item fileserver or partition offline

The fileserver or the instance's data partition is currently unavailable.
Possible reasons are network problems, the server crashed or the server is
just turned off.

=item I<busy>

The instance is currently being snapshotted, restored, dumped, ... which
means, the B<volserver> process on the fileserver machine needs the instance
exclusively.

=item I<needssalvage>

The instance was taken offline because the fileserver thinks, it's broken.
This happens i.e.  because the fileserver was restarted after a crash and no
salvage run was done.  Another reason is an IO error caused i.e.  by a
failing RAID or a "Pending Block" (check S.M.A.R.T.  information) on the
fileserver's data disk.

=back

=item B<rolocations>

Locations of all RO instances which do not reside on the RW's fileserver. Format is the same
as for I<rwlocation> with the logical exception that there's at most one RO instance 
on a given fileserver. The notation of an additional instance is therefor not used here.

=back

=back

=cut

1;
