package AFS::Instant::ServerUsers;

use 5.006;
use strict;
use warnings;
use Carp;

use Expect;

use Fb::IO qw(iaexec);
use Fb::Data qw(check_passwd bash_quote);
use Fb::Out;

use AFS::Instant::Toolz qw(read_mount make_mount remove_mount);
use AFS::Instant::Distri;
use AFS::Instant::CellDB;
use AFS::Instant::Manage;

require Exporter;
our @ISA=qw(Exporter);
our $VERSION='1.0';

=head1 NAME

AFS::Instant::ServerUsers - An object represention of privileged user related information about an InstantAFS-cell.

=head1 SYNOPSIS

 
use AFS::Instant::ServerUsers;

my $serverusers=AFS::Instant::ServerUsers->new($users);


=head1 REQUIRES

AFS::Instant::Toolz, AFS::Instant::Distri

=head1 DESCRIPTION

This class represents all user related (public) data of an AFS cell. It is to be used
on servers only. The definition of servers in this case is:

=over 4

=item AFS-Key

The computer is an AFS-server and the AFS-key is located @ /etc/openafs/servers/KeyFile .

=item acall-keytab

The acall-principal's keytab is located @ /etc/instantafs/acall.keytab

=back

=head1 METHODS

=head2 Constructor

=over 4

=item $users=new(%options)

Creates a new user-management object.

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self={
		manage => undef,
		celldb => undef,
		kadmin_local => '/usr/sbin/kadmin.local',
	};
	if ( defined($options{'celldb'}) ) {
		$self->{celldb}=$options{'celldb'};
	} else {
		if ( ! defined($options{'manage'}) ) {
			$self->{celldb}=AFS::Instant::CellDB->new();
		} else {
			$self->{celldb}=$options{'manage'}->{celldb};
		}
	}
	if ( defined($options{'manage'}) ) {
		$self->{manage}=$options{'manage'};
	} else {
		$self->{manage}=AFS::Instant::Manage->new(celldb => $self->{celldb});
	}
	my $me=bless($self,$package);
	
	return $me;
}

=head2 Database Access

=over 4

=item $state=$serverusers->getUserState($user,$mode)

=cut

sub getUserState {
	my ($self,$user,$mode)=@_;
	if ( $mode eq 'db' ) {
		return $self->get_state_db($user);
	} elsif ( $mode eq 'krb5' ) {
		return $self->get_state_krb5($user);
	} elsif ( $mode eq 'homedir' ) {
		return $self->get_state_homedir($user);
	} else {
		out('e',"Invalid mode '$mode'");
		return undef;
	}
}

sub setUserState {
	my ($self,$user,$mode,$newstate,%options)=@_;
	if ( $mode eq 'db' ) {
		return $self->set_state_db($user,$newstate,$options{'id'});
	} elsif ( $mode eq 'krb5' ) {
		return $self->set_state_krb5($user,$newstate,$options{'passwd'});
	} elsif ( $mode eq 'homedir' ) {
		return $self->set_state_homedir($user,$newstate);
	} else {
		out('e',"Invalid mode '$mode'");
		return undef;
	}
}

sub setUserPassword {
	my ($self,$user,$mode,$newpassword)=@_;
	if ( $mode eq 'krb5' ) {
		return $self->set_passwd_krb5($user,$newpassword);
	} else {
		out('e',"Invalid mode '$mode'");
		return undef;
	}
}

=item $state=$serverusers->get_state_krb5($user);

Returns the state of given user I<$user> within the Kerberos5-Database.

=cut

sub get_state_krb5 {
	my ($self,$user)=@_;
	my $cmd="$self->{kadmin_local} -q 'getprinc $user'";
	out('v',"Reading krb5-state of user '$user'");
	my $kaoutput=iaexec(
		[$self->{kadmin_local},'-q',"getprinc $user"],
		'readpipe',
		description => "Reading principal data of '$user'",
		format => 'arrayref',
		stderr => 'hide',
	);
	if ( $kaoutput ) {
		for my $l (@{$kaoutput}) {
			if ( $l =~ /^Expiration date: \[(never|none)\]$/ ) {
				return 'ACTIVE';
			} elsif ( $l =~ /^Expiration date: [A-Z]/ ) {
				return 'NOTACTIVE';
			}
		}
		return 'NOTEXISTENT';
	} else {
		return 'UNKNOWN';
	}
}

=item $result=$serverusers->set_state_krb5($user,$newstate,{$passwd});

Sets the given user I<$user>'s state to I<$newstate>. If I<$passwd> is
given, it will be applied ( if the I<$newstate> ne 'B<NOTEXISTENT>' ).

I<$result> == 1, only if the new state was set successfully.

This function is idempotent.

=cut

sub set_state_krb5 {
	my ($self,$user,$newstate,$passwd)=@_;
	my $state=$self->get_state_krb5($user);

	# This is not necessary for all state changes but it doesn't hurt to do it...
	my $passwdparam='';

	if ( $newstate !~ /^(NOTEXISTENT|NOTACTIVE|ACTIVE)$/ ) {
		out('e',"set_state_krb5: unknown/unsupported target state '$newstate'");
		return 0;
	}
	$passwdparam='-randkey';
	if ( defined($passwd) ) {
		if ( ! check_passwd($passwd) ) {
			out('e',"Given password is invalid. Please choose another one.");
			return 0;
		}
		$passwdparam="-pw \"$passwd\"";
	}
  	if ( $state eq 'UNKNOWN' ) {
		out('e',"krb5-state of user '$user' is unknown, new state '$newstate' won't be applied.");
		return 0;
	} elsif ( $state eq 'NOTEXISTENT' ) {
		my $defaultparams="-maxlife 72hours +requires_preauth";
		if ( $newstate eq 'ACTIVE' ) {
			if ( iaexec("kadmin.local -q 'addprinc -policy user -expire never $passwdparam $defaultparams $user' 2>/dev/null | grep -q 'Principal.*created.'",'exec') ) {
				return 1;
			} else {
				out('e',"unable to set user '$user' to krb5-state 'ACTIVE' (create)");
				return 0;
			}
		} elsif ( $newstate eq 'NOTACTIVE' ) {
			if ( iaexec("kadmin.local -q 'addprinc -expire yesterday -policy user $passwdparam $defaultparams $user' 2>/dev/null | grep -q 'Principal.*created.'",'exec') ) {
				return 1;
			} else {
				out('e',"unable to set user '$user' to krb5-state 'NOTACTIVE' (create)");
				return 0
			}
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			return 1;
		}
	} elsif ( $state eq 'NOTACTIVE' ) {
		if ( defined($passwd) ) {
			iaexec("kadmin.local -q 'modprinc -policy user $user'",'exec');
			iaexec("kadmin.local -q 'cpw -randkey $user'",'exec');
			if ( ! iaexec("kadmin.local -q 'cpw $passwdparam $user'",'exec') ) {
				out('e',"unable to apply given krb5-password to user '$user'.");
				return 0;
			}
		}
		if ( $newstate eq 'ACTIVE' ) {
			if ( iaexec("kadmin.local -q 'modprinc -policy user -expire never $user' 2>/dev/null | grep -q 'Principal.*modified.'",'exec') ) {
				return 1;
			} else {
				out('e',"unable to set user '$user' to krb5-state 'ACTIVE' (enable)");
				return 0;
			}
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			if ( iaexec("kadmin.local -q 'delprinc -force $user' 2>/dev/null | grep -v 'Principal.*deleted.'",'exec') ) {
				return 1;
			} else {
				out('e',"unable to set user '$user' to krb5-state 'NOTEXISTENT' (remove)");
				return 0;
			}
		} elsif ( $newstate eq 'NOTACTIVE' ) {
			return 1;
		}
	} elsif ( $state eq 'ACTIVE' ) {
		if ( defined($passwd) ) {
			iaexec("kadmin.local -q 'modprinc -policy user $user'","exec",description => "Applying user policy");
			iaexec("kadmin.local -q 'cpw -randkey $user'","exec",description => "Cleaning up kerberos principal");
			if ( ! iaexec("kadmin.local -q 'cpw $passwdparam $user'",'exec',description => "Applying actual password") ) {
				out('e',"unable to apply given krb5-password to user '$user'.");
				return 0;
			}
		}
		if ( $newstate eq 'NOTACTIVE' ) {
			if ( iaexec("kadmin.local -q 'modprinc -expire yesterday $user' 2>/dev/null | grep -q 'Principal.*modified.'",'exec') ) {
				return 1;
			} else {
				out('e',"unable to set user '$user' to krb5-state 'NOTACTIVE' (disable)");
				return 0;
			}
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			if ( iaexec("kadmin.local -q 'delprinc -force $user' 2>/dev/null | grep -v 'Principal.*deleted.'",'exec') ) {
				return 1;
			} else {
				out('e',"Enable to set user '$user' to krb5-state 'NOTEXISTENT' (remove)");
				return 0;
			}
		} elsif ( $newstate eq 'ACTIVE' ) {
			return 1;
		}
	}
	out('e',"set_state_krb5: Don't know, what to do ( '$state' => '$newstate' ), internal error.");
	exit 1;
}

=item $result=set_passwd_krb5($user,$passwd)

Changes user I<$user>'s passwd within the Kerberos5-Database.

I<$result> == 1, only if the new password was applied successfully.

=cut

sub set_passwd_krb5 {
	my ($self,$user,$passwd)=@_;
	my $state=$self->get_state_krb5($user);

	if ( $state !~ /^(ACTIVE|NOTACTIVE)$/ ) {
		return 0;
	}
	
	if ( ! check_passwd($passwd) ) {
		out('e',"Given password is invalid. Please choose another one.");
		return 0;
	}
	
	# We won't use quotemeta here. The check_passwd()-sub will watch for '"'-characters
	# all others won't be a security risk as long as they are shell-quoted in '"'-chars.
	iaexec("kadmin.local -q 'modprinc -policy $user'",'exec');
	iaexec("kadmin.local -q 'cpw -randkey $user'",'exec');
	if ( iaexec("kadmin.local -q 'cpw -pw \"$passwd\" $user'",'exec') ) {
		return 1;
	} else {
		out('e',"unable to change password of principal '$user'");
		return 0;
	}
}


sub get_users_db {
	my ($self)=@_;
	my $group_users=$self->{celldb}->groupInfo('members','users');
	my $group_users_na=$self->{celldb}->groupInfo('members','users-na');
	my %userdb=();
	for my $user (keys %{$group_users},keys %{$group_users_na}) {
		my $users=exists $group_users->{$user};
		my $users_na=exists $group_users_na->{$user};
		my $state;
		if ( $users and $users_na ) {
			$state='VAGUE';
		} elsif ( $users and ! $users_na ) {
			$state='ACTIVE';
		} elsif ( ! $users and $users_na ) {
			$state='NOTACTIVE';
		}
		$userdb{$user}->{'state'}->{'db'}=$state;
	}
	return \%userdb;
}

=item $state=get_state_db($user)

Returns the current state of user I<$user> according to AFS' PTDB.

=cut

sub get_state_db {
	my ($self,$user)=@_;
	my $group_users=$self->{celldb}->userInfo('ismemberof',$user,'group' => 'users');
	my $group_users_na=$self->{celldb}->userInfo('ismemberof',$user,'group' => 'users-na');
	my $user_exists=$self->{celldb}->userInfo('name2id',$user);
	if ( ! $user_exists ) {
		return 'NOTEXISTENT';
	} elsif ( ! defined($group_users) or ! defined($group_users_na) or ! $user_exists ) {
		return 'UNKNOWN';
	} elsif ( $group_users and ! $group_users_na) {
		return 'ACTIVE';
	} elsif ( ( ! $group_users ) and ( $group_users_na ) ) {
		return 'NOTACTIVE';
	} elsif ( ! $group_users and ! $group_users_na ) {
		if ( $user_exists ) {
			return 'VAGUE';
		} else {
			return 'NOTEXISTENT';
		}
	} elsif ( ( $group_users == 1 ) and ( $group_users_na == 1 ) ) {
		return 'VAGUE';
	}
}



=item $result=$serverusers->set_state_db($user,$newstate{,$id});

Sets the given user I<$user>'s PTBD-state to I<$newstate>.

I<$result> == 1, only if the new state was set successfully.

This function is idempotent.

=cut

sub set_state_db {
	my ($self,$user,$newstate,$id)=@_;
	my $state=$self->get_state_db($user);
	if ( $newstate !~ /^(NOTEXISTENT|NOTACTIVE|ACTIVE)$/ ) {
		out('e',"set_state_db: unknown/unsupported target state '$newstate'");
		return 0;
	}
	my $result=1;
	if ( $state eq 'UNKNOWN' ) {
		out('e',"db-state of user '$user' is unknown, new state '$newstate' won't be applied.");
		return 0;
	} elsif ( $state eq $newstate ) {
		return 1;
	} elsif ( $state eq 'VAGUE' ) {
		if ( $newstate eq 'NOTEXISTENT' ) {
			return $self->{manage}->removeUser($user);
		}
		if ( $newstate eq 'ACTIVE' ) {
			if ( ! $self->{manage}->setPtdbMembership($user,'users-na',0) ) {
				out('v',"Unable to set membership($user,'users-na')");
				$result=0;
			}
			if ( ! $self->{manage}->setPtdbMembership($user,'users',1) ) {
				out('v',"Unable to set membership($user,'users')");
				$result=0;
			}
		}
		if ( $newstate eq 'NOTACTIVE' ) {
			if ( ! $self->{manage}->setPtdbMembership($user,'users',0) ) {
				out('v',"Unable to set membership($user,'users')");
				$result=0;
			}
			if ( ! $self->{manage}->setPtdbMembership($user,'users-na',1) ) {
				out('v',"Unable to set membership($user,'users-na')");
				$result=0;
			}
		}
		return $result;
	} elsif ( $state eq 'NOTEXISTENT' ) {
		$result &= $self->{manage}->createUser($user,'afsid' => $id);
		if ( $result ) {
			if ( $newstate eq 'ACTIVE' ) {
				$result &= $self->{manage}->setPtdbMembership($user,'users',1);
			} elsif ( $newstate eq 'NOTACTIVE' ) {
				$result &= $self->{manage}->setPtdbMembership($user,'users-na',1);
			}
		}
		return $result;
	} elsif ( $state eq 'NOTACTIVE' ) {
		if ( $newstate eq 'ACTIVE' ) {
			$result &= $self->{manage}->setPtdbMembership($user,'users-na',0);
			$result &= $self->{manage}->setPtdbMembership($user,'users',1);
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			$result &= $self->{manage}->removeUser($user);
		}
		return $result;
	} elsif ( $state eq 'ACTIVE' ) {
		if ( $newstate eq 'NOTACTIVE' ) {
			$result &= $self->{manage}->setPtdbMembership($user,'users',0);
			$result &= $self->{manage}->setPtdbMembership($user,'users-na',1);
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			$result &= $self->{manage}->removeUser($user);
		}
		return $result;
	}
}


=item $result=$serverusers->write_user_keytab($user,$passwd,$directory)

Creates a keytab "I<$directory>B</>I<$user>B<.keytab>" and puts keys for user
I<$user> (which are derived from I<$passwd>) there.

I<$result> == B<1>, only if no error occured.

=cut

sub write_user_keytab {
	my ($self,$user,$passwd,$filename,$chown)=@_;
	my $filenameq=bash_quote($filename);
	
	# Just in case the keytab existed before. Wa can't rely on
	# ktutil overwriting it - ktutil will always append new keys
	unlink($filename);

	my $success=0;
	out('v',"Generating a user keytab using expect on 'ktutil'");
	my $ktutil=Expect->spawn("ktutil");
	$Expect::Log_Stdout=0;
	if ( $ktutil->expect(1,"ktutil:") ) {
		print $ktutil "add_entry -password -p $user -k 1 -e des3-hmac-sha1\n";
		if ( $ktutil->expect(1,"Password for ") ) {
			print $ktutil "$passwd\n";
			if ( $ktutil->expect(1,"ktutil:") ) {
				print $ktutil "write_kt $filename\n";
				if ( $ktutil->expect(2,"ktutil:") ) {
					print $ktutil "exit\n";
					if ( defined($chown) ) {
						iaexec("chown $user $filenameq",'exec');
					}
					$success=1;
				} else {
					out('e',"ktutil-session: no prompt after write_kt.");
				}
			} else {
				out('e',"ktutil-session: no prompt after password was entered.");
			}
		} else {
			out('e',"ktutil-session: no password-prompt.");
		}
	} else {
		out('e',"ktutil-session: no prompt after program start.");
	}
	$ktutil->soft_close();
	return $success;
}


=item $state=$serverusers->get_state_homedir($username,$check_vmp);

Reads the state of user I<$username> according to his homedirectory.
If I<$check_vmp> == B<1>, the volume mountpoints within the cell's
homedirectory-rootvolumes are tested against InstantAFS-defaults and
I<$state> eq 'UNKNOWN' is returned in case of mismatch.

=cut

sub get_state_homedir {
	my ($self,$username,$check_vmp)=@_;
	my $user_volume=$self->{celldb}->getVolume("user.$username");
	my $userna_volume=$self->{celldb}->getVolume("user-na.$username");
	my $userne_volume=$self->{celldb}->getVolume("user-ne.$username");
	if ( $userne_volume ) {
		out('e',"Found old user volume 'user-ne.$username' - please delete that one.");
		return 'UNKNOWN';
	}
	if ( $user_volume and $userna_volume ) {
		out('e',"Found volumes 'user.$username' and 'user-na.$username' - this is a contradiction.");
		return 'UNKNOWN';
	}
	if ( ! $user_volume and ! $userna_volume ) {
		out('v',"Homedirectory volume not found");
		return 'NOTEXISTENT';
	}
	
	my %mountpoints=(
		'user-na' => "user-na.$username",
		'user' => "user.$username",
		'backup/user/backup' => "user.$username.backup",
		'backup/user/ro' => "user.$username.readonly"
	);
	my @vmps=();
	
	foreach my $mp ('user-na','user','backup/user/backup','backup/user/ro') {
		if ( read_mount("/afs/$self->{celldb}->{cellname}/$mp/$username") ) {
			push(@vmps,1);
		} else {
			push(@vmps,0);
		}
	}
	if ( ( $vmps[0] + $vmps[1] + $vmps[2] + $vmps[3] ) == 0 ) {
		if ( $user_volume or $userna_volume ) {
			return 'VAGUE';
		} else {
			return 'NOTEXISTENT';
		}
	}
	if ( $vmps[0] == 1 ) {
		if ( ( $vmps[1] + $vmps[2] + $vmps[3] ) > 0 ) {
			return 'VAGUE';
		} else {
			if ( $user_volume ) {
				return 'VAGUE';
			} else {
				return 'NOTACTIVE';
			}
		}
	} else {
		if ( ( $vmps[1] * $vmps[2] * $vmps[3] ) == 0 ) {
			return 'VAGUE';
		} else {
			if ( $userna_volume ) {
				return 'VAGUE';
			} else {
				return 'ACTIVE';
			}
		}
	}
}

=item $result=$serverusers->set_state_homedir($username,$newstate,$enforce_vmp);

Set the state of user I<$username> according to his/her homedirectory.
Executing this method requires acall-priviledges (use I<get_acallrights()>
before calling).

I<$enforce_vmp> : When == B<1>, set_state_homedir() will enforce the correct
                  mountpoint at /a/u/$username .

=cut

sub set_state_homedir {
	my ($self,$username,$newstate,$enforce_vmp)=@_;
	my $state=$self->get_state_homedir($username,$enforce_vmp);
	if ( $state eq 'UNKNOWN' ) {
		return 0;
	}
	my $user_volume=$self->{celldb}->getVolume("user.$username");
	my $userna_volume=$self->{celldb}->getVolume("user-na.$username");
	if ( $state eq 'NOTEXISTENT' ) {
		if ( $newstate eq 'ACTIVE' ) {
			if ( $userna_volume ) {
				if ( ! $self->{manage}->renameVolume("user-na.$username","user.$username") ) {
					out('e',"set_state_homedir(): Unable to rename user volume");
					return 0;
				}
			}
		} elsif ( $newstate eq 'NOTACTIVE' ) {
			if ( $user_volume ) {
				if ( ! $self->{manage}->renameVolume("user.$username","user-na.$username") ) {
					out('e',"set_state_homedir(): Unable to rename user volume");
					return 0;
				}
			}
		}
	} elsif ( $state eq 'NOTACTIVE' ) {
		if ( $newstate eq 'ACTIVE' ) {
			if ( ! $self->{manage}->renameVolume("user-na.$username","user.$username") ) {
				out('e',"set_state_homedir(): Unable to rename user volume");
				return 0;
			}
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			if ( ! $self->{manage}->renameVolume("user.$username","user-ne.$username") ) {
				out('e',"set_state_homedir(): Unable to rename user volume");
				return 0;
			}
			out('v',"Please remove 'user-ne.$username' asap.");
		}
	} elsif ( $state eq 'ACTIVE' ) {
		if ( $newstate eq 'NOTACTIVE' ) {
			if ( ! $self->{manage}->renameVolume("user.$username","user-na.$username") ) {
				out('e',"set_state_homedir(): Unable to rename user volume");
				return 0;
			}
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			if ( ! $self->{manage}->renameVolume("user.$username","user-ne.$username") ) {
				out('e',"set_state_homedir(): Unable to rename user volume");
				return 0;
			}
			out('v',"Please remove 'user-ne.$username' asap.");
		}
	} elsif ( $state eq 'VAGUE' ) {
		if ( $newstate eq 'NOTACTIVE' ) {
			if ( $user_volume ) {
				if ( ! $self->{manage}->renameVolume("user.$username","user-na.$username") ) {
					out('e',"set_state_homedir(): Unable to rename user volume");
					return 0;
				}
			}
		} elsif ( $newstate eq 'ACTIVE' ) {
			if ( $userna_volume ) {
				if ( ! $self->{manage}->renameVolume("user-na.$username","user.$username") ) {
					out('e',"set_state_homedir(): Unable to rename user volume");
					return 0;
				}
			}
		} elsif ( $newstate eq 'NOTEXISTENT' ) {
			if ( $user_volume ) {
				if ( ! $self->{manage}->renameVolume("user.$username","user-ne.$username") ) {
					out('e',"set_state_homedir(): Unable to rename user volume");
					return 0;
				}
			} elsif ( $userna_volume ) {
				if ( ! $self->{manage}->renameVolume("user-na.$username","user-ne.$username") ) {
					out('e',"set_state_homedir(): Unable to rename user volume");
					return 0;
				}
			}
			out('v',"Please remove 'user-ne.$username' asap.");
		}
	}
	my %mountpoints=(
		'user-na' => {
			'target' => "user-na.$username",
			'must_exist' => ( $newstate eq 'NOTACTIVE' ),
			'mountrw' => '1',
			'release' => 'user-na',
			'rw' => '.user-na'
		},
		'user' => {
			'target' => "user.$username",
			'mountrw' => '1',
			'must_exist' => ( $newstate eq 'ACTIVE' ),
			'release' => 'user',
			'rw' => '.user'
		},
		'backup/user/backup' => {
			'target' => "user.$username.backup",
			'must_exist' => ( $newstate eq 'ACTIVE' ),
			'mountrw' => 0,
			'release' => 'backup.user',
			'rw' => 'backup/.user/backup'
		},
		'backup/user/ro' => {
			'target' => "user.$username.readonly",
			'must_exist' => ( $newstate eq 'ACTIVE' ),
			'mountrw' => 0,
			'release' => 'backup.user',
			'rw' => 'backup/.user/readonly'
		},
	);
	my @release_volumes=();
	my $success=1;
	foreach my $mp (keys %mountpoints) {
		my $rewrite_vmp=0;
		my $exists=0;
		my $path="/afs/$self->{celldb}->{cellname}/$mountpoints{$mp}->{'rw'}";
		my $pathu="$path/$username";
		if ( read_mount($pathu) ) {
			$exists=1;
		}
		if ( $exists != $mountpoints{$mp}->{'must_exist'} ) {
			$rewrite_vmp=1;
		} else {
			if ( ! $enforce_vmp ) {
				out('v',"Mountpoints won't be enforced.");
			} else {
				$enforce_vmp=1;
			}
		}
		if ( $rewrite_vmp ) {
			#  We'll remove the filesystem object in all possible ways
			remove_mount($pathu);
			rmdir($pathu);
			unlink($pathu);
			if ( $mountpoints{$mp}->{'must_exist'} ) {
				if ( ! make_mount($pathu,$mountpoints{$mp}->{'target'},'rw' => $mountpoints{$mp}->{'mountrw'}) ) {
					out('e',"set_state_homedir(): Unable to create mountpoint '$pathu'");
					$success=0;
				}
			}
			push(@release_volumes,$mountpoints{$mp}->{'release'});
		}
	}
	if ( ! $self->{manage}->syncVolumes(\@release_volumes,1) ) {
		out('e',"Error releasing crucial volumes");
		$success=0;
	}

	my $id;
	if ( $newstate eq 'ACTIVE' ) {
		$id='user';
	} elsif ( $newstate eq 'NOTACTIVE' ) {
		$id='user-na';
	}
	if ( defined($id) ) {
		my $dir="/afs/$self->{celldb}->{cellname}/$mountpoints{$id}->{'rw'}/$username";
		# Create new homedirectory skeleton, if the homedirectory is empty.
		my $create=0;
		if ( opendir(DIR,$dir) ) {
			my @entries=sort readdir(DIR);
			if ( @entries == 2 ) {
				if ( ( $entries[0] eq '.' ) and ( $entries[1] eq '..' ) ) {
					# The homedir is empty, we'll create the skeleton...
					$create=1;
				}
			}
		} else {
			out('e',"set_state_homedir(): Unable to read new homedirectory.");
			$success=0;
		}
		if ( $create ) {
			if ( ! iaexec("instantafs.homedir $username " . bash_quote($dir),'exec') ) {
				out('e',"Unable to create template homedirectory.");
				$success=0;
			}
		}
	}
	return $success;
}



=back

=head1 APPENDIX

=head2 User state types

InstantAFS knows 4 different states about each user:

=over 4

=item B<db>

The PTDB (AFS-Protection-DataBase). The user name is associated with
it's AFSID/UID here. Group membership in either 'B<users>' or 'B<users-na>'
defines the user's state as either ACTIVE (-> 'B<users'>) or NOTACTIVE
(-> 'B<users-na'>)

=item B<krb5>

The Kerberos5-Principal database. The user's name is associated with
a Kerberos key (which is comparable to a password) here. The user's state
is encoded in the B<expired>-field of the user's principal: 'B<[never]>' means
ACTIVE, anything else means NOTACTIVE.

=item B<samba>

The Samba user database stores the LANMAN- and NT- hashes (which are
essentially the encoded password) associated to the user name. The disabled-flag
is used to represent the user's state (ACTIVE vs. NOTACTIVE).

=item B<homedir>

The user homedirectory volume. It's mounted either at /a/user (ACTIVE) or
//a/user-na (NOTACTIVE). Furthermore the homedirectory volume is named
either B<user.>I<username> (ACTIVE) or B<user-na.>I<username> (NOTACTIVE).

=back

=head2 User states

All users have an associated state in each "database" used by InstantAFS.

=over 4

=item B<NOTEXISTENT>

The user doesn't exist.

=item B<NOTACTIVE>

The user exists but is disabled (-> unable to use her/his account)

=item B<ACTIVE>

The user exists and is enabled

=back


There are some special states:

=over 4

=item B<VAGUE>

The user's state doesn't make sense. Resetting the state to one of the
3 main states will fix that.

=item B<UNKNOWN>

The user's state can't be determined. This is either a technical or
a permission problem. Maybe a required package is not installed.

=back

=head1 AUTHOR

Frank Burkhardt <burk@cbs.mpg.de>

=cut


1;
__END__

