package AFS::Instant::Toolz;

use 5.006;
use strict;
use warnings;
use Carp;
use MIME::Base64;
use Errno qw(:POSIX);
use Data::Dumper;
use filetest 'access';
use Errno;
use Socket;
use Fb::Out qw(out initOut makeOptionsHash);
use Fb::Data qw($MAXINT32 min);
use Fb::IO qw(iaexec readFile readXmlFile);
use Net::DNS;

require Exporter;

our @ISA=qw(
	Exporter
);


our @EXPORT=qw(
	make_mount
	remove_mount
	read_mount

	get_acl
	set_acl
	copy_acls
	mergeAcls
	unmergeAcls
	trimAcls
	aclH2S
	aclS2H
	printAcl
	
	test_acl
	instantafs_init
	instantafs_help
	parse_storagesize
	parse_timeinterval
	scanafs
	dir2vol
	k4name
	get_right_description
	get_right_from_word
	check_name
	int_to_flipbase64
	flipbase64_to_int
	get_scannermode
	add_exec_on_exit
	exec_on_exit
	@AFS_ACL_RIGHTS
	%AFS_ACL_RIGHTS
	$DEFAULTSCANNERMODE
	$DEFAULTSCANNERFALLBACK
	cm_check_volumes
	getCellName
	getRealmName
);

our @EXPORT_OK=qw(
	dot
	getUserPref
	cleanup_mountpoint_basedir
	update_mountpoint_basedir_family
	update_mountpoint_basedir
	prepareFtests
	getCommandBlock
	getDbServers
	getDbServersDns
	getDbServersFile
	getAfsdbMode
	checkXmlAuthorisation
	create_temp
	remove_temp
);

our @AFS_ACL_RIGHTS=('r','l','w','i','d','k','a');

our $cellname_override;
our $realmname_override;
my $cellname_cache;
my $realmname_cache;

our %AFS_ACL_RIGHTS=(
	'r'	=>	[
		'read',
		"Reading files",
		"This right is necessary for reading files in the given directory."
	],
	'w'	=>	[
		'write',
		"Changing Content of existing Files",
		"For changing the content of a file in the given directory. Holding this right is not sufficient to create files (see i-right)."
	],
	'a'	=>	 [
		'admin',
		"Changing ACLs, creating Mountpoints",
		"For changing the given directory's ACLs and for creating/removing volume mountpoints"
	],
	'i'	=>	[
		'insert',
		"Creating new files/directories",
		"For creating new filesystem objects (files, directories, symlinks) in the given directory. This right is not sufficient to creeate volume mountpoints."
	],
	'k'	=>	[
		'lock',
		"Locking files",
		"For cellwide file-locking (using flock() - lockf() won't work inside AFS) of files in the given directory."
	],
	'l'	=>	[
		'list',
		'Entering the directory',
		"For entering (and for walking through) the given directory. Without this right, all others are useless."
	],
	'd'	=>	[
		'delete',
		"Deleting files/directories",
		"This right is necessary for deleting filesystem Objects (files, symlinks, directories) in the given directory."
	]
);

# Functions listed here will be executed when the script terminates
my @exec_on_exit=(\&remove_temp_signal);
my $exec_on_exit_done=0;

# Tempfiles created by the script but never deleted
my %tempfiles=();

my $SSHASKPASS;
my @SSHASKPASS=('/usr/bin/ssh-askpass','/usr/lib/openssh/gnome-ssh-askpass');

our $xmlnsup='http://instantafs.cbs.mpg.de/xmlns/userpref/1.0';

=head1 NAME

AFS::Instant::Toolz - various Functions needed by InstantAFS

=head1 SYNOPSIS


=head1 REQUIRES

-

=head1 DESCRIPTION

This module contains some helper functions used by InstantAFS that
don't fit into one of the AFS::Instant-classes.

=head1 FUNCTIONS

=over 4

=item $cellname=getCellName(%options);

Returns the workstation cell (the local AFS clients default cell). This is
done via the respective fs subcommand or by reading the ThisCell file as
fallback. Possible I<%options> are:

=over 4

=item B<test>

Suppress error message if no cell can be determined.

=back

I<$cellname> is the name of the AFS cell or B<undef> if it cannot be determined.

=cut

sub getCellName {
	my (%options)=@_;
	if ( defined($cellname_override) ) {
		return $cellname_override;
	}
	if ( defined($cellname_cache) ) {
		return $cellname_cache;
	}
	
	my $output=iaexec(
		['/usr/bin/fs','wscell'],
		'readpipe',
		format => 'string',
		stderr => 'hide',
	);
	if ( defined($output) ) {
		if ( $output =~ /^This workstation belongs to cell '(.*)'$/ ) {
			$cellname_cache=$1;
			return $1;
		}
	}
	my $l=readFile('/etc/openafs/ThisCell',format => 'string');
	if ( ! defined($l) ) {
		if ( ! $options{test} ) {
			out('e',"Unable to read cellname from 'ThisCell': $!");
		}
		return undef;
	}
	chomp($l);
	if ( $l eq '' ) {
		if ( ! $options{test} ) {
			out('e',"Empty cellname in 'ThisCell'");
		}
		return undef;
	}
	$cellname_cache=$l;
	return $l;
}

sub getRealmName {
	my (%options)=@_;
	if ( defined($realmname_override) ) {
		return $realmname_override;
	}
	if ( defined($realmname_cache) ) {
		return $realmname_cache;
	}
	my $krb5conf=readFile("/etc/krb5.conf",test => 1,format => 'arrayref');
	if ( $krb5conf ) {
		my $libdefaults=0;
		for my $l (@{$krb5conf}) {
			if ( $l eq '[libdefaults]' ) {
				$libdefaults=1;
			}
			if ( $libdefaults ) {
				if ( $l =~ /^\s*default_realm\s*=\s*(\S+)\s*$/ ) {
					$realmname_cache=$1;
					return $realmname_cache;
				}
			}
		}
	}
	
	my $resolver=Net::DNS::Resolver->new();
	my $reply=$resolver->query('_kerberos','TXT');
	if ( $reply ) {
		my @rrs=$reply->answer();
		if ( @rrs ) {
			$realmname_cache=$rrs[0]->txtdata();
			return $realmname_cache;
		}
	}
	return undef;
}

=back

=head2 Temp file functions

=over 4

=item @tempfilenames=create_temp($count)

Safely creates I<$count> temporary files, returning their names.

=cut

sub create_temp {
	my $count=shift;
	my @tempfilelist=();
	for(my $n=0;$n<$count;$n++) {
		my $tempfile=iaexec(["/bin/tempfile",'-p','ia_'],'readpipe',format => 'string',description => "Getting temporary file");
		if ( defined($tempfile) ) {
			exit 1;
		}
		chomp($tempfile);
		$tempfiles{$tempfile}=$$;
		push(@tempfilelist,$tempfile);
	}
	return @tempfilelist;
}

=item remove_temp(@tempfilelist)

Removes created temporary files in I<@tempfilelist> or removes
all temporary files created by the script (if I<@tempfilelist> is
empty).

=cut


sub remove_temp {
	my @tempfilelist=@_;
	if ( ! @tempfilelist ) {
		@tempfilelist=keys %tempfiles;
	}
	foreach my $tempfile (@tempfilelist) {
		if ( defined($tempfiles{$tempfile}) ) {
			if ( $tempfiles{$tempfile} == $$ ) {
				if ( -e $tempfile ) {
					if ( ! unlink($tempfile) ) {
						out('e',"Unable to delete temporary file '$tempfile'");
					}
				} else {
					out('d',"Hmm ... temporary file '$tempfile' was removed by someone else");
				}
			}
			delete($tempfiles{$tempfile});
		}
	}
}


=item remove_temp_signal($signal)

Signal handler for removing all remaining tempfiles of the script.

=cut

sub remove_temp_signal {
	remove_temp();
}


=item add_exec_on_exit(\&function)

Addes &function to list of functions that will be executed when the
script ends - even when it ends by signal.

=cut


sub add_exec_on_exit {
	my $function=shift;
	push(@exec_on_exit,$function);
	return 1;
}

=item exec_on_exit()

This function will be executed when a signal occures or when the
script terminates.

=cut

sub exec_on_exit {
	my $signal=shift;
	if ( $exec_on_exit_done ) { return 1; }
	out('dd',"Executing " . @exec_on_exit . " exit()-triggerd functions...");
	for(my $n=$#exec_on_exit;$n >= 0;$n--) {
		my $function=$exec_on_exit[$n];
		&$function($signal);
	}
	out('dd',"Finished executing exit()-triggered functions...");
	my %deadly_signals=(
		'INT' => 2,
		'TERM' => 15,
		'HUP' => 1
	);
	$exec_on_exit_done=1;
	if ( defined($deadly_signals{$signal}) ){
		exit (128+$deadly_signals{$signal});
	}
}

=back



=head2 String manipulation

=cut



=item $result=check_name($type,$name);
 
Checks, if $name is valid as a given kind of name ($type)
$type is one of these:

B<user> A normal AFS-User (Kerberos4-Name).
 Remember: There's an additional filter
 level available via AFS::Instant::User->isUntouchable($name).
   
B<host> A hostname (no FQDN, just the host part). All names used as aliases in
   InstantAFS are filtered.

B<cell> The name of an AFS-cell.
 
B<volume> The name of an AFS-volume.

B<volset> The name of a AFS-backup-volset

B<afsid> An AFSID (OK, it's not a name but it's handy to check it here)

Given names are matched against regular expressions that will be printed whenever a name is invalid.
I<$result> == 1, only if the name is valid

=cut

sub check_name {
	my $type=shift;
	my $name=shift;
	my $ncurses=shift;
	if ( ! defined($ncurses) ) { $ncurses=0; }
	if ( $type eq 'user' ) {
		if ( $name !~ /^[a-z][-a-z0-9\.]{1,13}$/ ) {
			out('e',"Invalid username '$name', must match /^[a-z][-a-z0-9\.]{1,13}\$/");
			return 0;
		}
		# Those usernames are technically forbidden (to avoid problems with volume-names)
		if ( $name =~ /\.readonly|backup$/ ) {
			out('e',"Invalid username '$name'. It would cause problems with the resulting Homedirectory-Volume-name.");
			return 0;
		}
	} elsif ( $type eq 'host' ) {
		if ( $name !~ /^[a-z][-a-z0-9]{1,39}$/ ) {
			out('e',"Invalid hostname '$name', must match /^[a-z][-a-z0-9]{1,39}\$/");
			return 0;
		}
		# Those hostname are forbidden because InstantAFS uses them as aliases
		if ( $name =~ /^(afsdb|kerberos|timeserv|sambaserv)[0-9]*|acall-\S+|_kerberos.*|_kpasswd.*$/ ) {
			out('e',"Invalid hostname: InstantAFS might use some aliases like that");
			return 0;
		}
	} elsif ( $type eq 'cell' ) {
		if ( $name !~ /^[a-z][-\.a-z0-9]+$/ ) {
			out('e',"Invalid cell name '$name', must match /^[a-z][-\.a-z0-9]+\$/ .");
			return 0;
		}
		# Those names are forbidden because we use the cellname as part of a filename in /etc/bind (db.[cellname])
		if ( $name =~ /^root|local$/ ) {
			out('e',"Invalid cell name: 'root' and 'local' are forbidden");
			return 0;
		}
	} elsif ( $type eq 'volume' ) {
		if ( $name !~ /^[-a-zA-Z0-9_\.]+$/ ) {
			out('e',"Invalid volume name '$name', must match /^[-a-zA-Z0-9_\.]{1,22}\$/");
			return 0;
		}
		# Those volume names are forbidden by AFS
		if ( $name =~ /\.(backup|readonly)$/ ) {
			out('e',"Forbidden volume name: suffixes .readonly and .backup must not be used.");
			return 0;
		}
	} elsif ( $type eq 'volset' ) {
		if ( $name !~ /^[-_a-z0-9\.]+$/ ) {
			out('e',"volset name '$name' contains invalid character. Valid characters are =~ [-_a-z0-9\.] .");
			return 0;
		}
	} elsif ( $type eq 'afsid' ) {
		if ( $name =~ /^[0-9]+$/ ) {
			if ( $name < 0 ) {
				out('e',"AFSID must not be < 0 (because < 0 means it's a PT-Group-ID");
			} elsif ( $name == 0 ) {
				out('e',"0 is not a valid AFSID.");
			} elsif ( $name == 32766 ) {
				out('e',"The AFSID 32766 is reserved for 'anonymous'");
			} else {
				return 1;
			}
			return 0;
		} else {
			out('e',"An AFSID must consist of digits only.");
			return 0;
		}
	} else {
		out('e',"check_name(): Unknown name type '$type', internal error, returning failure.");
		return 0;
	}
	return 1;
}



=item $k4name=k4name($k5name)

Returns the Kerberos5 principal name I<$k5name> converted to a Kerberos4
principal name ( I<$k4name> ).

=cut

sub k4name {
	my $k5name=shift;
	my $k4name=$k5name;
	$k4name =~ s/\//./g;
	return $k4name;
}

=item $text=get_right_description($right,$type)

Gets a textual description of the given AFS-ACL-Right. I<$right>
is one char or 'rlwidka', I<$type> is either 'B<word>', 'B<short>'
or 'B<long>'.

I<$text> : returns the description of the given right I<$right>.

Examples:

 get_right_description('r','long') eq 'Reading Files'
 
 get_right_description('r','word') eq 'read'

=cut

sub get_right_description {
	my ($right,$type)=@_;

	if ( ! defined($AFS_ACL_RIGHTS{$right}) ) {
		return undef;
	}
	if ( $type eq 'word' ) {
		return $AFS_ACL_RIGHTS{$right}->[0];
	} elsif ( $type eq 'short' ) {
		return $AFS_ACL_RIGHTS{$right}->[1];
	} elsif ( $type eq 'long' ) {
		return $AFS_ACL_RIGHTS{$right}->[2];
	} else {
		return undef;
	}
}




sub get_right_from_word {
	my $word=shift;
	foreach my $right (keys %AFS_ACL_RIGHTS) {
		if ( $AFS_ACL_RIGHTS{$right}->[0] eq $word ) {
			return $right;
		}
	}
	return undef;
}


=item $text=int_to_flipbase64($int)

Converts an integer value into a flibase64-representation
as used by OpenAFS-namei-fileservers for storing volumes
on storage partitions.

=cut

sub int_to_flipbase64 {
	my $a=shift;
	my $c_xlate="+=0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
	my $result='';
	if ( $a == 0 ) {
		$result=substr($c_xlate,0,1);
	} else {
		for(my $n=$a & 0x3f; $a != 0 ; $n=(( $a >>= 6) & 0x3f)) {
			$result .= substr($c_xlate,$n,1);
		}
	}
	return $result;
}


sub flipbase64_to_int {
	my $text=shift;
	my $result=0;
	my @c_reverse=(
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,  0, 99, 99, 99, 99,
     2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 99, 99, 99,  1, 99, 99,
    99, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26,
    27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 99, 99, 99, 99, 99,
    99, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52,
    53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99,
    99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99, 99
	);
	my $shift=0;
	for(my $i=0;$i < length($text);$i++) {
		my $value=$c_reverse[ord(substr($text,$i,1))];
		$result += ( ( $value & 63 ) << $shift );
		$shift += 6;
	}
	return $result;
}



=item $result=get_ssh_askpass()

Looks for an ssh-askpass executable to be used for prompting the user
for the Kerberos password.

I<$result> == B<1>, only if a (possibly) usable ssh-askpass was found.

=cut

sub get_ssh_askpass {
	if ( defined($SSHASKPASS) ) {
		return 1;
	} else {
		foreach my $s (@SSHASKPASS) {
			if ( -x $s ) {
				$SSHASKPASS=$s;
				return 1;
			}
		}
	}
	return 0;
}

sub dot {
	if ( $ENV{DOT} ) {
		return '.';
	} else {
		return '';
	}
}

=back


=back

=head2 InstantAFS default options

=over 4

=item instantafs_help()

Prints a help-message for cmdline-options common for all
instantafs-scripts.

=cut

sub instantafs_help {
	print "instantafs options:
 -v, --verbose         Be verbose (use twice to run in debug mode)
 -Y, --output-media=X  Set output media to X.
                       Valid media are 'console', 'syslog' and 'stdout'
                       (Defaults to 'console')
     --instantafs-cell=C
                       Work on cell Cinstead of local workstation's cell.
     --instantafs-realm=R
                       Use kerberos realm R instead of the local default
                       realm.

";
}


=item $result=instantafs_init(I<%options>)

Parses common cmdline-options for all instantafs-scripts. Certain
I<%options> might be given:

=over 4

=item B<cmdoptions>

A hashref containing cmdline definitions for I<GetOptions()>.

=item B<safemode>

Don't interpret "critical" commandline options. This is for acall-server-scripts.

=back

I<$result> will be false, when errors occured while parsing cmdline arguments. The script
should I<die()> in this case.

=cut

sub instantafs_init {
	my %options=@_;
	my %cmdoptions;
	if ( defined($options{'cmdoptions'}) ) {
		%cmdoptions=%{$options{'cmdoptions'}};
	}
	if ( defined($options{blocks}) ) {
		%cmdoptions=%{makeOptionsHash(blocks => $options{blocks})};
	}
	Getopt::Long::Configure("bundling","no_getopt_compat");
	if ( defined($ENV{'INSTANTAFS_CELL'}) ) {
		$cellname_override=$ENV{'INSTANTAFS_CELL'};
	}
	if ( defined($ENV{'INSTANTAFS_REALM'}) ) {
		$realmname_override=$ENV{'INSTANTAFS_REALM'};
	}
	my %addoptions;
	if ( ! $options{safemode} ) {
		$cmdoptions{'instantafs-cell=s'}=\$cellname_override;
		$cmdoptions{'instantafs-realm=s'}=\$realmname_override;
	}
	my $result=initOut(
		cmdoptions => \%cmdoptions,
		%options,
	);
	if ( ! $result ) {
		return 0;
	}
	foreach my $signal ('INT','TERM','HUP','__DIE__') {
		$SIG{$signal}=\&exec_on_exit;
	}
	return 1;
}

=back

=head2 Command execution

=over 4

=item $result=fexec($cmdline,$description,$hideonsuccess)

Executes a program using the system()-command.
fexec will show the I<$description> and the command executed if I<$VERBOSE> > B<1>.
The command's output will not be shown if the command terminates
successfully and I<$hideonsuccess> == 1.

I<$result> == B<1>, only if the command was executed successfully (exit code == 0).

=cut

sub fexec {
	my $cmd=shift;
	my $description=shift;
	my $hideonsuccess=shift;
	if ( ! defined($description) ) {
		$description='(no description available)';
	}
	if ( ! defined($hideonsuccess) ) {
		$hideonsuccess=0;
	}
	out('v',"Exec: $description");
	out('v',"Cmd: $cmd");
	my $result=0;
	if ( ! $hideonsuccess ) {
		system($cmd);
		if ( ($? >> 8) == 0 ) {
			$result=1;
		}
	} else {
		my $buf='';
		open(STDERRBACKUP,'>&STDERR');
		open(STDERR,'>&STDOUT');
		if ( open(X,'-|',$cmd) ) {
			while (my $l=<X>) {
				$buf .= $l;
			}
			close(X);
			if ( ($? >> 8) != 0 ) {
				print $buf;
			} else {
				$result=1;
			}
		}
		open(STDERR,'>&STDERRBACKUP');
		close(STDERRBACKUP);
	}
	return $result;
}   


=item check_volumes()

Executes fs checkvolumes.

=cut

sub cm_check_volumes {
	out('d',"Flushing AFS volume name to ID table");
	return iaexec(["/usr/bin/fs",'checkvolumes'],'test',description => "Checking volumes mappings");
}


=item $link=afs_readlink($path)

Returns the AFS-mountpoint link of the given AFS-mountpoint I<$path>. If I<$path>
is not an AFS-mountpoint, B<undef> is returned.

This function is only used internally.

=cut

sub afm_readlink {
	my $link=shift;
	my $line=iaexec(['/usr/bin/fs','lsmount',$link],'readpipe',stderr => 'hide',format => 'string',description => "Reading AFS MP target");
	if ( defined($line) ) {
		if ( $line =~ /^'.*' is a mount point for volume '([%#]\S+?)'/ ) {
			return $1;
		} else {
			out('d',"Unable to interpret output of 'fs lsmount'");
			return undef;
		}
	}
	return undef;
}

=item $mp=read_mount($path)

Reads information about the AFS-mountpoint given by its I<$path>. I<$mp>
contains a hashref with these data fields:

=over 4

=item B<volume>

the name of the referenced volume (without instance ID)

=item B<instance>

Name of the volume instance - either an empty string (RW instance) or an volume suffix (B<readonly> or B<backup>).

=item B<ext>

Appendable extention (either empty string, B<.readonly> or B<.backup>)

=item B<type>

Mountpoint class - either B<rw>, B<backup>, B<readonly> or B<magic> (Mountpoint to an RW without the B<-rw> option in fs sa).

=item B<cell>

The cell name. If non is explicitely given, B<undef> is the value.

=back

=cut

sub read_mount {
	my $link=shift;
	my $target=afm_readlink($link);
	if ( defined($target) ) {
		if ( $target =~ /^([#%])(?:(\S+):)?(\S+)$/ ) {
			my $mpclass='magic';
			if ( $1 eq '%' ) {
				$mpclass='rw';
			}
			my $cell=$2;
			my $instance=$3;
			my $volume=$instance;
			my $instext='';
			if ( $instance =~ /^(.*)\.(backup|readonly)$/ ) {
				$volume=$1;
				$instext=$2;
				if ( $2 eq 'readonly' ) {
					$mpclass='readonly';
				} else {
					$mpclass='backup';
				}
			}
			my %data=(
				'instance' => $instance,
				'volume' => $volume,
				'ext' => $instext,
				'type' => $mpclass,
				'cell' => $cell,
			);
			return \%data;
		} else {
			out('d',"read_mount(): Invalid mountpoint format '$target' of mp '$link'");
		}
	}
	return undef;
}

=item $result=make_mount($mountpoint,$volume,%options)

Creates a mountpoint in AFS pointing to the given volume.

Possible I<%options> are:

=over 4

=item B<rw>

Make the mountpoint explicitely RW (B<-rw> option in fs mkm)

=item B<cell>

Define cell to make the mountpoint point to. Default is to point to the cell of the volume that
contains the new mountpoint. In this case, no cell information will be embedded into the mountpoint.

=item B<idempotent> (default: B<1>)

Defines, if the call to make_mount() should be idempotent - and how.

Possible values are:

=over 4

=item 0

If a filesystem object exists at I<$mountpoint>, a error is returned.

=item 1

If a mountpoint I<$mountpoint> exists and points to the location requested by
the make_mount()-call, success is returned.

=item 2

If any AFS-mountpoint I<$mountpoint> exists, it will be made pointing to the
correct volume and success is returned.

=back

=item B<dryrun>

Simmulate creating the mountpoint without actually changing anything

=item B<checkvolumes>

A value != B<0> Enforces a 'fs checkvolumes' call after a mountpoint was changed. A value
of B<2> additionall enforces a 'fs checkvolumes' call after a mountpoint was just created.
B<3> additionally enforces 'fs checkvolumes' if the mounpoint was already there
pointing at the correct volume instance.

=back

I<$result> returns a bitmapped result of the operation. Those bits are of interest:

=over 4

=item B<0> (1)

Success. Everything went smooth.

=item B<1> (2)

A mountpoint was actually created or changed in the filesystem (this might be a simmulation caused by dryrun)

=item B<2> (4)

A mountpoint was replaced in the filesystem (this might be a simmulation caused by dryrun).

=back

=cut


sub make_mount {
	my ($mountpoint,$volume,%options)=@_;
	my $rw=0;
	if ( $options{'rw'} ) {
		$rw=1;
	}
	my $cell;
	if ( $options{'cell'} ) {
		$cell=$options{'cell'};
	}
	my $idempotent=1;
	if ( defined($options{'idempotent'}) ) {
		$idempotent=$options{'idempotent'};
	}
	my $checkvolumes=0;
	if ( defined($options{'checkvolumes'}) ) {
		$checkvolumes=$options{'checkvolumes'};
	}
	my $target='';
	if ( $rw ) {
		$target .= '%';
	} else {
		$target .= '#';
	}
	if ( defined($cell) ) {
		$target .= "$cell:";
	}
	$target .= $volume . ".";
	out('dd',"Creating mountpoint '$mountpoint', target '$target', idp=$idempotent");
	my $former_mp=afm_readlink($mountpoint);
	if ( defined($former_mp) ) {
		$former_mp .= ".";
		if ( $former_mp eq $target ) {
			if ( $idempotent ) {
				out('d',"Mountpoint '$mountpoint' already exists and points to correct volume '$target'");
				if ( $checkvolumes >= 3 ) {
					cm_check_volumes();
				}
				return 1;
			} else {
				out('e',"Mountpoint already exists, no idempotence check");
				return 0;
			}
		} else {
			if ( $idempotent >= 2 ) {
				if ( $options{'dryrun'} ) {
					out('i',"Would have re-created mountpoint '$mountpoint' to '$volume' now");
					return 7;
				}
				if ( remove_mount($mountpoint) ) {
					if ( make_mount($mountpoint,$volume,%options) ) {
						if ( $checkvolumes ) {
							cm_check_volumes();
						}
						out('d',"Replaced mountpoint '$mountpoint' by one pointing to '$target'");
						return 7;
					}
				} else {
					out('e',"Mountpoint '$mountpoint' already exists with wrong target, rmmount failed");
					return 0;
				}
			} else {
				out('e',"Mountpoint '$mountpoint' already exists with wrong target ('$former_mp'!='$target')");
				return 0
			}
		}
	}
	my @s=stat($mountpoint);
	if ( ! @s ) {
		my $mountpointq=quotemeta($mountpoint);
		 if ( $!{ENODEV} or $!{ETIMEDOUT} ) {
			# Mountpoint exists, volume instance most likely doesn't or is not accessible
			if ( ! $options{'dryrun'} ) {
				if ( ! remove_mount($mountpoint) ) {
					out('e',"Unable to remove former mountpoint to missing volume: $!");
					return 0;
				}
				# Mountpoint was successfully removed, continuing...
			} else {
				out('i',"Would've removed old mountpoint '$mountpoint' now");
			}
		}
		if ( ! $options{'dryrun'} ) {
			# Creating a mountpoint is actually a symlink() call
			my $result=symlink($target,$mountpoint);
			if ( $result ) {
				if ( $checkvolumes >= 2 ) {
					cm_check_volumes();
				}
				out('d',"Created mountpoint '$mountpoint' to '$target'");
				return 3;
			} else {
				out('e',"Unable to create volume-mountpoint-symlink '$mountpoint': $!");
				return 0;
			}
		} else {
			out('i',"Would've created mountpoint '$mountpoint' to '$target' now");
			return 3;
		}
	} else {
		out('e',"Filesystem object already exists on mointpoint position '$mountpoint'");
		return 0;
	}
}

=item $result=remove_mount($mountpoint,%options);

Removes a mountpoint in AFS. Possible I<%options> are:

=over 4

=item B<itempotent> (default: B<1>)

If B<1>, success will be returned in case of a non-existing filesystem object I<$mountpoint>.

=item B<dryrun>

Just simmulate removing the mountpoint.

=back

I<$result> == 1, if the mountpoint was removed successfully. I<$result> == 3, if an actual change was done to the filesystem. B<0> otherwise.

=cut

sub remove_mount {
	my ($mountpoint,%options)=@_;
	my $idempotent=1;
	if ( $options{'idempotent'} ) {
		$idempotent=$options{'idempotent'};
	}
	my @s=stat($mountpoint);
	if ( ! @s ) {
		if ( $!{ENODEV} or $!{ETIMEDOUT} ) {
			# Mountpoint exists but is dangling or volume instance is not accessible
		} elsif ( $idempotent ) {
			return 1;
		}
	}
	if ( $options{'dryrun'} ) {
		out('i',"Would've removed mountpoint '$mountpoint' now");
		return 1;
	}
	my $result=iaexec(
		['/usr/bin/fs','rmm',$mountpoint],
		'exec',
		description => "Removing AFS-mountpoint",
	);
	if ( $result ) {
		out('v',"Removed mountpoint '$mountpoint'");
		return 3;
	} else {
		out('e',"Unable to remove mountpoint '$mountpoint'");
		return 0;
	}
}

sub printAcl {
	my $acl=shift;
	my @acl;
	for my $user (sort keys %{$acl}) {
		my $p="$user=";
		my $perms=$acl->{$user};
		for my $perm (@AFS_ACL_RIGHTS) {
			if ( $perms->{$perm} ) {
				$p .= $perm;
			}
		}
		push(@acl,$p);
	}
	return join(' ',@acl);
}

sub aclS2H {
	my $string=shift;
	if ( ! defined($string) ) {
		return undef;
	}
	my %hash;
	for(my $n=0;$n<length($string);$n++) {
		my $perm=substr($string,$n,1);
		if ( $AFS_ACL_RIGHTS{$perm} ) {
			$hash{$perm}=1;
		}
	}
	return \%hash;
}

sub aclH2S {
	my $hash=shift;
	if ( ! defined($hash) ) {
		return undef;
	}
	my $string='';
	for my $perm (@AFS_ACL_RIGHTS) {
		if ( $hash->{$perm} ) {
			$string .= $perm;
		}
	}
	return $string;
}

sub mergeAcls {
	my ($acl,$diff,%options)=@_;
	my $hashref=1;
	if ( $options{mode} && $options{mode} eq 'string' ) {
		$hashref=0;
	}
	for my $user (keys %{$diff}) {
		my $permissions=$diff->{$user};
		if ( $hashref ) {
			if ( ! defined($acl->{$user}) ) {
				%{$acl->{$user}}=%{$permissions}
			}
			for my $perm (keys %{$permissions}) {
				$acl->{$user}->{$perm}=1;
			}
		} else {
			if ( ! defined($acl->{$user}) ) {
				$acl->{$user}=$permissions;
				next;
			}
			my %oldperms=();
			for my $perm (split //,$acl->{$user}) {
				$oldperms{$perm}=1;
			}
			for my $perm (split //,$permissions) {
				$oldperms{$perm}=1;
			}
			$acl->{$user}=join('',keys %oldperms);
		}
	}
	return 1;
}

=item $result=trimAcls($positive,$negative,%options)

Removes permission assignments in the I<$positive> ACL if they have
a counterpart in the I<$negative> ACL. This is the behaviour of AFS
when examining directory permissions.

Possible I<%options> are:

=over 4

=item B<trimnegative>

Remove negative permission if it just cancels out a removed positive one.

=back

=cut

sub trimAcls {
	my ($acl,$nacl,%options)=@_;
	my $change=0;
	for my $user (keys %{$acl}) {
		if ( $nacl->{$user} ) {
			for my $perm (keys %{$acl->{$user}}) {
				if ( $nacl->{$user}->{$perm} ) {
					delete $acl->{$user}->{$perm};
					if ( $options{trimnegative} ) {
						delete $nacl->{$user}->{$perm};
					}
					$change=1;
				}
			}
			if ( ! scalar(keys %{$acl->{$user}}) ) {
				delete $acl->{$user};
				$change=1;
			}
			if ( ! scalar(keys %{$nacl->{$user}}) ) {
				delete $nacl->{$user};
				$change=1;
			}
		}
	}
	return 1 + 2*$change;
}

sub unmergeAcls {
	my ($acl,$diff,%options)=@_;
	my $hashref=1;
	if ( $options{mode} && $options{mode} eq 'string' ) {
		$hashref=0;
	}
	for my $user (keys %{$diff}) {
		my $permissions=$diff->{$user};
		if ( $hashref ) {
			if ( ! defined($acl->{$user}) || ! scalar(keys %{$acl->{$user}}) ) {
				next;
			}
			for my $perm (keys %{$permissions}) {
				delete $acl->{$user}->{$perm};
			}
			if ( ! scalar(keys %{$acl->{$user}}) ) {
				delete $acl->{$user};
			}
		} else {
			if ( ! defined($acl->{$user}) ) {
				return;
			}
			my %oldperms=();
			for my $perm (split //,$acl->{$user}) {
				$oldperms{$perm}=1;
			}
			for my $perm (split //,$permissions) {
				delete $oldperms{$perm};
			}
			my $string=join('',keys %oldperms);
			if ( $string eq '' ) {
				delete $acl->{$user};
			} else {
				$acl->{$user}=$string;
			}
		}
	}
	return 1;
}

sub test_user_permissions {
	my ($acl,$user,$rights)=@_;
	
	my $result=1;
	my $text=$acl->{$user};
	if ( defined($text) ) {
		for(my $n=0;$n < length($rights);$n++) {
			my $c=substr($rights,$n,1);
			if ( $text !~ /$c/ ) {
				$result=0;
			}
		}
	} else {
		if ( $rights ne '' ) {
			$result=0;
		}
	}
	return $result;
}


=item $result=test_acl($dir,\%rules)

Tests, if the permissions given in I<%rules> (user/group in keys, permission as characters in values)
are present in directory I<$dir>. Possible I<%options> are:

=over 4

=item B<test>

Do not show errors, only the I<$result> is needed.

=back

I<$result> == 1, only if all permission rules are satisfied.

=cut

sub test_acl {
	my ($dir,$rules,%options)=@_;
	if ( ! -d $dir ) {
		out('w',"Directory '$dir' doesn't exist.");
		return undef;
	}
	my $acl=get_acl($dir);
	if ( ! $acl ) {
		out('e',"Unable to read ACL-rights from '$dir'");
		return undef;
	}
	my $result=1;
	foreach my $ug (keys %$rules) {
		if ( ! test_user_permissions($acl,$ug,$rules->{$ug}) ) {
			if ( ! $options{test} ) {
				out('e',"Insufficient rights: ('$ug','$dir'), needed: '$rules->{$ug}'");
			}
			$result=0;
		}
	}
	return $result;
}


=item $result=get_acl($dir{,$negative});

Reads the ACL entries of the given I<$dir> ectory. I<$negative> non-existent or
B<0> means, the positive ACL is read. Otherwise the negative one will be read.

I<$result> == undef, if there was some kind of error reading the directory's
              ACL. Otherwise the hash's keys are usernames, the values are
              strings containing 'rlwidka' characters.

=cut

sub get_acl {
	my $dir=shift;
	my $negative=shift;
	if ( ! defined($negative)) {
		$negative=0;
	}

	my %acls=();
	my $lmode='Normal';
	if ( $negative ) {
		$lmode='Negative';
	}
	my $output=iaexec(["/usr/bin/fs",'la',$dir],'readpipe',format => 'arrayref',stderr => 'hide',description => "Reading AFS ACL");
	if ( ! defined($output) ) {
		return undef;
	}
	my $mode='';
	for my $l (@{$output}) {
		if ( $l =~ /^(\S+)/ ) {
			$mode=$1;
		} elsif ( $mode eq $lmode ) {
			if ( $l =~ /^\s+(\S+)\s+(\S+)/ ) {
				$acls{$1}=$2;
			}
		}
	}
	return \%acls;
} 

=item I<$result>=set_acl(I<\%acl>,I<$dir>,I<%options>);

Sets the given ACLs I<%acl> on given directory I<$dir>. I<%acl> is a hash with keys
being AFS group or AFS user names and values being strings containing perission bits.
Possible I<%options> are:

=over 4

=item B<dryrun>

Only simmulate things.

=item B<negative> => I<\%negative_acls>

Hash of negative ACLs to be set. Format is the same as for I<%acl>.

=item B<clear> => (0|1)

Set to one, the AFS will be replaced instead of merged with the
current one.

=back

I<$result> == 1, only if all given directories were successfully ACLed.

=cut

sub set_acl {
	my ($acl,$dir,%options)=@_;
	
	my @parameters;
	if ( $options{'clear'} ) {
		push(@parameters,'-clear');
	}
	push(@parameters,$dir);
	for my $user (keys %{$acl}) {
		push(@parameters,$user,$acl->{$user});
	}
	my $result=1;
	
	if ( ! $options{'dryrun'} ) {
		$result &= iaexec(["/usr/bin/fs",'sa',@parameters,'-clear'],'test',description => "Setting new permission on '$dir'");
	} else {
		out('i',"Would've set normal acl on '$dir'");
	}
	if ( $options{negative} && scalar(keys %{$options{negative}}) ) {
		@parameters=();
		for my $user (keys %{$options{negative}}) {
			push(@parameters,$user,$options{negative}->{$user});
		}
		if ( ! $options{'dryrun'} ) {
			$result &= iaexec(["/usr/bin/fs",'sa',$dir,@parameters,'-negative'],'test',description => "Setting new permission on '$dir'");
		} else {
			out('i',"Would've set negative acl on '$dir'");
		}
	}

	return $result;
}

=item $result=copy_acls($sourcedir,$targetdir,%options)

Copies AFS ACLs from directory I<$sourcedir> to directory I<$targetdir>. Possible I<%options> are:

=over 4

=item B<dryrun>

Only simmulate things.

=back

=cut

sub copy_acls {
	my ($source,$target,%options)=@_;
	my $sq=quotemeta($source);
	my $tq=quotemeta($target);
	if ( ! $options{'dryrun'} ) {
		if ( ! iaexec(
			["/usr/bin/fs",'ca','-fromdir',$source,'-todir',$target,'-clear'],
			'test',
			description => "Copying ACLs from '$source' to '$target'"
		) ) {
			return 0;
		} else {
			out('d',"Successfully copied ACLs from '$source' to '$target'");
		}
	} else {
		out('i',"Would've copied ACLs from '$source' to '$target'");
	}
	return 1;
}

=back



=item $ok=scanafs(I<$directory>,I<%options>);

Scans through the AFS-namespace starting at I<$directory>.
Some options may be given as hashkeys in I<%options>:

=over 4

=item 'B<permission_errors>' => B<1>

Permission problems are considered to be errors (Default: B<1>)

=item 'B<max_mountdepth>' => B<0>

Maximum number of mountpoints that may be crossed

=item 'B<max_celldepth>' => B<0>

Maximum number of inter-cell-mountpoints that may be crossed.

=item 'B<max_depth>' => B<0>

Maximum directories to scan down the tree.

=item B<cb_dir> => \I<&callback>(I<$directory>,I<$currentcell>)

Defines a callback function to be run for regular directories. This includes
the base directory of the scan but not root-directories of mounted volumes - they
are handled via the B<cb_mountpoint> callback.

=item B<cb_mountpoint> => \I<&callback>(I<$mountpoint>,I<$currentcell>,I<\%mpinfo>)

Defines a callback function to be run for mountpoints. This doesn't include the
base directory of the scan. I<\%mpinfo> is a hashref as returned by L<read_mount()>.

=item B<stoppable> => I<bool>

Setting this option to B<1> will change scan behaviour to "don't scan down a
directory if the callback returned false".

=back

I<$ok> == B<1>, only if all calls to I<&sub> were successfully. The B<permission_errors> option
  controlls, if an error is reported, when the scan process hits a restricted directory.

=cut

sub scanafs {
	my ($directory,%options)=@_;
	my $deepscan=1;
	my $permission_errors=1;
	my $cellname;
	if ( $options{'permission_errors'} ) {
		$permission_errors=1;
	}
	my $max_mountdepth=0;my $max_celldepth=0;my $max_depth=$MAXINT32;
	if ( defined($options{max_depth}) ) {
		$max_depth=$options{max_depth};
	}
	if ( defined($options{'max_mountdepth'}) ) {
		$max_mountdepth=$options{'max_mountdepth'};
	}
	if ( defined($options{'max_celldepth'}) ) {
		$max_celldepth=$options{'max_celldepth'};
	}

	# We'll keep book of the volumes, we crossed to avoid double scanning
	# (afs is not necessarily a tree...)
	my %volumes_hit;
	my $basecell='?';
	
	my $voldata=dir2vol($directory);
	if ( $voldata ) {
		$volumes_hit{"$voldata->{cell}|$voldata->{volume}"}=1;
		$basecell=$voldata->{cell};
	} else {
		out('v',"scanafs(): Unable to get '$directory' AFS properties");
	}
	
	
	# max_* options are counted down during scanafs_wrapped()-iterations
	# depth-first-search is stopped if one of those limits hits -1
	return scanafs_wrapped($directory,
		'max_mountdepth' => $max_mountdepth,
		'max_celldepth' => $max_celldepth,
		'max_depth' => $max_depth,
		'permission_errors' => $permission_errors,
		'cb_mountpoint' => $options{cb_mountpoint},
		'cb_dir' => $options{cb_dir},
		'cellname' => $basecell,
		'volumes_hit' => \%volumes_hit,
	);
}

=item $ok=scanafs_wrapped($directory,%options)

Scans a single directory for subdirectories, calls itself recursively.
This function must never be used manually, use I<scanafs()> instead.

=cut

sub scanafs_wrapped {
	my ($directory,%opt)=@_;
	my %optdown=%opt;
	my $dirhandle;
	my $result=1;
	if ( $opt{cb_dir} ) {
		$result = &{$opt{cb_dir}}($directory,$opt{cellname});
		if ( ! $result && $opt{stoppable} ) {
			return 0;
		}
	}
	if ( opendir($dirhandle,$directory) ) {
		my %to_be_scanned=();
		while ( my $dir=readdir($dirhandle) ) {
			my $newdir=$directory . '/' . $dir;
			if ( $dir !~ /^\.\.?$/ ) {
				my @stat=lstat($newdir);
				my $mpinfo;
				my $scandown=0;
				if ( @stat ) {
					if ( ! ($stat[2] & 0120000 ) ) {
						# dentry isn't point to a symlink
						if ( $stat[2] & 0040000 ) {
							# dentry is a directory
							if ( $stat[1] % 2 ) {
								# directory is not a volume mountpoint ( -> odd inode#)
								$scandown=1;
							} else {
								$mpinfo=read_mount($newdir);
							}
						}
					}
				} else {
					# Might be a mountpoint...
					$mpinfo=read_mount($newdir);
				}
				if ( defined($mpinfo) ) {
					if ( $opt{cb_mountpoint} ) {
						my $lresult=&{$opt{cb_mountpoint}}("$directory/$dir",$opt{cellname},$mpinfo);
						if ( $lresult || ! $opt{stoppable} ) {
							$scandown=1;
						} elsif ( ! $lresult ) {
							$result=0;
						}
					}
					if ( @stat && $opt{'max_mountdepth'} > 0 ) {
						# This is a mountpoint and we want to cross at least
						# one more mountpoint and we were able to stat() the
						# MP which means, the volume instance exists and is online.
						$optdown{'max_mountdepth'}--;
						my $newcell=$opt{cellname};
						if ( defined($mpinfo->{cell}) ) {
							$newcell=$mpinfo->{cell};
						}
						my $volid="$newcell|$mpinfo->{volume}";
						if ( $opt{volumes_hit}->{$volid} ) {
							# Do not hit the volume more than once
							$scandown=0;
						} else {
							if ( $opt{'max_celldepth'} > 0) {
								if ( @stat ) { $scandown=1; }
								$optdown{'max_celldepth'}--;
							}
							if ( $newcell eq $opt{'cellname'} ) {
								if ( @stat ) { $scandown=1; }
							}
							# It's called 'hit' not scanned :-)
							$opt{volumes_hit}->{$volid}=1;
						}
					} else {
						out('d',"No scanning below mountpoint '$newdir'");
					}
				}
				if ( $scandown ) {
					$to_be_scanned{$dir}=$stat[1];
				}
			} elsif ( $dir eq '.' ) {
				if ( defined($opt{'expected_inode'}) ) {
					my @dirstat=stat($directory);
					if ( $dirstat[1] != $opt{'expected_inode'} ) {
						out('w',"Unexpected inode# change at '$directory', skipping this subtree.");
						out('i',"This is either an attack or we just prevented a race condition from happening.");
						out('d'," Inode changed from $opt{'expected_inode'} to $dirstat[1]");
						close($dirhandle);
						return 0
					}
				}
			}
		}
		closedir($dirhandle);
		if ( $optdown{max_depth} > 0 ) {
			# Max directory depth for scanning not reached, yet
			$optdown{max_depth}--;
			foreach my $dir (sort keys %to_be_scanned) {
				my $newdir=$directory . '/' . $dir;
				$optdown{'expected_inode'}=$to_be_scanned{$dir};
				if ( ! scanafs_wrapped($newdir,%optdown) ) {
					$result=0;
				}
			}
		}
	} elsif ( ! $opt{'permission_errors'} and $!{EACCESS} ) {
		return 1;
	} else {
		return 0;
	}
	return $result;
}

=item \%result=dir2vol($directory)

Determines the volume a given directory I<$directory> resides in. I<%result> is returned as
a hashref containing these keys:

=over 4

=item B<volume>

Name of the volume

=item B<instance>

Instancetype ( 'B<readonly>', 'B<readwrite>' or 'B<backup>' )

=item B<cell>

Name of the cell the volume resides in

=item B<quota>

Quota limit of the volume (in kiBytes)

=item B<size>

Size of the volume (in kiBytes)

=item B<partfree>

Free space on the volume's partition (in kiBytes)

=item B<partsize>

Size of the volume's partition (in kiBytes)

=back

B<undef> is returned if there was an error determining the volume name
or the given I<$directory> doesn't reside in AFS.

=cut

sub dir2vol {
	my ($dir,%options)=@_;
	my $volume='';my $instance='';my $cell='';my $quota;my $size;my $partfree;my $partsize;my $vid;
	my $output=iaexec(
		['/usr/bin/fs','examine',$dir],
		'readpipe',
		stderr => 'hide',
		format => 'arrayref',
		description => "Getting information about volume containing '$dir'"
	) || return undef;
	for my $l (@{$output}) {
		if ( $l =~ /^Volume status for vid = ([0-9]+) named (\S+)$/ ) {
			($vid,$volume)=($1,$2);
		} elsif ( $l =~ /^Current disk quota is ([0-9]+)$/ ) {
			$quota=$1;
		} elsif ( $l =~ /^Current blocks used are ([0-9]+)$/ ) {
			$size=$1;
		} elsif ( $l =~ /^The partition has ([0-9]+) blocks available out of (-[0-9]+)$/ ) {
			($partfree,$partsize)=($1,$2);
			if ( $partsize < 0 ) {
				$partsize=undef;
			}
		}
	}
	$output=iaexec(
		['/usr/bin/fs','whichcell',$dir],
		'readpipe',
		stderr => 'hide',
		format => 'string',
		description => "Getting cellname of directory '$dir'",
	) || return undef;
	if ( $output =~ /^File.*lives in cell '(\S+)'$/ ) {
		$cell=$1;
	}
	if ( $volume ne '' ) {
		if ( $volume =~ /^(.*)\.backup$/ ) {
			$volume=$1;$instance='backup';
		} elsif ( $volume =~ /^(.*)\.readonly$/ ) {
			$volume=$1;$instance='readonly';
		} else {
			$instance='readwrite';
		}
		return {
			'volume' => $volume,
			'instance' => $instance,
			'cell' => $cell,
			'quota' => $quota,
			'size' => $size,
			'partfree' => $partfree,
			'partsize' => $partsize,
			'id' => $vid,
		};
	} else {
		return undef;
	}
}


=back

=cut

sub update_mountpoint_basedir_family {
	my ($bases,$valid,%options)=@_;
	my $result=1;
	for my $base (@{$bases}) {
		my %v2;
		for my $dentry (keys %{$valid}) {
			my $instance=$valid->{$dentry};
			if ( $base->[3] ne '' ) {
				$instance .= ".$base->[3]";
			}
			$v2{$dentry}=[$instance,$base->[2]];
		}
		$result &= update_mountpoint_basedir($base->[0],\%v2,$base->[1],%options);
	}
	return $result;
}

=item $result=update_mountpoint_basedir($path,\%valid,$volume,%options)

Makes sure, the given AFS directory I<$path> contains the AFS mountpoints defined in I<%valid>.
Format of I<%valid> is like this:

=over 4

=item Key is the dentry name

=item Value is an arrayref containing these information:

=over 4

=item I<0>: The volume instance name

=item I<1>: A boolean flag defining, if the volume should be mounted rw

=back

=back

Value being B<undef> causes the dentry to be ignored (->not removed) in the target directory.

I<%volume> is the name of the volume I<$path> resides in. Possible I<%options> are:

=over 4

=item B<volumeupdates> = I<\%volumeupdates>

Defines a hashref in which the key I<$volume> will be set to B<1> if any change occured.

=item B<dryrun>

Only simmulate things

=back

I<$result> == 1, only if no error occured.

=cut

sub update_mountpoint_basedir {
	my ($path,$valid,$volume,%options)=@_;
	my $result=1;
	for my $dentry (sort keys %{$valid}) {
		if ( ! defined($valid->{$dentry}) ) {
			next;
		}
		my $lresult=make_mount(
			"$path/$dentry",
			$valid->{$dentry}->[0],
			rw => $valid->{$dentry}->[1],
			idempotent => 2,
			dryrun => $options{dryrun},
		);
		if ( $lresult & 3 ) {
			if ( defined($options{volumeupdates}) ) {
				$options{volumeupdates}->{$volume}=1;
			}
		} elsif ( ! $lresult ) {
			$result=0;
		}
	}
	my $lresult=cleanup_mountpoint_basedir($path,$valid,$volume,%options);
	if ( $lresult & 2 ) {
		if ( defined($options{volumeupdates}) ) {
			$options{volumeupdates}->{$volume}=1;
		}
	}
	return $result;
}

=item $result=cleanup_mountpoint_basedir($path,\%valid,$volume,%options)

This function removes Dentries (AFS-Mountpoints) from the given directory I<$path>.
All dentries are removed that are not a key in I<%valid>.

=cut

sub cleanup_mountpoint_basedir {
	my ($path,$valid,$volume,%options)=@_;
	my $fh;
	if ( ! opendir($fh,$path) ) {
		out('e',"Unable to read directory '$path': $!");
		return 0;
	}
	my %tbr;
	while ( my $d=readdir($fh) ) {
		if ( $d ne '.' && $d ne '..' ) {
			if ( ! exists($valid->{$d}) ) {
				my @s=stat("$path/$d");
				if ( ! @s ) {
					if ( $!{ENODEV} or $!{ETIMEDOUT} ) {
						$tbr{$d}=1;
					}
				} elsif ( -l "$path/$d" ) {
					# Ignore links
				} elsif ( -d "$path/$d" ) {
					$tbr{$d}=1;
				}
			}
		}
	}
	out('d',"cleanup_mountpoint_basedir(): '" . scalar(keys %tbr) . "' dentries to be removed in '$path'");
	closedir($fh);
	my $success=1;
	for my $d (sort keys %tbr) {
		$success = $success and remove_mount("$path/$d",dryrun => $options{dryrun});
	}
	if ( scalar(keys %tbr) ) {
		if ( defined($options{volumeupdates}) ) {
			$options{volumeupdates}->{$volume}=1;
		}
		return $success*3;
	} else {
		return $success;
	}
}

sub prepareFtests {
	my %options=@_;
	if ( $options{cellname} ) {
		$cellname_override=$options{cellname};
	}
	if ( $options{realmname} ) {
		$realmname_override=$options{realmname}
	}
	return 1;
}

sub getCommandBlock {
	my $block=Fb::Out::getCommandBlock();
	push(@{$block->{parameters}},
		{},
		{name => 'instantafs-cell=s',arg => 'C',text => ["Work on cell C instead of local workstation's cell"],var => \$cellname_override},
		{name => 'instantafs-realm=s',arg => 'R',text => ["Work with kerberos realm R instead of local workstation's default realm"],var => \$realmname_override},
	);
	$block->{title}="Instantafs Options";
	return $block;
}

=item $xs=getUserPref()

Returns the XPath context of the current user's userpref.xml file or undef if there
is none.

=cut

my $userpref;
sub getUserPref {
	if ( $userpref ) {
		return $userpref;
	}
	if ( ! defined($ENV{HOME}) ) {
		return undef;
	}
	$userpref=readXmlFile("$ENV{HOME}/.instantafs/userpref.xml",ns => {'up' => $xmlnsup},test => 1);
	return $userpref;
}

=item \@dbservers=getDbServersDns($cell,%options);

Returns the associated AFS DB servers of AFS cell I<$cell>.
B<undef> is returned upon failure. Possible I<%options> are:

=over 4

=item B<timeout> = I<sec>

Return unsuccessfully if DNS didn't return a result after I<sec> seconds.

=item B<retry> = B<1>

Retry on errors if the timeout has not been triggered, yet.

=item B<test> = B<1>

Don't print errors on STDOUT.

=back

=cut

sub getDbServersDns {
	my ($cell,%options)=@_;
	my $dnstimeout=3; # 5 seconds timeout for DNS queries
	my $timeouttime;
	if ( $options{timeout} ) {
		$timeouttime=time() + $options{timeout};
	}
	my @timeoutadd;
	my @dbservers;
	my $cellq=quotemeta($cell);
	my $errors=0;

	RETRY:
	if ( $timeouttime ) {
		my $timeout=min($timeouttime-time(),$dnstimeout);
		if ( $timeout < 0 ) {
			if ( ! $options{test} ) {
				out('e',"No time left to look for DB servers in DNS.");
			}
			return undef;
		}
		@timeoutadd=("/usr/bin/timeout",$timeout);
	}
	my $dbservers=iaexec(
		[@timeoutadd,'/usr/bin/host','-t','SRV',"_afs3-vlserver._udp.$cell"],
		'readpipe',
		format => 'arrayref',
		description => "Reading AFS DB servers from DNS (SRV Records)"
	);
	if ( defined($dbservers) ) {
		for my $l (@{$dbservers}) {
			if ( $l =~ /^_afs3-vlserver._udp.$cellq has SRV record [0-9]+ [0-0] [0-9]+ (\S+?)\.?$/ ) {
				push(@dbservers,$1);
			} else {
				out('d',"Invalid/Unknown output from host command: '$l'");
				$errors++;
			}
		}
		if ( ! $errors ) {
			# No errors were found and a potential timeout was not triggered
			return \@dbservers;
		}
	}
	if ( $timeouttime ) {
		goto RETRY;
	}
	if ( ! $options{test} ) {
		out('e',"Unable to read DB servers from DNS");
	}
	return undef;
}

sub getDbServersFile {
	my ($cell,%options)=@_;
	my $cellservdb=readFile('/etc/openafs/CellServDB',format => 'arrayref');
	if ( ! defined($cellservdb) ) {
		if ( ! $options{test} ) {
			out('e',"Unable to read CellServDB file");
		}
		return undef;
	}
	my @dbservers;
	my $ccell='';
	for my $l (@{$cellservdb}) {
		if ( $l =~ /^>\s*([^\s#]+)/ ) {
			$ccell=$1;
		}
		if ( $ccell eq $cell ) {
			if ( $l =~ /^([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)/ ) {
				push(@dbservers,$1);
			}
		}
	}
	return \@dbservers;
}

=item $mode=getAfsdbMode()

Returns the local AFS client's mode of operation regarding DNS-based DB-server
discovery.

I<$mode> == B<1>, if the DNS is asked - otherwise B<0>

=cut

sub getAfsdbMode {
	my %options=@_;
	my $afsconf=readFile('/etc/openafs/afs.conf.client',format => 'arrayref');
	if ( ! defined($afsconf) ) {
		if ( ! $options{test} ) {
			out('w',"Unable to determine 'AFSDB-in-use'-state, assuming 'no'");
		}
		return 0;
	}
	my $afsdb=0;
	for my $l (@{$afsconf}) {
		if ( $l =~ /^\s*([^=]+?)\s*=\s*(\S+)\s*$/ ) {
			my ($key,$value)=($1,$2);
			if ( $key eq 'AFS_AFSDB' ) {
				if ( $value eq 'true' ) {
					$afsdb=1;
				}
			}
		}
	}
	out('d',"AFSDB mode is '$afsdb'");
	return $afsdb
}

sub getDbServers {
	my (%options)=@_;
	my $cell;
	if ( $options{cell} ) {
		$cell=$options{cell};
	} else {
		$cell=getCellName(%options) || return undef;
	}
	out('v',"Reading AFS DB server list of cell '$cell'");
	my $dbservers;
	if ( getAfsdbMode() == 1 ) {
		$dbservers=getDbServersDns($cell,%options);
	} else {
		$dbservers=getDbServersFile($cell,%options);
	}
	if ( ! defined($dbservers) ) {
		# Error message was already shown by submodule
		return undef;
	}
	if ( @{$dbservers} ) {
		return $dbservers;
	} else {
		if ( ! $options{test} ) {
			out('w',"Unable to determine any DB server of cell '$cell'");
		}
		return undef;
	}
} 

sub checkXmlAuthorisation {
	my ($k5user,$rule,$celldb)=@_;
	my $k4user=$k5user;
	$k4user =~ s/^host\//rcmd\//;
	$k4user =~ s/\//\./;
	for my $principal ( @{$rule->getElementsByTagName('principal')} ) {
		if ( $principal->hasAttribute('name') ) {
			my @names=split /\s+/,$principal->getAttribute('name');
			for my $name (@names) {
				if ( $name eq $k5user ) {
					return 1;
				}
			}
		}
		if ( $principal->hasAttribute('pattern') ) {
			my $pattern=$principal->getAttribute('pattern');
			if ( $k5user =~ /^$pattern$/ ) {
				return 1;
			}
		}
	}
	my $users_groups;
	for my $group ( @{$rule->getElementsByTagName('group')} ) {
		if ( ! $celldb ) {
			out('d',"checkXmlAuthorisation(): No celldb available. Skipping group matching");
			return 0;
		}
		if ( $group->hasAttribute('name') ) {
			if ( ! defined($users_groups) ) {
				$users_groups=$celldb->userInfo('memberships+',$k4user);
			}
			my @names=split /\s+/,$group->getAttribute('name');
			for my $name (@names) {
				if ( $users_groups->{$name} ) {
					return 1;
				}
			}
		}
		if ( $group->hasAttribute('pattern') ) {
			out('i',"Group membership patterns not supported, yet");
		}
	}
	return 0;
}


=head1 AUTHOR

Frank Burkhardt <burk@cbs.mpg.de>

=cut

END {
	exec_on_exit('__END__');
}

1;
__END__

