package AFS::Instant::Update;

use strict;
use warnings;
use 5.006;

use Carp qw(cluck);
use Carp;
use POSIX ":sys_wait_h";
use Data::Dumper;
use Socket;

use Fb::IO qw(iaexec);
use Fb::Out qw(out $VERBOSE);
use Fb::Network qw(getFQDN);

require Exporter;
our @ISA=qw(Exporter);
our $VERSION='3.0';
our $ONLINETIMEOUT=10*60;

=head1 NAME

AFS::Instant::Update - SQL-Database update for CellDB

=head1 SYNOPSIS

  use AFS::Instant::CellDB;
  use AFS::Instant::Update;
  
  my $celldb=AFS::Instant::CellDB->new();
  my $update=AFS::Instant::Update->new(celldb => $celldb);

  # Update everything, works only as root on an AFSDB-Server
  $update->run('*',ptdb => 1,partitiontimetolerance => 0);

=head1 REQUIRES

AFS::Instant::Toolz

=head1 METHODS

=head2 Constructor

=over 4

=item $update=new(%options)

Creates a celldb-update-object. Possible I<%options> are:

=over 4

=item B<celldb>

CellDB-Object to use for database-access.

=item B<ignorero>

Ignore Readonly-state of the CellDB-object. Otherwise RO-seeming access to the Database will
cause a critical error.

=back

=back

=cut

sub new {
	my ($package,%options)=@_;
	my $self={
		celldb => undef,
		vos => '/usr/bin/vos',
		rxdebug => '/usr/bin/rxdebug',
		pt_util => '/usr/sbin/pt_util',
	};
	my $me=bless($self,$package);
	if ( defined($options{'celldb'}) ) {
		$self->{celldb}=$options{'celldb'};

		if ( defined($self->{celldb}->{dbro}) ) {
			if ( ! $options{'ignorero'} ) {
				if ( $self->{celldb}->{dbro} ) {
					out('e',"DB access is readonly. Unable to initialize Update-object");
					return undef;
				}
			}
		}
	}
	return $me;
}

=head2 Object Methods

=over 4

The update instructure is used to populate the SQL Database from live information of the
connected AFS cell.

=over 4

=item $result=$update->ptdb(%options)

Updates group- and user information in the SQL DB from data read directly out of a PTDB
server's data file. This part of the update infrastructure is the only one with a
host-to-be-run-on-constraint (-> on one of the AFSDB-servers).

I<$result> == 1, only if there were no errors during the update.

Possible I<%options> are:

=over 4

=item B<ignoremissingptdb>

If set, existence of a local PTDB data file is not tested for.

=back

=cut

sub ptdb {
	my ($self,%options)=@_;
	out('v',"Updating PTDB entries and memberships");
	my $query;
	if ( ! $options{'ignoremissingptdb'} ) {
		if ( ! -r "/var/lib/openafs/db/prdb.DB0" ) {
			out('e',"PTDB file either doesn't exist or is unreadable");
			out('i',"updatePtdb() has to be called locally on an AFS-Database server with root permissions");
			return 0;
		}
	}

	my %name2id=();
	my $fh=iaexec(
		[$self->{pt_util},'-user'],
		'readpipe',
		stderr => 'hide',
		format => 'fh',
		description => "Reading PTDB user list from PTDB database",
	) || return 0;
	while ( my $l=<$fh> ) {
		chomp($l);
		my @f=split /\s+/,$l;
		$name2id{$f[0]}=int($f[2]);
	}
	close($fh);
	if ( $? ) {
		out('e',"Error reading list of PTDB users");
		return 0;
	}
	
	my $groupid;
	my %memberships=();
	$fh=iaexec(
		[$self->{pt_util},'-members'],
		'readpipe',
		stderr => 'hide',
		format => 'fh',
		description => "Reading PTDB memberships from PTDB database",
	) || return 0;
	while ( my $l=<$fh> ) {
		chomp($l);
		my @f=split /\s+/,$l;
		if ( $f[0] ne '' ) {
			$groupid=$f[2];
			$name2id{$f[0]}=$f[2];
		} else {
			$memberships{$groupid}->{$f[2]}=1;
		}
	}
	close($fh);
	if ( $? ) {
		out('e',"Error reading list of PTDB memberships");
		return 0;
	}
	
	my %dbname2id=();
	my %dbmemberships=();
	out('d',"Reading list of afsids from DB");
	$query=$self->{celldb}->sqlquery("SELECT name,afsid FROM afsids","Unable to read list afsids from DB");
	if ( ! $query ) {
		return 0;
	}
	while ( my $row=$query->fetchrow_arrayref() ) {
		$dbname2id{$row->[0]}=$row->[1];
	}
	out('d',"Reading group memberships from DB");
	$query=$self->{celldb}->sqlquery("SELECT afsgroup,member FROM memberships","Unable to read group memberships from DB");
	if ( ! $query ) {
		return 0;
	}
	while ( my $row=$query->fetchrow_arrayref() ) {
		$dbmemberships{$row->[0]}->{$row->[1]}=1;
	}
	my $success=1;
	for my $name (sort keys %name2id) {
		if ( ! defined($dbname2id{$name}) ) {
			my $nameq=$self->{celldb}->sqlquote($name);
			$query=$self->{celldb}->sqlqueryw("INSERT INTO afsids (name,afsid) VALUES ($nameq,$name2id{$name})","Unable to add user/group '$name' to DB");
			if ( ! $query ) {
				$success=0;
			} else {
				out('d',"Added user/group '$name' to DB");
			}
		} elsif ( $dbname2id{$name} != $name2id{$name} ) {
			# AFSID differs
			my $nameq=$self->{celldb}->sqlquote($name);
			$query=$self->{celldb}->sqlqueryw("UPDATE afsids SET afsid=$name2id{$name} WHERE name=$nameq","Unable to update AFSID of user/group '$name'");
			if ( ! $query ) {
				$success=0;
			} else {
				out('d',"Updated AFSID of user/group '$name' from $dbname2id{$name} to $name2id{$name}");
			}
		}
	}
	for my $name (sort keys %dbname2id) {
		if ( ! defined($name2id{$name}) ) {
			# user no longer exists
			$query=$self->{celldb}->sqlqueryw("DELETE FROM afsids WHERE afsid=$dbname2id{$name}","Unable to remove user/group '$name'");
			if ( ! $query ) {
				$success=0;
			} else {
				out('d',"Removed user/group '$name' from DB");
			}
		}
	}
	
	for my $group (sort keys %memberships) {
		for my $member (sort keys %{$memberships{$group}}) {
			if ( ! $dbmemberships{$group}->{$member} ) {
				$query=$self->{celldb}->sqlqueryw("INSERT INTO memberships (afsgroup,member) VALUES ($group,$member)","Unable to add membership of $member to group $group to DB");
				if ( ! $query ) {
					$success=0;
				} else {
					out('d',"Added membership of $member to group $group to DB");
				}
			}
		}
	}
	for my $group (sort keys %dbmemberships) {
		for my $member (sort keys %{$dbmemberships{$group}}) {
			if ( ! $memberships{$group}->{$member} ) {
				$query=$self->{celldb}->sqlqueryw("DELETE FROM memberships WHERE afsgroup=$group AND member=$member","Unable to remove membership of $member to group $group from DB");
				if ( ! $query ) {
					$success=0;
				} else {
					out('d',"Removed membership of $member to group $group from DB");
				}
			}
		}
	}
	return $success;
}


=item $result=$update->run($mode,%options)

Runs the synchronisation process on either a certain aspect of the AFS-cell's data or
on everything (once).

Possible I<$mode>s are:

=over 4

=item B<addresses>

read list of fileservers and their addresses

=item B<cleanup>

remove Instances residing on fileservers no longer existing

=item B<vldb>

read volumes information from VLDB

=item B<ptdb>

read user/group information from PTDB


=item B<server:>I<address>

update partition and local instances information of the filesserver given by its I<address>

=item B<*>

Update everything except of B<ptdb>.

=back

Available I<%options> are:

=over 4

=item B<parallel> = I<parallelity>

Defines the parallelity in which the update should be performed. B<1> means "none".

=item B<partitiontimetolerance>

This is forwarded to calls to the I<$update->instances()> method. (refer to this
method's doc for mor information).

=item B<debugdir>

This option will be forwarded to instances().

=item B<debuglevel>

This option will be forwarded to instances().

=back

=cut

sub run {
	my ($self,$mode,%options)=@_;
	my $parallel=1;
	if ( $options{'parallel'} ) {
		$parallel=$options{'parallel'};
	}
	my $success=1;
	my %jobs;

	# Server list has to be known first. Everything else doesn't
	# have any dependencies
	if ( $mode eq 'all' ) {
		$mode='*';
	}
	if ( $mode eq 'noptdb' ) {
		$mode='addresses,servers,vldb,cleanup';
	}
	for my $m (split /,/,$mode) {
		if ( $m eq '*' || $m eq 'addresses' ) {
			out('v',"Update run in progress, updating addresses.");
			$success &= $self->addresses(parallel => $parallel,dryrun => $options{dryrun});
			last;
		}
	}
	my $servers=$self->{celldb}->serverInfo('uuidlist');
	out('v',"Update run in progress, parallelity=$parallel");

	for my $m (split /,/,$mode) {
		my $ok=0;
		if ( $m eq 'all' ) {
			$m='*';
		}
		if ( $m eq 'cleanup' || $m eq '*' ) {
			$jobs{'cleanup'}=1;
			$ok=1;
		}
		if ( $m eq 'vldb' || $m eq '*' ) {
			$jobs{'vldb'}=1;
			$ok=1;
		}
		if ( $m eq 'ptdb' || $m eq '*' ) {
			$jobs{'ptdb'}=1;
			$ok=1;
		}
		if ( $m eq 'servers' || $m eq '*' ) {
			for my $uuid (@{$servers}) {
				$jobs{"server:$uuid"}=1;
			}
			$ok=1;
		}
		if ( $m =~ /^server:(\S+)$/ ) {
			my $serveruuid=$self->{celldb}->serverInfo('address2uuid',$1);
			if ( defined($serveruuid) ) {
				$jobs{"server:$serveruuid"}=1;
			} else {
				out('e',"Unable to determine UUID of server '$1'");
				return 0;
			}
			$ok=1;
		}
		if ( $m eq 'addresses' ) {
			$ok=1;
		}
		if ( ! $ok ) {
			out('e',"Invalid update mode '$m'");
			return 0;
		}
	}

	my @jobs=sort keys %jobs;

	my %pids;
	while ( scalar(@jobs) or scalar(keys %pids) ) {
		my $donesomething=0;
		if ( scalar(@jobs) and ( scalar(keys %pids) < $parallel ) ) {
			my $job=pop(@jobs);
			my $f=fork();
			if ( $f == 0 ) {
				exit ! $self->doJob($job,%options);
			} else {
				$pids{$f}=$job;
			}
			$donesomething=1;
		}
		if ( ( scalar(keys %pids) == $parallel ) or ( scalar(@jobs) == 0 && scalar(keys %pids) ) ) {
			my $pid=wait();
			if ( $? > 0 ) {
				$success=0;
			}
			delete $pids{$pid};
			$donesomething=1;
		}
		if ( ! $donesomething ) {
			sleep(3);
		}
	}
	return $success;
}

sub doJob {
	my ($self,$job,%options)=@_;
	if ( $job =~ /^server:(\S+)$/ ) {
		my $uuid=$1;
		my $servername=$self->{celldb}->serverInfo('uuid2name',$uuid);
		out('v',"Updating server '$servername' 's local volume instances");
		$0="server $servername";
		return $self->instances($uuid,%options);
	} elsif ( $job eq 'ptdb' ) {
		out('v',"Updating PTDB entries and memberships");
		$0=$job;
		return $self->ptdb(%options);
	} elsif ( $job eq 'vldb' ) {
		out('v',"Updating VLDB volumes");
		$0=$job;
		exit ! $self->vldb(%options);
	} elsif ( $job eq 'cleanup' ) {
		out('v',"Removing abandoned volume instances");
		$0=$job;
		exit ! $self->cleanup();
	} else {
		out('e',"Invalid job id '$job'");
		return 0;
	}
}


=item $result=$update->addresses()

Retrieve all active fileservers from VLDB incl. their addresses and the accessibility status of
all their interfaces. Fileservers that are no longer available will be removed automatically.

I<$result> == 1, only if no errors occured

=cut

sub addresses {
	my ($self,%options)=@_;
	my $fileservers={};
	my $parallel=$options{parallel};
	my $timeout=8;
	$self->{celldb}->ensureDBw() || return 0;
	out('v',"Updating Fileserver addresses");
		
	my @output;
	my $result=iaexec(
		[$self->{vos},'listaddrs','-cell',$self->{celldb}->{cellname},'-printuuid','-noresolve','-verbose','-noauth'],
		'readpipe',
		stderr => 'hide',
		format => \@output,
		description => "Reading numeric list of fileservers",
	);

	# Create a map (alias->fileserver) where alias is either an IP or a dns name
	# ip-addresses
	my $uuid='';my $addrcount=0;my $namecount=0;
	for my $l (@output) {
		if ( $l =~ /^UUID: (\S+)$/ ) {
			$uuid=$1;
			$fileservers->{$uuid}={
				'name' => undef,
				'addresses' => {},
			};
		} elsif ( $l =~ /^(\S+)$/ ) {
			my $ip=$1;
			$fileservers->{$uuid}->{'addresses'}->{$ip}={'io' => 0,'viewname' => 0};
			$addrcount++;
			if ( ! $fileservers->{$uuid}->{'name'} ) {
				# Reverse DNS lookup
				my $name=getFQDN($ip);
				if ( ! defined($name) ) {
					$name=$ip;
				}
				$fileservers->{$uuid}->{'name'}=$name;
				$fileservers->{$uuid}->{'addresses'}->{$name}={'io' => 0,'viewname' => 1};
				$namecount++;
			}
		}
	}
	out('d',"Found '$addrcount' IPs with '$namecount' names on '" . scalar(keys %{$fileservers}) . "' servers in VLDB.");
	
	my $fsaddresses={};
	my $tempcount=0;
	for my $uuid (sort keys %{$fileservers}) {
		for my $addr (sort keys %{$fileservers->{$uuid}->{'addresses'}}) {
			if ( ! $fileservers->{$uuid}->{'addresses'}->{$addr}->{'viewname'} ) {
				$fsaddresses->{$addr}={
					'uuid' => $uuid,
				};
				$tempcount++;
			}
		}
	}
	out('d',"About to test '" . scalar(keys %{$fsaddresses}) . "' IPs.");
	
	# Only check IP-addresses, resolve symbolic names and map ip-test-results
	my %pids=();my $running=0;my @addresses=sort keys %{$fsaddresses};
	my %starttime=();
	my $success=0;my $error=0;
	while ( $running || scalar(@addresses) ) {
		my $didsomething=0;
		if ( scalar(@addresses) && ( $running < $parallel ) ) {
			my $addr=shift @addresses;
			my $f=fork();
			if ( $f == 0 ) {
				$0="rxdebug-$addr";
				exit ! iaexec(
					['/usr/bin/timeout','-s','9','6s',$self->{rxdebug},$addr,7000,'-version'],
					'test',
					description => "'rxdebug'ing address '$addr'",
				);
			} else {
				$pids{$f}=$addr;
				$starttime{$f}=time();
			}
			$running++;
			$didsomething=1;
		}
		if ( ( scalar(keys %pids) == $parallel ) or ( scalar(@addresses) == 0 && scalar(keys %pids) ) ) {
			my $pid=waitpid(-1,WNOHANG);
			if ( $pid > 0 ) {
				my $addr=$pids{$pid};
				my $uuid=$fsaddresses->{$addr}->{'uuid'};
				my $name=$fileservers->{$uuid}->{'name'};
				if ( $? == 0 ) {
					$success++;
					my $diff=time()-$starttime{$pid};
					out('d',"Accessible address '$addr' for server '$name' determined in '${diff}s'");
					$fileservers->{$uuid}->{'addresses'}->{$addr}->{'io'}=1;
				} else {
					out('d',"Inaccessible address '$addr' for server '$name'.");
					$error++;
				}
				delete $pids{$pid};
				delete $starttime{$pid};
				out('d',"Still running: " . scalar(keys %pids) . ", left to run: " . scalar(@addresses));
				$didsomething=1;
				$running--;
			}
		}
		my $now=time();
		for my $pid (sort keys %starttime) {
			if ( ($now-$starttime{$pid}) > $timeout ) {
				my $addr=$pids{$pid};
				my $uuid=$fsaddresses->{$addr}->{'uuid'};
				my $name=$fileservers->{$uuid}->{'name'};
				out('d',"Timeout testing address '$addr' for server '$name' pid '$pid'");
				kill(15,$pid);
				$didsomething=1;
			}
		}
		if ( ! $didsomething ) {
			sleep(1);
		}
	}

	# Read state of servers/addresses from db
	my $query;
	my $dbfs={};
	my $dbuuid={};
	$query=$self->{celldb}->sqlqueryw("SELECT * FROM addresses");
	if ( ! $query ) {
		out('e',"Unable to read server/addresses list from DB");
		return 0;
	}
	while ( my $row=$query->fetchrow_hashref() ) {
		$dbfs->{$row->{'uuid'}}->{$row->{'address'}}=$row;
		$dbuuid->{$row->{'uuid'}}=1;
	}
	# compare database content to reality, start with db-content as reference
	my $infotime=time();
	$success=1;
	for my $uuid (sort keys %{$dbfs}) {
		for my $addr (sort keys %{$dbfs->{$uuid}}) {
			if ( $fileservers->{$uuid} ) {
				# given fileserver still exists in reality
				if ( $fileservers->{$uuid}->{'addresses'}->{$addr} ) {
					# address of given fileserver still exists in reality
					my $isviewname=int($fileservers->{$uuid}->{'name'} eq $addr);
					my $isaccessible=$fileservers->{$uuid}->{'addresses'}->{$addr}->{'io'};
					my @modify=("infotime=$infotime");
					if ( $isviewname != $dbfs->{$uuid}->{$addr}->{'viewname'} ) {
						push(@modify,"viewname=$isviewname");
					}
					if ( $isaccessible != $dbfs->{$uuid}->{$addr}->{'io'} ) {
						push(@modify,"io=$isaccessible");
					}
					if ( @modify ) {
						my $uuidq=$self->{celldb}->sqlquote($uuid);
						$query=$self->{celldb}->sqlqueryw("UPDATE addresses SET " . join(", ",@modify) . " WHERE uuid=$uuidq AND address='$addr'");
						if ( ! $query ) {
							out('e',"Unable to update DB record of server '$uuid'/addr='$addr'");
							$success=1;
						} else {
							out('d',"Updated DB record for server '$uuid'/addr='$addr'/io=$isaccessible");
						}
					}
				} else {
					# address of given fileserver was removed
					$query=$self->{celldb}->sqlqueryw("DELETE FROM addresses WHERE uuid='$uuid' AND address='$addr'");
					if ( ! $query ) {
						out('e',"Unable to remove address '$addr' of fileserver '$uuid' from DB");
						$success=0;
					} else {
						out('d',"Removed DB record of server '$uuid'/addr='$addr'");
					}
				}
			} else {
				# given fileserver is gone, remove all traces
				$query=$self->{celldb}->sqlqueryw("DELETE FROM addresses WHERE uuid='$uuid'");
				if ( ! $query ) {
					out('e',"Unable to remove address fileserver '$uuid' from DB");
					$success=0;
				} else {
					out('d',"Removed DB records of removed server '$uuid' 's addresses");
				}
				$query=$self->{celldb}->sqlqueryw("DELETE FROM partitions WHERE serveruuid='$uuid'");
				if ( ! $query ) {
					out('e',"Unable to remove partitions of fileserver '$uuid' from DB");
					$success=0;
				} else {
					out('d',"Removed DB records of removed server '$uuid' 's partitions");
				}
			}
		}
	}
	# Now we'll add new addresses/fileservers to db
	for my $uuid (sort keys %{$fileservers}) {
		for my $addr (sort keys %{$fileservers->{$uuid}->{'addresses'}}) {
			if ( ! $dbfs->{$uuid}->{$addr} ) {
				# address of this fileserver (or the whole fileserver) doesn't exist in DB
				my $isviewname=int($fileservers->{$uuid}->{'name'} eq $addr);
				my $isaccessible=$fileservers->{$uuid}->{'addresses'}->{$addr}->{'io'};
				my $addrq=$self->{celldb}->sqlquote($addr);
				$query=$self->{celldb}->sqlqueryw("INSERT INTO addresses (uuid,address,viewname,io,infotime) VALUES ('$uuid',$addrq,$isviewname,$isaccessible,$infotime)");
				if ( ! $query ) {
					out('e',"Unable to add server '$uuid'/addr='$addr' to DB");
					$success=0;
				} else {
					out('v',"Added server '$uuid' 's address '$addr' to DB");
				}
			}
		}
	}
	
	my @offline_servers=();
	for my $uuid (sort keys %{$fileservers}) {
		my $online=0;
		for my $addr (sort keys %{$fileservers->{$uuid}->{'addresses'}}) {
			$online |= $fileservers->{$uuid}->{'addresses'}->{$addr}->{'io'};
		}
		if ( ! $online ) {
			push(@offline_servers,$self->{celldb}->sqlquote($uuid));
		}
	}
	if ( @offline_servers ) {
		$query=$self->{celldb}->sqlqueryw(
			"UPDATE partitions SET ponline=0 WHERE serveruuid IN (" . join(',',@offline_servers) . ")",
			"Unable to mark non-responsive fileservers' partitions offline"
		);
		$success &= defined($query);
	}
	
	return $success;
}

=item $update->volumes()

Reads list of volumes and volume instances from VLDB and stores aquired data into the DB.

=cut

sub vldb {
	my ($self,%options)=@_;
	$self->{celldb}->ensureDBw() || return 0;
	my $debuglevel=0;
	if ( defined($options{'debuglevel'}) ) {
		$debuglevel=$options{'debuglevel'};
	}
	my $debug=0;
	if ( defined($options{'debugdir'}) ) {
		$debug=1;
	}
	my $query;
	my $fh=iaexec(
		[$self->{vos},'listvldb','-cell',$self->{celldb}->{cellname},'-nosort','-noauth','-noresolve'],
		'readpipe',
		format => 'fh',
		description => "Reading VLDB content",
	);
	if ( ! $fh ) {
		out('e',"##### Unable to exec() vos listvldb '[$self->{vos},'listvldb','-cell',$self->{celldb}->{cellname},'-nosort','-noauth','-noresolve']");
		return 0;
	}
	my $fhd;
	if ( $debug ) {
		# Protect against attacks via DNS
		my $filename='_vldb_volumes';
		my $file="$options{'debugdir'}/$filename";
		if ( -e $file ) {
			rename($file,"$file.1");
		}
		if ( ! open($fhd,"> $file") ) {
			out('e',"volumes(): Unable to open debug log '$file'");
			close($fh);
			return 0;
		} else {
			out('i',"Writing 'vos listvldb'-logfile to '$file'");
		}
	}

	my %volumes;my %instances;
	my $volname;my $volref;
	while ( my $l=<$fh> ) {
		chomp($l);
		if ( $l =~ /^(VLDB entries for all servers|Total entries:)/ ) {
			# Junk! We'll ignore that.
		} elsif ( $l eq "" ) {
			if ( defined($volname) ) {
				$volumes{$volname}=$volref;
				$volref->{'name'}=$volname;
				if ( @{$volref->{'ro'}} == 0 ) {
					$volref->{'rostate'}=2;
				} else {
					my $notzero=0;
					for my $ro (@{$volref->{'ro'}}) {
						if ( $ro->{'state'} != 0 ) {
							$notzero=1;
						}
					}
					$volref->{'rostate'}=$notzero;
				}
				if ( $debug ) {
					print $fhd Dumper(\$volref);
				}
			}
		} elsif ( $l =~ /^(\S+)\s*$/ ) {
			$volname=$1;
			$volref={
				'locked' => 0,
				'ro' => [],
				'backupid' => 0,
				'roid' => 0,
				'rwid' => 0,
			};
		} elsif ( $l =~ /^\s+RWrite/ ) {
			$l =~ /RWrite: ([0-9]+)/;
			$volref->{'rwid'}=$1;
			if ( $l =~ /Backup: ([0-9]+)/ ) {
				$volref->{'backupid'}=$1;
			}
			if ( $l =~ /ROnly: ([0-9]+)/ ) {
				$volref->{'roid'}=$1;
			}
		} elsif ( $l =~ /^\s+server (\S+)\s+partition \/vicep([a-z]+)\s+(RO|RW) Site(.*)$/ ) {
			# we found an instance of this volume...
			my $server=$1;my $partition=$2;my $copy=$3;my $comment=$4;
			
			# Server and Partition are registered
			my $serveruuid=$self->{celldb}->serverInfo('address2uuid',$server);
			if ( ! defined($serveruuid) ) {
				out('e',"Found an instance on a fileserver which is not registered in VLDB.");
				out('i'," Instance's volume is '$volname', fileserver address is '$server'");
				out('i'," This is an inconsistency in the VLDB, there'll probably be more errors later.");
			}
			
			# Check state of ro-copy
			if ( $copy eq 'RO' ) {
				# We'll determine the "quality" of this ro-instance
				# The smaller $state is, the better. $state > 1 means, the
				# ro-instance doesn't contain useable data.
				
				# Without a comment, all ro-instances are equally recent
				my $state=0;
				
				if ( $comment =~ /Not released/ ) {
					# This instance has never been released 
					$state=2;
				} elsif ( $comment =~ /Old release/ ) {
					# This is not an up to date RO-instance (other ROs are more recent)
					$state=1;
				} elsif ( $comment =~ /New release/ ) {
					# This is an up to date RO-instance
					$state=0;
				}
				push(@{$volref->{'ro'}},{
					'serveruuid' => $serveruuid,
					'partition' => $partition,
					'state' => $state,
				});
				$instances{$volname}->{$serveruuid}->{1}={
					'rostate' => $state,
					'partition' => $partition,
				}
			} else {
				$volref->{'rw'}={
					'serveruuid' => $serveruuid,
					'partition' => $partition,
				};
				$instances{$volname}->{$serveruuid}->{0}={
					'rostate' => 0,
					'partition' => $partition
				};
			}
		} elsif ( $l =~ /Volume is currently LOCKED/ ) {
			$volref->{'locked'}=1;
		}
		if ( $debug ) {
			print $fhd "$l\n";
		}

	}
	if ( $debug ) {
		close($fhd);
	}
	close($fh);
	if ( $? != 0 ) {
		out('e',"vos listvldb exited unsuccessfully '$?'");
		return 0;
	}

	my $success=1;	
	my %dbvolumes;my %dbinstances;
	$query=$self->{celldb}->sqlquery("SELECT * FROM volumes");
	if ( ! $query ) {
		out('e',"Unable to read list of volumes from DB");
		return 0;
	}
	while ( my $row=$query->fetchrow_hashref() ) {
		$dbvolumes{$row->{'volumename'}}=$row;
	}
	$query=$self->{celldb}->sqlquery("SELECT * FROM vldbinstances");
	if ( ! $query ) {
		out('e',"Unable to read list of instances from VLDB's Point of view from DB");
		return 0;
	}
	while ( my $row=$query->fetchrow_hashref() ) {
		$dbinstances{$row->{'volumename'}}->{$row->{'serveruuid'}}->{$row->{'instancetype'}}={
			'rostate' => $row->{'rostate'},
			'partition' => $row->{'partition'},
		};
	}
	my $vadded=0;my $vremoved=0;my $vmodified=0;
	my $vpredb=scalar(keys %dbvolumes);my $vfound=scalar(keys %volumes);
	# Remove volumes that no longer exist
	for my $volume (sort keys %dbvolumes) {
		if ( ! $volumes{$volume} ) {
			# volume no longer exists
			my $nameq=$self->{celldb}->sqlquote($volume);
			$query=$self->{celldb}->sqlqueryw("DELETE FROM volumes WHERE volumename=$nameq");
			if ( ! $query ) {
				out('e',"Unable to remove volume '$volume'");
				$success=0;
			} else {
				out('d',"Removed volume record of volume '$volume' from DB");
			}
			$vremoved++;
		} else {
			# volumes exists in DB, test for differences to reality
			my @modify;
			my $vref=$volumes{$volume};
			for my $field ('rwid','roid','backupid','rostate','locked') {
				if ( $vref->{$field} != $dbvolumes{$volume}->{$field} ) {
					push(@modify,"$field=$vref->{$field}");
				}
			}
			if ( @modify ) {
				$vmodified++;
				my $nameq=$self->{celldb}->sqlquote($volume);
				$query=$self->{celldb}->sqlqueryw(
					"UPDATE volumes SET " . join(", ",@modify) . 
						" WHERE volumename=$nameq",
					"Unable to update volume '$volume' in DB");
				if ( ! $query ) {
					$success=0;
				} else {
					out('d',"Updated volume record of '$volume' in DB");
				}
			}
		}
	}
	for my $volume (sort keys %volumes) {
		if ( ! $dbvolumes{$volume} ) {
			# volume is not it DB, yet
			my $nameq=$self->{celldb}->sqlquote($volume);
			my $vref=$volumes{$volume};
			$query=$self->{celldb}->sqlqueryw("INSERT INTO `volumes` (volumename,rwid,roid,backupid,rostate,locked,synctry_ro,synctry_backup) VALUES ($nameq,$vref->{'rwid'},$vref->{'roid'},$vref->{'backupid'},$vref->{'rostate'},$vref->{'locked'},0,0)");
			if ( ! $query ) {
				out('e',"Unable to add new volume '$volume' to DB");
				$success=0;
			} else {
				out('d',"Added record of volume '$volume' to DB");
			}
			$vadded++;
		}
	}
	out('v',"Updated volume records, pre-db=$vpredb, found=$vfound, added=$vadded, modified=$vmodified, removed=$vremoved");
	
	my $ipredb=scalar(keys %dbinstances);my $ifound=scalar(keys %instances);
	my $iadded=0;my $iremoved=0;my $imodified=0;
	for my $volume (sort keys %dbinstances) {
		for my $serveruuid (sort keys %{$dbinstances{$volume}}) {
			for my $instancetype (sort keys %{$dbinstances{$volume}->{$serveruuid}}) {
				if ( defined($instances{$volume}->{$serveruuid}->{$instancetype}) ) {
					# instance record already exists
					if (
						( $dbinstances{$volume}->{$serveruuid}->{$instancetype}->{'partition'} ne $instances{$volume}->{$serveruuid}->{$instancetype}->{'partition'} ) or
						( $dbinstances{$volume}->{$serveruuid}->{$instancetype}->{'rostate'} != $instances{$volume}->{$serveruuid}->{$instancetype}->{'rostate'} )
					) {
						my $partitionq=$self->{celldb}->sqlquote($instances{$volume}->{$serveruuid}->{$instancetype}->{'partition'});
						my $volumeq=$self->{celldb}->sqlquote($volume);
						my $serveruuidq=$self->{celldb}->sqlquote($serveruuid);
						$query=$self->{celldb}->sqlqueryw("UPDATE vldbinstances SET `partition`=$partitionq, rostate=$instances{$volume}->{$serveruuid}->{$instancetype}->{'rostate'} WHERE serveruuid=$serveruuidq AND volumename=$volumeq AND instancetype=$instancetype");
						if ( ! $query ) {
							out('e',"Unable to update vldb instance of '$volume'/$instancetype in DB");
							$success=0;
						} else {
							out('d',"Successfully updated vldb instance of '$volume'/$instancetype in DB");
						}
						$imodified++;
					}
				} else {
					# instance was removed in reality
					my $volumeq=$self->{celldb}->sqlquote($volume);
					my $serveruuidq=$self->{celldb}->sqlquote($serveruuid);
					$query=$self->{celldb}->sqlquery("DELETE FROM vldbinstances WHERE volumename=$volumeq AND serveruuid=$serveruuidq AND instancetype=$instancetype");
					if ( ! $query ) {
						out('e',"Unable to remove vldb instance of '$volume'/$instancetype from DB");
						$success=0;
					} else {
						out('d',"Successfully removed vldb instance record of '$volume'/$instancetype from DB");
					}
					$iremoved++;
				}
			}
		}
	}
	for my $volume (sort keys %instances) {
		for my $serveruuid (sort keys %{$instances{$volume}}) {
			for my $instancetype (sort keys %{$instances{$volume}->{$serveruuid}}) {
				if ( ! defined($dbinstances{$volume}->{$serveruuid}->{$instancetype}) ) {
					# instance is not in DB, yet
					my $volumeq=$self->{celldb}->sqlquote($volume);
					my $serveruuidq=$self->{celldb}->sqlquote($serveruuid);
					my $partitionq=$self->{celldb}->sqlquote($instances{$volume}->{$serveruuid}->{$instancetype}->{'partition'});
					$query=$self->{celldb}->sqlqueryw("INSERT INTO vldbinstances (volumename,instancetype,serveruuid,`partition`,rostate) VALUES ($volumeq,$instancetype,$serveruuidq,$partitionq,$instances{$volume}->{$serveruuid}->{$instancetype}->{'rostate'})");
					if ( ! $query ) {
						out('e',"Unable to add vldb instance of '$volume'/$instancetype to DB");
						$success=0;
					} else {
						out('d',"Sucessfully added instance of '$volume'/$instancetype to DB");
					}
					$iadded++;
				}
			}
		}
	}
	out('v',"Updated by-VLDB volume instance (only RW and RO) records, pre-db=$ipredb, found=$ifound, added=$iadded, modified=$imodified, removed=$iremoved");
	return $success;
}

=item $update->instances($serveruuid,%options)

Reads information from the fileserver specified by its I<$serveruuid>,
updates partition information (free space, ...) + information about volume
instances on partitions on this server in the SQL DB. These options are accepted:

=over 4

=item B<partitiontimetolerance>

Defines how long (in seconds) a partition intotime-stamp is not considered outdated.
B<0> means, there'll always be an update.

Main reason for this parameter is to avoid unnesesarry updates on DB records if there
are multiple update instances running on the same DB for fault tolerance.

=item B<debugdir>

Enables writing information read from fileservers to the given directory. One file
per fileserver will be created containing the output of the 'vos listvol' command.
Filename is the name of the fileserver as provided by $celldb->serverInfo('uuid2name').

=item B<debuglevel>

Defines the depth of debugging. Possible values are:

=over 4

=item B<0>

Just write the 'vos listvol' output to the debug file.

=item B<1>

Write a dump of the instance record which was extracted from an instance block
of the output after each instance block.

=back

=back

=cut

sub instances {
	my ($self,$serveruuid,%options)=@_;
	my $ptt=0;
	if ( defined($options{'partitiontimetolerance'}) ) {
		$ptt=$options{'partitiontimetolerance'};
	}
	my $debuglevel=0;
	if ( defined($options{'debuglevel'}) ) {
		$debuglevel=$options{'debuglevel'};
	}
	my $debug=0;
	if ( defined($options{'debugdir'}) ) {
		$debug=1;
	}
	my $query;	
	$self->{celldb}->ensureDBw() || return 0;
	my $infotime=time();
	my $server=$self->{celldb}->serverInfo('uuid2ioaddress',$serveruuid);
	my $servername=$self->{celldb}->serverInfo('uuid2name',$serveruuid);
	if ( ! defined($server) ) {
		if ( defined($servername) ) {
			$servername=" ('$servername')";
		} else {
			$servername='';
		}
		out('d',"Unable to determine io address of server '$serveruuid'$servername, seems to be offline");
		return 0;
	}
	out('d',"Updating partitions of server '$serveruuid', addr='$server'");
	
	my %partitions;
	my @output;
	my $result=iaexec(
		[$self->{vos},'partinfo',$server,'-noauth'],
		'readpipe',
		stderr => 'hide',
		format => \@output,
		description => "Reading partition information from server '$servername' via '$server'",
	);
	if ( ! $result ) {
		return 0;
	}
	my %scanlengths;
	for my $l (@output ) {
		if ( $l =~ /^Free space on partition \/vicep([a-z]+): ([0-9]+) K blocks out of total ([0-9]+)/ ) {
			$partitions{$1}={
				'free' => $2*1024,
				'size' => $3*1024,
				'quotasum' => 0,
				'quotasumrw' => 0,
				'sizesum' => 0,
				'sizesumrw' => 0,
				'filessum' => 0,
				'filessumrw' => 0,
				'partinstancesrw' => 0,
				'partinstances' => 0,
				'partvolumes' => 0,
				'maxmtimerw' => 0,
			};
			$scanlengths{$1}=0;
		}
	}

	my $fh=iaexec(
		[$self->{vos},'listvol',$server,'-format','-noresolve','-noauth'],
		'readpipe',
		format => 'fh',
		description => "Reading volume instances from server '$servername' via '$server'",
	);
	if ( ! $fh ) {
		return 0;
	}
	my $fhd;
	if ( $debug ) {
		my $filename=$servername;
		# Protect against attacks via DNS
		$filename =~ s/[^a-z0-9\.-]/-/g;
		my $file="$options{'debugdir'}/$filename";
		if ( -e $file ) {
			rename($file,"$file.1");
		}
		if ( ! open($fhd,"> $file") ) {
			out('e',"instances(): Unable to open debug log '$file'");
			close($fh);
			return 0;
		} else {
			out('i',"Writing 'vos listvol'-logfile to '$file'");
		}
	}

	my %instances=();
	my %emptyiref=(
		'creationDate' => undef,
		'accessDate' => undef,
		'updateDate' => undef,
		'_size' => undef,
		'_quota' => undef,
		'dayUse' => undef,
		'filecount' => undef,
		'id' => undef,
		'_partition' => undef,
		'_online' => undef,
		'_needssalvage' => undef,
		'status' => undef,
		'_volumename' => undef,
		'_instancetype' => undef,
		'_isbusy' => 0,
	);
	my $isalvage=0;

	my $iref;
	my $prevvol='(first)';
	my $recordend=0;

	# For measuring partition scanlength
	my $starttime=time();
	my $currentpartition='';

	while ( my $l=<$fh> ) {
		chomp($l);
		if ( $l ) {
			my @f=split /\s+/,$l;
			if ( @f >= 2 ) {
				if ( exists $iref->{$f[0]} ) {
					$iref->{$f[0]}=$f[1];
					if ( $f[0] eq 'status' ) {
						if ( $f[1] eq 'BUSY' ) {
							$iref->{'_isbusy'}=1;
						}
					}
				} elsif ( $f[0] eq 'name' ) {
					if ( $f[1] =~ /^(.*)\.(backup|readonly)$/ ) {
						# This is a RO or backup instance
						$iref->{'_volumename'}=$1;
						$prevvol=$f[1];
						$iref->{'_instancetype'}=int(1 + ($2 eq 'backup'));
					} else {
						# This is a RW instances
						$iref->{'_volumename'}=$f[1];
						$prevvol=$f[1];
						$iref->{'_instancetype'}=0;
					}
				} elsif ( $f[0] eq 'diskused' ) {
					$iref->{'_size'}=$f[1]*1024;
				} elsif ( $f[0] eq 'maxquota' ) {
					$iref->{'_quota'}=$f[1]*1024;
				} elsif ( $f[0] eq 'part' ) {
					$iref->{'_partition'}=substr($f[1],6);
					if ( $currentpartition ne $iref->{'_partition'} ) {
						$currentpartition=$iref->{'_partition'};
						$scanlengths{$currentpartition}=time() - $starttime;
					}
				} elsif ( $f[0] eq 'inUse' ) {
					$iref->{'_online'}=int($f[1] eq 'Y');
				} elsif ( $f[0] eq 'needsSalvaged' ) {
					$iref->{'_needssalvage'}=0;
					if ( $f[1] eq 'Y' ) {
						$iref->{'_needssalvage'}=1;
					}
				}
			} elsif ( $f[0] eq 'BEGIN_OF_ENTRY' ) {
				$iref={%emptyiref};
			} elsif ( $f[0] eq 'END_OF_ENTRY' ) {
				if ( ! defined($iref->{'_partition'}) ) {
					out('e',"Undefined partition spotted on '$servername'. Aborting for safety reasons. Previous volume was '$prevvol'");
					out('e',"Record is" . Dumper($iref));
					close($fh);
					return 0;
				}
				if ( $iref->{'_isbusy'} ) {
					my $result=$self->{celldb}->resolveVolumeId($iref->{'id'});
					if ( defined($result) ) {
						$iref->{'_volumename'}=$result->[0];
						$iref->{'_instancetype'}=$result->[1];
					} else {
						out('d',"Found instance without volumename:" . Dumper($iref));
					}
				}
				if ( defined($iref->{'_volumename'}) ) {
					$recordend=1;
					if ( $iref->{'status'} eq 'UNATTACHABLE' ) {
						$iref->{'_needssalvage'}=1;
						# Not zero - this might cause problems with Software checking for unlimited quota
						$iref->{'_quota'}=1;
						for my $f ('updateDate','_size','accessDate','filecount','dayUse','creationDate','_online') {
							$iref->{$f}=0;
						}
					}
					if ( $iref->{'_needssalvage'} ) {
						$isalvage++;
					}
					$instances{$iref->{'_volumename'}}->{$iref->{'_instancetype'}}=$iref;
					if ( ( $VERBOSE >= 2 ) and ! $iref->{'_isbusy'} ) {
						my @undefined=();
						for my $key (keys %emptyiref) {
							if ( ! defined($iref->{$key}) ) {
								push(@undefined,$key);
							}
						}
						if ( @undefined ) {
							out('d',"Undefined fields in iref of instance '$iref->{'_volumename'}' on server '$serveruuid");
							print Dumper($iref);
						}
					}
				}
			}
		}
		if ( $debug ) {
			print $fhd "$l\n";
			if ( $recordend ) {
				$recordend=0;
				print $fhd Dumper($iref) . "\n";
			}
		}
	}
	close($fh);
	out('d',"Reading list of volume instances of server '$serveruuid' from DB");
	my $serveruuidq=$self->{celldb}->sqlquote($serveruuid);
	$query=$self->{celldb}->sqlquery("SELECT * FROM instances WHERE serveruuid=$serveruuidq","Unable to read list of instances from on partitions from DB");
	if ( ! $query ) {
		if ( $debug ) {
			close($fhd);
		}
		return 0;
	}
	my %dbinstances;
	my %doubles=();
	while ( my $row=$query->fetchrow_hashref() ) {
		if ( $dbinstances{$row->{'volumename'}}->{$row->{'instancetype'}} ) {
			if ( $dbinstances{$row->{'volumename'}}->{$row->{'instancetype'}}->{'partition'} eq $row->{'partition'} ) {
				$doubles{$row->{'volumename'}}->{$row->{'instancetype'}}=1;
			}
		}
		$dbinstances{$row->{'volumename'}}->{$row->{'instancetype'}}=$row;
	}
	my $idbpre=scalar(keys %dbinstances);
	my $iserver=scalar(keys %instances);
	my $imodified=0;
	my $ibusy=0;
	my $iremoved=0;
	my $iadded=0;
	my $idremove=0;
	my @added;my @removed;my @modified;my @busy;my %vldbpartitions=();

	# Read list of partitions that VLDB knows about. This is used to check for missing
	# data partitions.
	$query=$self->{celldb}->sqlquery("SELECT DISTINCT `partition` FROM `vldbinstances` WHERE `serveruuid`=$serveruuidq");
	while ( my $row=$query->fetchrow_arrayref() ) {
		$vldbpartitions{$row->[0]}=1;
	}

	# This will fix inconsistencies in the database where multiple records
	# exist for a single physical volume instance
	for my $volume (sort keys %doubles) {
		for my $instancetype (sort keys %{$doubles{$volume}}) {
			my $volumeq=$self->{celldb}->sqlquote($volume);
			my $query=$self->{celldb}->sqlqueryw(
				"DELETE FROM instances WHERE serveruuid=$serveruuidq AND volumename=$volumeq AND instancetype=$instancetype",
				"Unable to remove multiple instances of the same volume+type+serveruuid"
			);
			# We'll make the DB-record disappear so it will be rewritten
			$idremove++;
			delete $dbinstances{$volume}->{$instancetype};
		}
	}

	# Check known instances from DB, removing records of no longer existing
	# instances and updating the others.
	for my $volume (sort keys %dbinstances) {
		for my $instancetype (sort keys %{$dbinstances{$volume}}) {
			# Iterate over all existing database volume instance entries
			if ( defined($instances{$volume}->{$instancetype}) ) {
				my $direc=$dbinstances{$volume}->{$instancetype};
				my $irec=$instances{$volume}->{$instancetype};
				if ( ! $irec->{'_isbusy'} ) {
					# instance record already exists, updating necessary fields
					my @modify;
					if ( $direc->{'size'} != $irec->{'_size'} ) {
						push(@modify,"size=$irec->{'_size'}");
					}
					if ( $direc->{'quota'} != $irec->{'_quota'} ) {
						push(@modify,"quota=$irec->{'_quota'}");
					}
					if	( $direc->{'files'} != $irec->{'filecount'} ) {
						push(@modify,"files=$irec->{'filecount'}");
					}
					if ( $direc->{'ctime'} != $irec->{'creationDate'} ) {
						push(@modify,"ctime=$irec->{'creationDate'}");
					}
					if ( $direc->{'mtime'} != $irec->{'updateDate'} ) {
						push(@modify,"mtime=$irec->{'updateDate'}");
					}
					if ( $direc->{'atime'} != $irec->{'accessDate'} ) {
						push(@modify,"atime=$irec->{'accessDate'}");
					}
					if ( $direc->{'id'} != $irec->{'id'} ) {
						# Rather unlikely but nevertheless possible...
						push(@modify,"id=$irec->{'id'}");
					}
					if ( $direc->{'partition'} ne $irec->{'_partition'} ) {
						push(@modify,"`partition`=" . $self->{celldb}->sqlquote($irec->{'_partition'}));
					}
					if ( $direc->{'online'} != $irec->{'_online'} ) {
						push(@modify,"online=$irec->{'_online'}");
					}
					if ( $direc->{'needssalvage'} != $irec->{'_needssalvage'} ) {
						push(@modify,"needssalvage=$irec->{'_needssalvage'}");
					}
					if ( $direc->{'isbusy'} ) {
						push(@modify,"isbusy=0");
					}
					if ( @modify ) {
						my $volumeq=$self->{celldb}->sqlquote($volume);
						$query=$self->{celldb}->sqlqueryw(
							"UPDATE instances SET " . join(", ",@modify) .
								" WHERE serveruuid=$serveruuidq AND volumename=$volumeq AND instancetype=$instancetype",
							"Unable to update '$volume'/$instancetype on '$servername' in DB"
						);
						if ( $query ) {
							if ( $VERBOSE > 2 ) {
								out('d',"Updated instance '$volume'/$instancetype in DB");
							}
							$imodified++;
							push(@modified,"$volume/$instancetype");
						}
					}
				} else {
					# We take the instance's last known data from DB
					$partitions{$irec->{'_partition'}}->{'quotasum'} += $direc->{'quota'};
					$partitions{$irec->{'_partition'}}->{'sizesum'} += $direc->{'size'};
					$partitions{$irec->{'_partition'}}->{'filessum'} += $direc->{'files'};
					$partitions{$irec->{'_partition'}}->{'partinstances'} += 1;
					if ( $irec->{instancetype} == 0 ) {
						$partitions{$irec->{'_partition'}}->{'quotasumrw'} += $direc->{'quota'};
						$partitions{$irec->{'_partition'}}->{'sizesumrw'} += $direc->{'size'};
						$partitions{$irec->{'_partition'}}->{'filessumrw'} += $direc->{'files'};
						$partitions{$irec->{'_partition'}}->{'partinstancesrw'} += 1;
						if ( $partitions{$irec->{'_partition'}}->{'maxmtimerw'} < $direc->{'mtime'} ) {
							$partitions{$irec->{'_partition'}}->{'maxmtimerw'}=$direc->{'mtime'};
						}
					}
					
					# Now we set the instance to busy if it isn't already according to DB
					if ( ! $direc->{'isbusy'} ) {
						my $volumeq=$self->{celldb}->sqlquote($volume);
						$query=$self->{celldb}->sqlqueryw("UPDATE instances SET isbusy=1 WHERE volumename=$volumeq AND serveruuid=$serveruuidq AND instancetype=$instancetype","Unable to set ' $volume '/$instancetype to busy in DB");
						if ( $query ) {
							out('d',"Set '$volume'/$instancetype to busy in DB");
						}
					}
					push(@busy,"$volume/$instancetype");
					$ibusy++;
				}
			} else {
				# instance was removed, purge DB entry
				my $volumeq=$self->{celldb}->sqlquote($volume);
				$query=$self->{celldb}->sqlqueryw("DELETE FROM instances WHERE volumename=$volumeq AND serveruuid=$serveruuidq AND instancetype=$instancetype","Unable to remove instance of '$volume'/$instancetype from DB");
				if ( $query ) {
					out('d',"Removed instance of '$volume'/$instancetype from DB");
					$iremoved++;
					push(@removed,"$volume/$instancetype");
				}
			}
		}
	}
	for my $volume (sort keys %instances) {
		for my $instancetype (sort keys %{$instances{$volume}}) {
			my $irec=$instances{$volume}->{$instancetype};
			$partitions{$irec->{'_partition'}}->{volumes}->{$irec->{'_volumename'}}=1;
			if ( ! $irec->{'_isbusy'} ) {
				$partitions{$irec->{'_partition'}}->{'quotasum'} += $irec->{'_quota'};
				$partitions{$irec->{'_partition'}}->{'sizesum'} += $irec->{'_size'};
				$partitions{$irec->{'_partition'}}->{'filessum'} += $irec->{'filecount'};
				$partitions{$irec->{'_partition'}}->{'partinstances'} += 1;
				if ( $irec->{_instancetype} == 0 ) {
					$partitions{$irec->{'_partition'}}->{'quotasumrw'} += $irec->{'_quota'};
					$partitions{$irec->{'_partition'}}->{'sizesumrw'} += $irec->{'_size'};
					$partitions{$irec->{'_partition'}}->{'filessumrw'} += $irec->{'filecount'};
					$partitions{$irec->{'_partition'}}->{'partinstancesrw'} += 1;
					if ( $partitions{$irec->{'_partition'}}->{'maxmtimerw'} < $irec->{'updateDate'} ) {
						$partitions{$irec->{'_partition'}}->{'maxmtimerw'}=$irec->{'updateDate'};
					}
				}
			}
			if ( ! defined($dbinstances{$volume}->{$instancetype}) ) {
				# instance is not in DB, yet
				my $volumeq=$self->{celldb}->sqlquote($volume);
				my $partitionq=$self->{celldb}->sqlquote($irec->{'_partition'});
				if ( ! $irec->{'_isbusy'} ) {
					$query=$self->{celldb}->sqlqueryw("INSERT INTO instances (volumename,serveruuid,instancetype,`partition`,size,quota,files,ctime,mtime,atime,id,online,needssalvage,isbusy) VALUES ($volumeq,$serveruuidq,$instancetype,$partitionq,$irec->{'_size'},$irec->{'_quota'},$irec->{'filecount'},$irec->{'creationDate'},$irec->{'updateDate'},$irec->{'accessDate'},$irec->{'id'},$irec->{'_online'},$irec->{'_needssalvage'},0)","Unable to add non-busy volume instance of '$volume'/$instancetype to DB");
					if ( $query ) {
						out('d',"Successfully added volume instance of '$volume'/$instancetype to DB");
						$iadded++;
						push(@added,"$volume/$instancetype");
					} else {
						print Dumper($irec);
					}
				} else {
					# We can't add a busy instance since we do not have all the information (size,quota,...) until it's
					# non-busy again. This is guarantee that volume instances in DB always contain
					# Useful filecount/quota/size/... information
					out('d',"Skipped adding busy volume instance with id '$irec->{'id'}' on '$servername/$irec->{'_partition'}'");
				}
			}
		}
	}
	out('v',"Updated instances on '$servername', pre-db=$idbpre, found=$iserver, added=$iadded, modified=$imodified, removed=$iremoved, busy=$ibusy, tbsalvaged=$isalvage, idremove=$idremove");
	if ( $debug ) {
		print $fhd "Added: " . join(' ',@added) . "\n" .
			"Modified: " . join(' ',@modified) . "\n" .
			"Removed: " . join(' ',@removed) . "\n";
	}
	
	# Look for partitions in the database
	my %dbpartitions=();
	$query=$self->{celldb}->sqlquery("SELECT * FROM partitions WHERE `serveruuid`=$serveruuidq","Unable to read list of partitions of server '$serveruuid' from DB");
	if ( ! $query ) {
		if ( $debug ) {
			close($fhd);
		}
		return 0;
	}
	while ( my $row=$query->fetchrow_hashref() ) {
		$dbpartitions{$row->{'partition'}}=$row;
	}
	# Remove partitions that no longer exist, write changes for modified partitions
	for my $partition (sort keys %dbpartitions) {
		my $partitionq=$self->{celldb}->sqlquote($partition);
		my $prec=$partitions{$partition};
		if ( $partitions{$partition} ) {
			# Partition record already exists, testing if an update is necessary
			my @modify;
			my $dprec=$dbpartitions{$partition};
			if ( abs($infotime-$dprec->{'infotime'}) >= $ptt ) {
				push(@modify,"infotime=$infotime");
				push(@modify,"scanlength=$scanlengths{$partition}");
			}
			if ( $dprec->{'ponline'} != 1 ) {
				push(@modify,"ponline=1");
			}
			if ( $dprec->{'partsize'} != $prec->{'size'} ) {
				push(@modify,"partsize=$prec->{'size'}");
			}
			if ( $dprec->{'partfree'} != $prec->{'free'} ) {
				push(@modify,"partfree=$prec->{'free'}");
			}
			if ( $dprec->{'quotasum'} != $prec->{'quotasum'} ) {
				push(@modify,"quotasum=$prec->{'quotasum'}");
			}
			if ( $dprec->{'quotasumrw'} != $prec->{'quotasumrw'} ) {
				push(@modify,"quotasumrw=$prec->{'quotasumrw'}");
			}
			if ( $dprec->{'sizesum'} != $prec->{'sizesum'} ) {
				push(@modify,"sizesum=$prec->{'sizesum'}");
			}
			if ( $dprec->{'sizesumrw'} != $prec->{'sizesumrw'} ) {
				push(@modify,"sizesumrw=$prec->{'sizesumrw'}");
			}
			if ( $dprec->{'filessum'} != $prec->{'filessum'} ) {
				push(@modify,"filessum=$prec->{'filessum'}");
			}
			if ( $dprec->{'filessumrw'} != $prec->{'filessumrw'} ) {
				push(@modify,"filessumrw=$prec->{'filessumrw'}");
			}
			if ( $dprec->{'partinstances'} != $prec->{'partinstances'} ) {
				push(@modify,"partinstances=$prec->{'partinstances'}");
			}
			if ( $dprec->{'partinstancesrw'} != $prec->{'partinstancesrw'} ) {
				push(@modify,"partinstancesrw=$prec->{'partinstancesrw'}");
			}
			my $partvolumes=scalar(keys %{$prec->{volumes}});
			if ( $dprec->{'partvolumes'} != $partvolumes ) {
				push(@modify,"partvolumes=$partvolumes");
			}
			if ( $dprec->{'maxmtimerw'} != $prec->{'maxmtimerw'} ) {
				push(@modify,"maxmtimerw=$prec->{'maxmtimerw'}");
			}
			if ( @modify ) {
				$query=$self->{celldb}->sqlqueryw(
					"UPDATE partitions SET " . join(", ",@modify) . 
						" WHERE serveruuid=$serveruuidq AND `partition`=$partitionq",
					"Unable to update partition '$serveruuid':$partition");
				if ( $query ) {
					out('d',"Successfully updated partition '$serveruuid':$partition");
				}
			}
		} else {
			if ( $vldbpartitions{$partition} ) {
				# Partitions is referred to by VLDB, assuming missing mount on fileserver
				# If the partition hasn't been seen before, this update will do no harm.
				$query=$self->{celldb}->sqlqueryw("UPDATE partitions SET ponline=0 WHERE serveruuid=$serveruuidq AND `partition`=$partitionq","Unable to flag partition '$serveruuid':$partition as 'offline'");
			} else {
				# Partition doesn't exist physically, yet. -> Remove DB record
				$query=$self->{celldb}->sqlqueryw("DELETE FROM partitions WHERE serveruuid=$serveruuidq AND `partition`=$partitionq","Unable to remove partition '$serveruuid':$partition");
				if ( $query ) {
					out('v',"Successfully removed partition '$serveruuid':$partition");
				}
			}
		}
	}
	for my $partition (sort keys %partitions) {
		if ( ! $dbpartitions{$partition} ) {
			# Partition record doesn't exist, yet. -> Create
			my $partitionq=$self->{celldb}->sqlquote($partition);
			my $prec=$partitions{$partition};
			my $partvolumes=scalar(keys %{$prec->{volumes}});
			$query=$self->{celldb}->sqlqueryw("INSERT INTO partitions " . 
				"(serveruuid,`partition`,partsize,partfree,infotime,quotasum,quotasumrw,sizesum,sizesumrw,filessum,filessumrw,partinstances,partinstancesrw,ponline,partvolumes,maxmtimerw,scanlength) " .
				"VALUES ($serveruuidq,$partitionq,$prec->{'size'},$prec->{'free'},$infotime,$prec->{'quotasum'},$prec->{'quotasumrw'},$prec->{'sizesum'},$prec->{'sizesumrw'},$prec->{'filessum'},$prec->{'filessumrw'},$prec->{'partinstances'},$prec->{'partinstancesrw'},1,$partvolumes,$prec->{'maxmtimerw'},$scanlengths{$partition})",
				"Unable to add partition '$serveruuid':$partition");
			if ( $query ) {
				out('v',"Successfully added partition '$serveruuid':$partition");
			}
		}
	}
	if ( $debug ) {
		close($fhd);
	}
	return 1;
}

=item $update->cleanup()

Removes volume instance records and partition records of fileservers that are no longer
available from SQL DB. Availability is determined by the addresses() method.

=cut

sub cleanup {
	my $self=shift;
	$self->{celldb}->ensureDBw() || return 0;
	my $query;
	$query=$self->{celldb}->sqlqueryw("DELETE FROM instances WHERE serveruuid NOT IN (SELECT DISTINCT uuid FROM addresses)","Unable to removing abandoned volume instances");
	if ( ! $query ) {
		return 0;
	}
	$query=$self->{celldb}->sqlquery("DELETE FROM partitions WHERE serveruuid NOT IN (SELECT DISTINCT uuid FROM addresses)","Unable to removing abandoned partitions");
	if ( ! $query ) {
		return 0;
	}
	return 1;
}

=back

=cut

1;
