package AFS::Instant::aCall;

use 5.006;
use strict;
use warnings;

use IO::Socket;
use Storable qw(nfreeze thaw);
use XML::LibXML;
use XML::LibXML::XPathContext;
use GSSAPI;

use Fb::Out qw(out $VERBOSE);
use Fb::Network qw(getFQDN);
use AFS::Instant::Toolz qw(getCellName dot);

require Exporter;
our @ISA=qw(Exporter);
our $VERSION='3.0';
our $krb5object;

my $defaultacallversion=3;

=head1 NAME

AFS::Instant::aCall - Encrypted/Authenticated shell command execution

=head1 SYNOPSIS

  use AFS::Instant::aCall;

  my $acall=AFS::Instant::aCall->new('db');
  $acall->call('instantafs.acall-release',['root.cell']);
  $acall->call('instantafs.acall-release',['root.afs']);

=head1 REQUIRES

GSSAPI, AFS::Instant::Toolz, IO::Socket, Storable
XML::LibXML, strict :-)

=head1 DESCRIPTION

This class' purpose is to call shell commands on remote servers.
It's like RPC but very easy to set up.

This class is a rather dirty because I don't know how to work with this
Kerberos-module correctly. It should be used only once per process.

=head1 METHODS

=head2 Constructor

=over 4

=item $acall=new(%options)

Creates a new aCall-Object. Some options may be given:

B<hostoverride> => I<server> Override acall-server's hostname
  (with the DNS-name I<server>)

=cut

sub new {
	my ($package,@conf)=@_;
	my %conf;
	if ( ( scalar(@conf) % 2 ) == 1 ) {
		my $role=shift @conf;
		%conf=@conf;
		$conf{'role'}=$role;
	} else {
		%conf=@conf;
	}
	my $self={
		role => $conf{'role'},
		hostname => undef,
		version => $defaultacallversion,
		fh_acall_lock => undef,
		acall_xml_mtime => 0,
		doc => undef,
		root => undef,
		xc => undef,
		error => 0,
	};
	if ( $conf{'cellname'} ) {
		$self->{'cellname'}=$conf{'cellname'};
	} else {
		$self->{cellname}=getCellName();
	}
	my $me=bless($self,$package);

	if ( defined($me->{role}) ) {
		# Get server's role without optional domainname
		if ( $me->{role} =~ /^([^\.]+)\.(.*)$/ ) {
			$me->{role}=$1;
			$me->{cellname}=$2;
		}
		# Get remote server's FQDN
		if ( defined($conf{'hostoverride'}) ) {
			$me->{hostname}=$conf{'hostoverride'};
		} else {
			$me->{hostname}="acall-" . $me->{role};
			if ( defined($me->{cellname}) ) {
				$me->{hostname} = $me->{hostname} . "." . $me->{cellname};
			}
		}
		my $hostnameu=getFQDN($me->{hostname});
		if ( ! defined($hostnameu) ) {
			if ( ! $conf{test} ) {
				out('e',"Unable to resolve acall server hostname '$me->{hostname}'");
			}
			return undef;
		}
		$me->{hostname}=$hostnameu;
	}
	if ( defined($conf{'version'}) ) {
		$me->{version}=$conf{'version'};
	}
	if ( defined($conf{'key_timeout'}) ) {
		$me->{key_timeout}=$conf{'key_timeout'};
	} else {
		$me->{key_timeout}=-1;
	}
	return $me;
}

=item ($root,$xc,$doc)=acallXml(%options);

Reads the acall.xml file from network wide etc folder. Possible I<%options> are:

=over 4

=item B<bw> => B<1>

Open file in RW location and lock it for writing later.

=item B<filename> => I<filename>

Use I<filename> to read acall rules from.

=back

=cut

sub acallXml {
	my ($self,%options)=@_;

	my $filename;
	my $rw=0;
	if ( defined($options{filename}) ) {
		$filename=$options{filename};
	} else {
		my $dot=dot();
		if ( $options{rw} ) {
			$dot=".";
			$rw=1;
		}
		$filename="/afs/$self->{cellname}/admin/${dot}etc/acall.xml";
	}
	
	# Determine, if the access file was modified
	my @stat=stat($filename);
	my $mtime=0;
	if ( @stat ) {
		$mtime=$stat[9];
	} else {
		out('e',"Unable to read mtime of '$filename'");
	}
	if ( $self->{acall_xml_mtime} and $mtime ) {
		if ( $mtime == $self->{acall_xml_mtime} ) {
			if ( ( $rw and defined($self->{fh_acallxml_lock}) ) or ! $rw ) {
				return ($self->{root},$self->{xc},$self->{doc});
			}
		}
	}

	if ( $rw ) {
		out('d',"Locking acall.xml file");
		my $lockfile=$filename . ".lock";
		if ( open($self->{fh_acallxml_lock},">> $lockfile") ) {
			if ( ! flock($self->{fh_acallxml_lock},2) ) {
				out('e',"Unable to lock acall.xml");
				close($self->{fh_acallxml_lock});
				return (undef,undef,undef);
			}
		} else {
			out('e',"Unable to open lock file for acall.xml");
			return (undef,undef,undef);
		}
	}
	$self->{acall_xml_mtime}=$mtime;
	my $parser=XML::LibXML->new();
	$parser->no_network(1);
	$parser->load_ext_dtd(0);
	$parser->validation(0);
	$self->{doc}=$parser->parse_file($filename);
	$self->{root}=$self->{doc}->getDocumentElement;
	$self->{xc}=XML::LibXML::XPathContext->new($self->{doc});
	$self->{xc}->registerNs('ac','http://instantafs.cbs.mpg.de/xmlns/acallconfig/1.0');
	
	return ($self->{root},$self->{xc},$self->{doc});
}

sub acallXmlWrite {
	my ($self,$keeplock)=@_;
	if ( ! $self->{fh_acallxml_lock} ) {
		out('e',"No lock was set for acall.xml");
		return 0;
	}
	if ( ! $self->{doc}->toFile("/a/admin/.etc/acall.xml") ) {
		out('e',"Unable to write to acall.xml");
		return 0;
	} else {
		out('v',"Successfully wrote acall.xml");
	}
	if ( $keeplock ) {
		return 1;
	}
	if ( ! flock($self->{fh_acallxml_lock},8) ) {
		out('e',"Unable to lift lock from acall.xml");
		return 0;
	}
	$self->{fh_acallxml_lock}=undef;
	return 1;
}



=item $key=$acall->get_key()

Generates a random key by reading entropy from B</dev/urandom> and
returns it in I<$key>.

=cut


sub get_key {
	my ($self,$length,$timeout)=@_;
	my $key;
	my $prefix='';
	if ( ! defined($timeout) ) {
		$timeout=0;
	}
	if ( $timeout == -1 ) {
		$prefix='u';
	}
	if ( open(I, "/dev/${prefix}random") ) {
		sysread(I,$key,$length); 
		close(I);
	} else {
		out('e',"Unable to get random bits, aborting for security reasons.");
		return undef;
	}
	return($key);
}

=item ($result,$exit_code,$output)=$acall->call($command,\@params)

Sends a single acall-request to the server defined while creating
I<$acall>. The request's command is I<$command>, parameters are defined in I<@params>.
The remotely executed command's output (STDERR,STDOUT) is returned, too.

I<$result> == 1, only if the acall was successfully transmitted

I<$exit_code> : the exit code of the program called remotely
                (-> 0 == success )

I<$output>    : the output (STDERR + STDOUT of the executed command)

=cut


sub call {
	my ($self,$command,$params)=@_;

	my $socket = new IO::Socket::INET(
		PeerAddr => $self->{hostname},
		PeerPort => 6999,
		Proto => 'tcp'
	);

	if ( ! defined($socket) ) {
		out('e',"acall(): Unable to open socket: $!");
		return (0,0,'');
	}
	out('dd',"acall: Opened UDP socket");

	my $header=pack("v",65535) . pack("v",$self->{version});
	print $socket $header;
	
	my $comm;
	if ( $self->{version} == 3 ) {
		return $self->doProtocol3($socket,$command,$params);
	} else {
		out('e',"Protocol version '$self->{version}' is not supported");
		return (0,0,'');
	}
}

sub doProtocol3 {
	my ($self,$socket,$command,$params)=@_;
	if ( $VERBOSE ) {
		my $securedparams='';
		my $dontshow=0;
		for my $p (@{$params}) {
			if ( $dontshow ) {
				$securedparams .= ' [hidden]';
				$dontshow=0;
			} else {
				if ( $p =~ /^--passwd/ ) {
					if ( $p =~ /(.*?)=.*$/ ) {
						$securedparams .= " $1=[hidden]";
					} else {
						$dontshow=1;
						$securedparams .= " $p";
					}
				} else {
					$securedparams .= " $p";
				}
			}
		}
		out('v',"Cmd: $command $securedparams");
	}
	my $gss_error=sub {
		my ($message,$status)=@_;
		my @major_errors = $status->generic_message();
		my @minor_errors = $status->specific_message();
		out('e',$message);
		for my $e (@major_errors) {
			out('i'," Major: $e");
		}
		for my $e (@minor_errors) {
			out('i'," Minor: $e");
		}
	};

	my $server_name="acall\@$self->{hostname}";
	my $status=GSSAPI::Name->import(my $gss_server_name, $server_name, gss_nt_service_name);
	if ( ! $status ) {
		&$gss_error("aCall::GSSAPI: Unable to import server name: $server_name", $status);
		return (0,0,'');
	}
	$status = $gss_server_name->display(my $display_name);
	out('dd',"aCall::GSSAPI: communicating with '$display_name'");

	my $gss_auth_flags = GSS_C_MUTUAL_FLAG; # Use mutual authentication - why not :-)
	my $counter=0;
	my $gss_input_token;
	my $client_context=GSSAPI::Context->new();
	my $out_flags=0;
	my $out_mech=0;
	my $out_time=0;
	do {
		$counter++;
		no warnings 'uninitialized'; # Suppress warning when in the first run there's no input token
		my $gss_output_token=0;
		$status = $client_context->init(
			GSS_C_NO_CREDENTIAL,
			$gss_server_name,          # authenticate to this name
			GSS_C_NO_OID,              # use default mechanism (krb5)
			$gss_auth_flags,           # input flags
			0,                         # input time
			GSS_C_NO_CHANNEL_BINDINGS, # no channel binding
			$gss_input_token,          # input token
			0,
			$gss_output_token,
			$out_flags,
			$out_time
		);
		if ( ! $status ) {
			&$gss_error("aCall::GSSAPI: Unable to initialize security context",$status);
			return (0,0,'');
		}
		out('dd',"aCall::GSSAPI: Context init flags='$out_flags', time='$out_time', mech='$out_mech'");

		if ($counter == 1) {
			# Client auth
		} elsif ( $counter == 2) {
			my $server_name;
			$status = $gss_server_name->display($server_name);
			if ( ! $status ) {
				&$gss_error("aCall::GSSAPI: Unable to read server name",$status);
				return (0,0,'');
			}
			out('dd',"Confirmed server identity '$server_name'");
		} else {
			out('w',"Unexpected GSSAPI iteration '$counter'");
		}
		my $encoded_token_length;
		if( $gss_output_token ) {
			out('d',"Sending token (length='" . length($gss_output_token) . "')");
			$encoded_token_length=pack("v",length($gss_output_token));
			print $socket $encoded_token_length . $gss_output_token;
		}
		if ($status->major & GSS_S_CONTINUE_NEEDED) {
			read($socket,$encoded_token_length,2);
			if ( length($encoded_token_length) == 0 ) {
				out('e',"acall was interrupted (1)");
				return (0,0,'');
			}
			my $token_length=unpack("v",$encoded_token_length);
			read($socket,$gss_input_token,$token_length);
			out('dd',"Read GSSAPI token (length=$token_length)");
		}
	} while ( $status->major & GSS_S_CONTINUE_NEEDED );
	my %packet=(
		'serverrole' => $self->{role},
		'command' => $command,
		'params' => $params,
	);
	my $packet;my $encrypted_packet;
	$packet=nfreeze(\%packet);
	my $confstate;
	$status=$client_context->wrap(1,GSS_C_QOP_DEFAULT,$packet,$confstate,$encrypted_packet);
	if ( ! $status ) {
		&$gss_error("Unable to encrypt/sign RPC request",$status);
		return (0,0,'');
	}
	if ( ! $confstate ) {
		out('e',"Unable to encrypt RPC request");
		return (0,0,'');
	}
	out('dd',"GSSAPI-Encryption complete. state='$confstate', length='" . length($encrypted_packet) . "'");
	my $packed_encrypted_packet_length=pack("V",length($encrypted_packet));
	print $socket $packed_encrypted_packet_length . $encrypted_packet;
	
	read($socket,$packed_encrypted_packet_length,4);
	my $encrypted_packet_length=unpack("V",$packed_encrypted_packet_length);
	read($socket,$encrypted_packet,$encrypted_packet_length);

	my $packet_buffer;
	$status=$client_context->unwrap($encrypted_packet,$packet_buffer,$confstate,GSS_C_QOP_DEFAULT);
	if ( ! $status ) {
		&$gss_error("Unable to decrypt RPC response",$status);
		return (0,0,'');
	}
	$packet=thaw($packet_buffer);
	return (1,$packet->{exitcode},$packet->{text});
}


=back


=head1 APPENDIX

=head2 Environment variables

Some variables are set by the acall-server to give called programs/scripts
a hint about what they are expected to do:

=over 4

=item I<$ACALL_PRINCIPAL>

The Kerberos5-name of the calling user (i.e. root/admin)

=item I<$ACALL_USER>

The Kerberos4-name (= the AFS-name) of the calling user (i.e. root.admin)

=item I<$ACALL_SERVER>

The acall-server-context specified by the user (i.e. 'db')

=item I<$ACALL_INUSE>

Will be set to '1' as a hint for subprocesses that they are used
in an acall-operation.

=back

=head2 aCall Protocol

Current RPC version is 2. All numbers are send big endian.

=over 4

=item 1. B<Kerberos Negotiation>

We don't have influence on that. The client uses its/her/his key to authenticate
against the server using an B<acall>/I<serverfqdn>@I<REALM> key which is read
from B</etc/instantafs/acall.local.keytab> .

=item 2. B<Session Init>

A Sessionkey is generated and encoded using a shared secret of client and
server based on kerberos:

C: I<$sessionkey>=randomstring(length => 32)

C: I<$encodedSessionkey>=kerberoscrypt($sessionkey)

C: send(0xffff) # 2 Bytes

C: send($rpcversion) # 2 Bytes

C: send(length(I<$encodedSessionkey>)) # 2 Bytes

C: send(I<$encodedSessionkey>)

S: $_=receive(2) # 0xffff header

S: I<$rpcversion>=receive(2) # only B<2> is allowed

S: I<$encodedkeylength>=receive(2)

S: I<$encodedSessionkey>=receive(I<$encodedkeylength>)

S: I<$sessionkey>=kerberosdecrypt(I<$encodedSessionkey>)


=item 3. B<RPC Request>

C: I<$RPC>=Frontier::RPC2::encode(I<$acall_role>,I<$command>,I<@params>)

C: I<$encodedRPC>=AESencrypt(I<$RPC>,I<$sessionkey>)

C: send(length(I<$encodedRPC>)) # 4 Bytes

C: send(sha1(I<$RPC>)) # 20 Bytes

=item 4 B<RPC Processing>

S: I<$encodedRPClength>=receive(4)

S: I<$encodedRPC>=receive(I<$encodedRPClength>)

S: I<$RPCsha>=receive(20)

S: I<$RPC>=AESdecrypt(I<$encodedRPC>,I<$sessionkey>)

S: if ( sha1(I<$RPC>) ne I<$RPCsha> ) { die "Attack!"; }

S: (I<$exitcode>,I<$output>)=processRPC(I<$RPC>)

S: I<$Response>=Frontier::RPC::encode(I<$exitcode>,I<$output>)

S: I<$encodedResponse>=AESencrypt(I<$Response>,$I<$sessionkey>)

S: send(length(I<$encodedResponse>)) # 4 Bytes

S: send(I<$encodedResponse>)

=item 5 B<Request processing>

C: I<$Responselength>=receive(4)

C: I<$Responsesha>=receive(20)

C: I<$encodedResponse>=receive(I<$Responselength>)

C: I<$Response>=AESdecrypt(I<$encodedResponse>,I<$sessionkey>)

C: if ( sha1(I<$Response>) ne $Responsesha ) { die "Attack!"; }

C: (I<$exitcode>,I<$output>)=Frontier::RPC::decode(I<$Response>)

=back

=head1 AUTHOR

Frank Burkhardt <burk@cbs.mpg.de>

=cut

1;

